/*
"(c) Copyright 1993-1997 Pad++ Consortium {University of New Mexico (UNM),
and New York University (NYU)}, All Rights Reserved."  
Licensee can not remove or obscure any of the
copyright or trademark notices in this software.

IN NO EVENT SHALL THE AUTHORS OR DISTRIBUTORS BE LIABLE TO ANY PARTY
FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
ARISING OUT OF THE USE OF THIS SOFTWARE, ITS DOCUMENTATION, OR ANY
DERIVATIVES THEREOF, EVEN IF THE AUTHORS HAVE BEEN ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.

THE AUTHORS AND DISTRIBUTORS SPECIFICALLY DISCLAIM ANY WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE, AND NON-INFRINGEMENT.  THIS SOFTWARE
IS PROVIDED ON AN "AS IS" BASIS, AND THE AUTHORS AND DISTRIBUTORS HAVE
NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR
MODIFICATIONS.

See the file "License" for general information on usage and
redistribution, and the file "LicenseTerms" for the specific license
agreement on usage and redistribution of this file, and the Pad++
software in general.
*/

#include "defs.h"
#include "alias.h"
#include "pad.h"
#include "global.h"
#include "win.h"

#include <stdlib.h>

//////////////////////////////////////////////
//              Alias definitions
//////////////////////////////////////////////  

#define ALIAS_DEFAULT_REFERENCE    NULL

Pad_Alias::~Pad_Alias()
{
    Generate_delete();
    if (reference) {
	reference->Remove_alias(this);
	reference = NULL;
    }
}

Pad_Alias::Pad_Alias(Pad *pad) :
Pad_Object(pad)
{
    _type = PAD_ALIAS;
    aliasFlags = PAD_NO_MASK;
    reference = NULL;
}

//
// Set the object's <option> to the value specified by <sval>.
// Note that each option has a specific setter and getter method
// if the correct argument type is available.
//
Pad_Bool
Pad_Alias::Set_option(Pad_Option *option, char *sval)
{
    Pad_Bool rc = TRUE;
    Pad_Object *obj;

    switch (option->id) {
      case PAD_REFERENCE:
	obj = pad->Get_object_from_id(atoi(sval));
	rc = Set_reference(obj);
	break;
      default:
	rc = Pad_Object::Set_option(option, sval);
	break;
    }

    return(rc);
}

//
// Set the object's <option> to the default value.
//
Pad_Bool
Pad_Alias::Set_option_default(Pad_Option *option)
{
    Pad_Bool rc = TRUE;

    switch (option->id) {
      case PAD_REFERENCE:
	Set_reference_default();
	break;
      default:
	rc = Pad_Object::Set_option_default(option);
	break;
    }

    return(rc);
}

//
// Return the value of the object's <option> as a string
// in <result>.
//
Pad_Bool
Pad_Alias::Get_option(Pad_Option *option, Pad_String &result, Pad_Bool &modified)
{
    Pad_Bool rc = TRUE;
    Pad_Object *obj;

    switch (option->id) {
      case PAD_REFERENCE:
	modified = (aliasFlags & ALIAS_REFERENCE_SET) ? TRUE : FALSE;
	obj = Get_reference();
	if (obj) {
	    result = obj->id;
	} else {
	    result = "{}";
	}
	break;
      default:
	rc = Pad_Object::Get_option(option, result, modified);
	break;
    }

    return(rc);
}

//
// Setters and getters for reference
//
Pad_Bool
Pad_Alias::Set_reference(Pad_Object *obj)
{
    if (obj && (obj->Type() == PAD_ALIAS)) {
	Pad_errorString = "Can't make an alias of an alias";
	return(FALSE);
    }

    aliasFlags |= ALIAS_REFERENCE_SET;
    if (reference) {
	reference->Remove_alias(this);
    }
    if (obj) {
	obj->Add_alias(this);
    }
    reference = obj;
    Update();
    
    return(TRUE);
}

void
Pad_Alias::Set_reference_default(void)
{
    Set_reference(ALIAS_DEFAULT_REFERENCE);
    aliasFlags &= ~ALIAS_REFERENCE_SET;
}

Pad_Object *
Pad_Alias::Get_reference(void)
{
    return(reference); 
}

void
Pad_Alias::Compute_bounding_box() 
{
    Pad_BBox bb;

    if (reference) {
	reference->Get_bbox(bb);
	if (!(optionFlags & PAD_WIDTH_SET)) {
	    Set_bbox_xmin(bb.Xmin());
	    Set_bbox_xmax(bb.Xmax());
	}
	if (!(optionFlags & PAD_HEIGHT_SET)) {
	    Set_bbox_ymin(bb.Ymin());
	    Set_bbox_ymax(bb.Ymax());
	}

	Pad_Object::Compute_bounding_box();
    } else {
        Set_bbox(0, 0, 0, 0);
    }
}

//
// Compute the dimensions of the alias based on the dimensions of
// its reference.  Also set the reference's dimensions to that
// of the alias (for proper rendering, etc.).
//
float
Pad_Alias::Compute_dimensions(float &global_width, float &global_height)
{
    float dz;

    if (reference) {
	dz = transform.Get_scale() / reference->transform.Get_scale();
	reference->Compute_dimensions(global_width, global_height);
	pixelDim = reference->pixelDim * dz;
	percentDim = reference->percentDim * dz;
	reference->pixelDim = pixelDim;
	reference->percentDim = percentDim;
    } else {
	percentDim = 0;
	pixelDim = 0;
    }

    return(pixelDim);
}

float
Pad_Alias::Compute_dimensions(Pad_Event *event, float &global_width, float &global_height)
{
    float dz;

    if (reference) {
	dz = transform.Get_scale() / reference->transform.Get_scale();
	reference->Compute_dimensions(event, global_width, global_height);
	pixelDim = reference->pixelDim * dz;
	percentDim = reference->percentDim * dz;
	reference->pixelDim = pixelDim;
	reference->percentDim = percentDim;
    } else {
	percentDim = 0;
	pixelDim = 0;
    }

    return(pixelDim);
}

float
Pad_Alias::Compute_dimensions(Pad_List &views, float &global_width, float &global_height)
{
    float dz;

    if (reference) {
	dz = transform.Get_scale() / reference->transform.Get_scale();
	reference->Compute_dimensions(views, global_width, global_height);
	pixelDim = reference->pixelDim * dz;
	percentDim = reference->percentDim * dz;
	reference->pixelDim = pixelDim;
	reference->percentDim = percentDim;
    } else {
	percentDim = 0;
	pixelDim = 0;
    }

    return(pixelDim);
}

//
// Determines if the specified point is within halo pixels of the 
// reference of this alias.
// Coordinates are local.
//
Pad_Bool
Pad_Alias::Pick(Pad_Event *event, float halo)
{
    Pad_Bool rc;

    if (reference) {
	rc = reference->Pick(event, halo);
    } else {
	rc = FALSE;
    }

    return(rc);
}

//
// Return true to continue rendering other objects, or false if rendering
// was interrupted.
//
Pad_Bool
Pad_Alias::Render(void)
{
    Pad_Bool rc;
    float globalWidth, globalHeight;

    Compute_dimensions(globalWidth, globalHeight);
    if (reference) {
				// There's a tricky issue here related to bboxes.
				// If the reference is a group, its members might
				// not get rendered because their global bounding
				// boxes might not be visible (despite the fact
				// that this alias is visible).  Thus, to correct
				// this situation, we will temporarily shift the
				// active bbox to be related to the reference instead
				// of this alias.
	Pad_BBox *activeBBox = &Pad_prc->activeBBox;
	Pad_BBox tmpBBox(*activeBBox);
	float dx = reference->Get_global_bbox()->Xmin() - Get_global_bbox()->Xmin();
	float dy = reference->Get_global_bbox()->Ymin() - Get_global_bbox()->Ymin();
	activeBBox->Translate(dx, dy);
	rc = reference->Render();
	Pad_prc->activeBBox = tmpBBox;
    } else {
	rc = TRUE;
    }

    return(rc);
}

void
Pad_Alias::Write_obj(char *padname, Pad_String &str)
{
    Pad_Object *obj = Get_writeobj();
    if (obj == NULL) {
        Pad_Object::Write_obj(padname, str);
    } else {
        Pad_StreamData sd;
	obj->Write_obj_wrapper(PAD_SCRIPT_FORMAT, padname, str, sd);
        Pad_Object::Write_obj(padname, str);
    }
}

//
// Write the -reference option of the alias.
// Assume the Pad_NewID entry of the ref. object is already set.
//
void
Pad_Alias::Write_obj_args(char *, Pad_String &str)
{
    if (reference) {
        str += "-reference $Pad_NewID(";
	str += reference->id;
	str += ") ";
    }
}

Pad_Bool
Pad_Alias::Write_obj(char *padname, Pad_StreamData &sd)
{
    Pad_Bool rc=TRUE;
    Pad_Object *obj = Get_writeobj();
    if (obj == NULL) {
        rc = Pad_Object::Write_obj(padname, sd);
	sd << aliasFlags;
	_Options_stream(sd, TRUE);
	sd << PAD_ENDOBJ_OPTIONS;
    } else {
        Pad_String str;
	obj->Write_obj_wrapper(sd.Get_rep(), padname, str, sd);
	rc = Write_obj(padname, sd);
    }

    return rc;
}

Pad_Bool
Pad_Alias::Read_obj(Pad_StreamData &sd)
{
   Pad_Bool rc = Pad_Object::Read_obj(sd);

   rc &= !sd.Reconstruct(aliasFlags);
   if (!rc) {
     Pad_errorString << "ERR: problems reading alias object: " << id << "\n";
   }

   rc &= _Options_stream(sd, FALSE);

   return rc;
}

Pad_Bool
Pad_Alias::_Options_stream(Pad_StreamData &sd, Pad_Bool write) 
{
    Pad_Bool rc=TRUE;

    if (write) {
      Pad_Type *type = (Pad_Type *)pad->typeTable->Get(Pad_aliasUid);
      Pad_Iterator oi;
      Pad_Option *option;
      DOLIST(oi, type->options, Pad_Option, option) {      
	rc &= _Option_stream(sd, option->id, write, option);
      }
      sd << PAD_ENDOBJ_OPTIONS;

    } else {
      int optionid=0;
      while ((sd.Reconstruct(optionid) == noErr) &&
	     (optionid != PAD_ENDOBJ_OPTIONS)) {
	rc &= _Option_stream(sd, optionid, write);
      }

      if (optionid != PAD_ENDOBJ_OPTIONS) {
	Pad_errorString << "Could not get to end of options of object: "
			<< id << " of type: " << Type_name() << "\n";
	rc = FALSE;
      }
    }

    return rc;
}

Pad_Bool
Pad_Alias::_Option_stream(Pad_StreamData &sd, int optionid, Pad_Bool write,
			  Pad_Option *option)
{
    Pad_Bool rc=write;
    Pad_String str;
    int oldRefid;
    int newRefid;
    char *newRef;
    Pad_StreamData opsd(sd.Get_rep(), 0, sd.Get_istream(), sd.Get_ostream());
    int optLength=0;

    if (write) {
      rc = TRUE;
    } else {
      rc = FALSE;
      // get the options total length
      sd.Reconstruct(optLength);
    }  

    switch (optionid) {
    case PAD_REFERENCE:
      if (write) {
	if (aliasFlags & ALIAS_REFERENCE_SET) {
	  sd << PAD_REFERENCE;
	  opsd << reference->id;
	}
      } else {
	if (!sd.Reconstruct(oldRefid)) {
	  str = oldRefid;
	  newRef = Tcl_GetVar2(pad->view->win->interp, "Pad_NewID", str.Get(), 
			       TCL_GLOBAL_ONLY | TCL_LEAVE_ERR_MSG);
	  if (newRef) {
	    newRefid = atoi(newRef);
	    reference = pad->Get_object_from_id(newRefid);
	    rc = (reference) ? TRUE : FALSE;
	  } else {
	    rc = FALSE;
	  }
	}
      }
      break;
    case PAD_USER_OPTION:
      rc = _UserOption_stream(sd, opsd, optionid, write, option, 
			      Pad_aliasUid, optLength);
      break;

    default:
      Pad_errorString << "Unknown option id: " << optionid
		      << " of object: " << id << " of type: " << Type_name();
      if (!write) {
	Pad_errorString << " -- SKIPPING " << optLength << " BYTES";
	sd.SkipInput(optLength);
      }
      Pad_errorString << '\n';
      rc = FALSE;
    }

    if (write && rc) {
      if (opsd.Items()) {
	sd << opsd.ItemsLength();
	sd.MoveItems(opsd);
      }
    }
        
    return rc;
}

//
// Find the object which should get written out before the alias
// object can be written.  If the reference object is already written then
// NULL is returned.  Otherwise the reference object is returned (unless it's
// in a group).
//
//
Pad_Object*
Pad_Alias::Get_writeobj(void)
{
    Pad_Object *obj=NULL;
    if (reference) {
        if ( !reference->Get_written() ) {
	    for (obj = reference; obj->group; obj = (Pad_Object*) obj->group);
	    if (obj == (Pad_Object*) group) {
	        obj = reference;
	    }
	}
    }
    
    return obj;
}

