/*
"(c) Copyright 1993-1997 Pad++ Consortium {University of New Mexico (UNM),
and New York University (NYU)}, All Rights Reserved."  
Licensee can not remove or obscure any of the
copyright or trademark notices in this software.

IN NO EVENT SHALL THE AUTHORS OR DISTRIBUTORS BE LIABLE TO ANY PARTY
FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
ARISING OUT OF THE USE OF THIS SOFTWARE, ITS DOCUMENTATION, OR ANY
DERIVATIVES THEREOF, EVEN IF THE AUTHORS HAVE BEEN ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.

THE AUTHORS AND DISTRIBUTORS SPECIFICALLY DISCLAIM ANY WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE, AND NON-INFRINGEMENT.  THIS SOFTWARE
IS PROVIDED ON AN "AS IS" BASIS, AND THE AUTHORS AND DISTRIBUTORS HAVE
NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR
MODIFICATIONS.

See the file "License" for general information on usage and
redistribution, and the file "LicenseTerms" for the specific license
agreement on usage and redistribution of this file, and the Pad++
software in general.
*/

#include "api.h"
#include "global.h"
#include "group.h"
#include "pad.h"
#include "bind.h"
#include "anim-channel.h"
#include "anim.h"
#include "animatable.h"
#include "callback.h"
#include "pad-sound.h"
#include "pad-tcl-sound.h"

#include <ctype.h>
#include <stdlib.h>
#include <string.h>

#define UCHAR(c) ((unsigned char) (c))
////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////
//
// Pad_Handle Class definition
//
////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////


//
// Alternative memory allocation for Pad_Handle's
// in an attempt to reduce memory fragmentation.
//
// Instead of allocating these small handles one at a time
// which would result in lots of little memory pieces, we
// allocated larger blocks of memory, and then dole them
// out for handles one at a time.  Note that this requires
// that all handles are the same size.  Also, note that
// handles are never returned to the system.  We assume
// that roughly a similar number of handles are used over
// and over.
//

static Pad_Handle *freelist = NULL;

void *
Pad_Handle::operator new(size_t size)
{
#define BLOCK_SIZE 512

    Pad_Handle *result;

    static int nused = BLOCK_SIZE;
    static Pad_Handle *freeblock = NULL;

    if (freelist) {
	// If there is a handle on the free list, use it
	result = freelist;
	freelist = freelist->_next;
    } else {
	// Otherwise, if we don't have a block of handles available, alloc up a
	// new block of them
	if (nused >= BLOCK_SIZE) {
	    freeblock = (Pad_Handle *)malloc(BLOCK_SIZE * size);
	    nused = 0;
	}
	// Take a handle off the block
	result = &freeblock[nused++];
    }

    return result;

#undef BLOCK_SIZE
}

void
Pad_Handle::operator delete(void *object)
{
    Pad_Handle *handle = (Pad_Handle *)object;

    handle->_obj = NULL;
    handle->_tag = NULL;
    handle->_next = freelist;
    freelist = handle;
}

/*
 *----------------------------------------------------------------------
 *
 * Pad_Handle::Pad_Handle --
 *
 *   Constructor for Pad_Handle
 *
 * Results:
 *   None
 *
 *----------------------------------------------------------------------
 */

Pad_Handle::Pad_Handle()
{
    _obj = NULL;
    _next = NULL;
    _tag = NULL;
    _data = NULL;
}

Pad_Handle::Pad_Handle(Pad_Handle &handle)
{
    *this = handle;
}

Pad_Handle::~Pad_Handle(void)
{
}

Pad_Handle & 
Pad_Handle::operator=(Pad_Handle &handle)
{
				// If assigning to self, then do nothing
    if (this == &handle) {
	return(*this);
    }

    _obj = handle._obj;
    _tag = handle._tag;
    _data = handle._data;
    _next = handle._next;
    handle._next = this;

    return (*this);
}

Pad_Bool
Pad_Handle::operator==(Pad_Handle &handle)
{
    if ((_obj == handle._obj) &&
	(_data == handle._data) &&
	((_tag && handle._tag && (*_tag == *handle._tag)) ||
	 (!_tag && !handle._tag))) {
	return(TRUE);
    } else {
	return(FALSE);
    }
}

Pad_Bool
Pad_Handle::operator!=(Pad_Handle &handle)
{
    return (!(*this == handle));
}

////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////
//
// Pad_ObjHandle Class definition
//
//    A handle to one or more Pad++ Objects.  A handle can specify either a
//    single object with very quick access, or a group of objects specified
//    by a tag with access through a hash table.  Handles are memory-safe,
//    even when the object the handle references has been deleted.
//
////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////


Pad_ObjHandle::Pad_ObjHandle()
{
}

Pad_ObjHandle::Pad_ObjHandle(Pad_ObjHandle &objHandle)
{
    Attach(objHandle);
}

Pad_ObjHandle::Pad_ObjHandle(Pad_Object *obj)
{
    Attach(obj);
}

Pad_ObjHandle::Pad_ObjHandle(Pad_SurfaceHandle &surface, char *tag)
{
    Attach(surface, tag);
}

Pad_ObjHandle::Pad_ObjHandle(Pad_SurfaceHandle &surface, int objectId)
{
    Attach(surface, objectId);
}

Pad_ObjHandle::Pad_ObjHandle(char *surfaceName, char *tag)
{
    Attach(surfaceName, tag);
}

Pad_ObjHandle::Pad_ObjHandle(char *surfaceName, int objectId)
{
    Attach(surfaceName, objectId);
}

Pad_ObjHandle & 
Pad_ObjHandle::operator=(Pad_ObjHandle &objHandle)
{
    Attach(objHandle);

    return (*this);
}

/*
 *----------------------------------------------------------------------
 *
 * Pad_ObjHandle::~Pad_ObjHandle --
 *
 *   Destructor for Pad_ObjHandle
 *   Remove this handle from linked list of handles that 
 *   share its object.  Leave object alone.
 *
 * Results:
 *   None
 *
 *----------------------------------------------------------------------
 */

Pad_ObjHandle::~Pad_ObjHandle(void)
{
    Attach((Pad_Object *)NULL);
}

/*
 *----------------------------------------------------------------------
 *
 * Pad_ObjHandle::Get_handle_type --
 *
 *   Returns the type of this particular handle.
 *
 * Results:
 *   Handle type
 *
 *----------------------------------------------------------------------
 */

Pad_HandleType
Pad_ObjHandle::Get_handle_type(void) const
{
    return(Pad_ObjHandleType);
}

/*
 *----------------------------------------------------------------------
 *
 * Pad_ObjHandle::Get_object_type --
 *
 *   Returns the type of the first object specified by this handle.
 *   If this handle does not specify any objects, then Pad_NoType
 *   is returned.
 *
 * Results:
 *   Object type
 *
 *----------------------------------------------------------------------
 */

Pad_ObjectType
Pad_ObjHandle::Get_object_type(void) const
{
    Pad_Object *obj;
    Pad_ObjectType type = Pad_NoType;

    FIRSTHANDLEOBJ(this, obj);
    if (obj) {
	type = (Pad_ObjectType)obj->Type();
    }

    return(type);
}

/*
 *----------------------------------------------------------------------
 *
 * Pad_ObjHandle::Get_id --
 *
 *   Returns the unique integer id of the object the first object this
 *   handle references (in display-list order), or 0 if there is no
 *   object.  Note that id's are only unique per pad surface.
 *
 * Results:
 *   Object pointer
 *
 *----------------------------------------------------------------------
 */

int
Pad_ObjHandle::Get_id(void) const
{
    int id = 0;
    Pad_Object *obj;

    FIRSTHANDLEOBJ(this, obj);
    if (obj) {
	id = obj->id;
    }

    return(id);
}

/*
 *----------------------------------------------------------------------
 *
 * Pad_ObjHandle::Attach --
 *
 *   Set the object this handle refers to.
 *
 *   There are five methods:
 *
 *      one using an object pointer.
 *
 *      the other four are the permutations of the two ways of
 *      specifying an boject---either by integer ID, or by char *
 *      tagorid---and the two ways of specifying a surface---either by
 *      a surfaceHandle or by a char * surfaceName.
 *
 *   In any case, if the handle previously referenced an object, that
 *   reference will be properly removed.  If it is set to NULL, it
 *   will not reference any object.
 *
 * Results:
 *   Returns TRUE or FALSE. It is always TRUE for tags. For object ids
 *   it is TRUE if the id matches the id of an object an the surface,
 *   and for a Pad_Object* it is TRUE if the pointer is not NULL.
 *
 *----------------------------------------------------------------------
 */

Pad_Bool
Pad_ObjHandle::Attach(char *surfaceName, char *tagOrId)
{
    Pad_SurfaceHandle surface;

    surface.Attach(surfaceName);// make surface handle from name
    return Attach(surface, tagOrId);
}

Pad_Bool
Pad_ObjHandle::Attach(Pad_SurfaceHandle &surface, char *tagOrId)
{
    Pad_Bool isInt;
    int stringLen;
    char *tagOrIdCpy;
    int digitCount = 0;
    Pad *pad;

    stringLen  = strlen(tagOrId);
    tagOrIdCpy = tagOrId;

    while (isspace(UCHAR(*tagOrId))) {
	tagOrId++;
	digitCount++;
    }
    while (isdigit(UCHAR(*tagOrId))) {
	tagOrId++;
	digitCount++;
    }

    if (digitCount == stringLen) {
	isInt = TRUE;
    } else {
	isInt = FALSE;
    }

    if (isInt) {
	return Attach(surface, atoi(tagOrIdCpy));
    } else {
	Attach((Pad_Object *)NULL);   // First dereference any objects
	pad = surface._Get_pad();
	if (pad) {
	    _obj = pad;
	    _tag = new Pad_String(tagOrId);
	    return TRUE;
	} else {
	    return FALSE;
	}
    }
}

Pad_Bool
Pad_ObjHandle::Attach(Pad_SurfaceHandle &surface, int id)
{
    Pad *padPtr;
    Pad_Bool rc;

    padPtr = surface._Get_pad();
    if (padPtr) {
	rc = Attach(padPtr->Get_object_from_id(id));
    } else {
	rc = Attach((Pad_Object *)NULL);
    }

    return(rc);
}

Pad_Bool
Pad_ObjHandle::Attach(Pad_ObjHandle &objHandle)
{
    Pad_Bool rc;
    char *tag;

				// If assigning to self, then do nothing
    if (this == &objHandle) {
	return(TRUE);
    }

    tag = objHandle._Get_tag();
    if (tag) {
	Pad_SurfaceHandle surface;
	objHandle.Get_surface(surface);
	rc = Attach(surface, tag);
    } else {
	rc = Attach(objHandle._Get_object());
    }

    return(rc);
}

Pad_Bool
Pad_ObjHandle::Attach(Pad_Object *obj)
{
				// First, dereference any objects
    if (_tag) {
	delete _tag;
	_tag = NULL;
    } else if (_obj) {
	Pad_Handle *prev, *cur;

	prev = NULL;
	cur = _obj->_handle;
	while (cur) {
	    if (cur == this) {
		if (prev == NULL) {
		    _obj->_handle = _next;
		} else {
		    prev->_next = _next;
		}
		break;
	    }
	    prev = cur;
	    cur = cur->_next;
	}
	_obj = NULL;
	_next = NULL;
    }

				// Then, if there is a new one, reference it
    if (obj) {
	_obj = obj;
	_next = obj->_handle;
	obj->_handle = this;

	_tag = NULL;
        return TRUE;
    }
        // obj was NULL
    return FALSE;
}

Pad_Bool
Pad_ObjHandle::Attach(char *surfaceName, int id)
{
    Pad *padPtr;
    Pad_SurfaceHandle surface;

    surface.Attach (surfaceName);
    padPtr = surface._Get_pad();
    if (padPtr) {
	return Attach(padPtr->Get_object_from_id(id));
    } else {
	return Attach((Pad_Object *)NULL);
    }
}

/*
 *----------------------------------------------------------------------
 *
 * Pad_ObjHandle::_Get_object --
 *
 *   PRIVATE:
 *
 *   If this handle specifies an object directly, this returns
 *   that object.  If this handle specifies a tag or nothing, this 
 *   returns NULL.
 *
 * Results:
 *   Tag or NULL
 *
 *----------------------------------------------------------------------
 */

Pad_Object *
Pad_ObjHandle::_Get_object(void) const
{
    return(_obj);
}

/*
 *----------------------------------------------------------------------
 *
 * Pad_ObjHandle::_Get_tag --
 *
 *   PRIVATE:
 *
 *   If this handle specifies a tag, this returns that tag.  
 *   If this handle specifies an object or nothing, this returns
 *   NULL.
 *
 * Results:
 *   Object pointer
 *
 *----------------------------------------------------------------------
 */

char *
Pad_ObjHandle::_Get_tag(void) const
{
    if (_tag) {
	return(_tag->Get());
    } else {
	return(NULL);
    }
}

/*
 *----------------------------------------------------------------------
 *
 * Pad_ObjHandle::Get_surface --
 *
 *   PRIVATE:
 *
 *   Returns the surface this handle is associated with, or 
 *   resets the surface parameter if none.
 *
 * Results:
 *   Fills in surface handle
 *
 *----------------------------------------------------------------------
 */

void
Pad_ObjHandle::Get_surface(Pad_SurfaceHandle &surface) const
{
    if (_tag) {
	surface.Attach((Pad *)_obj);
    } else {
	Pad_Object *obj = _Get_object();
	if (obj) {
	    surface.Attach(obj->pad);
	} else {
	    surface.Attach((Pad *)NULL);
	}
    }
}

/*
 *----------------------------------------------------------------------
 *
 * Pad_ObjHandle::Get_first_object --
 *
 *   This returns the first object specified by this handle
 *   in display-list order, or NULL if there are no objects.
 *
 * Results:
 *   Object pointer
 *
 *----------------------------------------------------------------------
 */

Pad_Object *
Pad_ObjHandle::_Get_first_object(void) const
{
    Pad_Object *obj;

    FIRSTHANDLEOBJ(this, obj);

    return(obj);
}

/*
 *----------------------------------------------------------------------
 *
 * Pad_ObjHandle::Is_mapped --
 *
 *   Determines if this handle is mapped to an object either directly
 *   or via a tag.
 *
 * Results:
 *   TRUE if handle is mapped to one or more objects.
 *
 *----------------------------------------------------------------------
 */
Pad_Bool
Pad_ObjHandle::Is_mapped(void)
{
    return ((_obj || _tag) ? TRUE : FALSE);
}

/*
 *----------------------------------------------------------------------
 *
 * Pad_ObjHandle::Get_tags --
 *
 *   This returns the tags that first object specified by this
 *   handle has.  The tags are returned by filling in a list.
 *
 * Results:
 *   None
 *
 *----------------------------------------------------------------------
 */

void
Pad_ObjHandle::Get_tags(Pad_List &tags) const
{
    Pad_Object *obj;

    obj = _Get_first_object();
    if (obj) {
	obj->Get_tags(tags);
    } else {
	tags.Make_empty();
    }
}

/*
 *----------------------------------------------------------------------
 *
 * Pad_ObjHandle::Set_tags --
 *
 *   Sets the tags of each object specified by handle.
 *   The new tags replace any tags the objects may have had.
 *   Tags can be specified by either a list of tags, or
 *   a single string where the string contains a list of
 *   space-separated tags.
 *
 * Results:
 *   True if successfull
 *
 *----------------------------------------------------------------------
 */

Pad_Bool
Pad_ObjHandle::Set_tags(Pad_List &tags)
{
    Pad_Object *obj;
    Pad_Bool rc = TRUE;

    DOHANDLE(this, obj, 
	     if (rc) {
		 rc = obj->Set_tags(tags);
	     }
    )

    return(rc);
}

Pad_Bool
Pad_ObjHandle::Set_tags(char *tags)
{
    Pad_Object *obj;
    Pad_Bool rc = TRUE;

    DOHANDLE(this, obj, 
	     if (rc) {
		 rc = obj->Set_tags(tags);
	     }
    )

    return(rc);
}

/*
 *----------------------------------------------------------------------
 *
 * Pad_ObjHandle::Append_coords --
 *
 *   Append the coordinates onto each object specified by the handle.
 *
 * Results:
 *   True if successfull
 *
 *----------------------------------------------------------------------
 */

Pad_Bool
Pad_ObjHandle::Append_coords(Pad_PList &coords)
{
    Pad_Object *obj;
    Pad_Bool rc = TRUE;

    DOHANDLE(this, obj, 
	     if (rc) {
		 rc = obj->Append_coords(coords, FALSE);
	     }
    )

    return(rc);
}

/*
 *----------------------------------------------------------------------
 *
 * Pad_ObjHandle::Set_coords --
 *
 *   Sets the coordinates of each object specified by the handle.
 *
 * Results:
 *   True if successfull
 *
 *----------------------------------------------------------------------
 */

Pad_Bool
Pad_ObjHandle::Set_coords(Pad_PList &coords)
{
    Pad_Object *obj;
    Pad_Bool rc = TRUE;

    DOHANDLE(this, obj, 
	     if (rc) {
		 rc = obj->Set_coords(coords, FALSE);
	     }
    )

    return(rc);
}

/*
 *----------------------------------------------------------------------
 *
 * Pad_ObjHandle::Get_coords --
 *
 *   Gets the coordinates of the first object specified by this handle.
 *   If no objects are specified by the handle, then no coordinates
 *   are returned.
 *
 * Results:
 *   None
 *
 *----------------------------------------------------------------------
 */

void
Pad_ObjHandle::Get_coords(Pad_PList &coords) const
{
    Pad_Object *obj;

    FIRSTHANDLEOBJ(this, obj);
    if (obj) {
	obj->Get_coords(coords, FALSE);
    } else {
	coords.Make_empty();
    }
}

/*
 *----------------------------------------------------------------------
 *
 * Pad_ObjHandle::Get_bbox --
 *
 *   This returns the global bounding box of all the objects
 *   specified by this handle.
 *
 * Results:
 *   None
 *
 *----------------------------------------------------------------------
 */

void
Pad_ObjHandle::Get_bbox(Pad_BBox &bbox) const
{
    Pad_Object *obj;
    Pad_Bool first = TRUE;
    Pad_BBox bb;
    
    DOHANDLE(this, obj, 
	     obj->Get_global_bbox(bb);
	     if (first) {
		 first = FALSE;
		 bbox = bb;
	     } else {
		 bbox.Union(bb);
	     }
    )

    if (first) {   // No objects
	bbox.Set_zero();
    }
}

/*
 *----------------------------------------------------------------------
 *
 * Pad_ObjHandle::Get_group --
 *
 *   This returns the group the first object specified by this handle
 *   is a member of.  If this handle does not specify any objects, or
 *   if the object specified is not a member of a group, then the
 *   handle will be emptied.
 *
 * Results:
 *   None
 *
 *----------------------------------------------------------------------
 */

void
Pad_ObjHandle::Get_group(Pad_ObjHandle &group) const
{
    Pad_Object *obj;
    Pad_Object *groupObj = NULL;

    FIRSTHANDLEOBJ(this, obj);
    if (obj) {
	groupObj = obj->group;
    }
    group.Attach(groupObj);
}

/*
 *----------------------------------------------------------------------
 *
 * Pad_ObjHandle::Set_transform --
 *
 *   Sets the transform of each of the objects specified by this handle.
 *   <x, y> specifes the position that the anchor point of each object
 *   should have, and <s> specifies the scale of each object.
 *
 * Results:
 *   True if successfull
 *
 *----------------------------------------------------------------------
 */

Pad_Bool
Pad_ObjHandle::Set_transform(float x, float y, float s)
{
    Pad_Object *obj;
    Pad_Bool rc = TRUE;

    DOHANDLE(this, obj, 
	     if (rc) {
		 rc = obj->Set_rel_position(x, y, s, FALSE);
	     }
    )

    return(rc);
}

/*
 *----------------------------------------------------------------------
 *
 * Pad_ObjHandle::Damage --
 *
 *   This procedure schedules the bounding box of each object 
 *   specified by the handle to be rendered.  The rendering doesn't
 *   actually occur until the next time the system returns to 
 *   the main loop.
 *
 * Results:
 *   None
 *
 *----------------------------------------------------------------------
 */

void
Pad_ObjHandle::Damage(void)
{
    Pad_Bool prevDamageEnabled;
    Pad_Object *obj;

    DOHANDLE(this, obj, 
				// Normally, damage is disabled within renders
				// Force damage to occur within renders if called here
	prevDamageEnabled = obj->pad->view->win->damageEnabled;
	obj->pad->view->win->damageEnabled = TRUE;
	obj->Damage();
	obj->pad->view->win->damageEnabled = prevDamageEnabled;
    )
}

/*
 *----------------------------------------------------------------------
 *
 * Pad_ObjHandle::Scale --
 *
 * Scale each object specified by the handle around the point 
 * (x, y) by <zmult>
 *
 * Results:
 *   None
 *
 *----------------------------------------------------------------------
 */

Pad_Bool
Pad_ObjHandle::Scale(float x, float y, float zmult)
{
    Pad_Object *obj;
    Pad_Bool rc = TRUE;

    DOHANDLE(this, obj, 
	     if (rc) {
		 rc = obj->Scale(x, y, zmult, TRUE);
	     }
    )

    return(rc);
}

/*
 *----------------------------------------------------------------------
 *
 * Pad_ObjHandle::Slide --
 *
 * Slide each object specified by the handle in the plane by (dx, dy)
 *
 * Results:
 *   None
 *
 *----------------------------------------------------------------------
 */

Pad_Bool
Pad_ObjHandle::Slide(float dx, float dy)
{
    Pad_Object *obj;
    Pad_Bool rc = TRUE;

    DOHANDLE(this, obj, 
	     if (rc) {
		 rc = obj->Slide(dx, dy, TRUE);
	     }
    )

    return(rc);
}

/*
 *----------------------------------------------------------------------
 *
 * Pad_ObjHandle::Rotate --
 *
 * Rotate each object specified by this handle <deg> degrees.
 * If a center point is specified, then rotate each object
 * about that point.  Otherwise, rotate each object about
 * their anchors.
 *
 * Results:
 *   None
 *
 *----------------------------------------------------------------------
 */

Pad_Bool
Pad_ObjHandle::Rotate(float theta)
{
    Pad_Object *obj;
    Pad_Bool rc = TRUE;

				// If rotating everything, then don't want to 
				// modify group members since they will get
				// rotated by their groups
    if (_tag && (!strcmp(_tag->Get(), "all"))) {
	DOHANDLENOGROUPS(this, obj, 
	     if (rc) {
		 rc = obj->Rotate(theta);
	     }
	)
    } else {
	DOHANDLE(this, obj, 
	     if (rc) {
		 rc = obj->Rotate(theta);
	     }
	)
    }

    return(rc);
}

Pad_Bool
Pad_ObjHandle::Rotate(float theta, Pad_Point &center)
{
    Pad_Object *obj;
    Pad_Bool rc = TRUE;

				// If rotating everything, then don't want to 
				// modify group members since they will get
				// rotated by their groups
    if (_tag && (!strcmp(_tag->Get(), "all"))) {
        DOHANDLENOGROUPS(this, obj, 
	     if (rc) {
		 rc = obj->Rotate(theta, center);
	     }
        )
    } else {
        DOHANDLE(this, obj, 
	     if (rc) {
		 rc = obj->Rotate(theta, center);
	     }
        )
    }

    return(rc);
}

/*
 *----------------------------------------------------------------------
 *
 * Pad_ObjHandle::Lower --
 *
 * Lower all objects specified by the handle to the bottom of the
 * drawing order.  Keep the relative order of the objects the same.
 * If <ref> is specified, then lower objects just below lowest
 * object within <ref> instead of lowering all the way to the bottom.
 * If objects to be lowered are group members, then only lower 
 * them within their group.
 *
 * Results:
 *   None
 *
 *----------------------------------------------------------------------
 */

Pad_Bool
Pad_ObjHandle::Lower(void)
{
    Pad_Object *obj;
    Pad_Bool rc = TRUE;

    obj = _Get_object();
    if (obj) {
	rc = obj->Lower();
    } else {
	Pad_List objs;
	Pad_Iterator oi;

	DOHANDLE(this, obj, 
		 objs.Push(obj);
	     )
	DOLIST(oi, objs, Pad_Object, obj) {
	    rc = obj->Lower();
	    if (!rc) {
		break;
	    }
	}    
    }

    return(rc);
}

Pad_Bool
Pad_ObjHandle::Lower(Pad_ObjHandle &ref)
{
    Pad_Object *obj, *refObj;
    Pad_Bool rc = TRUE;

				// First, get reference
    refObj = ref._Get_first_object();
				// Now, lower objects below reference
    obj = _Get_object();
    if (obj) {
	rc = obj->Lower(refObj);
    } else {
	Pad_List objs;
	Pad_Iterator oi;

	DOHANDLE(this, obj, 
		 objs.Push(obj);
	     )
	DOLIST(oi, objs, Pad_Object, obj) {
	    rc = obj->Lower(refObj);
	    if (!rc) {
		break;
	    }
	}    
    }

    return(rc);
}

/*
 *----------------------------------------------------------------------
 *
 * Pad_ObjHandle::Raise --
 *
 * Raise all objects specified by the handle to the top of the
 * drawing order.  Keep the relative order of the objects the same.
 * If <ref> is specified, then raise objects just above the highest
 * object within <ref> instead of raising all the way to the top.
 * If objects to be raised are group members, then only raise 
 * them within their group.
 *
 * Results:
 *   None
 *
 *----------------------------------------------------------------------
 */

Pad_Bool
Pad_ObjHandle::Raise(void)
{
    Pad_Object *obj;
    Pad_Bool rc = TRUE;

    DOHANDLE(this, obj, 
	     if (rc) {
		 rc = obj->Raise();
	     }
	 )

    return(rc);
}

Pad_Bool
Pad_ObjHandle::Raise(Pad_ObjHandle &ref)
{
    Pad_Object *obj, *refObj;
    Pad_Bool rc = TRUE;

				// First, get reference
    refObj = ref._Get_first_object();
				// Now, raise objects above reference

    DOHANDLE(this, obj, 
	     if (rc) {
		 rc = obj->Raise(refObj);
	     }
	 )

    return(rc);
}

/*
 *----------------------------------------------------------------------
 *
 * Pad_ObjHandle::Add_tag --
 *
 * Add <tag> to each object specified by this handle.
 * If an object already has <tag>, then nothing is done to that object.
 *
 * Results:
 *   None
 *
 *----------------------------------------------------------------------
 */

void
Pad_ObjHandle::Add_tag(char *tag)
{
    Pad_Object *obj;

    DOHANDLE(this, obj, 
	     obj->Add_tag(tag);
    )
}

/*
 *----------------------------------------------------------------------
 *
 * Pad_ObjHandle::Delete_tag --
 *
 * Delete <tag> from each object specified by this handle.
 * If an object doesn't have <tag>, then nothing is done to that object.
 *
 * Results:
 *   None
 *
 *----------------------------------------------------------------------
 */

void
Pad_ObjHandle::Delete_tag(char *tag)
{
    Pad_Object *obj;

    DOHANDLE(this, obj, 
	     obj->Delete_tag(tag);
    )
}

/*
 *----------------------------------------------------------------------
 *
 * Pad_ObjHandle::Delete_obj --
 *
 *   Deletes the object associated with this handle,
 *   or does nothing if there is no object.
 *
 * Results:
 *   None
 *
 *----------------------------------------------------------------------
 */

void
Pad_ObjHandle::Delete_obj(void)
{
    Pad_Object *obj;

    DOHANDLESAFE(this, obj, 
				// Object destructor will set handle
				// pointer to object to NULL, so it is
				// not necessary to do that here.
	delete obj;
    )
}

/*
 *----------------------------------------------------------------------
 *
 * Pad_ObjHandle::Set_option --
 *
 *   This procedure sets an option of one or more objects on the Pad++ surface
 *   referenced by this handle.  <optionType> specifies which option is being set.  
 *   The type of the value that is passed in depends on the option being set.
 *   Not all options are valid for all object types.  See the reference
 *   documentation for a complete description of all options and option
 *   types.  
 *
 *   If an error ocurrs when setting an option, the error is recorded, and
 *   the option is still set on any other objects referenced by this handle.
 *   If multiple errors ocurr, only the last one is returned.
 * 
 *   Briefly, the legal options are:
 *
 *   For all object types:
 *     Pad_AlwaysRenderOption, value = (Pad_Bool)
 *     Pad_AnchorOption,       value = (Pad_Anchor)
 *     Pad_AngleOption,        value = (double)
 *     Pad_AngleCtrOption,     value = (double, double, double)
 *     Pad_EventsOption,       value = (Pad_Bool)
 *     Pad_FadeRangeOption,    value = (double)
 *     Pad_HeightOption,       value = (double)
 *     Pad_InfoOption,         value = (NULL terminated string)
 *     Pad_LayerOption,        value = (NULL terminated string)
 *     Pad_LockOption,         value = (Pad_Bool)
 *     Pad_MaxSizeAbsOption,   value = (double)
 *     Pad_MaxSizeRelOption,   value = (double)
 *     Pad_MinSizeAbsOption,   value = (double)
 *     Pad_MinSizeRelOption,   value = (double)
 *     Pad_Pposition,          value = (double, double, double)
 *     Pad_RPosition,          value = (double, double, double)
 *     Pad_StickyOption,       value = (int)
 *     Pad_TransparencyOption, value = (double)
 *     Pad_WidthOption,        value = (double)
 *
 *   For specific object types:
 *     Pad_FillOption,         value = (int, int, int)
 *     Pad_PenOption,          value = (int, int, int)
 *     Pad_PenWidthOption,     value = (double)
 *     Pad_TextOption,         value = (NULL terminated string)
 *     Pad_ViewOption,         value = (double, double, double)
 *
 * Results:
 *   TRUE if option is set succesfully, or
 *   FALSE if there was an error in which case the error will
 *   be stored in Pad_errorString.
 *
 *----------------------------------------------------------------------
 */

static Pad_Bool
_Set_anchor_option(Pad_Object *obj, Pad_OptionType optionType, Pad_Anchor anchorVal)
{
    Pad_Bool rc = TRUE;

    switch(optionType) {
      case Pad_AnchorOption:
	rc = obj->Set_anchor((Tk_Anchor)anchorVal);
	break;
      default:
	break;
    }

    return(rc);
}

static Pad_Bool
_Set_bool_option(Pad_Object *obj, Pad_OptionType optionType, Pad_Bool boolVal)
{
    Pad_Bool rc = TRUE;

    switch(optionType) {
      case Pad_AlwaysRenderOption:
	rc = obj->Set_alwaysrender(boolVal);
	break;
      case Pad_EventsOption:
	rc = obj->Set_events(boolVal);
	break;
      case Pad_LockOption:
	rc = obj->Set_lock(boolVal);
	break;
      default:
	break;
    }

    return(rc);
}

static Pad_Bool
_Set_double_option(Pad_Object *obj, Pad_OptionType optionType, double doubleVal)
{
    Pad_Bool rc = TRUE;

    switch(optionType) {
      case Pad_AngleOption:
	rc = obj->Set_angle(doubleVal);
	break;
      case Pad_FadeRangeOption:
	rc = obj->Set_faderange(doubleVal);
	break;
      case Pad_HeightOption:
	rc = obj->Set_height(doubleVal);
	break;
      case Pad_MaxSizeAbsOption:
	rc = obj->Set_maxsize_abs(doubleVal);
	break;
      case Pad_MaxSizeRelOption:
	rc = obj->Set_maxsize_rel(doubleVal);
	break;
      case Pad_MinSizeAbsOption:
	rc = obj->Set_minsize_abs(doubleVal);
	break;
      case Pad_MinSizeRelOption:
	rc = obj->Set_minsize_rel(doubleVal);
	break;
      case Pad_PenWidthOption:
	rc = obj->Set_penwidth(doubleVal);
	break;
      case Pad_TransparencyOption:
	rc = obj->Set_transparency(doubleVal);
	break;
      case Pad_WidthOption:
	rc = obj->Set_width(doubleVal);
	break;
      default:
	break;
    }

    return(rc);
}

static Pad_Bool
_Set_int_option(Pad_Object *obj, Pad_OptionType optionType, int intVal)
{
    Pad_Bool rc = TRUE;

    switch(optionType) {
      case Pad_StickyOption:
	rc = obj->Set_sticky(intVal);
	break;
      default:
	break;
    }

    return(rc);
}

static Pad_Bool
_Set_charptr_option(Pad_Object *obj, Pad_OptionType optionType, char *charPtr)
{
    Pad_Bool rc = TRUE;

    switch(optionType) {
      case Pad_InfoOption:
	rc = obj->Set_info(charPtr);
	break;
      case Pad_LayerOption:
	rc = obj->Set_layer(charPtr);
	break;
      case Pad_TextOption:
	rc = obj->Set_text(charPtr);
	break;
      default:
	break;
    }

    return(rc);
}

static Pad_Bool
_Set_triple_double_option(Pad_Object *obj, Pad_OptionType optionType, double doubleVal1, double doubleVal2, double doubleVal3)
{
    Pad_Bool rc = TRUE;

    switch(optionType) {
      case Pad_ViewOption:
	if (obj->Type() == PAD_PAD) {
	    obj = ((Pad *)obj)->view;
	}
	rc = obj->Set_view(doubleVal1, doubleVal2, doubleVal3);
	break;
      case Pad_PositionOption:
	rc = obj->Set_abs_position(doubleVal1, doubleVal2, doubleVal3);
	break;
      case Pad_RPositionOption:
	rc = obj->Set_rel_position(doubleVal1, doubleVal2, doubleVal3);
	break;
      case Pad_AngleCtrOption:
	{
	    Pad_Point ctr(doubleVal2, doubleVal3);
	    rc = obj->Set_angle(doubleVal1, &ctr);
	}
	break;
      default:
	break;
    }

    return(rc);
}

static Pad_Bool
_Set_triple_int_option(Pad_Object *obj, Pad_OptionType optionType, int intVal1, int intVal2, int intVal3)
{
    Pad_Bool rc = TRUE;

    switch(optionType) {
      case Pad_PenOption:
	rc = obj->Set_pen(intVal1, intVal2, intVal3);
	break;
      case Pad_FillOption:
	rc = obj->Set_fill(intVal1, intVal2, intVal3);
	break;
      default:
	break;
    }

    return(rc);
}

Pad_Bool
Pad_ObjHandle::_Set_option(Pad_OptionType optionType, va_list &argPtr)
{
    Pad_Object *obj;
    Pad_Bool rc = TRUE;
    char *charPtr;
    Pad_Bool boolVal;
    Pad_Anchor anchorVal;
    double doubleVal, doubleVal1, doubleVal2, doubleVal3;
    int intVal, intVal1, intVal2, intVal3;

    enum _OptionArgType {NoArg, AnchorArg, BoolArg, CharArg, DoubleArg, IntArg, 
			 TripleDoubleArg, TripleIntArg} argType;

				// First, parse the parameters, and convert
				// it to the right type.
    switch(optionType) {
      case Pad_AnchorOption:
	argType = AnchorArg;
	anchorVal = (Pad_Anchor)va_arg(argPtr, int);
	break;
      case Pad_AlwaysRenderOption:
      case Pad_EventsOption:
      case Pad_LockOption:
	argType = BoolArg;
	boolVal = (Pad_Bool)va_arg(argPtr, int);
	break;
      case Pad_InfoOption:
      case Pad_LayerOption:
      case Pad_TextOption:
	argType = CharArg;
	charPtr = va_arg(argPtr, char *);
	break;
      case Pad_AngleOption:
      case Pad_FadeRangeOption:
      case Pad_HeightOption:
      case Pad_MaxSizeAbsOption:
      case Pad_MaxSizeRelOption:
      case Pad_MinSizeAbsOption:
      case Pad_MinSizeRelOption:
      case Pad_PenWidthOption:
      case Pad_TransparencyOption:
      case Pad_WidthOption:
	argType = DoubleArg;
	doubleVal = va_arg(argPtr, double);
	break;

      case Pad_StickyOption:
	argType = IntArg;
	intVal = va_arg(argPtr, int);
	break;

      case Pad_ViewOption:
      case Pad_PositionOption:
      case Pad_RPositionOption:
      case Pad_AngleCtrOption:
	argType = TripleDoubleArg;
	doubleVal1 = va_arg(argPtr, double);
	doubleVal2 = va_arg(argPtr, double);
	doubleVal3 = va_arg(argPtr, double);
	break;

      case Pad_PenOption:
      case Pad_FillOption:
	argType = TripleIntArg;
	intVal1 = va_arg(argPtr, int);
	intVal2 = va_arg(argPtr, int);
	intVal3 = va_arg(argPtr, int);
	break;

      case Pad_NoiseDataOption:
      case Pad_RenderScriptOption:
      case Pad_TagsOption:
      case Pad_TimerScriptOption:
      case Pad_TimerRateOption:
      case Pad_ViewScriptOption:
      case Pad_ZoomActionOption:
      case Pad_ArrowOption:
      case Pad_ArrowshapeOption:
      case Pad_BBoxOption:
      case Pad_BorderOption:
      case Pad_BorderWidthOption:
      case Pad_CapStyleOption:
      case Pad_CommandOption:
      case Pad_DitherOption:
      case Pad_DivisibleOption:
      case Pad_DoneScriptOption:
      case Pad_EchoCharOption:
      case Pad_EditableOption:
      case Pad_ErrorScriptOption:
      case Pad_FileOption:
      case Pad_FontOption:
      case Pad_FromOption:
      case Pad_HTMLAnchorsOption:
      case Pad_ImageDataOption:
      case Pad_IsMapOption:
      case Pad_JoinStyleOption:
      case Pad_LineSizeOption:
      case Pad_LookOnOption:
      case Pad_MembersOption:
      case Pad_OrientationOption:
      case Pad_PageSizeOption:
      case Pad_ReferenceOption:
      case Pad_ReliefOption:
      case Pad_SizeOption:
      case Pad_StateOption:
      case Pad_TitleOption:
      case Pad_ToOption:
      case Pad_UpdateScriptOption:
      case Pad_URLOption:
      case Pad_ValueOption:
      case Pad_VisibleLayersOption:
      case Pad_WriteFormatOption:
	Pad_errorString = "Not implemented yet";
	rc = FALSE;
	break;

      default:
	Pad_errorString = "Invalid optionType";
	rc = FALSE;
	break;
    }

				// Now, for each object, call the function
				// to set that option.  If there is an error, report
				// it, but keep going through the objects.
    DOHANDLE(this, obj, 
	 switch (argType) {
	   case AnchorArg:
	     rc &= _Set_anchor_option(obj, optionType, anchorVal);
	     break;
	   case BoolArg:
	     rc &= _Set_bool_option(obj, optionType, boolVal);
	     break;
	   case CharArg:
	     rc &= _Set_charptr_option(obj, optionType, charPtr);
	     break;
	   case DoubleArg:
	     rc &= _Set_double_option(obj, optionType, doubleVal);
	     break;
	   case IntArg:
	     rc &= _Set_int_option(obj, optionType, intVal);
	     break;
	   case TripleDoubleArg:
	     rc &= _Set_triple_double_option(obj, optionType, doubleVal1, doubleVal2, doubleVal3);
	     break;
	   case TripleIntArg:
	     rc &= _Set_triple_int_option(obj, optionType, intVal1, intVal2, intVal3);
	     break;
	   default:
	     break;
	 }
         if (rc) {
	     obj->Generate_event(Pad_ModifyNotify, NULL, "itemconfigure");	     
	 }
    )

    return (rc);
}

Pad_Bool
Pad_ObjHandle::Set_option(Pad_OptionType optionType, ...)
{
    Pad_Bool rc;
    va_list argPtr;

    va_start(argPtr, optionType);
    rc = _Set_option(optionType, argPtr);
    va_end(argPtr);

    return(rc);
}

/*
 *----------------------------------------------------------------------
 *
 * Pad_ObjHandle::Set_options --
 *
 *   This procedure sets any number of options by parsing the parameters
 *   as option,value pairs.  Set_option is called with each pair until
 *   the Pad_NoOption type is seen.
 *
 *   If an error ocurrs when setting an option, the error is recorded, and
 *   no further options are set.
 *
 * Results:
 *   TRUE if all options were set succesfully, or
 *   FALSE if there was an error in which case the error will
 *   be stored in Pad_errorString.
 *
 *----------------------------------------------------------------------
 */

Pad_Bool
Pad_ObjHandle::Set_options(Pad_OptionType optionType, ...)
{
    va_list argPtr;
    Pad_Bool rc = TRUE;
    Pad_Bool done = FALSE;

    va_start(argPtr, optionType);

    do {
	switch(optionType) {
	case Pad_NoOption:
	    done = TRUE;
	    break;
	default:
	    rc = _Set_option(optionType, argPtr);
	    if (!rc) {
		done = TRUE;
		break;
	    }
	    optionType = (Pad_OptionType)va_arg(argPtr, int);
	}
    } while (!done);

    va_end(argPtr);
    return(rc);
}

/*
 *----------------------------------------------------------------------
 *
 * Pad_ObjHandle::Set_render_callback --
 *
 *   Set a callback that gets executed every time the object gets
 *   rendered (in place of the actual render).
 *
 *   Note that this callback is assigned to the object at the
 *   time of this call, so if this handle refers to multiple objects
 *   via a tag, then all of those objects independently will use
 *   this callback.
 * 
 *   If <callback> is NULL, then the callback is removed from
 *   all objects specified by this handle.
 *
 * Results:
 *   None
 *
 *----------------------------------------------------------------------
 */
void
Pad_ObjHandle::Set_render_callback(Pad_Callback *callback)
{
    Pad_Object *obj;

    DOHANDLE(this, obj, 
	     obj->Set_renderscript(callback);
    )
}

/*
 *----------------------------------------------------------------------
 *
 * Pad_ObjHandle::Bind --
 *
 *   Creates an event binding so that whenever an event of type <event>
 *   occurs on <object>, <callback> is executed.  <event> is a standard
 *   Tcl-style description of an event, such as "<ButtonPress-1>".
 *   <object> is a string that can specify a tag, the special tag "all",
 *   or the string representation of an object id (i.e., "347").
 *   <callback> is a Pad++ callback which encapsulates a callback that
 *   can be in any supported language. 
 *
 * Results:
 *   sets Pad_errorString on Failure.
 *   Return's FALSE on failure, and TRUE on success. If the handle represents
 *   a tag, the return value is the success value for the last object in
 *   the list.
 *
 *----------------------------------------------------------------------
 */
Pad_Bool
Pad_ObjHandle::Bind(char *event, Pad_Callback *callback)
{
    ClientData objectPtr = NULL;
    Pad_Bool rc = TRUE;
    Pad_Win *win;

    if (_tag) {
        objectPtr = Tk_GetUid((char *)_tag->Get());
        win = ((Pad*) _obj)->Get_win();
    } else if (_obj) {
        objectPtr = _obj;
        win = ((Pad_Object*) objectPtr)->pad->Get_win();
    }
    
    if (objectPtr) {
        rc = Pad_CreateBinding(win->interp,
                               win->bindingTable,
                               objectPtr,
                               event,
                               callback);
    }
    if (!rc) {
        Pad_errorString = win->interp->result;
        Tcl_ResetResult (win->interp);
    }
    return rc;
}


/*
 *----------------------------------------------------------------------
 *
 * Pad_ObjHandle::Get_binding --
 *
 *   retrieves the Pad++ binding associated with <object> and <event>
 *
 * Results:
 *   Sets Pad_resultString on success.
 *   Returns a Pad_Bool to indicating success value.
 *
 *----------------------------------------------------------------------
 */
Pad_Bool
Pad_ObjHandle::Get_binding(char *event) const
{
    ClientData objectPtr = NULL;
    Pad_Win *win;
    
    if (_tag) {
        objectPtr = Tk_GetUid((char *)_tag->Get());
        win = ((Pad*) _obj)->Get_win();
    } else if (_obj) {
        objectPtr = _obj;
        win = ((Pad_Object*) objectPtr)->pad->Get_win();
    }
    
    if (objectPtr) {
        Pad_resultString = Pad_GetBinding(win->interp,
                                          win->bindingTable,
                                          objectPtr,
                                          event);
    }
    
    if (Pad_resultString == "") {
        Pad_errorString = win->interp->result;
        Tcl_ResetResult (win->interp);
        return FALSE;
    } else {
        return TRUE;
    }
}

/*
 *----------------------------------------------------------------------
 *
 * Pad_ObjHandle::Get_all_bindings --
 *
 *   retrieves all Pad++ bindings associated with <object> 
 *
 * Results:
 *   Sets Pad_resultString on success.
 *   returns Pad_bool indicating success value.
 *
 *----------------------------------------------------------------------
 */
Pad_Bool
Pad_ObjHandle::Get_all_bindings(void) const
{
    ClientData objectPtr = NULL;
    Pad_Win *win;
    
    if (_tag) {
        objectPtr = Tk_GetUid((char *)_tag->Get());
        win = ((Pad*) _obj)->Get_win();
    } else if (_obj) {
        objectPtr = _obj;
        win = ((Pad_Object*) objectPtr)->pad->Get_win();
    }
    
    if (objectPtr) {
        Pad_GetAllBindings(win->interp,
                           win->bindingTable,
                           objectPtr);
    }
    
    Pad_resultString = win->interp->result;
    Tcl_ResetResult (win->interp); // So Tcl doesn't echo the result
    if (Pad_resultString == "") {
        return FALSE;
    } else {
        return TRUE;
    }
}

/*
 *----------------------------------------------------------------------
 *
 * Pad_ObjHandle::Delete_binding --
 *
 *   Deletes event bindings.  One version deletes the event
 *   binding for the <event> associated with <object>.
 *   The other version deletes all events associated with <object>
 *
 * Results:
 *   Sets Pad_errorString on error.
 *   returns Pad_Bool indicating success value.
 *
 *----------------------------------------------------------------------
 */
Pad_Bool
Pad_ObjHandle::Delete_binding(char *event)
{
    ClientData objectPtr = NULL;
    Pad_Win *win;
    Pad_Bool rc;
    
    if (_tag) {
        objectPtr = Tk_GetUid((char *)_tag->Get());
        win = ((Pad*) _obj)->Get_win();
    } else if (_obj) {
        objectPtr = _obj;
        win = ((Pad_Object*) objectPtr)->pad->Get_win();
    }
    
    if (objectPtr) {
        rc = Pad_DeleteBinding(win->interp,
                               win->bindingTable,
                               objectPtr,
                               event);
    }
    
    if (!rc) {
        Pad_errorString = win->interp->result;
        return FALSE;
    } else {
        return TRUE;
    }
}

/*
 *----------------------------------------------------------------------
 *
 * Pad_ObjHandle::Delete_all_bindings --
 *
 *   Deletes all event bindings for the <event> associated with <object>.
 *
 * Results:
 *   None.
 *
 *----------------------------------------------------------------------
 */
void
Pad_ObjHandle::Delete_all_bindings(void)
{
    ClientData objectPtr = NULL;
    Pad_Win *win;
    
    if (_tag) {
        objectPtr = Tk_GetUid((char *)_tag->Get());
        win = ((Pad*) _obj)->Get_win();
    } else if (_obj) {
        objectPtr = _obj;
        win = ((Pad_Object*) objectPtr)->pad->Get_win();
    }
    
    if (objectPtr) {
        Pad_DeleteAllBindings(win->bindingTable, objectPtr);
    }
}

////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////
//
// Pad_SurfaceHandle Class definition
//
////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////

Pad_SurfaceHandle::Pad_SurfaceHandle()
{
}

Pad_SurfaceHandle::Pad_SurfaceHandle(Pad_SurfaceHandle &surfaceHandle)
{
    Attach(surfaceHandle);
}

Pad_SurfaceHandle::Pad_SurfaceHandle(char *surfaceName)
{
    Attach(surfaceName);
}

Pad_SurfaceHandle::Pad_SurfaceHandle(Pad_WinHandle &winHandle)
{
    Attach(winHandle);
}

Pad_SurfaceHandle::~Pad_SurfaceHandle(void)
{
    Attach((Pad *)NULL);
}

Pad_SurfaceHandle & 
Pad_SurfaceHandle::operator=(Pad_SurfaceHandle &surfaceHandle)
{
    Attach(surfaceHandle);

    return (*this);
}

/*
 *----------------------------------------------------------------------
 *
 * Pad_SurfaceHandle::_Get_pad --
 *
 *   PRIVATE:
 *
 *   Return Pad object specified by handle, or NULL if none
 *
 * Results:
 *   Pad pointer
 *
 *----------------------------------------------------------------------
 */
Pad *
Pad_SurfaceHandle::_Get_pad(void)
{
    return((Pad *)_obj);
}

/*
 *----------------------------------------------------------------------
 *
 * Pad_SurfaceHandle::Is_mapped --
 *
 *   Determines if this handle is mapped to a Pad++ Surface.
 *
 * Results:
 *   TRUE if handle is mapped to a Surface.
 *
 *----------------------------------------------------------------------
 */
Pad_Bool
Pad_SurfaceHandle::Is_mapped(void)
{
    return (_obj ? TRUE : FALSE);
}

/*
 *----------------------------------------------------------------------
 *
 * Pad_SurfaceHandle::Get_handle_type --
 *
 *   Returns the type of this particular handle.
 *
 * Results:
 *   Type of this handle
 *
 *----------------------------------------------------------------------
 */
Pad_HandleType
Pad_SurfaceHandle::Get_handle_type(void) const
{
    return(Pad_SurfaceHandleType);
}

/*
 *----------------------------------------------------------------------
 *
 * Pad_SurfaceHandle::Get_option_type --
 *
 *   Given an option string, return the Pad_OptionType
 *
 * Results:
 *   Option type, or Pad_NoOption if invalid option, or surface handle not mapped.
 *
 *----------------------------------------------------------------------
 */
Pad_OptionType
Pad_SurfaceHandle::Get_option_type(char *optionName)
{
    Pad *pad = _Get_pad();
    if (pad == NULL) {
	return(Pad_NoOption);
    } else {
	int type = int(pad->optionIdTable->Get(optionName));
	return Pad_OptionType(type);
    }

}

/*
 *----------------------------------------------------------------------
 *
 * Pad_SurfaceHandle::Get_name --
 *
 *   returns the name of the pad
 *
 * Results:
 *   name of the pad
 *
 *----------------------------------------------------------------------
 */
char *
Pad_SurfaceHandle::Get_name(void)
{
    Pad *pad = _Get_pad();
    if (pad == NULL) {
	return(NULL);
    } else {
	return(pad->Get_win()->Get_name());
    }
}

/*
 *----------------------------------------------------------------------
 *
 * Pad_SurfaceHandle::Attach --
 *
 *   Specifies the Pad++ surface that this handle references
 *
 * Results:
 *   TRUE if successful, or FALSE otherwise
 *
 *----------------------------------------------------------------------
 */
Pad_Bool
Pad_SurfaceHandle::Attach(char *surfaceName)
{
    Pad *pad;

    if (surfaceName) {
	pad = Pad_Win::Get_pad_from_name(Tk_GetUid(surfaceName));
    } else {
	pad = NULL;
    }
    Attach(pad);

    return(pad ? TRUE : FALSE);
}

Pad_Bool
Pad_SurfaceHandle::Attach(Pad_SurfaceHandle &surfaceHandle)
{
    return(Attach(surfaceHandle._Get_pad()));
}

Pad_Bool
Pad_SurfaceHandle::Attach(Pad_WinHandle &winHandle)
{
    return(Attach(winHandle.Get_name()));
}

Pad_Bool
Pad_SurfaceHandle::Attach(Pad *pad)
{
				// First, dereference any pads this references
    if (_obj) {
	Pad_Handle *prev, *cur;

	prev = NULL;
	cur = _obj->_handle;
	while (cur) {
	    if (cur == this) {
		if (prev == NULL) {
		    _obj->_handle = _next;
		} else {
		    prev->_next = _next;
		}
		break;
	    }
	    prev = cur;
	    cur = cur->_next;
	}
	_obj = NULL;
	_next = NULL;
    }
				// Now, set to new pad
    if (pad) {
	_obj = pad;
	_next = pad->_handle;
	pad->_handle = this;
    }

    return(pad ? TRUE : FALSE);
}

/*
 *----------------------------------------------------------------------
 *
 * Pad_SurfaceHandle::Damage --
 *
 *   This procedure schedules the Surface to be re-rendered.
 *   The rendering doesn't actually occur until the next time the 
 *   system returns to the main loop.
 *
 * Results:
 *   None
 *
 *----------------------------------------------------------------------
 */

void
Pad_SurfaceHandle::Damage(void)
{
    Pad *pad = (Pad *)_obj;

    if (pad) {
	pad->view->Damage();
    }
}

/*
 *----------------------------------------------------------------------
 *
 * Pad_SurfaceHandle::Create_obj --
 *
 *   This creates an object of the specified type with all default 
 *   properties.  There are three versions of this method.  The first
 *   does not return any reference to the object.  The second takes
 *   an empty handle as a parameter and fills it in with a reference
 *   to the created object.  The third fills in a handle, *and* takes
 *   an arbitrary list of option-value pairs to set options on the
 *   newly created object.  The list is terminated by the Pad_NoOption option.
 *
 * Results:
 *   TRUE if successful, or FALSE otherwise
 *
 * Side Effects;
 *   A Pad++ object is created.
 *
 *----------------------------------------------------------------------
 */

Pad_Bool
Pad_SurfaceHandle::Create_obj(Pad_ObjectType type)
{
    Pad_Bool rc = FALSE;
    Pad *pad = (Pad *)_obj;
    Pad_Object *obj;

    if (pad) {
	obj = pad->Create_object(type);
	if (obj) {
	    rc = TRUE;
	}
    }

    return(rc);
}

Pad_Bool
Pad_SurfaceHandle::Create_obj(Pad_ObjectType type, Pad_ObjHandle &handle)
{
    Pad_Bool rc = FALSE;
    Pad *pad = (Pad *)_obj;
    Pad_Object *obj;

    if (pad) {
	obj = pad->Create_object(type);
	handle.Attach(obj);
	if (obj) {
	    rc = TRUE;
	}
    }

    return(rc);
}

Pad_Bool
Pad_SurfaceHandle::Create_obj(Pad_ObjectType type, Pad_ObjHandle &handle, Pad_OptionType optionType, ...)
{
    Pad_Bool rc = FALSE;
    Pad *pad = (Pad *)_obj;
    Pad_Object *obj;

    if (pad) {
	obj = pad->Create_object(type);
	handle.Attach(obj);
	if (obj) {
	    va_list argPtr;
	    Pad_Bool done = FALSE;
	    
	    rc = TRUE;
	    va_start(argPtr, optionType);
	    
	    do {
		switch(optionType) {
		case Pad_NoOption:
		    done = TRUE;
		    break;
		default:
		    rc = handle._Set_option(optionType, argPtr);
		    if (!rc) {
			done = TRUE;
			break;
		    }
		    optionType = (Pad_OptionType)va_arg(argPtr, int);
		}
	    } while (!done);
	    
	    va_end(argPtr);
	}
    }

    return(rc);
}

/*
 *----------------------------------------------------------------------
 *
 * Pad_SurfaceHandle::Create_anim_path --
 *
 *   This creates a path with all default properties.  
 *   The first method creates a path and adds it to the hashtable of paths.
 *   The second method takes an empty handle as a parameter and fills 
 *   it in with a reference to the created path.  
 *
 * Results:
 *   TRUE if successful, or FALSE otherwise
 *
 * Side Effects;
 *   A Pad_AnimPath is created.
 *
 *----------------------------------------------------------------------
 */
Pad_Bool
Pad_SurfaceHandle::Create_anim_path()
{
    Pad_Bool rc=FALSE;
    Pad_AnimPath *animpath;
    Pad_String tokenName;

    animpath = new Pad_AnimPath();
    if (animpath) {
	sprintf(Pad_result, "path%d", Pad_AnimPath::tokenIndex++);
	tokenName = Pad_result;
	animpath->Set_token(tokenName);
	Pad_AnimPath::tokenTable.Set(Pad_result, animpath);
	rc = TRUE;
    }

    return(rc);
}

Pad_Bool
Pad_SurfaceHandle::Create_anim_path(Pad_AnimPathHandle &pathHandle)
{
    Pad_Bool rc=FALSE;
    Pad_AnimPath *animpath;
    Pad_String tokenName;

    animpath = new Pad_AnimPath();
    if (animpath) {
	sprintf(Pad_result, "path%d", Pad_AnimPath::tokenIndex++);
	tokenName = Pad_result;
	animpath->Set_token(tokenName);
	Pad_AnimPath::tokenTable.Set(tokenName.Get(), animpath);
	rc = TRUE;
    }

    pathHandle.Attach(animpath);
    return(rc);
}

/*
 *----------------------------------------------------------------------
 *
 * Pad_SurfaceHandle::Create_anim_channel --
 *
 *   This creates a channel with all default properties.  
 *   One method just creates a Pad_AnimChannel.
 *   The other method takes an empty channel handle as a parameter 
 *   and fills it in with a reference to the created channel.  
 *
 * Results:
 *   TRUE if successful, or FALSE otherwise
 *
 * Side Effects;
 *   A Pad_AnimChannel is created.
 *
 *----------------------------------------------------------------------
 */
class Pad_AnimChannel;

Pad_Bool
Pad_SurfaceHandle::Create_anim_channel()
{
    Pad_Bool         rc=FALSE;
    Pad_AnimChannel *animchannel;

    animchannel = new Pad_AnimChannel;
    if (animchannel) {
	sprintf(Pad_result, "chan%d", Pad_Animatable::tokenIndex++);
	animchannel->Set_type(Pad_Animatable::CHAN);
	animchannel->Set_token(Pad_result);
	Pad_Animatable::tokenTable.Set(Pad_result, animchannel);
	rc = TRUE;
    } 

    return(rc);
}

Pad_Bool
Pad_SurfaceHandle::Create_anim_channel(Pad_AnimChannelHandle &chanHandle)
{
    Pad_Bool         rc=FALSE;
    Pad_AnimChannel *animchannel;

    animchannel = new Pad_AnimChannel;
    if (animchannel) {
	sprintf(Pad_result, "chan%d", Pad_Animatable::tokenIndex++);
	animchannel->Set_type(Pad_Animatable::CHAN);
	animchannel->Set_token(Pad_result);
	Pad_Animatable::tokenTable.Set(Pad_result, animchannel);
	rc = TRUE;
    } 

    chanHandle.Attach(animchannel);
    return(rc);
}

/*
 *----------------------------------------------------------------------
 *
 * Pad_SurfaceHandle::Create_animation --
 *
 *   This creates an animation with all default properties.  
 *   One method just creates a Pad_Animation.
 *   The other method takes an empty animtaion handle as a parameter 
 *   and fills it in with a reference to the created animation.  
 *
 * Results:
 *   TRUE if successful, or FALSE otherwise
 *
 * Side Effects;
 *   A Pad_Animation is created.
 *
 *----------------------------------------------------------------------
 */
Pad_Bool
Pad_SurfaceHandle::Create_animation()
{
    Pad_Bool         rc=FALSE;
    Pad_Animation   *animation;

    animation = new Pad_Animation;
    if (animation) {
	sprintf(Pad_result, "anim%d", Pad_Animation::tokenIndex++);
	animation->Set_type(Pad_Animatable::ANIM);
	animation->Set_token(Pad_result);
	Pad_Animation::tokenTable.Set(Pad_result, animation);
	Pad_Animatable::tokenTable.Set(Pad_result, (Pad_Animatable*)animation);
	rc = TRUE;
    } 

    return(rc);
}

Pad_Bool
Pad_SurfaceHandle::Create_animation(Pad_AnimationHandle &animHandle)
{
    Pad_Bool         rc=FALSE;
    Pad_Animation   *animation;

    animation = new Pad_Animation;
    if (animation) {
	sprintf(Pad_result, "anim%d", Pad_Animation::tokenIndex++);
	sprintf(Pad_result, "anim%d", Pad_Animatable::tokenIndex++);
	animation->Set_type(Pad_Animatable::ANIM);
	animation->Set_token(Pad_result);
	// add the animation to both the animation hash table 
	// and the animatable hash table
	Pad_Animation::tokenTable.Set(Pad_result, animation);
	Pad_Animatable::tokenTable.Set(Pad_result, (Pad_Animatable*)animation);
	rc = TRUE;
    } 

    animHandle.Attach(animation);
    return(rc);
}

/*
 *----------------------------------------------------------------------
 *
 * Pad_SurfaceHandle::Find_with_tagorid --
 *
 *   Finds all the objects that contain the specified <tagOrId>.
 *   If <groupMembers> is FALSE, then only top-level objects and not
 *   group members are found.
 *
 * Results:
 *   Fills in the list of handles.
 *
 *----------------------------------------------------------------------
 */

void
Pad_SurfaceHandle::Find_with_tagorid(char *tagOrId, Pad_HandleList &handles, Pad_Bool groupMembers)
{
    Pad_ObjHandle *handle;
    Pad_List objs;
    Pad *pad = (Pad *)_obj;
    Pad_Object *obj;
    Pad_Iterator oi;

    if (pad) {
	pad->Find_with_tagorid(tagOrId, objs, groupMembers);
	DOLIST(oi, objs, Pad_Object, obj) {
	    handle = new Pad_ObjHandle(obj);
	    handles.Push_last(handle);
	}
    }
}

/*
 *----------------------------------------------------------------------
 *
 * Pad_SurfaceHandle::Add_binding_modifier --
 *
 *   Add <modifier> to the list of user-defined event modifiers available.
 *
 * Results:
 *   None
 *
 *----------------------------------------------------------------------
 */
void
Pad_SurfaceHandle::Add_binding_modifier(char *modifier)
{
    Pad_AddBindingModifier(modifier);
}

/*
 *----------------------------------------------------------------------
 *
 * Pad_SurfaceHandle::Delete_binding_modifier --
 *
 *   Delete <modifier> from the list of user-defined event modifiers.
 *
 * Results:
 *   None
 *
 *----------------------------------------------------------------------
 */
void
Pad_SurfaceHandle::Delete_binding_modifier(char *modifier)
{
    Pad_DeleteBindingModifier(modifier);
}

/*
 *----------------------------------------------------------------------
 *
 * Pad_SurfaceHandle::Get_binding_modifier --
 *
 *   Return the current user-defined event modifier.
 *
 * Results:
 *   None
 *
 *----------------------------------------------------------------------
 */
char *
Pad_SurfaceHandle::Get_binding_modifier(void)
{
    Pad *pad = (Pad *)_obj;
    char *modifier = NULL;

    if (pad) {
	modifier = pad->Get_modifier();
    }

    return(modifier);
}

/*
 *----------------------------------------------------------------------
 *
 * Pad_SurfaceHandle::Set_binding_modifier --
 *
 *   Make <modifier> the current user-defined event modifier.
 *
 * Results:
 *   TRUE if successful or FALSE otherwise (could be modifier hasn't
 *   been added).
 *
 *----------------------------------------------------------------------
 */
Pad_Bool
Pad_SurfaceHandle::Set_binding_modifier(char *modifier)
{
    Pad *pad = (Pad *)_obj;
    Pad_Bool rc = FALSE;

    if (pad) {
	rc = pad->Set_modifier(modifier);
    }

    return(rc);
}

////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////
//
// Pad_ViewHandle Class definition
//
////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////

Pad_ViewHandle::Pad_ViewHandle()
{
}

Pad_ViewHandle::Pad_ViewHandle(Pad_ViewHandle &viewHandle)
{
    Attach(viewHandle);
}

Pad_ViewHandle::Pad_ViewHandle(char *surfaceName)
{
    Attach(surfaceName);
}

Pad_ViewHandle::~Pad_ViewHandle(void)
{
    Attach((Pad_View *)NULL);
}

Pad_ViewHandle & 
Pad_ViewHandle::operator=(Pad_ViewHandle &viewHandle)
{
    Attach(viewHandle);

    return (*this);
}

/*
 *----------------------------------------------------------------------
 *
 * Pad_ViewHandle::_Get_view --
 *
 *   PRIVATE:
 *
 *   Return Pad_View object specified by handle, or NULL if none
 *
 * Results:
 *   Pad_View pointer
 *
 *----------------------------------------------------------------------
 */
Pad_View *
Pad_ViewHandle::_Get_view(void)
{
    return((Pad_View *)_data);
}

/*
 *----------------------------------------------------------------------
 *
 * Pad_ViewHandle::Get_handle_type --
 *
 *   Returns the type of this particular handle.
 *
 * Results:
 *   Type of this handle
 *
 *----------------------------------------------------------------------
 */
Pad_HandleType
Pad_ViewHandle::Get_handle_type(void) const
{
    return(Pad_ViewHandleType);
}

/*
 *----------------------------------------------------------------------
 *
 * Pad_ViewHandle::Attach --
 *
 *   Specifies the Pad++ View that this handle references
 *
 * Results:
 *   TRUE if successful, or FALSE otherwise
 *
 *----------------------------------------------------------------------
 */
Pad_Bool
Pad_ViewHandle::Attach(char *surfaceName)
{
    Pad *pad;
    Pad_View *view;

    if (surfaceName) {
	pad = Pad_Win::Get_pad_from_name(Tk_GetUid(surfaceName));
	view = pad->view;
    } else {
	pad = NULL;
	view = NULL;
    }

    Attach(view);

    return(view ? TRUE : FALSE);
}

//
// Set this view handle based on the specified object.
// This only works if the object is a portal or a surface.
// This attaches the view to the first object specified 
// by objHandle.
//
Pad_Bool
Pad_ViewHandle::Attach(Pad_ObjHandle &objHandle)
{
    Pad_Bool rc;
    Pad_Object *obj;
    Pad_ObjectType type;

    type = objHandle.Get_object_type();
    if (type == Pad_PadType) {
	obj = objHandle._Get_first_object();
	Attach(((Pad *)obj)->view);
    } else if (type == Pad_PortalType) {
	obj = objHandle._Get_first_object();
				// Because a portal is a view, we can downcast safely
	Attach((Pad_View *)obj);
    } else {
	rc = FALSE;
    }

    return(rc);
}

Pad_Bool
Pad_ViewHandle::Attach(Pad_ViewHandle &viewHandle)
{
    return(Attach(viewHandle._Get_view()));
}

Pad_Bool
Pad_ViewHandle::Attach(Pad_View *view)
{
				// First, dereference any views this references
    if (_data) {
	Pad_Handle *prev, *cur;

	prev = NULL;
	cur = ((Pad_View *)_data)->_handle;
	while (cur) {
	    if (cur == this) {
		if (prev == NULL) {
		    ((Pad_View *)_data)->_handle = _next;
		} else {
		    prev->_next = _next;
		}
		break;
	    }
	    prev = cur;
	    cur = cur->_next;
	}
	_data = NULL;
	_next = NULL;
    }
				// Now, set to new view
    if (view) {
	_data = view;
	_next = view->_handle;
	view->_handle = this;
    }

    return(view ? TRUE : FALSE);
}

/*
 *----------------------------------------------------------------------
 *
 * Pad_ViewHandle::Set_view --
 *
 *   Specifies what this view sees by setting the position.
 *   <xview> and <yview> specify the center of the view,
 *   and <zoom> specifies the magnification.
 *
 * Results:
 *   TRUE if successful
 *
 *----------------------------------------------------------------------
 */
Pad_Bool
Pad_ViewHandle::Set_view(float xview, float yview, float zoom, int animationTime, Pad_Bool twoStep)
{
    Pad_Bool rc;
    Pad_View *view = _Get_view();

    if (!view) {
	Pad_errorString = "View handle not associated with a view";
	return(FALSE);
    }

    if (animationTime == 0.0) {
	rc = view->Set_view(xview, yview, zoom);
    } else {
	rc = view->Animate_to(xview, yview, zoom, animationTime, twoStep);
    }

    return(rc);
}

/*
 *----------------------------------------------------------------------
 *
 * Pad_ViewHandle::Get_view --
 *
 *   Returns the current view coordinates.  (xview, yview) specify
 *   the point at the center of the screen, and (zoom) specifies
 *   the magnification of the view.
 *
 * Results:
 *   TRUE if successful
 *
 *----------------------------------------------------------------------
 */
Pad_Bool
Pad_ViewHandle::Get_view(float &xview, float &yview, float &zoom)
{
    Pad_View *view = _Get_view();

    if (!view) {
	Pad_errorString = "View handle not associated with a view";
	return(FALSE);
    }

    view->Get_view(xview, yview, zoom);

    return(TRUE);
}

/*
 *----------------------------------------------------------------------
 *
 * Pad_ViewHandle::Set_visiblelayers --
 *
 *   Set the layers this view sees and doesn't see.  Syntax of layers is:
 *   "layer layer ... [-]layer ..."  Only the layers specified are visible,
 *   unless the special layer "all" is specified, in which case, all layers
 *   are visible except layers preceeded with a '-'.  Layers preceeded with
 *   a '-' when "all" is not specified are ignored.
 *
 * Results:
 *   None
 *
 *----------------------------------------------------------------------
 */
void
Pad_ViewHandle::Set_visiblelayers(char *layers)
{
    Pad_View *view = _Get_view();

    if (view) {
	view->Set_visiblelayers(layers);
    }
}

////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////
//
// Pad_WinHandle Class definition
//
////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////

Pad_WinHandle::Pad_WinHandle()
{
}

Pad_WinHandle::Pad_WinHandle(Pad_WinHandle &winHandle)
{
    Attach(winHandle);
}

Pad_WinHandle::Pad_WinHandle(char *surfaceName)
{
    Attach(surfaceName);
}

Pad_WinHandle::~Pad_WinHandle(void)
{
    Attach((Pad_Win *)NULL);
}

Pad_WinHandle & 
Pad_WinHandle::operator=(Pad_WinHandle &winHandle)
{
    Attach(winHandle);

    return (*this);
}

/*
 *----------------------------------------------------------------------
 *
 * Pad_WinHandle::_Get_win --
 *
 *   PRIVATE:
 *
 *   Return Pad_Win object specified by handle, or NULL if none
 *
 * Results:
 *   Pad_Win pointer
 *
 *----------------------------------------------------------------------
 */
Pad_Win *
Pad_WinHandle::_Get_win(void)
{
    return((Pad_Win *)_data);
}

/*
 *----------------------------------------------------------------------
 *
 * Pad_WinHandle::Get_handle_type --
 *
 *   Returns the type of this particular handle.
 *
 * Results:
 *   Type of this handle
 *
 *----------------------------------------------------------------------
 */
Pad_HandleType
Pad_WinHandle::Get_handle_type(void) const
{
    return(Pad_WinHandleType);
}

/*
 *----------------------------------------------------------------------
 *
 * Pad_WinHandle::Attach --
 *
 *   Specifies the Pad++ Window that this handle references
 *
 * Results:
 *   TRUE if successful, or FALSE otherwise
 *
 *----------------------------------------------------------------------
 */
Pad_Bool
Pad_WinHandle::Attach(char *surfaceName)
{
    Pad *pad;
    Pad_Win *win;

    if (surfaceName) {
	pad = Pad_Win::Get_pad_from_name(Tk_GetUid(surfaceName));
	win = pad->view->win;
    } else {
	pad = NULL;
	win = NULL;
    }

    Attach(win);

    return(win ? TRUE : FALSE);
}

Pad_Bool
Pad_WinHandle::Attach(Pad_WinHandle &winHandle)
{
    return(Attach(winHandle._Get_win()));
}

Pad_Bool
Pad_WinHandle::Attach(Pad_Win *win)
{
				// First, dereference any windows this references
    if (_data) {
	Pad_Handle *prev, *cur;

	prev = NULL;
	cur = ((Pad_Win *)_data)->_handle;
	while (cur) {
	    if (cur == this) {
		if (prev == NULL) {
		    ((Pad_Win *)_data)->_handle = _next;
		} else {
		    prev->_next = _next;
		}
		break;
	    }
	    prev = cur;
	    cur = cur->_next;
	}
	_data = NULL;
	_next = NULL;
    }
				// Now, set to new window
    if (win) {
	_data = win;
	_next = win->_handle;
	win->_handle = this;
    }

    return(win ? TRUE : FALSE);
}

/*
 *----------------------------------------------------------------------
 *
 * Pad_WinHandle::Get_name --
 *
 *   Returns the name of the window
 *
 * Results:
 *   name of the pad
 *
 *----------------------------------------------------------------------
 */
char *
Pad_WinHandle::Get_name(void)
{
    Pad_Win *win = _Get_win();
    if (win == NULL) {
	return(NULL);
    } else {
	return(win->Get_name());
    }
}

////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////
//
// Pad_AnimPathHandle Class definition
//
////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////
Pad_AnimPathHandle::Pad_AnimPathHandle()
{
}

Pad_AnimPathHandle::Pad_AnimPathHandle(Pad_AnimPathHandle &pathHandle)
{
    Attach(pathHandle);
}

Pad_AnimPathHandle::Pad_AnimPathHandle(Pad_AnimPath *path)
{
    Attach(path);
}

Pad_AnimPathHandle::~Pad_AnimPathHandle(void)
{
    Attach((Pad_AnimPath *)NULL);
}

Pad_AnimPathHandle & 
Pad_AnimPathHandle::operator=(Pad_AnimPathHandle &pathHandle)
{
    Attach(pathHandle);

    return (*this);
}

/*
 *----------------------------------------------------------------------
 *
 * Pad_AnimPathHandle::Attach --
 *
 *   Specifies the Pad_AnimPath  that this handle references
 *
 * Results:
 *   TRUE if successful, or FALSE otherwise
 *
 *----------------------------------------------------------------------
 */
Pad_Bool
Pad_AnimPathHandle::Attach(Pad_AnimPath *path)
{
				// First, dereference any path this references
    if (_data) {
	Pad_Handle *prev, *cur;

	prev = NULL;
	cur = ((Pad_AnimPath *)_data)->_handle;
	while (cur) {
	    if (cur == this) {
		if (prev == NULL) {
		    ((Pad_AnimPath *)_data)->_handle = _next;
		} else {
		    prev->_next = _next;
		}
		break;
	    }
	    prev = cur;
	    cur = cur->_next;
	}
	_data = NULL;
	_next = NULL;
    }
				// Now, set to new path
    if (path) {
	_data = path;
	_next = path->_handle;
	path->_handle = this;
    }

    return(path ? TRUE : FALSE);
}

Pad_Bool
Pad_AnimPathHandle::Attach(Pad_AnimPathHandle &pathHandle)
{
    Pad_Bool rc;

    rc = Attach(pathHandle._Get_path());
    
    return(rc);
}

/*
 *----------------------------------------------------------------------
 *
 * Pad_AnimPathHandle::Get_handle_type --
 *
 *   Returns the type of this particular handle.
 *
 * Results:
 *   Type of this handle
 *
 *----------------------------------------------------------------------
 */
Pad_HandleType
Pad_AnimPathHandle::Get_handle_type(void) const
{
    return(Pad_AnimPathHandleType);
}

/*
 *----------------------------------------------------------------------
 *
 * Pad_AnimPathHandle::Is_mapped --
 *
 *   Determines if this handle is mapped to a Pad_AnimPath.
 *
 * Results:
 *   TRUE if handle is mapped to a Pad_AnimPath.
 *
 *----------------------------------------------------------------------
 */
Pad_Bool
Pad_AnimPathHandle::Is_mapped(void) const
{
    return (_data ? TRUE : FALSE);
}

/*
 *----------------------------------------------------------------------
 *
 * Pad_AnimPathHandle::Delete_path --
 *
 *   Deletes Pad_AnimPath associtate with this handle.
 *  
 *
 * Results:
 *  TRUE if delete is successful, FALSE otherwise
 *
 *----------------------------------------------------------------------
 */
Pad_Bool
Pad_AnimPathHandle::Delete_path()
{
    Pad_Bool rc = FALSE;
    Pad_String  pathToken;

    if (Is_mapped()) {
	pathToken = ((Pad_AnimPath*)_data)->Get_token();
    }
    if (pathToken.Get()) {
	((Pad_AnimPath*)_data)->~Pad_AnimPath();
	rc = (Pad_Bool)Pad_AnimPath::tokenTable.Remove(pathToken.Get()); // Remove token from hash table
    } 

    return(TRUE);
}

/*
 *----------------------------------------------------------------------
 *
 * Pad_AnimPathHandle::Get_interp_value(Pad_AnimPath::datum3 &interpVal) --
 *
 *   Stuffs the interpolated Pad_AnimPath value/s calculated from the time, 
 *   Into the user provided Pad_AnimPath::datum3 array interpVal. 
 *   (This array is defined in Pad_AnimPath as:  typedef float datum3[3] )
 *   If the path is one dimensional, only the first value of the array is
 *   stuffed, if two dimensional, the first and second values are stuffed,
 *   etc. for three dimensional paths.
 * 
 * Results:
 *  If handle is mapped, stuffs Pad_AnimPath::datum3  with the 
 *  interpolated result, and returns TRUE. 
 *  Else Pad_AnimPath::datum3 is NULL and method returns FALSE.
 *
 *----------------------------------------------------------------------
 */
Pad_Bool           
Pad_AnimPathHandle::Get_interp_value(float time, Pad_AnimPath::datum3 &interpVal)
{
    Pad_AnimPath *animpath=NULL;

    if (Is_mapped()) {
	animpath = ((Pad_AnimPath*)_data);
    }
    if (animpath && animpath->Has_data()) {
	animpath->Interpolate(time, interpVal);
    } else {
	return(FALSE);
    }

    return(TRUE);
}

/*
 *----------------------------------------------------------------------
 *
 * Pad_AnimPathHandle::Get_order --
 *
 *   Determine if the path in the path handle as data set for it.
 *
 * Results:
 *  Returns TRUE if has data. Else, returns FALSE;
 *
 *----------------------------------------------------------------------
 */
Pad_Bool           
Pad_AnimPathHandle::Has_data()
{
    Pad_AnimPath *animpath=NULL;
    Pad_Bool      hasData=FALSE;

    if (Is_mapped()) {
	animpath = ((Pad_AnimPath*)_data);
    }
    if (animpath && animpath->Has_data()) {
	hasData = TRUE;
    } 

    return(hasData);
}

/*
 *----------------------------------------------------------------------
 *
 * Pad_AnimPathHandle::Get_order --
 *
 *   Returns the order of the path i.e. 1, 2, or 3
 *
 * Results:
 *  If mapped, returns integer order of path. Else, returns 0
 *
 *----------------------------------------------------------------------
 */
int
Pad_AnimPathHandle::Get_order() const
{
    int order = 0;
    Pad_AnimPath *animpath;

    if (Is_mapped()) {
	animpath = ((Pad_AnimPath*)_data);
    }
    if (animpath) {
	order = animpath->Get_order();
    } 

    return(order);
}


/*
 *----------------------------------------------------------------------
 *
 * Pad_AnimPathHandle::Set_begin_time() --
 *
 *   Set the animation path begin time
 *
 * Results:
 *  If mapped, sets path begin time and returns TRUE. 
 *  Else, returns FALSE.
 *
 *----------------------------------------------------------------------
 */
Pad_Bool
Pad_AnimPathHandle::Set_begin_time(float beginTime)
{
    Pad_AnimPath *animpath;
    Pad_Bool rc = FALSE;

    if (Is_mapped()) {
	animpath = ((Pad_AnimPath*)_data);
    }
    if (animpath) {
	rc = animpath->Set_begt(beginTime);
    }

    return rc;
}

/*
 *----------------------------------------------------------------------
 *
 * Pad_AnimPathHandle::Set_end_time() --
 *
 *   Returns the animation path end time
 *
 * Results:
 *  If mapped, returns path end time. Else, returns 0.0 .
 *
 *----------------------------------------------------------------------
 */
Pad_Bool
Pad_AnimPathHandle::Set_end_time(float endTime)
{
    Pad_AnimPath *animpath;
    Pad_Bool rc = FALSE;

    if (Is_mapped()) {
	animpath = ((Pad_AnimPath*)_data);
    }

    if (animpath) {
	rc = animpath->Set_endt(endTime);
    }

    return rc;
}

/*
 *----------------------------------------------------------------------
 *
 * Pad_AnimPathHandle::Set_data --
 *
 *   Defines the data for the path to be Pad_AnimPath::anim_data* data. 
 *   The struct Pad_AnimPath::anim_data is defined as:
 *       struct anim_data { // To hold the data that defines the path
 *	datum* data[3];
 *	int    order;
 *       };
 *   Stuff the data array and set the order which is the number
 *   of values in each data point (1 2 or 3). 
 *   One can also specify time value pairs. To do this, the
 *   first value of the array such as the values 1 and 2 in:
 *   {{1,3,3,4},{2,4,5,6}} ,
 *   specifies the time and the other numbers specify
 *   either a 1,2, or 3 dimensional point value. It is the 
 *   number of values, not including the time value, that
 *   determines that order of the points. The example is order
 *   three. When you want to specify that the path is a set of time-
 *   value pairs, set timeData to TRUE, otherwise, it is FALSE.
 *   
 * Results:
 *  If Handle is mapped, path's data is set and the length of the
 *  data is returned (ie the number of data points).
 *  Else a length of 0 is returned.
 *
 *----------------------------------------------------------------------
 */
int
Pad_AnimPathHandle::Set_path_data(int num, Pad_AnimPath::anim_data* data, int pathOrder, Pad_Bool timeData)
{
    int len;
    Pad_AnimPath* animpath;

    if (Is_mapped()) {
	animpath = ((Pad_AnimPath*)_data);
    }
    if (animpath) {
	animpath->Set_data(num, data, pathOrder, timeData);
	len  = animpath->Get_data_len();
    } else {
	len = 0;
    }

    return(len);
}

/*
 *----------------------------------------------------------------------
 *
 * Pad_AnimPathHandle::Set_pre() --
 *
 *   Sets the pre condition for the interpolation.
 *   This is an enum value that tells the interpolator
 *   how to handle times that come before the begin time.
 *   Possible values are CONSTANT, CYCLE, and OSCILLATE.
 *   The enum is defined in Pad_AnimPath:
 *   Pad_AnimPath::PRE_POST. (See Set_post() below) 
 *   DEFAULT: CONSTANT 
 *
 * Results:
 *  If mapped, sets path pre condition. Else, returns FALSE.
 *
 *----------------------------------------------------------------------
 */
Pad_Bool
Pad_AnimPathHandle::Set_pre(Pad_AnimPath::PRE_POST pre)
{
    Pad_AnimPath *animpath;
    Pad_Bool rc = FALSE;
    
    if (Is_mapped()) {
	animpath = ((Pad_AnimPath*)_data);
    }
    if (animpath) {
	rc = animpath->Set_pre(pre);
    }

    return rc;
}

/*
 *----------------------------------------------------------------------
 *
 * Pad_AnimPathHandle::Set_post() --
 *
 *   Sets the post condition for the interpolation.
 *   This is an enum value that tells the interpolator
 *   how to handle times that come after the end time.
 *   Possible values are CONSTANT, CYCLE, and OSCILLATE.
 *   The enum is defined in Pad_AnimPath:
 *   Pad_AnimPath::PRE_POST. (See Set_pre() above)
 *   DEFAULT: CONSTANT 
 *
 * Results:
 *  If mapped, sets path pre condition. Else, returns FALSE.
 *
 *----------------------------------------------------------------------
 */
Pad_Bool
Pad_AnimPathHandle::Set_post(Pad_AnimPath::PRE_POST post)
{
    Pad_AnimPath *animpath;
    Pad_Bool rc = FALSE;

    if (Is_mapped()) {
	animpath = ((Pad_AnimPath*)_data);
    }
    if (animpath) {
	rc = animpath->Set_post(post);
    }

    return rc;
}

/*
 *----------------------------------------------------------------------
 *
 * Pad_AnimPathHandle::Set_siso --
 *
 *   Sets the flag that applies a slow-in-slow-out effect
 *   to regular data paths (i.e. not used for paths that 
 *   specify time-data pairs).
 *   DEFAULT: FALSE
 *
 * Results:
 *  If mapped, turns siso on it TRUE, turns siso off if FALSE
 *  Returns true if set operation is successful.
 *
 *----------------------------------------------------------------------
 */
Pad_Bool
Pad_AnimPathHandle::Set_siso(Pad_Bool siso)
{
    Pad_AnimPath *animpath;
    Pad_Bool      rc = FALSE;

    if (Is_mapped()) {
	animpath = ((Pad_AnimPath*)_data);
    }
    if (animpath) {
	rc = animpath->Set_siso(siso);
    }

    return rc;
}

/*
 *----------------------------------------------------------------------
 *
 * Pad_AnimPathHandle::Set_anim_in --
 *
 *   Sets the time, within an defined path, where the animation should
 *   begin playing.
 *
 * Results:
 *  If mapped, sets the animation in time.
 *
 *----------------------------------------------------------------------
 */
Pad_Bool
Pad_AnimPathHandle::Set_in_time(float inTime)
{
    Pad_AnimPath *animpath;
    Pad_Bool      rc = FALSE;

    if (Is_mapped()) {
	animpath = ((Pad_AnimPath*)_data);
    }
    if (animpath && Has_data()) {
	rc = animpath->Set_in_time(inTime);
    }

    return rc;
}

/*
 *----------------------------------------------------------------------
 *
 * Pad_AnimPathHandle::Set_anim_out --
 *
 *   Sets the time, within an defined path, where the animation should
 *   end playing.
 *
 * Results:
 *  If mapped, sets the animation out time.
 *
 *----------------------------------------------------------------------
 */
Pad_Bool
Pad_AnimPathHandle::Set_out_time(float outTime)
{
    Pad_AnimPath *animpath;
    Pad_Bool      rc = FALSE;

    if (Is_mapped()) {
	animpath = ((Pad_AnimPath*)_data);
    }
    if (animpath && Has_data()) {
	rc = animpath->Set_out_time(outTime);
    }

    return rc;
}


/*
 *----------------------------------------------------------------------
 *
 * Pad_AnimPathHandle::Get_begin_time() --
 *
 *   Returns the animation path begin time
 *
 * Results:
 *  If mapped, returns path begin time. Else, returns 0.0 .
 *
 *----------------------------------------------------------------------
 */
float              
Pad_AnimPathHandle::Get_begin_time() const
{
    Pad_AnimPath *animpath;
    float beginTime = 0.0;
    
    if (Is_mapped()) {
	animpath = ((Pad_AnimPath*)_data);
    }

    if (animpath) {
	beginTime = animpath->Get_begt();
    } else {
	beginTime = 0.0;
    }

    return(beginTime);
}

/*
 *----------------------------------------------------------------------
 *
 * Pad_AnimPathHandle::Get_end_time() --
 *
 *   Returns the animation path end time
 *
 * Results:
 *  If mapped, returns path end time. Else, returns 0.0 .
 *
 *----------------------------------------------------------------------
 */
float              
Pad_AnimPathHandle::Get_end_time() const
{
    Pad_AnimPath *animpath;
    float endTime = 0.0;
    
    if (Is_mapped()) {
	animpath = ((Pad_AnimPath*)_data);
    }

    if (animpath) {
	endTime = animpath->Get_endt();
    } else {
	endTime = 0.0;
    }

    return(endTime);
}

/*
 *----------------------------------------------------------------------
 *
 * Pad_AnimPathHandle::Get_post() --
 *
 *   Gets the post condition for the interpolation.
 *   This is an enum value that tells the interpolator
 *   how to handle times that come after the end time.
 *   Possible values are CONSTANT, CYCLE, and OSCILLATE.
 *   The enum is defined in Pad_AnimPath:
 *   Pad_AnimPath::PRE_POST. (See Get_post() below) 
 *   DEFAULT: CONSTANT 
 *
 * Results:
 *  If mapped, gets path pre condition. Else, returns UNDEFINED.
 *
 *----------------------------------------------------------------------
 */
Pad_AnimPath::PRE_POST
Pad_AnimPathHandle::Get_pre() const
{
    Pad_AnimPath *animpath;
    Pad_AnimPath::PRE_POST pre;

    if (Is_mapped()){
	animpath = ((Pad_AnimPath*)_data);
    }

    if (animpath) {
	pre = animpath->Get_pre();
    }else {
	pre = Pad_AnimPath::UNDEFINED; // a dummy value used to signal errora
    } 
	
    return(pre);
}

/*
 *----------------------------------------------------------------------
 *
 * Pad_AnimPathHandle::Get_post() --
 *
 *   Gets the post condition for the interpolation.
 *   This is an enum value that tells the interpolator
 *   how to handle times that come after the end time.
 *   Possible values are CONSTANT, CYCLE, and OSCILLATE.
 *   The enum is defined in Pad_AnimPath:
 *   Pad_AnimPath::PRE_POST. (See Get_post() below) 
 *   DEFAULT: CONSTANT 
 *
 * Results:
 *  If mapped, gets path pre condition. Else, returns UNDEFINED.
 *
 *----------------------------------------------------------------------
 */
Pad_AnimPath::PRE_POST
Pad_AnimPathHandle::Get_post() const
{
    Pad_AnimPath *animpath;
    Pad_AnimPath::PRE_POST post;

    if (Is_mapped()) {
	animpath = ((Pad_AnimPath*)_data);
    }

    if (animpath) {
	post = animpath->Get_post();
    } else {
	post = Pad_AnimPath::UNDEFINED; // a dummy value used to signal errora
    } 

    return(post);
}

/*
 *----------------------------------------------------------------------
 *
 * Pad_AnimPathHandle::Get_siso --
 *
 *   Returns the flag that applies a slow-in-slow-out effect
 *   to regular data paths (i.e. not used for paths that 
 *   specify time-data pairs).
 *   DEFAULT: FALSE
 *
 * Results:
 *  If mapped and siso is on return TRUE, else return FALSE
 *
 *----------------------------------------------------------------------
 */
Pad_Bool
Pad_AnimPathHandle::Get_siso() const
{
    Pad_AnimPath *animpath;
    Pad_Bool      siso = FALSE;

    if (Is_mapped()) {
	animpath = ((Pad_AnimPath*)_data);
    }
    if (animpath) {
	siso = animpath->Get_siso();
    }

    return siso;
}

/*
 *----------------------------------------------------------------------
 *
 * Pad_AnimPathHandle::Get_in_time --
 *
 *   Gets the time, within an defined path, where the animation should
 *   begin playing.
 *
 * Results:
 *  If mapped, sets the animation in time.
 *
 *----------------------------------------------------------------------
 */
float
Pad_AnimPathHandle::Get_in_time() 
{
    Pad_AnimPath *animpath;
    float         in = 0.0;

    if (Is_mapped()) {
	animpath = ((Pad_AnimPath*)_data);
    }
    if (animpath && Has_data()) {
	in = animpath->Get_in_time();
    }

    return in;
}

/*
 *----------------------------------------------------------------------
 *
 * Pad_AnimPathHandle::Get_out_time --
 *
 *   Gets the time, within an defined path, where the animation should
 *   end playing.
 *
 * Results:
 *  If mapped, sets the animation out time.
 *
 *----------------------------------------------------------------------
 */
float
Pad_AnimPathHandle::Get_out_time()
{
    Pad_AnimPath *animpath;
    float         out = 0.0;

    if (Is_mapped()) {
	animpath = ((Pad_AnimPath*)_data);
    }
    if (animpath && Has_data()) {
	out = animpath->Get_out_time();
    }

    return out;
}

/*
 *----------------------------------------------------------------------
 *
 * Pad_AnimPathHandle::_Get_data --
 *
 *   Returns pointer to Pad_AnimPath::anim_data. This points to the
 *   actual data for the path. Be careful if you choose to use this 
 *   method.
 *
 * Results:
 *  If Handle is mapped,Pad_AnimPath::anim_data* is set to path's data, 
 *  the length of the path is returned. Else a length of 0 is returned.
 *
 *----------------------------------------------------------------------
 */
int                
Pad_AnimPathHandle::Get_path_data(Pad_AnimPath::anim_data *&pathData) const
{
    int len;
    Pad_AnimPath* animpath;

    if (Is_mapped()) {
	animpath = ((Pad_AnimPath*)_data);
	pathData = animpath->Get_data();
	len  = animpath->Get_data_len();
    } else {
	pathData = NULL;
	len = 0;
    }

    return(len);
}

Pad_Bool 
Pad_AnimPathHandle::Path_exists(Pad_String const pathToken) 
{
    Pad_AnimPath *animpath;
    Pad_Bool      exists = FALSE;

    if (pathToken.Get()) {
	animpath = (Pad_AnimPath *)Pad_AnimPath::tokenTable.Get(pathToken.Get());
	if (animpath) {
	    exists = TRUE;
	}
    }
    return exists;

}

/*
 *----------------------------------------------------------------------
 *
 * Pad_AnimPathHandle::Get_time_path_flag --
 *
 *   Returns TRUE if path includes time data
 *   FALSE other wise
 *
 * Results:
 *  
 *----------------------------------------------------------------------
 */
Pad_Bool
Pad_AnimPathHandle::Get_time_path_flag() const
{
    Pad_AnimPath*    animpath;
    Pad_Bool         flag = FALSE;

    if (Is_mapped()) {
	animpath = ((Pad_AnimPath*)_data);
    }
    if (animpath) { // If valid path token 
	flag = animpath->Get_time_path_flag();
    } 

    return flag;
}

/*
 *----------------------------------------------------------------------
 *
 * Pad_AnimPathHandle::_Get_path --
 *
 *   Returns pointer to Pad_AnimPath. If handle is not mapped
 *   the pointer will be NULL.
 *
 * Results:
 *  Pad_AnimPath* if mapped, else NULL.
 *
 *----------------------------------------------------------------------
 */
Pad_AnimPath * 
Pad_AnimPathHandle::_Get_path(void) const
{
    return ((Pad_AnimPath*)_data);
}

////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////
//
// Pad_AnimChannelHandle Class definition
//
////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////
Pad_AnimChannelHandle::Pad_AnimChannelHandle()
{
}

Pad_AnimChannelHandle::Pad_AnimChannelHandle(Pad_AnimChannelHandle &channelHandle)
{
    Attach(channelHandle);
}

Pad_AnimChannelHandle::Pad_AnimChannelHandle(Pad_AnimChannel *channel)
{
    Attach(channel);
}

Pad_AnimChannelHandle::~Pad_AnimChannelHandle(void)
{
    Attach((Pad_AnimChannel *)NULL);
}

Pad_AnimChannelHandle & 
Pad_AnimChannelHandle::operator=(Pad_AnimChannelHandle &channelHandle)
{
    Attach(channelHandle);

    return (*this);
}

/*
 *----------------------------------------------------------------------
 *
 * Pad_AnimChannelHandle::Attach --
 *
 *   Specifies the Pad_AnimChannel  that this handle references
 *
 * Results:
 *   TRUE if successful, or FALSE otherwise
 *
 *----------------------------------------------------------------------
 */
Pad_Bool
Pad_AnimChannelHandle::Attach(Pad_AnimChannel *channel)
{
				// First, dereference any channel this references
    if (_data) {
	Pad_Handle *prev, *cur;

	prev = NULL;
	cur = ((Pad_AnimChannel *)_data)->_handle;
	while (cur) {
	    if (cur == this) {
		if (prev == NULL) {
		    ((Pad_AnimChannel *)_data)->_handle = _next;
		} else {
		    prev->_next = _next;
		}
		break;
	    }
	    prev = cur;
	    cur = cur->_next;
	}
	_data = NULL;
	_next = NULL;
    }
				// Now, set to new channel
    if (channel) {
	_data = channel;
	_next = channel->_handle;
	channel->_handle = this;
    }

    return(channel ? TRUE : FALSE);
}

Pad_Bool
Pad_AnimChannelHandle::Attach(Pad_AnimChannelHandle &channelHandle)
{
    Pad_Bool rc;

    rc = Attach(channelHandle._Get_channel());
    
    return(rc);
}

/*
 *----------------------------------------------------------------------
 *
 * Pad_AnimChannelHandle::Get_handle_type --
 *
 *   Returns the type of this particular handle.
 *
 * Results:
 *   Type of this handle
 *
 *----------------------------------------------------------------------
 */
Pad_HandleType
Pad_AnimChannelHandle::Get_handle_type(void) const
{
    return(Pad_AnimChannelHandleType);
}

/*
 *----------------------------------------------------------------------
 *
 * Pad_AnimChannelHandle::Is_mapped --
 *
 *   Determines if this handle is mapped to a Pad_AnimChannel.
 *
 * Results:
 *   TRUE if handle is mapped to a Pad_AnimChannel.
 *
 *----------------------------------------------------------------------
 */
Pad_Bool
Pad_AnimChannelHandle::Is_mapped(void) const
{
    return (_data ? TRUE : FALSE);
}

/*
 *----------------------------------------------------------------------
 *
 * Pad_AnimChannelHandle::Delete_channel --
 *
 *   Deletes Pad_AnimChannel associtate with this handle.
 *  
 *
 * Results:
 *  TRUE if delete is successful, FALSE otherwise
 *
 *----------------------------------------------------------------------
 */
Pad_Bool
Pad_AnimChannelHandle::Delete_channel()
{
    Pad_Bool rc   = FALSE;
    Pad_String      channelToken;
    Pad_Iterator    animationi;
    Pad_Animation  *animation;

    if (Is_mapped()) {
	channelToken = ((Pad_AnimChannel*)_data)->Get_token();
    }
    if (channelToken.Get()) {
	DOLIST(animationi, ((Pad_Animatable*)_data)->parents, Pad_Animation, animation) {
	    animation->Remove_animatable((Pad_Animatable*)_data);
	}

	((Pad_AnimChannel*)_data)->~Pad_AnimChannel();
	rc = (Pad_Bool)Pad_Animatable::tokenTable.Remove(channelToken.Get()); // Remove token from hash table
	rc = (Pad_Bool)Pad_AnimChannel::tokenTable.Remove(channelToken.Get()); // Remove token from hash table
    } 

    return(TRUE);
}


/*
 *----------------------------------------------------------------------
 *
 * Pad_AnimChannelHandle::Get_path_order --
 *
 *   Returns the order of the path owned by the channel (i.e. 1, 2, or 3)
 *
 * Results:
 *  If mapped, returns integer order of channel's path. Else, returns 0
 *  if handle is not mapped or if channel has no path.
 *
 *----------------------------------------------------------------------
 */
// Return the order of an animation channel given its token
int                
Pad_AnimChannelHandle::Get_path_order() const
{
    int order = 0;
    Pad_AnimChannel *animchannel;

    if (Is_mapped()) { 
	animchannel = ((Pad_AnimChannel *)_data);
    }
    // necessary to check for path at this point
    if (animchannel && animchannel->Has_path() && animchannel->Get_path()->Has_data()) {
	order = animchannel->Get_path()->Get_data()->order;
    } 

    return(order);
}

/*
 *----------------------------------------------------------------------
 *
 * Pad_AnimChannelHandle::Set_path_begin_time() --
 *
 *   Set the end time of the path owned by the channel
 *
 * Results:
 *  If mapped, sets channel begin time and returns TRUE. 
 *  Else, returns FALSE if not mapped or has no path.
 *
 *----------------------------------------------------------------------
 */
Pad_Bool
Pad_AnimChannelHandle::Set_path_begin_time(float beginT)
{
    Pad_AnimChannel *animchannel;
    Pad_Bool         set = FALSE;

    if (Is_mapped() && Has_path()){
	animchannel = ((Pad_AnimChannel*)_data);
    }
    if (animchannel) {
	set = animchannel->Set_begin_time(beginT);
    }
    return set;
}

/*
 *----------------------------------------------------------------------
 *
 * Pad_AnimChannelHandle::Set_path_end_time() --
 *
 *   Returns the end time of the path owne by the channel
 *
 * Results:
 *  If mapped, sets path begin time and returns TRUE.
 *  Else, FALSE if not mapped or has no path.
 *
 *----------------------------------------------------------------------
 */
Pad_Bool
Pad_AnimChannelHandle::Set_path_end_time(float endT)
{
    Pad_AnimChannel *animchannel;
    Pad_Bool         set = FALSE;

    if(Is_mapped() && Has_path()) {
	animchannel = ((Pad_AnimChannel*)_data);
    }
    if (animchannel) {
	set = animchannel->Set_end_time(endT);
    }
    return set;
}

/*
 *----------------------------------------------------------------------
 *
 * Pad_AnimChannelHandle::Get_path_begin_time() --
 *
 *   Returns begin time of path owned by channel
 *
 * Results:
 *
 *  If mapped, returns path begin time. 
 *  Else, returns FALSE if not mapped or has no path.
 *
 *----------------------------------------------------------------------
 */
float              
Pad_AnimChannelHandle::Get_path_begin_time() const
{
    Pad_AnimChannel *animchannel;
    float beginTime = -1.0;// an unreasonable time
    
    if (Is_mapped() && Has_path()) {
	animchannel = ((Pad_AnimChannel*)_data);
    }
    if (animchannel && animchannel->Has_path()) {
	beginTime = animchannel->Get_path()->Get_begt();
    } 

    return(beginTime);
}

/*
 *----------------------------------------------------------------------
 *
 * Pad_AnimChannelHandle::Get_path_end_time() --
 *
 *   Returns the end time of the path owned by the channel
 *
 * Results:
 *  If mapped, returns path end time. 
 *  Else, returns FALSE if not mapped or has no path.
 *
 *----------------------------------------------------------------------
 */
float              
Pad_AnimChannelHandle::Get_path_end_time() const
{
    Pad_AnimChannel *animchannel;
    float endTime = -1.0;// an unreasonable time
    
    if (Is_mapped() && Has_path()) {
	animchannel = ((Pad_AnimChannel*)_data);
    }
    if (animchannel && animchannel->Has_path()) {
	endTime = animchannel->Get_path()->Get_endt();
    } 

    return(endTime);
}

/*
 *----------------------------------------------------------------------
 *
 * Pad_AnimChannelHandle::_Get_channel --
 *
 *   Returns pointer to Pad_AnimChannel. If handle is not mapped
 *   the pointer will be NULL.
 *
 * Results:
 *  Pad_AnimChannel* if mapped, else NULL.
 *
 *----------------------------------------------------------------------
 */
Pad_AnimChannel * 
Pad_AnimChannelHandle::_Get_channel(void) const
{
    return ((Pad_AnimChannel*)_data);
}

/*
 *----------------------------------------------------------------------
 *
 * Pad_AnimChannelHandle::Has_path() --
 *
 *   Checks to see that if the channel to which this channel is attached
 *   has a path associated with it
 *
 * Results:
 *  Returns TRUE if handle is mapped to a channels and the channel
 *  has a path associated with it. Else, returns FALSE.
 *----------------------------------------------------------------------
 */
Pad_Bool
Pad_AnimChannelHandle::Has_path() const
{
    Pad_AnimChannel *animchannel;
    Pad_Bool         hasPath = FALSE;

    if (Is_mapped()) {
	animchannel = ((Pad_AnimChannel*)_data);
    }
    if (animchannel) {
	hasPath = animchannel->Has_path();
    }

    return hasPath;
}

/*
 *----------------------------------------------------------------------
 *
 * Pad_AnimChannelHandle::Has_object() --
 *
 *   Checks to see that if the channel to which this channel is attached
 *   has an object specified
 *
 * Results:
 *  Returns TRUE if handle is mapped to a channels and the channel
 *  has a object specified. Else, returns FALSE.
 *----------------------------------------------------------------------
 */
Pad_Bool
Pad_AnimChannelHandle::Has_object() const
{
    Pad_AnimChannel *animchannel;
    Pad_Bool         hasObject = FALSE;

    if (Is_mapped()) {
	animchannel = ((Pad_AnimChannel*)_data);
    }
    if (animchannel) {
	hasObject = animchannel->Has_object_handle();
    }
    return hasObject;
}

/*
 *----------------------------------------------------------------------
 *
 * Pad_AnimChannelHandle::Has_option() --
 *
 *   Checks to see that if the channel to which this channel is attached
 *   has an option specified
 *
 * Results:
 *  Returns TRUE if handle is mapped to a channels and the channel
 *  has a option specified. Else, returns FALSE.
 *----------------------------------------------------------------------
 */
Pad_Bool
Pad_AnimChannelHandle::Has_option() const
{
    Pad_AnimChannel *animchannel;
    Pad_Bool         hasOption = FALSE;

    if (Is_mapped()) {
	animchannel = ((Pad_AnimChannel*)_data);
    }
    if (animchannel) {
	hasOption = animchannel->Has_option();
    }
    return hasOption;
}


/*
 *----------------------------------------------------------------------
 *
 * Pad_AnimChannelHandle::Play() --
 *
 *   Checks to see that if the channel has all of its components for
 *   playing. If so, animates the channel
 *
 * Results:
 *  Returns TRUE if animation is started. Else, FALSE;
 *
 *----------------------------------------------------------------------
 */
Pad_Bool
Pad_AnimChannelHandle::Play()
{
    Pad_AnimChannel *animchannel;
    Pad_Bool         started=FALSE;

    if (Is_mapped()) {
	animchannel = ((Pad_AnimChannel*)_data);
    }
    if (animchannel && Has_path() && Has_option() && Has_object()) {
	started = animchannel->Start_animation();
    }

    return started;
}

/*
 *----------------------------------------------------------------------
 *
 * Pad_AnimChannelHandle::Set_path() --
 *   
 *   If channel handle is mapped,
 *   sets the path that is associated with the channel.
 *
 * Results:
 *  Returns TRUE if set is successful. Else, returns FALSE.
 *
 *----------------------------------------------------------------------
 */
Pad_Bool      
Pad_AnimChannelHandle::Set_path(Pad_AnimPathHandle const &pathHandle)
{
    Pad_AnimChannel* animchannel;
    Pad_Bool         set = FALSE;
    
    if (Is_mapped() && pathHandle.Is_mapped()) {
	animchannel = ((Pad_AnimChannel*)_data);
    }
    if (animchannel) { // If valid channel token 
	set = animchannel->Set_path(pathHandle._Get_path());
    } 

    return set;
}

/*
 *----------------------------------------------------------------------
 *
 * Pad_AnimChannelHandle::Set_option() --
 *   
 *   If channel handle is mapped,
 *   sets the option that is associated with the channel.
 *
 * Results:
 *  Returns TRUE if set is successful. Else, returns FALSE.
 *
 *----------------------------------------------------------------------
 */
Pad_Bool 
Pad_AnimChannelHandle::Set_option(Pad_OptionType option)
{
    Pad_AnimChannel* animchannel;
    Pad_Bool         set = FALSE;
    
    if (Is_mapped()) {
	animchannel = ((Pad_AnimChannel*)_data);
    }
    if (animchannel) { // If valid channel token 
	set = animchannel->Set_option(option);
    } 

    return set;
}

/*
 *----------------------------------------------------------------------
 *
 * Pad_AnimChannelHandle::Set_object() --
 *   
 *   If channel handle is mapped,
 *   sets the object that is associated with the channel.
 *
 * Results:
 *  Returns TRUE if set is successful. Else, returns FALSE.
 *
 *----------------------------------------------------------------------
 */
Pad_Bool
Pad_AnimChannelHandle::Set_object(Pad_ObjHandle &objHandle)
{
    Pad_AnimChannel* animchannel;
    Pad_Bool         set = FALSE;

    if (Is_mapped()) {
	animchannel = ((Pad_AnimChannel*)_data);
    }
    if (animchannel) { // If valid channel token 
	set = animchannel->Set_object_handle(objHandle);
    }
    return set;
}

/*
 *----------------------------------------------------------------------
 *
 * Pad_AnimChannelHandle::Set_interrupt() --
 *   
 *   If channel handle is mapped,
 *   sets the interrupt value that is associated with the channel.
 *
 * Results:
 *  Returns TRUE if set is successful. Else, returns FALSE.
 *
 *----------------------------------------------------------------------
 */
Pad_Bool
Pad_AnimChannelHandle::Set_interrupt(Pad_Bool intpt)
{
    Pad_AnimChannel* animchannel;
    Pad_Bool         set = FALSE;

    if (Is_mapped()) {
	animchannel = ((Pad_AnimChannel*)_data);
    }
    if (animchannel) { // If valid channel token 
	set = animchannel->Set_interrupt(intpt);
    } 

    return set;
}

/*
 *----------------------------------------------------------------------
 *
 * Pad_AnimChannelHandle::Set_time_resolution() --
 *   
 *   If channel handle is mapped,
 *   sets the time resolution that is associated with the channel.
 *
 * Results:
 *  Returns TRUE if set is successful. Else, returns FALSE.
 *
 *----------------------------------------------------------------------
 */
Pad_Bool
Pad_AnimChannelHandle::Set_time_resolution(int res)
{
    Pad_AnimChannel* animchannel;
    Pad_Bool         set = FALSE;

    if (Is_mapped()) {
	animchannel = ((Pad_AnimChannel*)_data);
    }
    if (animchannel) { // If valid channel token 
	set = animchannel->Set_time_resolution(res);
    } 

    return set;
}

/*
 *----------------------------------------------------------------------
 *
 * Pad_AnimChannelHandle::Get_path_token() --
 *   
 *   If channel handle is mapped,
 *   return the string token that is associated with the
 *   object in this channel handle. 
 *
 * Results:
 *  Returns TRUE if get is successful. Else, returns FALSE if
 *  no object is associalted with this channel.
 *
 *----------------------------------------------------------------------
 */
Pad_String     
Pad_AnimChannelHandle::Get_path_token() const
{
    Pad_AnimChannel* animchannel;
    Pad_String       token;


    if (Is_mapped() && Has_path()) {
	animchannel = ((Pad_AnimChannel*)_data);
    }
    if (animchannel && animchannel->Has_path()) { 
	token =  animchannel->Get_path()->Get_token();
    } 

    return token;
}

/*
 *----------------------------------------------------------------------
 *
 * Pad_AnimChannelHandle::Get_object_tag_or_id() --
 *   
 *   If channel handle is mapped,
 *   return the string tagOrId that is associated with the
 *   object/s in this channel handle. 
 *
 * Results:
 *  Returns TRUE if get is successful. Else, returns FALSE if
 *  no object is associalted with this channel.
 *
 *----------------------------------------------------------------------
 */
Pad_String
Pad_AnimChannelHandle::Get_object_tag_or_id() const
{
    Pad_AnimChannel* animchannel;
    Pad_String       tagOrId;

    if (Is_mapped()) {
	animchannel = ((Pad_AnimChannel*)_data);
    }
    if (animchannel && Has_object()) { 
	tagOrId =  animchannel->Get_object_tag_or_id();
    } 

    return tagOrId;
}

/*
 *----------------------------------------------------------------------
 *
 * Pad_AnimChannelHandle::Get_object() --
 *   
 *   If channel handle is mapped,
 *   return the handle to the object that is associated with the
 *   this channel handle.
 *
 * Results:
 *  Returns TRUE if get is successful. Else, returns FALSE if
 *  no object is associalted with this channel.
 *
 *----------------------------------------------------------------------
 */
Pad_Bool
Pad_AnimChannelHandle::Get_object(Pad_ObjHandle &objHandle) 
{
    Pad_AnimChannel* animchannel;
    Pad_Bool         set=FALSE;

    if (Is_mapped()) {
	animchannel = ((Pad_AnimChannel*)_data);
    }
    if (animchannel && Has_object()) { 
	set =  animchannel->Get_object_handle(objHandle);
    } 

    return set;
}

/*
 *----------------------------------------------------------------------
 *
 * Pad_AnimChannelHandle::Get_object() --
 *   
 *   If channel handle is mapped, and has a path
 *   return the the time resolution of the path
 *   associated with this handle.
 *
 * Results:
 *  Returns resolution if get is successful. Else, returns 0 if
 *  no path is associalted with this channel or if handle is not mapped.
 *
 *----------------------------------------------------------------------
 */
int
Pad_AnimChannelHandle::Get_time_resolution() const
{
    Pad_AnimChannel* animchannel;
    int              res=0;

    if (Is_mapped()) {
	animchannel = ((Pad_AnimChannel*)_data);
    }
    if (animchannel) { 
	res =  animchannel->Get_time_resolution();
    } 

    return res;
}

/*
 *----------------------------------------------------------------------
 *
 * Pad_AnimChannelHandle::Get_interp_val() --
 *   
 *   If channel handle is mapped, has path, and has option,
 *   stuff the interpolated option value into interpVal
 *   that cooresponds to the given time.
 *
 * Results:
 *  Returns TRUE if get is successful. Else, returns FALSE.
 *
 *----------------------------------------------------------------------
 */
Pad_Bool           
Pad_AnimChannelHandle::Get_interp_val(float time, Pad_AnimPath::datum3 &interpVal)
{
    Pad_AnimChannel* animchannel;
    Pad_Bool         gotten=FALSE;

    if (Is_mapped()) {
	animchannel = ((Pad_AnimChannel*)_data);
    }
    if (animchannel && Has_object()) { 
	gotten =  animchannel->Get_option_val_from_time(time, interpVal);
    } 

    return gotten;
}

/*
 *----------------------------------------------------------------------
 *
 * Pad_AnimChannelHandle::Get_option() --
 *   
 *   If channel handle is mapped, and has option,
 *   return option associate with this channel handle.
 *
 * Results:
 *  Returns TRUE if get is successful. Else, returns FALSE.
 *
 *----------------------------------------------------------------------
 */
Pad_OptionType
Pad_AnimChannelHandle::Get_option() const
{
    Pad_AnimChannel* animchannel;
    Pad_OptionType   opType;

    if (Is_mapped()) {
	animchannel = ((Pad_AnimChannel*)_data);
    }
    if (animchannel && Has_option()) { 
	opType =  animchannel->Get_option();
    } 

    return opType;
}

Pad_Bool 
Pad_AnimChannelHandle::Channel_exists(Pad_String chanToken)
{
    Pad_AnimChannel *animchannel;
    Pad_Bool         exists = FALSE;

    if (chanToken.Get()) {
	animchannel = (Pad_AnimChannel *)Pad_Animatable::tokenTable.Get(chanToken.Get());
	if (animchannel) {
	    exists = TRUE;
	}
    }
    return exists;

}

////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////
//
// Pad_AnimationHandle Class definition
//
////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////
Pad_AnimationHandle::Pad_AnimationHandle()
{
}

Pad_AnimationHandle::Pad_AnimationHandle(Pad_AnimationHandle &animHandle)
{
    Attach(animHandle);
}

Pad_AnimationHandle::Pad_AnimationHandle(Pad_Animation *animation)
{
    Attach(animation);
}

Pad_AnimationHandle::~Pad_AnimationHandle(void)
{
    Attach((Pad_Animation *)NULL);
}

Pad_AnimationHandle & 
Pad_AnimationHandle::operator=(Pad_AnimationHandle &animHandle)
{
    Attach(animHandle);

    return (*this);
}

/*
 *----------------------------------------------------------------------
 *
 * Pad_AnimationHandle::Attach --
 *
 *   Specifies the Pad_AnimPath  that this handle references
 *
 * Results:
 *   TRUE if successful, or FALSE otherwise
 *
 *----------------------------------------------------------------------
 */
Pad_Bool
Pad_AnimationHandle::Attach(Pad_Animation *animation)
{
    // First, dereference any animation this references
    if (_data) {
	Pad_Handle *prev, *cur;

	prev = NULL;
	cur = ((Pad_Animation *)_data)->_handle;
	while (cur) {
	    if (cur == this) {
		if (prev == NULL) {
		    ((Pad_Animation *)_data)->_handle = _next;
		} else {
		    prev->_next = _next;
		}
		break;
	    }
	    prev = cur;
	    cur = cur->_next;
	}
	_data = NULL;
	_next = NULL;
    }
    // Now, set to new animation
    if (animation) {
	_data = animation;
	_next = animation->_handle;
	animation->_handle = this;
    }

    return(animation ? TRUE : FALSE);
}

Pad_Bool
Pad_AnimationHandle::Attach(Pad_AnimationHandle &animHandle)
{
    Pad_Bool rc;

    rc = Attach(animHandle._Get_animation());
    
    return(rc);
}

/*
 *----------------------------------------------------------------------
 *
 * Pad_AnimPathHandle::_Get_animation --
 *
 *   Returns pointer to Pad_Animation. If handle is not mapped
 *   the pointer will be NULL.
 *
 * Results:
 *  Pad_Animation* if mapped, else NULL.
 *
 *----------------------------------------------------------------------
 */
Pad_Animation * 
Pad_AnimationHandle::_Get_animation(void) const
{
    return ((Pad_Animation*)_data);
}

/*
 *----------------------------------------------------------------------
 *
 * Pad_AnimationHandle::Is_mapped --
 *
 *   Determines if this handle is mapped to a Pad_Animation.
 *
 * Results:
 *   TRUE if handle is mapped to a Pad_Animation.
 *
 *----------------------------------------------------------------------
 */
Pad_Bool
Pad_AnimationHandle::Is_mapped(void) const
{
    return (_data ? TRUE : FALSE);
}

/*
 *----------------------------------------------------------------------
 *
 * Pad_AnimationHandle::Get_handle_type --
 *
 *   Returns the type of this particular handle.
 *
 * Results:
 *   Type of this handle
 *
 *----------------------------------------------------------------------
 */
Pad_HandleType
Pad_AnimationHandle::Get_handle_type(void) const
{
    return(Pad_AnimationHandleType);
}

/*
 *----------------------------------------------------------------------
 *
 * Pad_AnimationHandle::Add_member--
 *
 *   Adds an animatable to the animatable list owned by the animation
 *   associated with the handle. 
 *   This method will need to be overloaded to accept each
 *   type of animatable.
 *
 * Results:
 *   TRUE successful. FALSE otherwise.
 *
 *----------------------------------------------------------------------
 */
Pad_Bool                
Pad_AnimationHandle::Add_member(Pad_AnimChannelHandle &chanHandle)
{
    Pad_Animation   *animation;
    Pad_Animatable  *animatable;
    Pad_Bool         set = FALSE;
    
    if (Is_mapped() && chanHandle.Is_mapped()) {
	animation   = ((Pad_Animation *)_data);
	animatable  = ((Pad_Animatable *)chanHandle._data);
    }
    if (animatable && animation) { // If valid animatable token 
	set = animation->Add_animatable(animatable);
    } 

    return set;
}

Pad_Bool                
Pad_AnimationHandle::Add_member(Pad_AnimationHandle &animHandle)
{
    Pad_Animation   *animation;
    Pad_Animatable  *animatable;
    Pad_Bool         set = FALSE;
    
    if (Is_mapped() && animHandle.Is_mapped()) {
	animation   = ((Pad_Animation *)_data);
	animatable  = ((Pad_Animatable *)animHandle._data);
    }
    if (animatable && animation) { // If valid animatable token 
	if (animation != animatable) {
	    set = animation->Add_animatable(animatable);
	}
    } 

    return set;
}

/*
 *----------------------------------------------------------------------
 *
 * Pad_AnimationHandle::Add_member--
 *
 *   Removes an animatable from the animatable list owned by the animation
 *   associated with the handle. 
 *
 * Results:
 *   TRUE successful. FALSE otherwise.
 *
 *----------------------------------------------------------------------
 */
Pad_Bool                
Pad_AnimationHandle::Remove_member(Pad_AnimChannelHandle &chanHandle)
{
    Pad_Animation   *animation;
    Pad_Animatable  *animatable;
    Pad_Bool         rem = FALSE;
    
    if (Is_mapped() && chanHandle.Is_mapped()) {
	animation   = ((Pad_Animation *)_data);
	animatable  = ((Pad_Animatable *)_data);
    }
    if (animatable && animation) { // If valid animatable token 
	rem = animation->Remove_animatable(animatable);
    } 

    return rem;
}

/*
 *----------------------------------------------------------------------
 *
 * Pad_AnimationHandle::Set_begin_time--
 *
 *   Sets the animation begin time.
 *
 * Results:
 *   TRUE successful. FALSE otherwise.
 *
 *----------------------------------------------------------------------
 */
Pad_Bool
Pad_AnimationHandle::Set_begin_time(float beginTime)
{
    Pad_Animation   *animation;
    Pad_Bool         set = FALSE;
    
    if (Is_mapped()) {
	animation   = ((Pad_Animation *)_data);
    }
    if (animation) { // If valid animatable token 
	set = animation->Set_begin_time(beginTime);
    } 

    return set;
}

/*
 *----------------------------------------------------------------------
 *
 * Pad_AnimationHandle::Set_begin_time--
 *
 *   Sets the animation end time.
 *
 * Results:
 *   TRUE successful. FALSE otherwise.
 *
 *----------------------------------------------------------------------
 */
Pad_Bool
Pad_AnimationHandle::Set_end_time(float endTime)
{
    Pad_Animation   *animation;
    Pad_Bool         set = FALSE;
    
    if (Is_mapped()) {
	animation   = ((Pad_Animation *)_data);
    }
    if (animation) { // If valid animatable token 
	set = animation->Set_end_time(endTime);
    } 

    return set;
}

/*
 *----------------------------------------------------------------------
 *
 * Pad_AnimationHandle::Set_speed_factor--
 *
 *   Default value of speed factor is 1.0; If set to 2.0 the
 *   play rate for the animation is doubled etc.
 *
 * Results:
 *   Sets speed of animation play to argument's value.
 *
 *----------------------------------------------------------------------
 */
Pad_Bool
Pad_AnimationHandle::Set_speed_factor(float speedF)
{
    Pad_Animation   *animation;
    Pad_Bool         set = FALSE;
    
    if (Is_mapped()) {
	animation   = ((Pad_Animation *)_data);
    }
    if (animation) { // If valid animatable token 
	set = animation->Set_speed_factor(speedF);
    } 

    return set;
}

/*
 *----------------------------------------------------------------------
 *
 * Pad_AnimationHandle::Set_begin_time--
 *
 *   Sets the animation interrupt flag.
 *
 * Results:
 *   TRUE successful. FALSE otherwise.
 *
 *----------------------------------------------------------------------
 */
Pad_Bool
Pad_AnimationHandle::Set_interrupt(Pad_Bool intpt)
{
    Pad_Animation   *animation;
    Pad_Bool         set = FALSE;
    
    if (Is_mapped()) {
	animation   = ((Pad_Animation *)_data);
    }
    if (animation) { // If valid animatable token 
	set = animation->Set_interrupt(intpt);
    }

    return set;
}

/*
 *----------------------------------------------------------------------
 *
 * Pad_AnimationHandle::Get_begin_time--
 *
 *   Gets the animation begin time.
 *
 * Results:
 *   Returns begin time if successful. -1.0 otherwise.
 *
 *----------------------------------------------------------------------
 */
float              
Pad_AnimationHandle::Get_begin_time() const
{
    Pad_Animation *animation;
    float beginTime = -1.0;

    if (Is_mapped()) {
	animation   = ((Pad_Animation *)_data);
    }
    if (animation) { // If valid animatable token 
	beginTime = animation->Get_begin_time();
    }

    return(beginTime);
}

/*
 *----------------------------------------------------------------------
 *
 * Pad_AnimationHandle::Get_end_time--
 *
 *   Gets the animation end time.
 *
 * Results:
 *   Returns end time if successful. -1.0 otherwise.
 *
 *----------------------------------------------------------------------
 */
float              
Pad_AnimationHandle::Get_end_time() const
{
    Pad_Animation *animation;
    float endTime = -1.0;

    if (Is_mapped()) {
	animation   = ((Pad_Animation *)_data);
    }
    if (animation) { // If valid animatable token 
	endTime = animation->Get_end_time();
    }

    return(endTime);
}

/*
 *----------------------------------------------------------------------
 *
 * Pad_AnimationHandle::Get_end_time--
 *
 *   Gets the animation speed factor (See Set_speed_factor, above)
 *
 * Results:
 *   Returns current animation speed factor.
 *
 *----------------------------------------------------------------------
 */
float              
Pad_AnimationHandle::Get_speed_factor() const
{
    Pad_Animation *animation;
    float speedFactor = 0.0;

    if (Is_mapped()) {
	animation   = ((Pad_Animation *)_data);
    }
    if (animation) { // If valid animatable token 
	speedFactor = animation->Get_speed_factor();
    }

    return(speedFactor);
}

/*
 *----------------------------------------------------------------------
 *
 * Pad_AnimationHandle::Get_interrupt--
 *
 *   Gets the animation interrupt flag value (TRUE or FALSE)
 *
 * Results:
 *   Returns TRUE if handle is mapped and flag is TRUE. FALSE otherwise 
 *
 *----------------------------------------------------------------------
 */
Pad_Bool
Pad_AnimationHandle::Get_interrupt() const
{
    Pad_Animation *animation;
    Pad_Bool intpt;

    if (Is_mapped()) {
	animation   = ((Pad_Animation *)_data);
    }
    if (animation) { // If valid animatable token 
	intpt = animation->Get_interrupt();
    }

    return(intpt);
}

/*
 *----------------------------------------------------------------------
 *
 * Pad_AnimationHandle::Get_animatable_tokens--
 *
 *   Stuffs membersTokens with a string of that token names associated
 *   with the animation to which handle is attached
 *
 * Results:
 *   Returns TRUE if successful. FALSE otherwise.
 *
 *----------------------------------------------------------------------
 */
Pad_Bool
Pad_AnimationHandle::Get_animatable_tokens(Pad_String &memberTokens) const
{
    Pad_Animation *animation;
 
    if (Is_mapped()) {
	animation   = ((Pad_Animation *)_data);
    }
    if (animation) { // If valid animatable token 
	animation->Get_animatable_tokens(memberTokens);
    }

    return(TRUE);
}

/*
 *----------------------------------------------------------------------
 *
 * Pad_AnimationHandle::Play()--
 *
 *   Starts the play of the animation to which the handle is 
 *   is attached.
 *
 * Results:
 *   Returns TRUE if successful. FALSE otherwise.
 *
 *----------------------------------------------------------------------
 */
Pad_Bool
Pad_AnimationHandle::Play()
{
    Pad_Animation   *animation;
    Pad_Bool        playStarted = FALSE;

    if (Is_mapped()) {
	animation   = ((Pad_Animation *)_data);
    }
    if (animation) { // If valid animatable token 
	playStarted = animation->Start();
    }

    return(playStarted);
}

/*
 *----------------------------------------------------------------------
 *
 * Pad_AnimationHandle::Play()--
 *
 *   Checks the Pad_Animation::tokenTable to see if animToken 
 *   is a key to a hashed animation. 
 *
 * Results:
 *   Returns TRUE key is mapped. FALSE otherwise.
 *
 *----------------------------------------------------------------------
 */
Pad_Bool 
Pad_AnimationHandle::Animation_exists(Pad_String animToken)
{
    Pad_Animation   *animation;
    Pad_Bool         exists = FALSE;

    if (animToken.Get()) {
	animation = (Pad_Animation *)Pad_Animation::tokenTable.Get(animToken.Get());
	if (animation) {
	    exists = TRUE;
	}
    }
    return exists;
}

Pad_Bool
Pad_AnimationHandle::Delete_animation()
{
    Pad_Bool rc   = FALSE;
    Pad_String      animToken;
    Pad_Iterator    animationi;
    Pad_Animation  *animation;

    if (Is_mapped()) {
	animToken = ((Pad_Animation*)_data)->Get_token();
    }
    if (animToken.Get()) {
	DOLIST(animationi, ((Pad_Animatable*)_data)->parents, Pad_Animation, animation) {
	    animation->Remove_animatable((Pad_Animatable*)_data);
	}

	((Pad_Animation*)_data)->~Pad_Animation();
	rc = (Pad_Bool)Pad_Animatable::tokenTable.Remove(animToken.Get()); // Remove token from hash table
	rc = (Pad_Bool)Pad_Animation::tokenTable.Remove(animToken.Get()); // Remove token from hash table
    } 

    return(TRUE);
}

  
////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////
//
// Pad_UniverseMgr Class definition
//
////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////

/*
 *----------------------------------------------------------------------
 *
 * Pad_UniverseMgr::Create_win --
 *
 *   Create a new Pad++ Window.  An associated View and Surface
 *   are also created that are associated with this Window.
 *
 * Results:
 *   TRUE if successful
 *
 *----------------------------------------------------------------------
 */
Pad_Bool
Pad_UniverseMgr::Create_win(Tcl_Interp *interp, char *surfaceName)
{
    Pad_Bool rc;

    rc = Pad_Create_win(interp, surfaceName);

    return(rc);
}

Pad_Bool
Pad_UniverseMgr::Create_win(Tcl_Interp *interp, char *surfaceName, Pad_WinHandle &win)
{
    Pad_Bool rc;

    rc = Pad_Create_win(interp, surfaceName);
    if (rc) {
	win.Attach(surfaceName);
    } else {
	win.Attach((Pad_Win *)NULL);
    }

    return(rc);
}

/*
 *----------------------------------------------------------------------
 *
 * Pad_UniverseMgr::Get_surfaces --
 *
 *   Retrieve all Pad++ surfaces
 *
 * Results:
 *   Fills in the list of handles.
 *
 *----------------------------------------------------------------------
 */
void
Pad_UniverseMgr::Get_surfaces(Pad_HandleList &handles)
{
    Pad_List padNames;
    Pad_Name padName;
    Pad_Iterator oi;
    Pad_SurfaceHandle *padHandle;

    Pad_Win::Get_pads(padNames);
    DOLISTI(oi, padNames, Pad_Name, padName) {
	padHandle = new Pad_SurfaceHandle(padName);
	handles.Push_last(padHandle);
    }
}

/*
 *----------------------------------------------------------------------
 *
 * Pad_UniverseMgr::Get_instance -- 
 *
 *   Returns the single instance of Pad_UniverseMgr
 *
 * Results:
 *   none
 *
 *----------------------------------------------------------------------
 */
Pad_UniverseMgr *Pad_UniverseMgr::_instance = NULL;

Pad_UniverseMgr *
Pad_UniverseMgr::Get_instance(void)
{
    if (!_instance) {
        _instance = new Pad_UniverseMgr;
    }
    return _instance;
}

/*
 *----------------------------------------------------------------------
 *
 * Pad_UniverseMgr::Init_Sound -- 
 *
 *   Initialize sound support 
 *
 * Results:
 *   Calls any language specific initialization functions
 *
 *----------------------------------------------------------------------
 */
int
Pad_UniverseMgr::Set_volume(int vol)
{
    SoundDevice *dev = SoundDevice::Get_instance();
    if (dev) {
        return dev->Set_volume (vol);
    }
    return -1;
}

int
Pad_UniverseMgr::Set_volume(int left, int right)
{
    SoundDevice *dev = SoundDevice::Get_instance();
    if (dev) {
        return dev->Set_volume (left,right);
    }
    return -1;
}

int
Pad_UniverseMgr::Get_volume(int &left, int &right)
{
    SoundDevice *dev = SoundDevice::Get_instance();
    if (dev) {
        return dev->Get_volume (left,right);
    }
    return -1;
}

Pad_Bool
Pad_UniverseMgr::Stop(void)
{
    SoundDevice *dev = SoundDevice::Get_instance();
    if (dev) {
        return dev->Stop();
    }
    return FALSE;
}

const Pad_List *
Pad_UniverseMgr::Get_sounds(void)
{
    SoundDevice *dev = SoundDevice::Get_instance();
    if (dev) {
        return dev->Get_sounds();
    }
    return NULL;
}



////////////////////////////////////////////////////////////////
//
// Pad_SoundHandle Class
//
//
////////////////////////////////////////////////////////////////

/*
 *----------------------------------------------------------------------
 *
 * Pad_SoundHandle::Pad_SoundHandle() --
 *
 *   Constructors for sound handles
 *
 * Results:
 *   Given a pathname, a file will be loaded in and the sound data
 *      will be associated with that handle
 *
 *----------------------------------------------------------------------
 */
Pad_SoundHandle::Pad_SoundHandle(void){}

Pad_SoundHandle::~Pad_SoundHandle(void){}

/*
 *----------------------------------------------------------------------
 *
 * Pad_SoundHandle::Get_handle_type --
 *
 *   Returns the type of this particular handle.
 *
 * Results:
 *   Type of this handle
 *
 *----------------------------------------------------------------------
 */
Pad_HandleType
Pad_SoundHandle::Get_handle_type(void) const
{
    return(Pad_SoundHandleType);
}

Pad_Bool
Pad_SoundHandle::Attach(Pad_Sound *sound)
{
        // Make a new sound
    _data = sound;
    return TRUE;
}

Pad_Bool
Pad_SoundHandle::Load(char *path)
{
        // Make a new sound
    _data = (Pad_Sound*) new Pad_FileSound();

        // Load the sound 
    if (! ((Pad_FileSound*)_data)->Load(path)){
        delete (Pad_FileSound*)_data;
        return FALSE;
    }
    return TRUE;
}

ostream&
operator<<(ostream &os, Pad_SoundHandle &sound){
    if (sound._data) {
        os << *(Pad_Sound*)sound._data;
    }
    return os;
}

int
Pad_SoundHandle::Get_channels(void){
    if (_data) {
        return ((Pad_Sound*)_data)->Get_channels();
    }
    return -1;
}

int
Pad_SoundHandle::Get_rate(void)
{
    if (_data) {
        return ((Pad_Sound*)_data)->Get_rate();
    }
    return -1;
}

int
Pad_SoundHandle::Get_bits_per_sample(void)
{
    if (_data) {
        return ((Pad_Sound*)_data)->Get_bits_per_sample();
    }
    return -1;    
}

int
Pad_SoundHandle::Get_length(void)
{
    if (_data) {
        return ((Pad_Sound*)_data)->Get_length();
    }
    return -1;    
}

Pad_Bool
Pad_SoundHandle::Get_configuration(Pad_List &conf) const
{
    if (_data) {
        return ((Pad_Sound*)_data)->Get_configuration(conf);
    }
    return FALSE;
}

Pad_Bool
Pad_SoundHandle::Get_configuration(Pad_SoundOption &option) const
{
    if (_data) {
        return ((Pad_Sound*)_data)->Get_configuration(option);
    }
    return FALSE;
}

Pad_Bool
Pad_SoundHandle::Play(Pad_SoundConfig *config)
{
    if (_data) {
        return ((Pad_Sound*)_data)->Play(config);
    }
    return FALSE;
}

Pad_Bool
Pad_SoundHandle::Stop(void)
{
    if (_data) {
        return ((Pad_Sound*)_data)->Stop();
    }
    return FALSE;
}

Pad_Bool
Pad_SoundHandle::Configure(Pad_List &options)
{
    if (_data) {
        return ((Pad_Sound*)_data)->Configure(options);
    }    
    return FALSE;
}

Pad_Bool
Pad_SoundHandle::Create(Pad_SoundType sound)
{
    switch (sound) {
        case SoundObject :
            _data = new Pad_ObjectSound();
            break;
        case SoundMachine :
            _data = new Pad_MachineSound();
            break;
        case SoundLift :
            _data = new Pad_LiftSound();
            break;
        default :
            Pad_errorString = "bad sound type";
            return FALSE;
    }
    return TRUE;
}
