/*
"(c) Copyright 1993-1997 Pad++ Consortium {University of New Mexico (UNM),
and New York University (NYU)}, All Rights Reserved."  
Licensee can not remove or obscure any of the
copyright or trademark notices in this software.

IN NO EVENT SHALL THE AUTHORS OR DISTRIBUTORS BE LIABLE TO ANY PARTY
FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
ARISING OUT OF THE USE OF THIS SOFTWARE, ITS DOCUMENTATION, OR ANY
DERIVATIVES THEREOF, EVEN IF THE AUTHORS HAVE BEEN ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.

THE AUTHORS AND DISTRIBUTORS SPECIFICALLY DISCLAIM ANY WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE, AND NON-INFRINGEMENT.  THIS SOFTWARE
IS PROVIDED ON AN "AS IS" BASIS, AND THE AUTHORS AND DISTRIBUTORS HAVE
NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR
MODIFICATIONS.

See the file "License" for general information on usage and
redistribution, and the file "LicenseTerms" for the specific license
agreement on usage and redistribution of this file, and the Pad++
software in general.
*/

/*
  File: api.h

  This defines the official published C++ API to Pad++.
  It contains a set of classes which define all functionality 
  that is available to applications.

  All communication between applications and Pad++ should occur through
  this API.  This includes applications written in C++, and applications
  written in other languages.  In the case of other languages (such as
  Tcl, Scheme, or Perl, etc.), they should communicate with a set of wrapper
  procedures for that language which communicate with Pad++ through these 
  procedures.  

  While there is nothing stopping an application from accessing Pad++
  internals directly, this is considered unsafe and is unsupported.
  In particular, Pad++ makes no promises about how internals may change
  in the future - but we will make a great effort to support this API
  in a backwards compatible way.

  There are several variable types that are used that the caller must know about.
  These include:
     Pad_Handle      - A handle to any type of item in Pad++ (object, surface, view,
                       sound, etc.)
     Pad_ObjHandle   - A handle to one or more Pad++ Objects.  A handle can specify 
                       either a single object with very quick access, or a group of
                       objects specified by a tag with access through a hash table.
                       Handles are memory-safe, even when the object the handle
                       references has been deleted.
     Pad_SurfaceHandle - A handle to a Pad++ Surface which contains objects.
     Pad_ViewHandle  - A handle to a Pad++ View which specifies what portion of
                       a Surface is visible.
     Pad_WinHandle   - A handle to a Pad++ Window which specifies a top-level
                       operating system dependent window.
     Pad_UniverseMgr - A single global class instance which oversees the Pad++ system.
     Pad_List        - A list of (void *)'s.  The user must cast these carefully.
     Pad_HandleList  - A list of (Pad_Handle *)'s.  Like a Pad_List, but the handles 
                       are deleted when the list is deleted.
     Pad_PList       - A list of (Pad_Point)'s.  Note that it is a list of actual
                       Pad_Point instances, and not pointers to Pad_Point.
     Pad_Bool        - A 1 byte boolean.  Can be TRUE or FALSE.
     Pad_String      - A dynamically growable string class.
     Pad_Anchor      - A anchor type specifying a point on a bounding box of an object
     Pad_Callback    - A reference to a callback that can be in any of the supported
                       languages.  This includes at least C++ and Tcl, but may also
                       include Scheme or Perl if they are compiled in.
		       See the callback API in generic/callback.h
     Pad_Renderer    - The system renderer that can be used in a render callback
                       to draw directly onto the screen.  When in a render callback,
		       the renderer can be accessed through the global variable,
		       Pad_renderer.  See the renderer API in generic/renderer.h

  The API is (will be...) completely documented in files in the PADHOME/doc directory.

  WARNING: This is a work in progress!
 */

#ifndef API_H
#define API_H 1

#include "defs.h"
#include "list.h"
#include "ilist.h"
#include "hlist.h"
#include "plist.h"
#include "pad-string.h"
#include "types.h"
#include "bbox.h"
#include "object.h"
#include "anim-path.h"

#include <stdarg.h>

//////////////////////////////////////////////////////////////////////
//
// Forward References
//
//////////////////////////////////////////////////////////////////////

class Pad;
class Pad_Object;
class Pad_Sound;
class Pad_Text;
class Pad_View;
class Pad_Win;
class Pad_SurfaceHandle;
class Pad_UniverseMgr;
class Pad_Callback;
class Pad_Event;
class Pad_Renderer;
class Pad_AnimPath;
class Pad_AnimChannel;
class Pad_Animation;
class Pad_AnimationHandle;
class Pad_SoundOption;
class Pad_SoundConfig;

//////////////////////////////////////////////////////////////////////
//
// Global variables used by the Pad++ API
//
//////////////////////////////////////////////////////////////////////

extern Pad_String      Pad_errorString; // Contains descriptive string of current error
extern Pad_Renderer   *Pad_renderer;    // Current system renderer


//////////////////////////////////////////////////////////////////////
//
// Types used by the Pad++ API
//
//////////////////////////////////////////////////////////////////////


         // Pad++ object types that are used to specify which type of
         // object to create, or othertimes when object types are
         // specified.  Applications should use the lower case
         // versions below.  The upper case versions are for
         // internal use only.
enum Pad_ObjectType {
    Pad_NoType        = 0,
    Pad_AliasType     = PAD_ALIAS,
    Pad_ButtonType    = PAD_BUTTON,
    Pad_CanvasType    = PAD_CANVAS,
    Pad_ContainerType = PAD_CONTAINER,
    Pad_FrameType     = PAD_FRAME,
    Pad_GridType      = PAD_GRID,
    Pad_GroupType     = PAD_GROUP,
    Pad_HTMLType      = PAD_HTML,
    Pad_ImageType     = PAD_IMAGE,
    Pad_KPLType       = PAD_KPL,
    Pad_LabelType     = PAD_LABEL,
    Pad_LineType      = PAD_LINE,
    Pad_PanelType     = PAD_PANEL,
    Pad_OvalType      = PAD_OVAL,
    Pad_PadType       = PAD_PAD,
    Pad_PolygonType   = PAD_POLYGON,
    Pad_PortalType    = PAD_PORTAL,
    Pad_RectangleType = PAD_RECTANGLE,
    Pad_ScrollbarType = PAD_SCROLLBAR,
    Pad_SplineType    = PAD_SPLINE,
    Pad_TCLType       = PAD_TCL,
    Pad_TextType      = PAD_TEXT,
    Pad_TextareaType  = PAD_TEXTAREA,
    Pad_TextfieldType = PAD_TEXTFIELD,
    Pad_TextfileType  = PAD_TEXTFILE,
    Pad_WindowType    = PAD_WINDOW
};

         // Pad++ object option types that are used to specify which option
         // of an object is being specified.
         // Applications should use the lower case versions below.  
         // The upper case versions are for internal use only.
enum Pad_OptionType {
         // These are options that every object type has.
    Pad_NoOption            = 0,
    Pad_AliasesOption       = PAD_ALIASES,
    Pad_AlwaysRenderOption  = PAD_ALWAYSRENDER,
    Pad_AnchorOption        = PAD_ANCHOR,
    Pad_AngleOption         = PAD_ANGLE,
    Pad_AngleCtrOption      = PAD_ANGLECTR,
    Pad_EventsOption        = PAD_EVENTS,
    Pad_FadeRangeOption     = PAD_FADERANGE,
    Pad_HeightOption        = PAD_HEIGHT,
    Pad_InfoOption          = PAD_INFO,
    Pad_LayerOption         = PAD_LAYER,
    Pad_LockOption          = PAD_LOCK,
    Pad_MaxSizeAbsOption    = PAD_LASTOPTION + 1,
    Pad_MaxSizeRelOption    = PAD_MAXSIZE,
    Pad_MinSizeAbsOption    = PAD_LASTOPTION + 2,
    Pad_MinSizeRelOption    = PAD_MINSIZE,
    Pad_NoiseDataOption     = PAD_NOISEDATA,
    Pad_RenderScriptOption  = PAD_RENDERSCRIPT,
    Pad_StickyOption        = PAD_STICKY,
    Pad_TagsOption          = PAD_TAGS,
    Pad_TimerScriptOption   = PAD_TIMERSCRIPT,
    Pad_TimerRateOption     = PAD_TIMERRATE,
    Pad_TransparencyOption  = PAD_TRANSPARENCY,
    Pad_ViewScriptOption    = PAD_VIEWSCRIPT,
    Pad_WidthOption         = PAD_WIDTH,
    Pad_ZoomActionOption    = PAD_ZOOMACTION,
    Pad_RPositionOption     = PAD_RPOSITION,
    Pad_PositionOption      = PAD_POSITION,

	// Options for derived types. These options are only valid
        // for the relevant object types.
    Pad_ArrowOption         = PAD_ARROW,
    Pad_ArrowshapeOption    = PAD_ARROWSHAPE,
    Pad_BBoxOption          = PAD_BBOX,
    Pad_BorderOption        = PAD_BORDER,
    Pad_BorderWidthOption   = PAD_BORDERWIDTH,
    Pad_CapStyleOption      = PAD_CAPSTYLE,
    Pad_CommandOption       = PAD_COMMAND,
    Pad_DitherOption        = PAD_DITHER,
    Pad_DivisibleOption     = PAD_DIVISIBLE,
    Pad_DoneScriptOption    = PAD_DONESCRIPT,
    Pad_EchoCharOption      = PAD_ECHOCHAR,
    Pad_EditableOption      = PAD_EDITABLE,
    Pad_ErrorScriptOption   = PAD_ERRORSCRIPT,
    Pad_FileOption          = PAD_FILE,
    Pad_FillOption          = PAD_FILL,
    Pad_FontOption          = PAD_FONT,
    Pad_FromOption          = PAD_FROM,
    Pad_HTMLAnchorsOption   = PAD_HTMLANCHORS,
    Pad_ImageDataOption     = PAD_IMAGEDATA,
    Pad_IsMapOption         = PAD_ISMAP,
    Pad_JoinStyleOption     = PAD_JOINSTYLE,
    Pad_LineSizeOption      = PAD_LINESIZE,
    Pad_LookOnOption        = PAD_LOOKON,
    Pad_MembersOption       = PAD_MEMBERS,
    Pad_OrientationOption   = PAD_ORIENTATION,
    Pad_PageSizeOption      = PAD_PAGESIZE,
    Pad_PenOption           = PAD_PEN,
    Pad_PenWidthOption      = PAD_PENWIDTH,
    Pad_ReferenceOption     = PAD_REFERENCE,
    Pad_ReliefOption        = PAD_RELIEF,
    Pad_SizeOption          = PAD_SIZE,
    Pad_StateOption         = PAD_STATE,
    Pad_TitleOption         = PAD_TITLE,
    Pad_TextOption          = PAD_TEXTOPTION,
    Pad_ToOption            = PAD_TO,
    Pad_UpdateScriptOption  = PAD_UPDATESCRIPT,
    Pad_URLOption           = PAD_URL,
    Pad_ValueOption         = PAD_VALUE,
    Pad_ViewOption          = PAD_VIEWOPTION,
    Pad_VisibleLayersOption = PAD_VISIBLELAYERS,
    Pad_WriteFormatOption   = PAD_WRITEFORMAT
};

         // Pad++ anchor type
enum Pad_Anchor {
    Pad_AnchorN  = TK_ANCHOR_N,
    Pad_AnchorNE = TK_ANCHOR_NE,
    Pad_AnchorE  = TK_ANCHOR_E,
    Pad_AnchorSE = TK_ANCHOR_SE,
    Pad_AnchorS  = TK_ANCHOR_S,
    Pad_AnchorSW = TK_ANCHOR_SW,
    Pad_AnchorW  = TK_ANCHOR_W,
    Pad_AnchorNW = TK_ANCHOR_NW,
    Pad_AnchorCenter = TK_ANCHOR_CENTER
};

         // Pad++ handle types
enum Pad_HandleType {
    Pad_ObjHandleType,
    Pad_SurfaceHandleType,
    Pad_ViewHandleType,
    Pad_TextHandleType,
    Pad_WinHandleType,
    Pad_AnimPathHandleType,
    Pad_AnimChannelHandleType,
    Pad_AnimationHandleType,
    Pad_SoundHandleType,
    Pad_SourceHandleType,
};

    // Configuration options for sound handles
enum Pad_SoundOptionType {
    SoundConfigNone,
    SoundConfigDirection,
    SoundConfigGain,
    SoundConfigMaterial,
    SoundConfigMaxWork,
    SoundConfigMinWork,
    SoundConfigRatio,
    SoundConfigSize,
    SoundConfigSpeed,
    SoundConfigHardness,
    SoundConfigImpact,
    SoundConfigLength,
    SoundConfigLoop,
    SoundConfigRoughness,
    SoundConfigScrape,
    SoundConfigVolume,
};

    // Types of parameterized sound
enum Pad_SoundType {
    SoundObject,
    SoundMachine,
    SoundLift,
};

/*
 *----------------------------------------------------------------------
 *
 * class Pad_Handle --
 *
 *    A class for referencing zero, one or more things within Pad++.
 *    There are different handle types to reference different kinds of
 *    things.  Handles have special code to reference Pad_Objects.  
 *    They can store either a pointer to an object, or a tag that 
 *    implicitly specifies all the objects that share that tag.  
 *    Derived handle types can use _data slot to reference any kind of data.
 * 
 *    This is a safe structure for storing a handle to an object where the 
 *    object may have been deleted by the time it is accessed.  A linked
 *    list of all handles that point to an object is maintained, and when
 *    an object is deleted, all handles sharing that object have their
 *    pointers NULLed.
 *
 *    Note that some data that is used only by derived types is in
 *    this base class so that all Pad_Handles are the same size, and
 *    we can take over memory management for them.
 *
 *    This is an abstract base class, and can not be instantiated.
 *    It is useful only because we can have types derived from this one.
 *
 *----------------------------------------------------------------------
 */

class Pad_Handle
{
  public:
    Pad_Handle();
    Pad_Handle(Pad_Handle &handle);
    virtual ~Pad_Handle();

    virtual Pad_HandleType Get_handle_type(void) const = 0; // Return type of this handle

    void *       operator new(size_t size);            // Do our own memory management to avoid memory fragmentation
    void         operator delete(void *object);
    Pad_Handle & operator=(Pad_Handle &handle);        // Assignment operator
    Pad_Bool     operator==(Pad_Handle &handle);       // Equality operator
    Pad_Bool     operator!=(Pad_Handle &handle);       // InEquality operator

  protected:
    friend class Pad_ObjHandle;
    friend class Pad_SoundHandle;
    friend class Pad_SurfaceHandle;
    friend class Pad_WinHandle;
    friend class Pad_ViewHandle;
    friend class Pad_Object;
    friend class Pad_AnimPathHandle;
    friend class Pad_AnimChannelHandle;
    friend class Pad_AnimationHandle;

				// A handle can specify either an object directly (with the _obj) slot,
				// or a group of objects indirectly using a tag in which case the
				// _obj slot is used to specify the pad widget the tag references.
				// Or, a handle can reference another kind of object specified by _data.
				// 
				// Note: All data for every handle type is stored in the base class
				// so that every Handle is the same size, and thus we can easily
				// do our own memory management.

    Pad_Object *_obj;	// Pointer to object.  This will be set to NULL when obj is deleted
    Pad_Handle *_next;	// Pointer to next handle sharing this object (or pad), or NULL if last one
    Pad_String *_tag;   // Tag that specifies objects.
    void       *_data;  // Pointer to type-specific data
};

/*
 *----------------------------------------------------------------------
 *
 * class Pad_ObjHandle --
 *
 *    A handle to one or more Pad++ Objects.  A handle can specify either a
 *    single object with very quick access, or a group of objects specified
 *    by a tag with access through a hash table.  Handles are memory-safe,
 *    even when the object the handle references has been deleted.
 *
 *    Use Get_id() to determine if the handle actually references any objects.
 *    It will return the id of the first object referenced, or 0 if none.
 *
 *---------------------------------------------------------------------- */

class Pad_ObjHandle : public Pad_Handle
{
  public:
    Pad_ObjHandle();
    Pad_ObjHandle(Pad_ObjHandle &objHandle);
    Pad_ObjHandle(Pad_Object *obj);
    Pad_ObjHandle(char *surfaceName, int objectId);
    Pad_ObjHandle(Pad_SurfaceHandle &surface, int objectId);  
    Pad_ObjHandle(Pad_SurfaceHandle &surface, char *tag);
    Pad_ObjHandle(char *surfaceName, char *tag);
    virtual ~Pad_ObjHandle();
    Pad_ObjHandle & operator=(Pad_ObjHandle &objHandle); // Assignment operator

            void           Add_tag(char *tag);          // Add the specified tag
            Pad_Bool       Append_coords(Pad_PList &pts);  // Append coordinates of objects
            Pad_Bool       Attach(Pad_Object *obj);     // Set object referred to by this handle
            Pad_Bool       Attach(Pad_ObjHandle &objHandle); // Attach to object(s) referenced by handle
            Pad_Bool       Attach(Pad_SurfaceHandle &surface, int id);    // Set object by id
            Pad_Bool       Attach(Pad_SurfaceHandle &surface, char *tagOrId); // Set object by tagOrId
            Pad_Bool       Attach(char *surfaceName, char *tagOrId);
            Pad_Bool       Attach(char *surfaceName, int id);    // Set object by id
    virtual void           Damage(void);                // Force this object to be redrawn
            void           Delete_obj(void);            // Delete object associated with this handle
            void           Delete_tag(char *tag);       // Delete the specified tag
            void           Get_bbox(Pad_BBox &bbox)   const; // Returns global bbox of objects
            void           Get_coords(Pad_PList &pts) const; // Returns coordinates of object
    virtual Pad_HandleType Get_handle_type(void) const;      // Return type of this handle
            int            Get_id(void)               const; // Return unique id reference for this object
            void           Get_group(Pad_ObjHandle &group) const; // Return group this object is a member of
            Pad_ObjectType Get_object_type(void)      const; // Return type of object
            void           Get_surface(Pad_SurfaceHandle &surface) const; // Return surface handle associated with
            void           Get_tags(Pad_List &tags)   const; // Fill in list with tags of object
    virtual Pad_Bool       Is_mapped(void);             // Returns true if handle is mapped to an object or tag
            Pad_Bool       Lower(void);	                // Lower objects to bottom of drawing order
            Pad_Bool       Lower(Pad_ObjHandle &ref);   // Lower objects below reference
            Pad_Bool       Raise(void);	                // Raise objects to top of drawing order
            Pad_Bool       Raise(Pad_ObjHandle &ref);   // Raise objects above reference
            Pad_Bool       Rotate(float theta);	        // Rotate objects about their anchor points
            Pad_Bool       Rotate(float theta, Pad_Point &center); // Rotate objects about <center>
            Pad_Bool       Scale(float x, float y, float zmult); // Scale objects
            Pad_Bool       Slide(float dx, float dy);   // Slide objects in the plane
            Pad_Bool       Set_coords(Pad_PList &pts);  // Set coordinates of objects
            Pad_Bool       Set_option(Pad_OptionType optionType, ...);    // Set an option of an object
            Pad_Bool       Set_options(Pad_OptionType optionType, ...);   // Set multiple options of an object
            void           Set_render_callback(Pad_Callback *callback);      // Set a callback to be executed when object is rendered
            Pad_Bool       Set_transform(float x, float y, float s);      // Set transform of objects
            Pad_Bool       Set_tags(Pad_List &tags);    // Set tags of objects
            Pad_Bool       Set_tags(char *tags);        // Set tags of objects

      // Methods related to event bindings
            Pad_Bool       Bind(char *event, Pad_Callback *callback);
            Pad_Bool       Get_binding(char *event) const;
            Pad_Bool       Get_all_bindings(void)   const;
            Pad_Bool       Delete_binding(char *event);
            void           Delete_all_bindings(void);


  protected:
    friend class Pad_SurfaceHandle;
    friend class Pad_ViewHandle;
    friend class Pad_AnimChannel;
    friend class Pad_Event;

    Pad_Bool     _Set_option(Pad_OptionType, va_list &);
    Pad_Object * _Get_first_object(void) const;     // Return first object referred to by this handle, or NULL if none
    Pad_Object * _Get_object(void)       const;     // Return object specified by handle, or NULL if handle specifies tag (or nothing)
    char *       _Get_tag(void)          const;     // Return tag specified by handle, or NULL if handle specifies object (or nothing)
};

/*
 *----------------------------------------------------------------------
 *
 * class Pad_SurfaceHandle --
 *
 *    A class for identifying and manipulating a Pad Surface.
 * 
 *    Note that the _obj slot in Pad_Handle actually stores
 *    a pointer to a Pad (which is a Pad_Object).
 *
 *----------------------------------------------------------------------
 */

class Pad_SurfaceHandle : public Pad_Handle
{
  public:
    Pad_SurfaceHandle();
    Pad_SurfaceHandle(char *surfaceName);
    Pad_SurfaceHandle(Pad_SurfaceHandle &surfaceHandle);
    Pad_SurfaceHandle(Pad_WinHandle &winHandle);
    virtual ~Pad_SurfaceHandle();
    Pad_SurfaceHandle & operator=(Pad_SurfaceHandle &surfaceHandle); // Assignment operator

            Pad_Bool       Attach(char *surfaceName);     // Set surface this handle references
            Pad_Bool       Attach(Pad_SurfaceHandle &surfaceHandle);
            Pad_Bool       Attach(Pad_WinHandle &winHandle);
            Pad_Bool       Attach(Pad *pad);
            Pad_Bool       Create_obj(Pad_ObjectType type);
            Pad_Bool       Create_obj(Pad_ObjectType type, Pad_ObjHandle &handle);
            Pad_Bool       Create_obj(Pad_ObjectType type, Pad_ObjHandle &handle, Pad_OptionType optionType, ...);

            Pad_Bool       Create_anim_path();
            Pad_Bool       Create_anim_path(Pad_AnimPathHandle &pathHandle);

            Pad_Bool       Create_anim_channel();
            Pad_Bool       Create_anim_channel(Pad_AnimChannelHandle &chanHandle);

            Pad_Bool       Create_animation();
            Pad_Bool       Create_animation(Pad_AnimationHandle &animHandle);

    virtual void           Damage(void);                  // Force this surface to be redrawn
            void           Find_with_tagorid(char *tagOrId, Pad_HandleList &handles, Pad_Bool groupMembers = FALSE);
    virtual Pad_HandleType Get_handle_type(void) const;   // Return type of this handle
    virtual char *         Get_name(void);                // Return the name of the pad
            Pad_OptionType Get_option_type(char *optionName);
    virtual Pad_Bool       Is_mapped(void);               // Returns true if handle is mapped to a surface

				// Methods related to event bindings
            void           Add_binding_modifier(char *modifier);
            void           Delete_binding_modifier(char *modifier);
            char *         Get_binding_modifier(void);
            Pad_Bool       Set_binding_modifier(char *modifier);

  private:
    friend class Pad_ObjHandle;

    Pad *  _Get_pad(void);           // Return pad surface specified by handle, or NULL if none
};

/*
 *----------------------------------------------------------------------
 *
 * class Pad_ViewHandle --
 *
 *    A class for identifying and manipulating a Pad++ View.
 *    A View specifies what portion of a Surface is visible.
 *    Every Window and every Portal has a view.
 * 
 *----------------------------------------------------------------------
 */

class Pad_ViewHandle : public Pad_Handle
{
  public:
    Pad_ViewHandle();
    Pad_ViewHandle(Pad_ViewHandle &viewHandle);
    Pad_ViewHandle(char *surfaceName);
    virtual ~Pad_ViewHandle();
    Pad_ViewHandle & operator=(Pad_ViewHandle &viewHandle); // Assignment operator

            Pad_Bool       Attach(char *surfaceName);       // Set view referred to by this handle
            Pad_Bool       Attach(Pad_ObjHandle &objHandle);
            Pad_Bool       Attach(Pad_ViewHandle &viewHandle);
            Pad_Bool       Attach(Pad_View *view);
    virtual Pad_HandleType Get_handle_type(void) const;     // Return type of this handle
            Pad_Bool       Get_view(float &xview, float &yview, float &zoom);
            Pad_Bool       Set_view(float xview, float yview, float zoom, 
				    int animationTime=0, Pad_Bool twoStep=FALSE); // Set the position that this view can see
            void           Set_visiblelayers(char *layers);

  private:
    Pad_View *_Get_view(void);           // Return Pad++ view specified by handle, or NULL if none
};

/*
 *----------------------------------------------------------------------
 *
 * class Pad_WinHandle --
 *
 *    A class for identifying and manipulating a Pad++ Window.
 *    A Window is the link to the operating system window that the
 *    Pad++ widget fits within.
 * 
 *----------------------------------------------------------------------
 */

class Pad_WinHandle : public Pad_Handle
{
  public:
    Pad_WinHandle();
    Pad_WinHandle(Pad_WinHandle &winHandle);
    Pad_WinHandle(char *surfaceName);
    virtual ~Pad_WinHandle();
    Pad_WinHandle & operator=(Pad_WinHandle &winHandle); // Assignment operator

            Pad_Bool       Attach(Pad_WinHandle &winHandle);
            Pad_Bool       Attach(char *surfaceName);      // Set window this handle references
            Pad_Bool       Attach(Pad_Win *win);
    virtual Pad_HandleType Get_handle_type(void) const;    // Return type of this handle
    virtual char *         Get_name(void);                 // Return the name of the pad

  private:
    Pad_Win *_Get_win(void);           // Return Pad++ window specified by handle, or NULL if none
};

/*
 *----------------------------------------------------------------------
 *
 * class Pad_UniverseMgr --
 *
 *    A class for interacting with the entire Pad system.
 *    There is a single global instance of the Pad_UniverseMgr
 *    that is accessed through the Get_instance() method.
 *
 *    Note that a window can be created via the Universe Manager.
 *    This currently only works when hooked up to Tcl/Tk.
 *    After the window is created, it must still be managed
 *    by a Tk Geometry Manager such as pack, or grid.
 *    This can be effected from C with a command such as:
 *      Tcl_Eval(interp, "pack .pad");
 * 
 *----------------------------------------------------------------------
 */

class Pad_UniverseMgr 
{
  public:
    static Pad_UniverseMgr *Get_instance(void);  // The access method for using the universe manager
    Pad_Bool Create_win(Tcl_Interp *interp, char *surfaceName);  // Create a new window
    Pad_Bool Create_win(Tcl_Interp *interp, char *surfaceName, Pad_WinHandle &win);
    void     Get_surfaces(Pad_HandleList &handles);    // Retrieve all Pad++ surfaces
            //
            // Sound Device related commands
            //
    int       Set_volume(int vol);               // Sets the sound volume right and left channels both to vol
    int       Set_volume(int left, int right);   // Sets the sound volume right and left channels seperately
    const Pad_List *Get_sounds(void);            // Get a list of the sounds currently loaded
    int       Get_volume(int &left, int &right); // set left and right to their current levels from the audio device
    Pad_Bool Stop(void);                         // Stops the currently playing sound
  private:
    static Pad_UniverseMgr * _instance;
};



//// Not implemented yet...

/*
class Pad_TextHandle : public Pad_ObjHandle
{
  public:
    Pad_TextHandle();
    Pad_TextHandle(Pad_Text *text);
    virtual ~Pad_TextHandle();

            void           Attach(Pad_Text *text);      // Set text object referred to by this handle
    virtual Pad_HandleType Get_handle_type(void);       // Return type of this handle
};
*/

/*
 *----------------------------------------------------------------------
 *
 * class Pad_AnimPathHandle --
 *
 *    A handle to an animtion path. 
 *
 *---------------------------------------------------------------------- */

// dt anchor to search
class Pad_AnimPathHandle : public Pad_Handle
{
  public:
    Pad_AnimPathHandle();
    Pad_AnimPathHandle(Pad_AnimPathHandle &pathHandle);
    Pad_AnimPathHandle(Pad_AnimPath *path);

    virtual ~Pad_AnimPathHandle();
    Pad_AnimPathHandle & operator=(Pad_AnimPathHandle &pathHandle); // Assignment operator

            Pad_Bool       Attach(Pad_AnimPath *path);              // Set path referred to by this handle
            Pad_Bool       Attach(Pad_AnimPathHandle &pathHandle);  // Attach to path referenced by handle

            Pad_Bool       Delete_path(void);                       // Delete path associated with this handle

    virtual Pad_HandleType Get_handle_type(void) const;             // Return type of this handle
    virtual Pad_Bool       Is_mapped(void) const;                   // Returns true if handle is 
                                                                    //  mapped to an path or tag

            Pad_Bool       Set_begin_time(float beginTime);
            Pad_Bool       Set_end_time(float endTime);
            Pad_Bool       Set_pre(Pad_AnimPath::PRE_POST pre);
            Pad_Bool       Set_post(Pad_AnimPath::PRE_POST post);
            Pad_Bool       Set_siso(Pad_Bool siso);  
            Pad_Bool       Set_in_time(float inTime);
            Pad_Bool       Set_out_time(float outTime);
            int            Set_path_data(int num, Pad_AnimPath::anim_data* data, int pathOrder, Pad_Bool timeData=FALSE);

            int            Get_path_data(Pad_AnimPath::anim_data *&data) const; // Returns data defining path
            Pad_Bool       Get_interp_value(float time, Pad_AnimPath::datum3 &interpVal);
            int            Get_order()          const;
            float          Get_begin_time()     const;
            float          Get_end_time()       const;
            Pad_Bool       Get_time_path_flag() const;
            Pad_Bool       Get_siso()           const;
            float          Get_in_time()        ;
            float          Get_out_time()       ;

    enum    Pad_AnimPath::PRE_POST    Get_pre()  const;
    enum    Pad_AnimPath::PRE_POST    Get_post() const;

            Pad_Bool       Has_data();
    
    static  Pad_Bool       Path_exists(Pad_String const pathToken);

  protected:
    friend class Pad_SurfaceHandle;
    friend class Pad_Event;
    friend class Pad_AnimChannelHandle;


    Pad_AnimPath*  _Get_path(void) const;  // Return path specified by handle, or NULL if handle is not mapped
                                           // Dangerous to use because you have a pointer to an internal object
                   
};

/*
 *----------------------------------------------------------------------
 *
 * class Pad_AnimChannelHandle --
 *
 *    A handle to an animtion channel. 
 *
 *---------------------------------------------------------------------- */

class Pad_AnimChannelHandle : public Pad_Handle
{
  public:
    Pad_AnimChannelHandle();
    Pad_AnimChannelHandle(Pad_AnimChannelHandle &channelHandle);
    Pad_AnimChannelHandle(Pad_AnimChannel *channel);

    virtual ~Pad_AnimChannelHandle();
    Pad_AnimChannelHandle & operator=(Pad_AnimChannelHandle &channelHandle); // Assignment operator

            Pad_Bool       Attach(Pad_AnimChannel *channel);              // Set path referred to by this handle
            Pad_Bool       Attach(Pad_AnimChannelHandle &channelHandle);  // Attach to path referenced by handle

            Pad_Bool       Delete_channel(void);           // Delete channel associated with this handle

    virtual Pad_HandleType Get_handle_type(void) const;    // Return type of this handle
    virtual Pad_Bool       Is_mapped(void) const;          // Returns true if handle is mapped 
                                                           //  to an channel or tag

            Pad_Bool       Set_path_begin_time(float beginTime);
            Pad_Bool       Set_path_end_time(float endTime);
            Pad_Bool       Set_path(Pad_AnimPathHandle const &pathHandle);  
            Pad_Bool       Set_option(Pad_OptionType option);
            Pad_Bool       Set_object(Pad_ObjHandle &objHandle);
            Pad_Bool       Set_interrupt(Pad_Bool intpt);
            Pad_Bool       Set_time_resolution(int res);


            int            Get_path_order()      const;
            float          Get_path_begin_time() const;
            float          Get_path_end_time()   const;

            Pad_String     Get_path_token()       const;
            Pad_String     Get_object_tag_or_id() const; // returns channel object tag or id in string form
            Pad_Bool       Get_object(Pad_ObjHandle &objHandle);
            Pad_OptionType Get_option()           const;
            int            Get_time_resolution()  const;
            Pad_Bool       Get_interp_val(float time, Pad_AnimPath::datum3 &interpVal);


            Pad_Bool       Has_path()   const; 
            Pad_Bool       Has_object() const;
            Pad_Bool       Has_option() const;
    
    static  Pad_Bool       Channel_exists(Pad_String chanToken); // checks to see if key (chanToken) cooresponds
                                                                 // to a channel in hash table
            Pad_Bool       Play();
            
  protected:
    friend class Pad_SurfaceHandle;
    friend class Pad_Event;
    friend class Pad_AnimPath;
    friend class Pad_AnimChannel;

    Pad_AnimChannel*  _Get_channel(void) const;  // Return pointer to channel specified by handle, 
                                                 //or NULL if handle is not mapped

};

/*
 *----------------------------------------------------------------------
 *
 * class Pad_AnimationHandle --
 *
 *    A handle to an animtion. 
 *
 *---------------------------------------------------------------------- */

class Pad_AnimationHandle : public Pad_Handle
{
  public:
    Pad_AnimationHandle();
    Pad_AnimationHandle(Pad_AnimationHandle &animHandle);
    Pad_AnimationHandle(Pad_Animation *animation);

    virtual ~Pad_AnimationHandle();
    Pad_AnimationHandle & operator=(Pad_AnimationHandle &animHandle); // Assignment operator

            Pad_Bool       Attach(Pad_Animation *animation);          // Set path referred to by this handle
            Pad_Bool       Attach(Pad_AnimationHandle &animHandle);   // Attach to path referenced by handle

            Pad_Bool       Delete_animation(void);         // Delete animation associated with this handle

    virtual Pad_HandleType Get_handle_type(void) const;    // Return type of this handle
    virtual Pad_Bool       Is_mapped(void) const;          // Returns true if handle is mapped 
                                                           //  to an animation

            Pad_Bool       Set_begin_time(float beginTime);
            Pad_Bool       Set_end_time(float endTime);
            Pad_Bool       Set_interrupt(Pad_Bool intpt);
            Pad_Bool       Set_time_resolution(int res);
            Pad_Bool       Set_speed_factor(float speedF);

            float          Get_begin_time()         const;
            float          Get_end_time()           const;
            Pad_String     Get_animatable_handles() const;
            int            Get_time_resolution()    const;
            Pad_Bool       Get_interrupt()          const;
            float          Get_speed_factor()       const;
            Pad_Bool       Get_animatable_tokens(Pad_String &memberTokens)  const;

            // The Add_member() and Remove_member() methods must be overloaded for each
            // animatable type. So far only channels are animatable.
            Pad_Bool       Add_member(Pad_AnimChannelHandle    &chanHandle);
            Pad_Bool       Add_member(Pad_AnimationHandle      &animHandle);
            Pad_Bool       Remove_member(Pad_AnimChannelHandle &chanHandle);

            Pad_Bool       Play();
    
    static  Pad_Bool       Animation_exists(Pad_String chanToken); // checks to see if key (chanToken) cooresponds


            
  protected:
    friend class Pad_SurfaceHandle;
    friend class Pad_Event;
    friend class Pad_AnimPath;
    friend class Pad_AnimChannel;
    friend class Pad_Animatable;

    Pad_Animation*  _Get_animation(void) const;  // Return pointer to channel specified by handle, 
                                                 //or NULL if handle is not mapped
};


////////////////////////////////////////////////////////////////
//
// Pad_SoundHandle Class
//
//
////////////////////////////////////////////////////////////////

class Pad_SoundHandle : public Pad_Handle {
public:
    friend ostream& operator<<(ostream&, Pad_SoundHandle&);
        
    Pad_SoundHandle();
    ~Pad_SoundHandle();

    Pad_Bool  Attach(Pad_Sound*); // Attach an existing Pad_Sound
    Pad_Bool  Load(char *path);   // Load a sound from disk and associate it with the handle
    virtual Pad_HandleType Get_handle_type(void) const; // Return type of this handle
    int   Get_channels(void); // Returns the number of channels that the sound was recorded in
    int   Get_rate(void); // Returns the sample rate the sound was recorded with
    int   Get_bits_per_sample(void); // Returns the sample format (8 or 16 bit) the sound was recorded with
    int   Get_length(void); // Returns the number of samples in the sound
    Pad_Bool  Play(Pad_Bool loop); // Play an already loaded sound
    Pad_Bool  Play(Pad_Bool loop, int volume); // Play a sound, temporarily resetting the volume first
    Pad_Bool  Create(Pad_SoundType); // Create a parameterized sound
    Pad_Bool  Configure(int, Pad_SoundOption*); // Configure a parameterized sound
    Pad_Bool Stop(void); // Stop playing all sounds associated with this sound
    Pad_Bool Play(Pad_SoundConfig*); // Play a sound
    Pad_Bool Configure(Pad_List&); // Configure a parameterized sound
    Pad_Bool Get_configuration(Pad_List&) const; // Get all the current configurations for a sound
    Pad_Bool Get_configuration(Pad_SoundOption&) const; // Get a single configuration for a sound
};


//////////////////////////////////////////////////////////////////////
//
// The following is an alphabetical list of all publicly available
// procedures for accessing Pad++.  All global symbols used by Pad++
// start with 'Pad_'.  All procedures start with 'Pad_' optionally
// followed by a short group name.
//
//////////////////////////////////////////////////////////////////////

//
// These two commands provide the string-based generic interface to Pad++ used by Tcl.
// These can also be used by other languages to avoid creating special code to access
// all the detailed API commands.  These are implemented by parsing the strings, and
// calling the API commands.  Pad_Command and Pad_Configure are in pad-tcl.cpp
//
int Pad_Command(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[]);
int Pad_Configure(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[]);

//
// DOHANDLE is a macro that is convenient for implementing the body
// of handle methods.  Called with a pointer to a handle, an empty
// pointer to an object, and some C++ code in <__body>, it executes
// <__body> once for each object specified by the handle where <__obj>
// is set to that object for each iteration.
//
#define DOHANDLE(__handle, __obj, __body) \
    { \
	if (__handle->_tag) { \
	    Pad_List __objs; \
	    Pad_Iterator __oi; \
	    Pad *__pad = (Pad *)__handle->_obj; \
	    if (__pad) { \
	        __pad->Find_with_tagorid(__handle->_tag->Get(), __objs, TRUE); \
	        DOLIST(__oi, __objs, Pad_Object, __obj) { \
 		    __body \
	        } \
	    } \
        } else if (__handle->_obj) { \
  	    __obj = __handle->_obj; \
	    __body \
	} \
    }

//
// DOHANDLENOGROUPS is just like DOHANDLE, but if the handle is specified
// by a tag, then group members are not iterated over.
//
#define DOHANDLENOGROUPS(__handle, __obj, __body) \
    { \
	if (__handle->_tag) { \
	    Pad_List __objs; \
	    Pad_Iterator __oi; \
	    Pad *__pad = (Pad *)__handle->_obj; \
	    if (__pad) { \
	        __pad->Find_with_tagorid(__handle->_tag->Get(), __objs, FALSE); \
	        DOLIST(__oi, __objs, Pad_Object, __obj) { \
 		    __body \
	        } \
	    } \
        } else if (__handle->_obj) { \
  	    __obj = __handle->_obj; \
	    __body \
	} \
    }

//
// DOHANDLESAFE is just like DOHANDLE, but is safe in that every object
// is checked for validity just before it is used.  This is somewhat
// slower than DOHANDLE, but should used if there is any chance that
// the operation one object could possibly delete another (I.e., if it
// could result in a callback being executed).
//
#define DOHANDLESAFE(__handle, __obj, __body) \
    { \
	if (__handle->_tag) { \
	    int __j; \
	    Pad_IList __ids; \
	    Pad *__pad = (Pad *)__handle->_obj; \
	    if (__pad) { \
	        __pad->Find_ids_with_tagorid(__handle->_tag->Get(), __ids, TRUE); \
	        for (__j=0; __j<__ids.Length(); __j++) { \
		    __obj = __pad->Get_object_from_id(__ids.Nth(__j)); \
		    if (__obj) { \
 		      __body \
		    } \
	        } \
	    } \
        } else if (__handle->_obj) { \
  	    __obj = __handle->_obj; \
	    __body \
	} \
    }

//
// FIRSTHANDLEOBJ is a macro that fills in <__obj> with the first
// object specified by the handle in display-list order, or NULL
// if there isn't one.
//
#define FIRSTHANDLEOBJ(__handle, __obj) \
    { \
	if (__handle->_tag) { \
	    Pad_List __objs; \
	    Pad *__pad = (Pad *)__handle->_obj; \
	    if (__pad) { \
	        __pad->Find_with_tagorid(__handle->_tag->Get(), __objs, TRUE); \
		__obj = (Pad_Object *)__objs.First(); \
	    } else { \
                __obj = NULL; \
	    } \
        } else if (__handle->_obj) { \
  	    __obj = __handle->_obj; \
        } else { \
            __obj = NULL; \
        } \
    }

#endif    // Endif API_H

