/*
"(c) Copyright 1993-1997 Pad++ Consortium {University of New Mexico (UNM),
and New York University (NYU)}, All Rights Reserved."  
Licensee can not remove or obscure any of the
copyright or trademark notices in this software.

IN NO EVENT SHALL THE AUTHORS OR DISTRIBUTORS BE LIABLE TO ANY PARTY
FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
ARISING OUT OF THE USE OF THIS SOFTWARE, ITS DOCUMENTATION, OR ANY
DERIVATIVES THEREOF, EVEN IF THE AUTHORS HAVE BEEN ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.

THE AUTHORS AND DISTRIBUTORS SPECIFICALLY DISCLAIM ANY WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE, AND NON-INFRINGEMENT.  THIS SOFTWARE
IS PROVIDED ON AN "AS IS" BASIS, AND THE AUTHORS AND DISTRIBUTORS HAVE
NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR
MODIFICATIONS.

See the file "License" for general information on usage and
redistribution, and the file "LicenseTerms" for the specific license
agreement on usage and redistribution of this file, and the Pad++
software in general.
*/

#include <fstream.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <fcntl.h>
#ifdef PAD_UNIX
#  include <dirent.h>
#endif
#include <errno.h>
#include "cache.h"
#include "global.h"
#include "image.h"
#include "imagedata.h"
#include "pad.h"

#ifdef PAD_UNIX
#  include <unistd.h>
#  include <sys/param.h>
#endif

#ifdef PAD_WIN
#  include <Direct.h>
#  define MAXPATHLEN      1024 
typedef unsigned long   pid_t;
#  include <process.h>
#endif

#define CACHEOUT_REQ (1<<1)
#define CACHEIN_REQ  (1<<2)
#define INMEMORY     (1<<3)
#define WRITTENOUT   (1<<4)

//
// Structure for info the cache manager maintains
// for objects it know about.
//
struct Pad_CacheInfo {
  int              cacheStatus;
  Pad_Time         time;
  Pad_Bool         fileWritten;
  Pad_String       file;
  Pad_Object       *obj;
};

static Pad_Bool _dirOk=FALSE;

//
// Cache timer callbak.
// Calls the update cache mehtod and reschedules another callback.
//
static void
Pad_UpdateCache(ClientData clientData)
{
  //cout << "in Pad_UpdateCache" << endl;
  Pad_CacheMgr *cacheMgr = (Pad_CacheMgr *)clientData;
  cacheMgr->UpdateCache(NULL, NULL);
  Pad_DeleteTimerHandler(cacheMgr->_updateTimer);
  cacheMgr->_updateTimer = Pad_CreateTimerHandler(cacheMgr->UpdateDelay(), 
						  Pad_UpdateCache, cacheMgr);
}

static int
CacheInfo_list_sort(const void *item1, const void *item2);

//
// The cache manager constructor.
// Create a cache directory for this process and initialize
// max cache size and the hashtable of cache info for objects.
// The cache directory is dir/pid.  Where dir is a given directory 
// (caching is disbaled if it is null).
//
Pad_CacheMgr::Pad_CacheMgr(float cachesize, char *dir, Pad_Bool disable, long delayTime)
  :_usedBytes(0),
   _cacheInfoTable(PAD_VOID_TABLE),
   _noCaching(disable)
{
  _updateTimer = NULL;
  Size(cachesize);
  _infoArrSize = 10;
  _infoArr = new Pad_CacheInfo*[_infoArrSize];
  _infoArrLen = 0;
  Dir(dir);
  ViewMultiple(2);
  UpdateDelay(delayTime);
}

//
// Remove the process cache dir and everything in it.
//
Pad_CacheMgr::~Pad_CacheMgr()
{
  Pad_HashTableIterator iter;
  Pad_CacheInfo *info;

  DOTABLE(iter, _cacheInfoTable, Pad_CacheInfo, info) {
    delete info;
  }

  if ( (_cacheDir.Length() > 0) && !RmDir(_cacheDir.Get()) ) {
    cerr << "ERR: COULD NOT REMOVE CACHE DIR: " << _cacheDir.Get() << endl;
#if defined(sun)
      cerr << strerror(errno) << endl;
#else
      cerr << sys_errlist[errno] << endl;
#endif
  }

  delete[] _infoArr;
  Pad_DeleteTimerHandler(_updateTimer);
}


//
// Handle the cachein request for the object.  
// Make sure the object is in memory.  If it is cached out then
// read its cache file and reconstruct the object.  If memory
// is needed then cache out other objects if possible.
//
Pad_Bool
Pad_CacheMgr::CacheIn(Pad_Object *obj)
{
  Pad_Bool rc=TRUE, firstReq=FALSE;

  _pad = obj->pad;

  // bail out if caching is disabled
  if ( _noCaching || !(obj->flags & PAD_MASK_CACHEDOUT)) {
    return TRUE;
  }

  // get caching info for this object
  Pad_CacheInfo *info = (Pad_CacheInfo*)_cacheInfoTable.Get(obj);
  if (!info) {
    info = Allocate_Info(obj, INMEMORY);
    firstReq = TRUE;
  }

  // update object cache info fields with this request and its time.
  info->cacheStatus |= CACHEIN_REQ;
  info->cacheStatus &= ~CACHEOUT_REQ;
  info->time.Update();

  // make sure there's enough memory for the the object and cache out other
  // objects if possible. If object is cached out then reconstruct it from 
  // its cache file.  Otherwise, just keep it in memory.

  UpdateCache(obj, info);
  if (obj->flags & PAD_MASK_CACHEDOUT) {  // bring it back
    if (Reconstruct(obj, info)) {
      obj->flags &= ~PAD_MASK_CACHEDOUT;
      info->cacheStatus |= INMEMORY;
      info->cacheStatus &= ~WRITTENOUT;
      _usedBytes += CacheSize(obj);
    }
  } else {
    info->cacheStatus |= INMEMORY;
    info->cacheStatus &= ~WRITTENOUT;
    if ( firstReq ) {
      _usedBytes += CacheSize(obj);
    }
  }

  return rc;
}

//
// Handle cacheout request for the object.
// The default algorthim is as follow:
//   - if there's enough memory available then keep object in memory.
//   - otherwise, pick the "best" objects for caching out.
//
Pad_Bool
Pad_CacheMgr::CacheOut(Pad_Object *obj)
{
  Pad_Bool rc=TRUE;

  _pad = obj->pad;

  if ( _noCaching ) {
    return TRUE;
  }

  // only do cache out for images
  if (obj->Type() != PAD_IMAGE) {
    rc = FALSE;

  } else if (obj->flags & PAD_MASK_CACHEDOUT) {
    // no sweat if object is already cached out
    rc = TRUE;

  } else {
    // if there's memory to spare then keep obj in memory and 
    // tag it as accounted for.  Otherwise, write it out and
    // deallocate its memory intensice parts.
    Pad_CacheInfo *info = (Pad_CacheInfo*) _cacheInfoTable.Get(obj);
    Pad_Bool firstReq=FALSE;

    if (info == NULL) {
      info = Allocate_Info(obj, INMEMORY);
      firstReq = TRUE;
    }

    info->cacheStatus |= CACHEOUT_REQ;
    info->cacheStatus &= ~CACHEIN_REQ;
    
    if (UpdateDelay() == 0) {
      UpdateCache(obj, info);
      if (firstReq) {
	if ( !(obj->flags & PAD_MASK_CACHEDOUT) ) {
	  _usedBytes += CacheSize(obj);
	}
      }
    }
  }

  return rc;
}

//
// Remove all cache related info for object.
// This should be called when object is deleted.
//
void
Pad_CacheMgr::Forget(Pad_Object *obj)
{
    Pad_CacheInfo *info = (Pad_CacheInfo*) _cacheInfoTable.Get(obj);
    if (info != NULL) {
      _cacheInfoTable.Remove(obj);
      if (info->fileWritten) {
	unlink(info->file.Get());
      }
      int i;
      for (i=0; (i < _infoArrLen) && (_infoArr[i]->obj != obj); i++);
      for (; i < (_infoArrLen-1); i++) {
	_infoArr[i] = _infoArr[i+1];
      }
      
      _infoArrLen--;
      delete info;
    }
}


//
// Update the cache for operation on the object.
// If caching in and there's not enough memory or caching out 
// and there's no memory left then select appropriate objects
// and cache them out.
//
Pad_Bool
Pad_CacheMgr::UpdateCache(Pad_Object *obj, Pad_CacheInfo *info)
{
    long bytesLeft = _cacheSize - _usedBytes;
    long objSize;
    Pad_Bool rc=TRUE;

    if ( _noCaching ) {
      return rc;
    }

    if (obj) {
      objSize = CacheSize(obj);
    }

    if (obj == NULL || 
	(((info->cacheStatus & CACHEIN_REQ) && (objSize > bytesLeft) && 
	  (obj->flags & PAD_MASK_CACHEDOUT)) ||

	 ((info->cacheStatus & CACHEOUT_REQ) && (bytesLeft < 0)))) {

      // cache out objects to make room for atleast the object size
      int i, n=0;
      Pad_Bool out=FALSE;
      Pad_CacheInfo **infos = SelectObjs(obj, n);

      for (i = 0; i < n; i++) {
	if (infos[i]->obj->Type() == PAD_IMAGE) {
	  out = CacheOut((Pad_Image*)((infos[i])->obj), infos[i]);
	} else {
	  out = FALSE;
	}
	if (out) {
	  _usedBytes -= CacheSize((infos[i])->obj);
	  (infos[i])->obj->flags |= PAD_MASK_CACHEDOUT;
	  (infos[i])->cacheStatus |= WRITTENOUT;
          (infos[i])->cacheStatus &= ~INMEMORY;
	}
      }

      delete[] infos;
    }

    return rc;
}

//
// Select objects for caching out that amount atleast to the size of a
// given object .  Selection criterias are:
//  - object is currently in memory but had a cache out request
//  - object is not visible within the cache view bbox.
// Selceted objects are considered in order of last render time.  This
// corresponds to a least-recently-used policy.
// 
Pad_CacheInfo**
Pad_CacheMgr::SelectObjs(Pad_Object *obj, int &numSelected)
{
  Pad_CacheInfo **infosSelected = new Pad_CacheInfo*[_infoArrLen];
  Pad_CacheInfo *info;
  int i=0;
  long needed, sum=0;
  Pad_BBox bbox;
  numSelected = 0;

  if (obj) {
    needed = CacheSize(obj);
  } else {
    long bytesLeft = _cacheSize - _usedBytes;
    if (bytesLeft < 0) {
      needed = -bytesLeft;
    } else {
      return NULL;
    }
  }

  // sort objects known by the cache manager by their last render time
  qsort(_infoArr, _infoArrLen, sizeof(Pad_CacheInfo*), CacheInfo_list_sort);

  // get the view bbox for visiblity criteria
  ViewBBox(bbox);

  // go through the objects and select objects that are not in the
  // the main view and nor in any portals.
  while ((i < _infoArrLen) && (sum < needed)) {
    info = _infoArr[i];
    if ((info->cacheStatus & INMEMORY) &&
	(info->cacheStatus & CACHEOUT_REQ)) {
      if (info->obj->Is_visible(bbox.Get(), info->obj->pad->view) == FALSE) {
	if (info->obj->Is_visible() == FALSE) {
	  infosSelected[numSelected] = info;
	  numSelected++;
	  sum += CacheSize(info->obj);
	}
      }
    }
    
    i++;
  }

  return infosSelected;
}

//
// Setup the view bbox for cache manager to use for
// object visibility crieteria.
//
void
Pad_CacheMgr::ViewBBox(Pad_BBox &bbox)
{
  // use the main view bbox
  //bbox = Pad_prc->win->view->viewBBox;
  if ( _pad ) {
    bbox = _pad->view->viewBBox;
  } else {
    bbox.Set_x(0,0);
    bbox.Set_y(0,0);
  }

  // apply the cache mananger's view multiple to it
  float w = bbox.Width() * _viewMultiple;
  float h = bbox.Height()  * _viewMultiple;
  float xmin = bbox.Xctr() - w/2.0;
  float xmax = bbox.Xctr() + w/2.0;
  float ymin = bbox.Yctr() - h/2.0;
  float ymax = bbox.Yctr() + h/2.0;
  bbox.Set_x(xmin, xmax);
  bbox.Set_y(ymin, ymax);
}

//
// Comparator for object info entries for cache out selection.
// Currently, just considers the last render time.  But this
// could be extended to take distance from view into consideration.
//
static int
CacheInfo_list_sort(const void *item1, const void *item2)
{
  Pad_CacheInfo *info1=*((Pad_CacheInfo**)item1);
  Pad_CacheInfo *info2=*((Pad_CacheInfo**)item2);

  // compare by render time second and u-second.
  long sec1 = info1->time.Get_sec();
  long sec2 = info2->time.Get_sec();
  long diff = sec1 - sec2;
  if (diff == 0) {
    long usec1 = info1->time.Get_usec();
    long usec2 = info2->time.Get_usec();
    diff = usec1 - usec2;
  }

  return (int)diff;
}

//
// Rebuild the object using its cache file.  Currently this is only supported
// for images.  Other types can be added though.
//
Pad_Bool
Pad_CacheMgr::Reconstruct(Pad_Object *obj, Pad_CacheInfo *info)
{
  Pad_Bool rc=TRUE;
  switch (obj->Type()) {
  case PAD_IMAGE:
    rc = CacheIn((Pad_Image*)obj, info);
    break;
  default:
    rc = FALSE;
    break;
  }

  return rc;
}

//
// Create and initialize a cache info structure for the object.
//
Pad_CacheInfo*
Pad_CacheMgr::Allocate_Info(Pad_Object *obj, int status)
{
  Pad_CacheInfo *info = new Pad_CacheInfo;
  info->cacheStatus = status;
  info->fileWritten = FALSE;
  Pad_String file;
  CacheFile(obj, file);
  info->file = file;
  info->obj = obj;
  memset(&info->time, 0, sizeof(Tcl_Time));
  _cacheInfoTable.Set(obj, info);

  if (_infoArrLen == _infoArrSize) {
    _infoArrSize += _infoArrSize;
    Pad_CacheInfo **newArr = new Pad_CacheInfo* [_infoArrSize];
    for (int i = 0; i < _infoArrLen; i++) {
      newArr[i] = _infoArr[i];
    }
    delete[] _infoArr;
    _infoArr = newArr;
  }

  _infoArr[_infoArrLen] = info;
  _infoArrLen++;
  
  return info;
}

//
// Compute the total cache size/worth of the object.
// Currently this is only done for images.
//
long
Pad_CacheMgr::CacheSize(Pad_Object *obj)
{
  long len=0;
  Pad_CacheInfo *info = (Pad_CacheInfo*)_cacheInfoTable.Get(obj);
  
  switch (obj->Type()) {
  case PAD_IMAGE:
    len = CacheSize((Pad_Image*)obj, info);
    break;

  default:
    break;
  }

  return len;
}

//
// Get the object cache file.
//
void
Pad_CacheMgr::CacheFile(Pad_Object *obj, Pad_String &file)
{
  file = _cacheDir;
  file += "/";
  file += obj->id;
}

//
// Set the caching enable/disable flag.
//
void
Pad_CacheMgr::Cache(Pad_Bool status)
{
  if ( !_noCaching && !status ) {
    Pad_DeleteTimerHandler(_updateTimer);
  } else if ( _noCaching && status ) {
    UpdateDelay(_updateDelay);
  }

  _noCaching = !status;
  if (!_noCaching && !_dirOk) {
    _noCaching = TRUE;
  }
}

//
// Setter and getter for total cache size.
//
void
Pad_CacheMgr::Size(float mbytes)
{
  _cacheSize = (long) (mbytes*(MEGABYTES));
  if (_cacheSize == 0) {
    Cache(FALSE);
  } else {
    Cache(TRUE);
  }
}

float
Pad_CacheMgr::Size(void)
{
  return (_cacheSize*1.0)/(MEGABYTES);
}

//
// Setter and getter for cache directory.
//
Pad_Bool
Pad_CacheMgr::Dir(char *dir)
{
  Pad_Bool rc=TRUE;
  pid_t pid = getpid();
  
  // For now only allow changing cache dir if there's nothing in it.
  if (_cacheDir.Length() > 0) {
    Pad_HashTableIterator iter;
    Pad_CacheInfo *info;
    DOTABLE(iter, _cacheInfoTable, Pad_CacheInfo, info) {
      if (info->fileWritten) {
	Pad_errorString << "Can not change cache dir now, current dir (" 
		 << _cacheDir.Get() << ") has something in it.  Sorry:(\n";
	cerr << Pad_errorString.Get();
	return FALSE;
      }
    }
    // remove current cache dir
    RmDir(_cacheDir.Get());
  }
   
  if (!dir || !strlen(dir)) {
    // dir not set, disable caching
    Cache(FALSE);
    _dirOk = FALSE;
    return TRUE;

  } else {
    // append process id to the path
    _cacheDir = dir;
    _cacheDir += "/";
    _cacheDir += (int)pid;

#ifdef PAD_UNIX
    if ((mkdir(dir, S_IRWXU) == -1 && errno != EEXIST) ||
	(mkdir(_cacheDir.Get(), S_IRWXU) == -1)) {
#endif
#ifdef PAD_WIN
      if ( errno != EEXIST) { // this may not work in win32, no solution 
	                      // has been found yet, just make it compiled	
#endif
      Pad_errorString << "ERR: COULD NOT CREATE CACHE DIR FOR THIS PROCESS: " 
		      << _cacheDir.Get() << "\n";
#if defined(sun)
      Pad_errorString << strerror(errno) << "\n";
#else
      Pad_errorString << sys_errlist[errno] << "\n";
#endif
      rc = FALSE;
    }

    if ( !rc ) {
      Pad_errorString << "CACHING DISABLED!" << "\n";
      cerr << Pad_errorString.Get();
      Cache(FALSE);
      _dirOk = FALSE;
    } else {
      _dirOk = TRUE;
      Cache(TRUE);
    }
  }

  return rc;
}

//
// Return the basepath of the cache dir (i.e. without the pid).
//
char*
Pad_CacheMgr::Dir(void)
{
  // would have used dirname(_cacheDir.Get()) but looks like
  // Linyx does not have dirname:(.

  static char path[MAXPATHLEN];

  path[0] = 0;
  strncpy(path, _cacheDir.Get(), MAXPATHLEN);
  int i = strlen(path), j;
  for (j = i-1; (j >= 0) && (path[j] != '/'); j--);
  if (j >= 0) {
    path[j] = 0;
  } else {
    path[0] = 0;
  }
  
  return path;
}

//
// Utitlity for removing a directory and its contents.
//
Pad_Bool
Pad_CacheMgr::RmDir(char *dir)
{
  Pad_Bool rc=TRUE;
#ifdef PAD_UNIX
  DIR *dirp = opendir(dir); 
  dirent *dp;
  Pad_String fname;

  if (dirp) {
    while ((dp = readdir(dirp)) != NULL) {
      fname << dir << "/" << dp->d_name;
      unlink(fname.Get());
    }
    closedir(dirp);
  }

  if (rmdir(dir) == -1) {
    rc = FALSE;
  }
#endif
  return rc;
}  


//
// Setter and getter for the cache manager's view bbox multiple.
//
void
Pad_CacheMgr::ViewMultiple(float multiple)
{
  _viewMultiple = multiple;
}

float
Pad_CacheMgr::ViewMultiple(void)
{
  return _viewMultiple;
}

//
// Setters and getters for the cache update delay time.
//
void
Pad_CacheMgr::UpdateDelay(long usecs)
{
  _updateDelay = usecs;
  Pad_DeleteTimerHandler(_updateTimer);
  if (_updateDelay && !_noCaching) {
    _updateTimer = Pad_CreateTimerHandler(_updateDelay, Pad_UpdateCache, this);
  }
}

long
Pad_CacheMgr::UpdateDelay(void)
{
  return _updateDelay;
}

//
// Cache out an image object and delete its memory.
// The imagedata rdbData and mask buffers are written out 
// and its pixels are deleted.
// This is only done of the imagedata is not shared with
// other image objects that are currently visible.
//
Pad_Bool
Pad_CacheMgr::CacheOut(Pad_Image *image, Pad_CacheInfo *info)
{
  Pad_Bool rc=TRUE;
  Pad_ImageData *imitem = image->image;
  Pad_Object *otherImage;
  Pad_CacheInfo *otherInfo;
  Pad_Iterator iter;

  Pad_View *view;
  if (Pad_prc) {
    view = (Pad_View*) Pad_prc->views.Last();
  } else {
    view = _pad->view;
  }

  DOLIST(iter, imitem->objects, Pad_Object, otherImage) {
    if (((Pad_Image*)otherImage) != image) {
      otherInfo = (Pad_CacheInfo*) _cacheInfoTable.Get(otherImage);
      if (otherInfo->cacheStatus & CACHEIN_REQ) {
	return FALSE;
      }
    }
  }

  //cout << "CacheOut: " << image->id << " View: " << view->id << endl;

  // only write to file if the file does not already exist.
  if ( !info->fileWritten ) {
    CacheFile(image, info->file);
    ofstream of(info->file.Get());
    long len = imitem->width * imitem->height;

    if (!of.good()) {
      return FALSE;
    }

    if (imitem->rgb) {
      of.write((char*)imitem->rgbData, len*sizeof(unsigned long));
    } else {
      of << (long) imitem->xpixels8 << ' ' << (long) imitem->xpixels;
      if (imitem->xpixels8) {
	of.write(imitem->xpixels8, len*sizeof(unsigned char));
      } else if (imitem->xpixels) {
	of.write((char*)imitem->xpixels,  len*sizeof(Pixel));
      }
    }

    of << imitem->mask;
    if (imitem->mask) {
      of.write(imitem->mask, len*sizeof(unsigned char));
    }
    of.close();
    info->fileWritten = TRUE;
  }

  // delete its memory
  delete[] imitem->rgbData;
  delete[] imitem->xpixels8;
  delete[] imitem->xpixels;
  delete[] imitem->mask;
  imitem->rgbData = NULL;
  imitem->xpixels8 = NULL;
  imitem->xpixels = NULL;
  imitem->mask = NULL;

  // update other images with we just did
  DOLIST(iter, imitem->objects, Pad_Object, otherImage) {
    if (((Pad_Image*)otherImage) != image) {
      otherInfo = (Pad_CacheInfo*) _cacheInfoTable.Get(otherImage);
      if (otherInfo->cacheStatus & INMEMORY) {
	otherInfo->cacheStatus |= WRITTENOUT;
	otherInfo->cacheStatus &= ~INMEMORY;
	otherInfo->file = info->file;
	otherImage->flags |= PAD_MASK_CACHEDOUT;
      }
    }
  }
  
  return rc;
}

//
// Reconstruct an image by reading its cache file.  The rgbData and
// mask buffers are read from the file and the pixels recomputed.
//
Pad_Bool
Pad_CacheMgr::CacheIn(Pad_Image *image, Pad_CacheInfo *info)
{
  Pad_Bool rc=TRUE;
  Pad_ImageData *item = image->image;

  Pad_View *view;
  if (Pad_prc) {
    view = (Pad_View*) Pad_prc->views.Last();
  } else {
    view = _pad->view;
  }
  //cout << "CacheIn: " << image->id << " View: " << view->id << endl;

  CacheFile(image, info->file);
  ifstream ifs(info->file.Get());
  long len = item->width * item->height;

  if (!ifs.good()) {
    return FALSE;
  }

  if (item->rgb) {
    item->rgbData = new unsigned long[len];
    ifs.read((char*)item->rgbData, len*sizeof(unsigned long));
  } else {
    long hasPixels8, hasPixels;
    ifs >> hasPixels8 >> hasPixels;
    if (hasPixels8) {
      item->xpixels8 = new unsigned char[len];
      ifs.read(item->xpixels8, len*sizeof(unsigned char));
    } else if (hasPixels) {
      item->xpixels = new Pixel[len];
      ifs.read((char*)item->xpixels, len*sizeof(Pixel));
    }
  }

  long hasMask;
  ifs >> hasMask;
  if (hasMask) {
    item->mask = new unsigned char[len];
    ifs.read(item->mask, len*sizeof(unsigned char));
  }

  if (item->rgb) {
    item->Compute_pixel_data();
  }

  Pad_Object *otherImage;
  Pad_CacheInfo *otherInfo;
  Pad_Iterator iter;

  // update other images with what we just did.
  DOLIST(iter, item->objects, Pad_Object, otherImage) {
    if (((Pad_Image*)otherImage) != image) {
      otherInfo = (Pad_CacheInfo*) _cacheInfoTable.Get(otherImage);
      otherInfo->cacheStatus |= INMEMORY;
      otherInfo->cacheStatus &= ~WRITTENOUT;
      otherInfo->obj->flags &= ~PAD_MASK_CACHEDOUT;
    }
  }

  return rc;
}

//
// Compute the total cache size/worth of an image.
// This includes the rgbData, mask and pixel buffers.
//
long
Pad_CacheMgr::CacheSize(Pad_Image *image, Pad_CacheInfo *)
{
  Pad_ImageData *imitem;
  long len, rgbData_len, xpixels8_len, xpixels_len, mask_len;

  imitem = image->image;
  len = rgbData_len = xpixels8_len = xpixels_len = mask_len = 0;
    
  len = imitem->width * imitem->height;
  if (imitem->rgbData) rgbData_len = len * sizeof(unsigned long);
  if (imitem->xpixels8) xpixels8_len = len * sizeof(unsigned char);
  if (imitem->xpixels) xpixels_len = len * sizeof(Pixel);
  if (imitem->mask) mask_len = len * sizeof(unsigned char);
  len = rgbData_len + xpixels8_len + xpixels_len + mask_len;

  return len;
}
