/*
"(c) Copyright 1993-1997 Pad++ Consortium {University of New Mexico (UNM),
and New York University (NYU)}, All Rights Reserved."  
Licensee can not remove or obscure any of the
copyright or trademark notices in this software.

IN NO EVENT SHALL THE AUTHORS OR DISTRIBUTORS BE LIABLE TO ANY PARTY
FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
ARISING OUT OF THE USE OF THIS SOFTWARE, ITS DOCUMENTATION, OR ANY
DERIVATIVES THEREOF, EVEN IF THE AUTHORS HAVE BEEN ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.

THE AUTHORS AND DISTRIBUTORS SPECIFICALLY DISCLAIM ANY WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE, AND NON-INFRINGEMENT.  THIS SOFTWARE
IS PROVIDED ON AN "AS IS" BASIS, AND THE AUTHORS AND DISTRIBUTORS HAVE
NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR
MODIFICATIONS.

See the file "License" for general information on usage and
redistribution, and the file "LicenseTerms" for the specific license
agreement on usage and redistribution of this file, and the Pad++
software in general.
*/

#include "callback.h"
#include "global.h"
#include "events.h"
#include "misc.h"
#include "hashtab.h"

#include <stdlib.h>

////////////////////////////////////////////////////////////////////////////
//
// Pad_Language definitions
//
////////////////////////////////////////////////////////////////////////////

Pad_Language::Pad_Language(Tcl_Interp *newInterp, char *newName, char *newCmdName, Pad_InitProc *newInitProc, 
			   Pad_CreateProc *newCreateProc, Pad_CommandProc *newCommandProc, 
			   Pad_PromptProc *newPromptProc, Pad_CompleteProc *newCompleteProc, 
			   Pad_EvalProc *newEvalProc)
{
    Tk_Uid languageName;

    name = newName;
    interp = newInterp;
    cmdName = newCmdName;
    initProc = newInitProc;
    commandProc = newCommandProc;
    completeProc = newCompleteProc;
    createProc = newCreateProc;
    promptProc = newPromptProc;
    evalProc = newEvalProc;
    
    languageName = Tk_GetUid(newName);
    Pad_languages.Set(languageName, this);
}

////////////////////////////////////////////////////////////////////////////
//
// Pad_Callback definitions
//
////////////////////////////////////////////////////////////////////////////

Pad_Callback::Pad_Callback(char *newScript, Pad_Language *newLanguage)
{
    _Init(newScript, newLanguage);
}

Pad_Callback::Pad_Callback(Pad_String &newScript, Pad_Language *newLanguage)
{
    _Init(newScript.Get(), newLanguage);
}

Pad_Callback::Pad_Callback(Pad_CallbackProc *newFunc, ClientData newClientData)
{
    script = NULL;
    func = newFunc;
    eventFunc = NULL;
    clientData = newClientData;
    language = NULL;
    refCount = 0;
}

Pad_Callback::Pad_Callback(Pad_EventCallbackProc((*newEventFunc)), ClientData newClientData)
{
    script = NULL;
    func = NULL;
    eventFunc = newEventFunc;
    clientData = newClientData;
    language = NULL;
    refCount = 0;
}

Pad_Callback::Pad_Callback(Pad_Callback &callback)
{
    if (callback.script) {
	script = new Pad_String(callback.script);
    } else {
	script = NULL;
    }
    func = callback.func;
    eventFunc = callback.eventFunc;
    clientData = callback.clientData;
    language = callback.language;
    refCount = 0;
}

Pad_Callback::~Pad_Callback()
{
    if (script) {
	delete script;
	script = NULL;
    }
}

void
Pad_Callback::_Init(char *newScript, Pad_Language *newLanguage)
{
    func = NULL;
    eventFunc = NULL;
    clientData = NULL;
    script = new Pad_String(newScript);
    if (newLanguage == NULL) {
	_Guess_language();
    } else {
	language = newLanguage;
    }
    refCount = 0;
}

//
// Guess what language the script is.
// If compiling with scheme, then if the first character is a left paren, assume it is scheme.
// Else, assume it is whatever the toplevel language is, except scheme.
//
void
Pad_Callback::_Guess_language(void)
{
    Tk_Uid lang_uid;

    language = NULL;
				// If Scheme isn't defined, then languages will get set to NULL
    if ((script->Length() >= 1) && (script->Get_char(0) == '(')) {
	lang_uid = Tk_GetUid("scheme");
	language = (Pad_Language *)Pad_languages.Get(lang_uid);
    }
    if (language == NULL) {
	language = Pad_topLevelLanguage;
	if (language->name == "scheme") {
	    lang_uid = Tk_GetUid("tcl");
	    language = (Pad_Language *)Pad_languages.Get(lang_uid);
	}
    }
}

//
// Maintain reference count - which is the number of times
// a callback is used in an event binding
//
void
Pad_Callback::Increment_refcount(void)
{
    refCount++;
}

void
Pad_Callback::Decrement_refcount(void)
{
    refCount--;
}

int
Pad_Callback::Get_refcount(void)
{
    return(refCount);
}

void
Pad_Callback::Append(char *newScript)
{
    if (!script) {
	script = new Pad_String;
    }
    *script += newScript;
}

Pad_Language *
Pad_Callback::Language(void)
{
    return(language);
}

int
Pad_Callback::Length(void)
{
    return(script->Length());
}

int
Pad_Callback::Eval(void)
{
    int rc = TCL_OK;

    if (script) {
	rc = language->evalProc(language, script->Get());
	if (rc == TCL_ERROR) {
	    Tcl_Interp *interp = language->interp;
	    if (interp) {
		Tcl_AddErrorInfo(interp, "\n    (command bound to event)");
		Pad_Background_error(interp);
	    }
	}
    } else {
	if (func) {
	    rc = func(clientData);
	} else if (eventFunc) {
	    rc = eventFunc((Pad_Object *)NULL, clientData, (Pad_Event *)NULL);
	}
    }

    return(rc);
}

int
Pad_Callback::Eval(Pad_Object *obj, Pad_Event *padEvent)
{
    int rc = TCL_OK;

    if (script) {
                           // Get interpreter and copy the script before
                           // evaluating it.  Because this could be
                           // deleted as a result of the script!
        Tcl_Interp *interp = language->interp;
	Pad_String scriptCopy(script);

	rc = language->evalProc(language, scriptCopy.Get());
	
	if (rc == TCL_ERROR) {
	    if (interp) {
		Tcl_AddErrorInfo(interp, "\n    (command bound to event)");
		Pad_Background_error(interp);
	    }
	}
    } else {
	if (func) {
	    rc = func(clientData);
	} else if (eventFunc) {
	    rc = eventFunc(obj, clientData, padEvent);
	}
    }

    return(rc);
}

char *   
Pad_Callback::Get(void)
{
    char *cmd;

    if (script) {
	cmd = script->Get();
    } else {
	cmd = "<C++ Event Handler>";
    }

    return(cmd);
}

void     
Pad_Callback::Set(char *newScript, Pad_Language *newLanguage)
{
    _Init(newScript, newLanguage);
}

void     
Pad_Callback::Set(Pad_String &newScript, Pad_Language *newLanguage)
{
    _Init(newScript.Get(), newLanguage);
}

void     
Pad_Callback::Set(Pad_Callback *newCallback)
{
    if (script) {
	delete script;
    }
    if (newCallback->script) {
	script = new Pad_String(newCallback->script);
    } else {
	script = NULL;
    }
    func = newCallback->func;
    eventFunc = newCallback->eventFunc;
    clientData = newCallback->clientData;
    language = newCallback->language;
}

Pad_Bool
Pad_Callback::Is_script(void)
{
    Pad_Bool rc;

    if (script) {
	rc = TRUE;
    } else {
	rc = FALSE;
    }

    return(rc);
}

////////////////////////////////////////////////////////////////////////////
//
// Other function definitions
//
////////////////////////////////////////////////////////////////////////////

//
// Evaluate the script in the specified language.
// The Tcl interpreter only needs to be specified
// if the languague is Tcl.
//
int
Pad_GlobalEval(Pad_Language *language, char *script)
{
    int rc;

    rc = language->evalProc(language, script);

    return(rc);
}

//
// Parse the language name, and set the language 
// or return an error if an invalid language was specified.
//
Pad_Bool
Pad_GetLanguage(char *languageStr, Pad_Language **newLanguage)
{
    Pad_Bool rc;
    Pad_Bool first;
    Tk_Uid languageUid;
    Pad_Language *language;
    Pad_HashTableIterator hi;

    languageUid = Tk_GetUid(languageStr);
    language = (Pad_Language *)Pad_languages.Get(languageUid);
    if (language) {
	*newLanguage = language;
	rc = TRUE;
    } else {
	first = TRUE;
	Pad_errorString =  "invalid language: try ";
	DOTABLE(hi, Pad_languages, Pad_Language, language) {
	    if (first) {
		first = FALSE;
	    } else {
		Pad_errorString += ", ";
	    }
	    Pad_errorString += language->name;
	}
	rc = FALSE;
    }

    return(rc);
}
