/*
"(c) Copyright 1993-1997 Pad++ Consortium {University of New Mexico (UNM),
and New York University (NYU)}, All Rights Reserved."  
Licensee can not remove or obscure any of the
copyright or trademark notices in this software.

IN NO EVENT SHALL THE AUTHORS OR DISTRIBUTORS BE LIABLE TO ANY PARTY
FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
ARISING OUT OF THE USE OF THIS SOFTWARE, ITS DOCUMENTATION, OR ANY
DERIVATIVES THEREOF, EVEN IF THE AUTHORS HAVE BEEN ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.

THE AUTHORS AND DISTRIBUTORS SPECIFICALLY DISCLAIM ANY WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE, AND NON-INFRINGEMENT.  THIS SOFTWARE
IS PROVIDED ON AN "AS IS" BASIS, AND THE AUTHORS AND DISTRIBUTORS HAVE
NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR
MODIFICATIONS.

See the file "License" for general information on usage and
redistribution, and the file "LicenseTerms" for the specific license
agreement on usage and redistribution of this file, and the Pad++
software in general.
*/

#include "defs.h"
#include "object.h"
#include "pad.h"
#include "text.h"
#include "html.h"
#include "image.h"
#include "object.h"
#include "pad-tcl.h"
#include "pad-string.h"
#include "line.h"
#include "win.h"
#include "restorer.h"
#include "group.h"
#include "callback.h"
#include "transform.h"
#include "global.h"
#include "imagedata.h"

#include <string.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>

#ifdef PAD_UNIX
#  include <unistd.h>
#endif

#define HTML_DEFAULT_FILL         "gray90"
#define HTML_DEFAULT_BORDER       "black"
#define HTML_DEFAULT_BORDERWIDTH  1
#define HTML_DEFAULT_DONESCRIPT   ""
#define HTML_DEFAULT_ERRORSCRIPT  ""
#define HTML_DEFAULT_FONT         "Line-1.2"
#define HTML_DEFAULT_URL          ""
#define HTML_DEFAULT_UPDATESCRIPT ""
#define HTML_DEFAULT_STATE        HTML_ANCHOR_UNVISITED

#define HTML_MIN_SIZE             100

// The html object itself is really just a wrapper for loading in 
// all kinds of sub-Objects (like text, images, etc.)  It gets
// rendered as the background of the html file.
//
// The html object and all of its components are created with pixel 
// coordinates.  Text is created one pixel high.  Images are
// scaled down by a factor of 10 (yielding text that is effectively
// 10 pixels high).
//
// The current Position (x_curr_pos, y_curr_pos) indicates the nw corner
// of each object.  x_left_marg represents the left margin.

static float IMAGE_SIZE = 0.1,
             DUMMY_IMAGE_SIZE  = 5.0;

static char *StartTagArray[] = {
    "html",
    "body",
    "address",
    "h1",
    "h2",
    "h3",
    "h4",
    "h5",
    "h6",
    "p",
    "a",
    "img",
    "ul",
    "li",
    "dl",
    "dt",
    "de",
    "dd",
    "hr",
    "br",
    "title",
    "head",
    "em",
    "i",
    "b",
    "center",
    "menu",
    "cite",
    "font",
    "pre",
    "kbd",
    "strong",
    "base",
    "code",
    "ol",
    "!--",
    "form",
    "input",
    ""
};

static HTML_TagOrKey tagArray[] = {
    {"html", TAG_HTML},
    {"body", TAG_BODY},
    {"address", TAG_ADDRESS},
    {"h1",   TAG_H1},
    {"h2",   TAG_H2},
    {"h3",   TAG_H3},
    {"h4",   TAG_H4},
    {"h5",   TAG_H5},
    {"h6",   TAG_H6},
    {"p",    TAG_P},
    {"a",    TAG_A},
    {"img",  TAG_IMG},
    {"ul",   TAG_UL},
    {"li",   TAG_LI},
    {"dl",   TAG_DL},
    {"dt",   TAG_DT},
    {"de",   TAG_DE},
    {"dd",   TAG_DD},
    {"hr",   TAG_HR},
    {"br",   TAG_BR},
    {"title", TAG_TITLE},
    {"head", TAG_HEAD},
    {"em",   TAG_EM},
    {"i",    TAG_I},
    {"b",    TAG_B},
    {"center", TAG_CENTER},
    {"menu", TAG_MENU},
    {"cite", TAG_CITE},
    {"font", TAG_FONT},
    {"pre",  TAG_PRE},
    {"kbd",  TAG_KBD},
    {"strong", TAG_STRONG},
    {"base", TAG_BASE},
    {"code", TAG_CODE},
    {"ol",   TAG_OL},
    {"!--",  TAG_COMMENT1},
    {"!---", TAG_COMMENT2},
    {"meta", TAG_META},
    {"textarea", TAG_TEXTAREA},
    {"form",  TAG_FORM},
    {"input", TAG_INPUT},
    {"table", TAG_TABLE},
    {"tr",   TAG_TR},
    {"td",   TAG_TD},
    {"th",   TAG_TH},
    {"blockquote", TAG_BLOCKQUOTE},
    {"blink",TAG_BLINK},
    {"link", TAG_LINK},
    {"caption", TAG_CAPTION},
    {"header", TAG_HEADER},
    {"",     TAG_UNKNOWN},
    {NULL, 0}
};

static HTML_TagOrKey keyArray[] = {
    {"href",   KEY_HREF},
    {"src",    KEY_SRC},
    {"alt",    KEY_ALT},
    {"name",   KEY_NAME},
    {"align",  KEY_ALIGN},
    {"border", KEY_BORDER},
    {"width",  KEY_WIDTH},
    {"height", KEY_HEIGHT},
    {"ismap",  KEY_ISMAP},
    {"bgcolor",KEY_BGCOLOR},
    {"text",   KEY_TEXT},
    {"link",   KEY_LINK},
    {"alink",  KEY_ALINK},
    {"vlink",  KEY_VLINK},
    {"pad_scale",  KEY_PAD_SCALE_BEGIN},
    {"/pad_scale",  KEY_PAD_SCALE_END},
    {"pad_tcl",  KEY_PAD_TCL},
    {"",     KEY_UNKNOWN},
    {NULL, 0}
};

static Pad_String Token;

float x_curr_pos,
  y_curr_pos,
  x_left_marg;

static Pad_Bool     Anchor_active = FALSE;  // True if creating an anchor
static HTML_Anchor *Anchor_first;   // First anchor component
static HTML_Anchor *Anchor_current; // Current anchor component
static Pad_String   Anchor_href;    // Current anchor href
static Pad_String   Anchor_name;    // Current anchor name
static int NewLine;	            // Number of recent newlines
static Pad_Bool Title;		    // True if inside Title tag
static Pad_Bool H1;		    // True if inside H1 tag
static Pad_Bool H2;		    // True if inside H2 tag
static Pad_Bool H3;		    // True if inside H3 tag
static Pad_Bool H4;		    // True if inside H4 tag
static Pad_Bool H5;		    // True if inside H5 tag
static Pad_Bool H6;		    // True if inside H6 tag
static Pad_Bool HTML_DD;	    // True if inside DD tag
static Pad_Bool PreFormatted;       // True if inside PRE tag

static Pad_List LineObjs;           // List of objects on the current line
static float DefaultLineSize;       // Default space to next line
static float LineSize;   	    // Height of text line
static float MaxLineSize;	    // Keeps track of how much room to add for a newline

static HTML_EntryTagTcl *Pad_tcl_entry;
static Pad_Bool LayingOut;	// Keeps track of whether we are in
                                // Create_display_list() or not. Used for
                                // adding Pad_Objects to HTML after creation

static Pad_Bool Initializing;	// Keeps track of whether we are in
                                // Set_url() or not. Used for
                                // adding Pad_Objects to HTML after creation

static Pad_List MSML_Scale_list;  // The stack of scale items for MSML
static Pad_List MSML_Left_margin; // The stack of left margins for MSML
static float MSML_Scale;	  // What the current MSML scale is
static float HTML_VERTICAL_IMAGE_OFFSET; // Space added to images for vertical padding

//////////////////////////////////////////////////////////////////
//
// Support Routines
//
//////////////////////////////////////////////////////////////////

//
// Maintain stack of scale tags for MSML extension.
//
static void
MSML_Reset_scale(void)
{
    float *f;

    while ((f = (float *)MSML_Scale_list.Pop())) {
	delete f;
    }
    MSML_Scale = 1.0;
}

static float 
MSML_Push_scale(float newScale)
{
    float *f = new float(newScale);
  
    MSML_Scale_list.Push(f);	// Add new value to top of stack
    return(newScale);		// Return top of stack value
}

static float 
MSML_Pop_scale(void)
{
    float *f;
    float tmp;
    
    f = (float *)MSML_Scale_list.Pop();
    if (f) {
	tmp = *f;
	delete f;
    } else {
	tmp = 1.0;
    }

    return(tmp);		// Return the old top of stack
}

//
// Maintain stack of left margins
//
static void
MSML_Reset_left_marg(void)
{
    float *f;

    while ((f = (float *)MSML_Left_margin.Pop())) {
	delete f;
    }
    x_left_marg = 0.0;
}

static float 
MSML_Push_left_marg(float new_margin)
{
    float *f = new float(new_margin);
  
    MSML_Left_margin.Push(f);	// Add new value to top of stack
    return(new_margin);		// Return top of stack value
}

static float 
MSML_Pop_left_marg(void)
{
    float *f;
    float tmp;
    
    f = (float *)MSML_Left_margin.Pop();
    if (f) {
	delete f;
    }

    f = (float *)MSML_Left_margin.First();
    if (f) {
	tmp = *f;
    } else {
	tmp = 0.0;
    }

    return(tmp);		// Return the resulting top of stack
}

//////////////////////////////////////////////////////////////////
//
// Start of Class Definitions
//
//////////////////////////////////////////////////////////////////

HTML::~HTML() 
{
    HTML_Display *d;
    HTML_Entry *e;
    Pad_Object *o;

    Generate_delete();

    if (urlItem) {
	pad->Url_cancel(urlItem);
    }
    
    if (updateScript) {
	delete updateScript;
	updateScript = NULL;
    }
    if (doneScript) {
	delete doneScript;
	doneScript = NULL;
    }
    if (errorScript) {
	delete errorScript;
	errorScript = NULL;
    }
    if (source) {
	delete [] source;
	source = NULL;
    }
    
    while ((d = (HTML_Display *)displayList.Pop())) {
	delete d;
    }
    while ((e = (HTML_Entry *)entryList.Pop())) {
	delete e;
    }
    while ((o = (Pad_Object *)members.Pop())) {
	delete o;
    }
}

HTML::HTML(Pad *pad) :
Pad_Group(pad)
{
    _type = PAD_HTML;
    htmlFlags = PAD_NO_MASK;
    width = HTML_DEFAULT_WIDTH; // Width of page in characters
    xmin = 0.0;
    ymin = 0.0;
    xmax = 0.0;
    ymax = 0.0;
    
    inlineImages = 0;

    Set_border_default();
    Set_borderwidth_default();
    Set_fill_default();
    Set_font_default();

    updateScript = NULL;
    doneScript = NULL;
    errorScript = NULL;
    source = NULL;
    sourceIndex = 0;
    sourceSpace = 0;
    urlItem = NULL;
}

//
// Callback for when a URL is updated (i.e., more data arrived)
//
static Pad_Bool
HTML_UpdateCB(ClientData clientData, int, char *new_data, int len)
{
    unsigned char *more_data;
    HTML *html = (HTML *)clientData;

				// First call, so get http header data from urlItem
    if (html->source == NULL) {
	html->docType =          html->urlItem->docType;
	html->docLastModified = html->urlItem->docLastModified;
	html->docLength =        html->urlItem->docLength;
	if (html->urlItem->docLength > 0) {
	    html->source = new unsigned char[html->urlItem->docLength + 1];
	    html->sourceSpace = html->urlItem->docLength;
	}
    }

				// Double space if not big enough
    if (len > (html->sourceSpace - html->sourceIndex + 1)) {
	html->sourceSpace = MAX(len, html->sourceSpace);
	html->sourceSpace *= 2;
	more_data = new unsigned char[html->sourceSpace];
	memcpy(more_data, html->source, html->sourceIndex);
	delete [] html->source;
	html->source = more_data;
    }
    memcpy(&html->source[html->sourceIndex], new_data, len);
    html->sourceIndex += len;
    html->source[html->sourceIndex] = '\0';

    return(TRUE);
}

//
// Callback for when a URL has completed loading
//
static Pad_Bool
HTML_DoneCB(ClientData clientData, int)
{
    int index, result;
    Pad_Bool eof;
    HTML *html = (HTML *)clientData;
    HTML_Entry *e;
    Pad_String cmd;

				// Don't do anything if document never retrieved.
				// Presumably, the error callback was already activated.
    if (!html->source) {
	return(TRUE);
    }
				// If no document type, then assume it is html
    if ((html->docType == "text/html") || (html->docType == "")) {
	index = 0;
	eof = FALSE;
	while (!eof && (e = html->Next_token(index, eof))) {
	    html->entryList.Push_last(e);
	} 
    } else if (html->docType == "image/gif" || 
	       html->docType == "image/jpeg") {
				// URL is a GIF or JPEG image
	HTML_EntryTagImage *ei;

	ei = new HTML_EntryTagImage(NULL, html);
	html->entryList.Push_last(ei);
	ei->Set_image(Pad_renderer->Alloc_image(html->source, html->sourceIndex));
				// No need to keep image source around
	delete [] html->source;
	html->source = NULL;
	html->sourceSpace = 0;
	html->sourceIndex = 0;
    }
    html->Create_display_list();
    
				// Call 'update' script.
    if (html->updateScript) {
	cmd = html->updateScript->Get();
	cmd += " ";
	cmd += html->id;
	result = Pad_GlobalEval(html->updateScript->Language(), cmd.Get());
	if (result == TCL_ERROR) {
	    Tcl_AddErrorInfo(html->pad->view->win->interp, "\n    (Pad URL Fetch)");
	    Pad_Background_error(html->pad->view->win->interp);
	}
    }

				// If no in-line images, then html is loaded, so call 
				// "done" script
    if ((html->inlineImages == 0) && (html->doneScript)) {
	cmd = html->doneScript->Get();
	cmd += " ";
	cmd += html->id;
	result = Pad_GlobalEval(html->doneScript->Language(), cmd.Get());
	if (result == TCL_ERROR) {
	    Tcl_AddErrorInfo(html->pad->view->win->interp, "\n    (Pad URL Fetch)");
	    Pad_Background_error(html->pad->view->win->interp);
	}
    }

    html->urlItem = NULL;

    return(TRUE);
}

//
// Callback for when there has been an error with the URL
//
static Pad_Bool
HTML_ErrorCB(ClientData clientData, int, char *, int)
{
    int result;
    HTML *html = (HTML *)clientData;
    Pad_String cmd;
				// Call 'error' script.
    if (html->errorScript) {
	cmd = html->errorScript->Get();
	cmd += " ";
	cmd += html->id;
	result = Pad_GlobalEval(html->errorScript->Language(), cmd.Get());
	if (result == TCL_ERROR) {
	    Tcl_AddErrorInfo(html->pad->view->win->interp, "\n    (Pad URL Fetch)");
	    Pad_Background_error(html->pad->view->win->interp);
	}
    }

    return(TRUE);
}

//
// Returns true if object needs more refinement.
//
Pad_Bool 
HTML::Continue_refinement(void)
{
    return(refineNeeded);
}

//
// Set the object's <option> to the value specified by <sval>.
// Note that each option has a specific setter and getter method
// if the correct argument type is available.
//
Pad_Bool
HTML::Set_option(Pad_Option *option, char *sval)
{
    float fval;
    Pad_Bool rc = TRUE;

    switch (option->id) {
      case PAD_BORDER:
	Set_border(sval);
	break;
      case PAD_BORDERWIDTH:
	fval = ATOVALUEF(pad->view->win, sval);
	Set_borderwidth(fval);
	break;
      case PAD_DONESCRIPT:
	Set_done_script(sval);
	break;
      case PAD_ERRORSCRIPT:
	Set_error_script(sval);
	break;
      case PAD_FILL:
	Set_fill(sval);
	break;
      case PAD_HTMLANCHORS:
	Pad_errorString = "Error: -htmlanchors is a read-only option";
	rc = FALSE;
	break;
      case PAD_UPDATESCRIPT:
	Set_update_script(sval);
	break;
      case PAD_URL:
	rc = Set_url(sval);
	break;
      default:
	rc = Pad_Group::Set_option(option, sval);
	break;
    }

    return(rc);
}

//
// Set the object's <option> to the default value.
//
Pad_Bool
HTML::Set_option_default(Pad_Option *option)
{
    Pad_Bool rc = TRUE;

    switch (option->id) {
      case PAD_BORDER:
	Set_border_default();
	break;
      case PAD_BORDERWIDTH:
	Set_borderwidth_default();
	break;
      case PAD_DONESCRIPT:
	Set_done_script_default();
	break;
      case PAD_ERRORSCRIPT:
	Set_error_script_default();
	break;
      case PAD_FILL:
	Set_fill_default();
	break;
      case PAD_HTMLANCHORS:
	Pad_errorString = "Error: -htmlanchors is a read-only option";
	rc = FALSE;
	break;
      case PAD_UPDATESCRIPT:
	Set_update_script_default();
	break;
      case PAD_URL:
	Set_url_default();
	break;
      default:
	rc = Pad_Group::Set_option_default(option);
	break;
    }

    return(rc);
}

//
// Return the value of the object's <option> as a string
// in <result>.
//
Pad_Bool
HTML::Get_option(Pad_Option *option, Pad_String &result, Pad_Bool &modified)
{
    Pad_Bool rc = TRUE;

    switch (option->id) {
      case PAD_BORDER:
	modified = (htmlFlags & HTML_BORDER_SET) ? TRUE : FALSE;
	Get_bordername(result);
	break;
      case PAD_BORDERWIDTH:
	modified = (htmlFlags & HTML_BORDERWIDTH_SET) ? TRUE : FALSE;
	result = Get_borderwidth();
	break;
      case PAD_DONESCRIPT:
	modified = (htmlFlags & HTML_DONESCRIPT_SET) ? TRUE : FALSE;
	result = Get_done_script();
	break;
      case PAD_ERRORSCRIPT:
	modified = (htmlFlags & HTML_ERRORSCRIPT_SET) ? TRUE : FALSE;
	result = Get_error_script();
	break;
      case PAD_FILL:
	modified = (htmlFlags & HTML_FILL_SET) ? TRUE : FALSE;
	Get_fillname(result);
	break;
      case PAD_HTMLANCHORS:
	{
	    Pad_Bool first;
	    HTML_Anchor *a;
	    Pad_List *anchors;
	    Pad_Iterator oi;

	    modified = FALSE;
	    first = TRUE;
	    anchors = Get_htmlanchors();
	    result = "";
	    DOLIST(oi, *anchors, HTML_Anchor, a)
	      {
		  if (first) {
		      first = FALSE;
		  } else {
		      result += " ";
		  }
		  result += a->id;
	      }
	}
	break;
      case PAD_UPDATESCRIPT:
	modified = (htmlFlags & HTML_UPDATESCRIPT_SET) ? TRUE : FALSE;
	result = Get_update_script();
	break;
      case PAD_URL:
	modified = (htmlFlags & HTML_URL_SET) ? TRUE : FALSE;
	result = Get_url();
	break;
      default:
	rc = Pad_Group::Get_option(option, result, modified);
	break;
    }

    return(rc);
}


//
// Setters and Getters for the url
//
Pad_Bool
HTML::Set_url(char *urlname)
{
    Pad_String strid;
    Pad_UrlItem *newUrlItem;
    Pad_Bool rc;
    Pad_String bind_cmd;
    HTML_Display *d;
    HTML_Entry *e;
    Pad_Object *o;

				// First clean out any existing entries, etc.
    while ((d = (HTML_Display *)displayList.Pop())) {
	delete d;
    }
    while ((e = (HTML_Entry *)entryList.Pop())) {
	delete e;
    }
    while ((o = (Pad_Object *)members.Pop())) {
	delete o;
    }


    htmlFlags |= HTML_URL_SET;
    Pad_tcl_entry = NULL;
    textColor.Set("black");
    linkColor.Set("blue3");
    alinkColor.Set("red");
    vlinkColor.Set("plum4");
    ulinkColor.Set("gray50");

    Pad_renderer->Set_font(_font);
    DefaultLineSize = 1.1 * Pad_renderer->Get_font_height();
    NewLine = 0;
    Title = FALSE;
    H1 = FALSE;
    H2 = FALSE;
    H3 = FALSE;
    H4 = FALSE;
    H5 = FALSE;
    H6 = FALSE;
    HTML_VERTICAL_IMAGE_OFFSET = 0.2 * DefaultLineSize;
    
    Token = "";
    url = urlname;

    Initializing = TRUE;
    strid = id;
				// Set Tcl global variables used by MSML callbacks
    Tcl_SetVar(pad->view->win->interp, "Pad_CurrentHTML", strid.Get(), TCL_GLOBAL_ONLY);
    Tcl_SetVar(pad->view->win->interp, "Pad_CurrentHTMLPad", 
	       pad->view->win->Get_name(), TCL_GLOBAL_ONLY);

				// Define event binding so objects that are created
				// with the "_html" tag are added to the current HTML object
    bind_cmd = pad->view->win->Get_name();
    bind_cmd += " bind _html <Create> {%P addgroupmember %O $Pad_CurrentHTML}";
    Tcl_GlobalEval(pad->view->win->interp, bind_cmd.Get());

				// Guarantee a reasonable starting bounding box
    xmin = 0.0   - HTML_XOFFSET;
    ymin = 0.0   - HTML_YOFFSET + DefaultLineSize;
    xmax = width + HTML_XOFFSET;
    ymax = 0.0   + HTML_YOFFSET + DefaultLineSize;
    Update();

				// Start URL fetch
    newUrlItem = pad->view->pad->Url_fetch(urlname, NULL, NULL, this, TRUE, 
						 HTML_UpdateCB, HTML_DoneCB, HTML_ErrorCB);
    if (newUrlItem) {
	rc = TRUE;
	urlItem = newUrlItem;
	url = urlItem->urlname; // Get this urlname since relative names are expanded
    } else {
	Create_display_list();
	Pad_errorString = "Error: Can't open URL: ";
	Pad_errorString += urlname;
	rc = FALSE;
    }

    Initializing = FALSE;
    
    return(rc);
}

void
HTML::Set_url_default(void)
{
    Set_url(HTML_DEFAULT_URL);
    htmlFlags &= ~HTML_URL_SET;
}

char *
HTML::Get_url(void)
{
    return(url.Get());
}

//
// Getters for htmlanchors
//
Pad_List *
HTML::Get_htmlanchors(void)
{
    return(&anchorList);
}


//
// Setters and Getters for the width of <this>
//
Pad_Bool
HTML::Set_width(float new_width)
{
				// Reset width flag so when Compute_bounding_box gets called,
				// it recomputes the width.
    optionFlags &= ~PAD_WIDTH_SET;

    width = new_width;
    Create_display_list();
    optionFlags |= PAD_WIDTH_SET;

    return(TRUE);
}

void
HTML::Set_width_default(void)
{
    Set_width(HTML_DEFAULT_WIDTH);
    optionFlags &= ~PAD_WIDTH_SET;
}

float
HTML::Get_width(void)
{
    return(width);
}

//
// Setters and getters for border color
//
void
HTML::Set_border(char *borderName)
{
    htmlFlags |= HTML_BORDER_SET;
    border.Set(borderName);
    Damage();
}

void
HTML::Set_border_default(void)
{
    htmlFlags &= ~HTML_BORDER_SET;
    border.Set(HTML_DEFAULT_BORDER);
    Damage();
}

void
HTML::Get_bordername(Pad_String &borderName)
{
    border.Get(borderName);
}

void
HTML::Set_borderwidth(float borderWidth)
{
    htmlFlags |= HTML_BORDERWIDTH_SET;
    lineWidth = borderWidth;

    Update();
}

void
HTML::Set_borderwidth_default(void)
{
    Set_borderwidth(HTML_DEFAULT_BORDERWIDTH);
    htmlFlags &= ~HTML_BORDERWIDTH_SET;
}

float
HTML::Get_borderwidth(void)
{
    return(lineWidth);
}

Pad_Bool
HTML::Set_fill(char *colorname)
{

    htmlFlags |= HTML_FILL_SET;
    fillColor.Set(colorname);

				// Set small color based on fill
    if (fillColor.Is_set()) {
	int red, green, blue;
	fillColor.Get(red, green, blue);
	smallColor.Set(
		MAX(0, (red - 60)),
		MAX(0, (green - 60)),
		MAX(0, (blue - 60)));
    }
    Damage();

    return(TRUE);
}

Pad_Bool
HTML::Set_fill(int red, int green, int blue)
{

    htmlFlags |= HTML_FILL_SET;
    fillColor.Set(red, green, blue);

				// Set small color based on fill
    if (fillColor.Is_set()) {
	smallColor.Set(
		MAX(0, (red - 60)),
		MAX(0, (green - 60)),
		MAX(0, (blue - 60)));
    }
    Damage();

    return(TRUE);
}

void 
HTML::Set_fill_default(void)
{
    Set_fill(HTML_DEFAULT_FILL);
    htmlFlags &= ~HTML_FILL_SET;
}

void
HTML::Get_fillname(Pad_String &fillname)
{
    fillColor.Get(fillname);
}

//
// Setter functions for Tcl callback scripts
//
void
HTML::Set_update_script(char *script)
{
    htmlFlags |= HTML_UPDATESCRIPT_SET;
    if (updateScript) {
	delete updateScript;
    }
    if (script[0] == '\0') {
	updateScript = NULL;
    } else {
	updateScript = new Pad_Callback(script, Pad_callbackLanguage);
    }
}

void
HTML::Set_update_script_default(void)
{
    Set_update_script(HTML_DEFAULT_UPDATESCRIPT);
    htmlFlags &= ~HTML_UPDATESCRIPT_SET;
}

char *
HTML::Get_update_script(void)
{
    if (updateScript) {
	return(updateScript->Get());
    } else {
	return(NULL);
    }
}

void
HTML::Set_done_script(char *script)
{
    htmlFlags |= HTML_DONESCRIPT_SET;
    if (doneScript) {
	delete doneScript;
    }
    if (script[0] == '\0') {
	doneScript = NULL;
    } else {
	doneScript = new Pad_Callback(script, Pad_callbackLanguage);
    }
}

void
HTML::Set_done_script_default(void)
{
    Set_done_script(HTML_DEFAULT_DONESCRIPT);
    htmlFlags &= ~HTML_DONESCRIPT_SET;
}

char *
HTML::Get_done_script(void)
{
    if (doneScript) {
	return(doneScript->Get());
    } else {
	return(NULL);
    }
}

void
HTML::Set_error_script(char *script)
{
    htmlFlags |= HTML_ERRORSCRIPT_SET;
    if (errorScript) {
	delete errorScript;
    }
    if (script[0] == '\0') {
	errorScript = NULL;
    } else {
	errorScript = new Pad_Callback(script, Pad_callbackLanguage);
    }
}

void
HTML::Set_error_script_default(void)
{
    Set_error_script(HTML_DEFAULT_ERRORSCRIPT);
    htmlFlags &= ~HTML_ERRORSCRIPT_SET;
}

char *
HTML::Get_error_script(void)
{
    if (errorScript) {
	return(errorScript->Get());
    } else {
	return(NULL);
    }
}

//
// Get the text from the HTML object
//
Pad_Bool 
HTML::Get_text(Pad_String &string)
{
    Pad_Iterator it;
    HTML_Entry *e;
    Pad_String tmpString;
    
    string = "";
    tmpString = "";
    DOLIST(it, entryList, HTML_Entry, e) {
	if (e->Get_text(tmpString)) {
            string += tmpString;
            string += " ";
            tmpString = "";
	}
    }   
    return(TRUE);
}

//
// Accessor functions for html info
// 
unsigned char *
HTML::Get_source(void)
{
    return(source);
}

char *
HTML::Get_type(void)
{
    return(docType.Get());
}

char *
HTML::Get_last_modified(void)
{
    return(docLastModified.Get());
}

int
HTML::Get_length(void)
{
    return(sourceIndex);
}

//
// Get rid of starting white space.
// Return the index of the first character that is not whitespace
// starting at index.
//
void
HTML::Skip_white_space(int &index)
{
    Pad_Bool found = FALSE;

    do {
	switch (source[index]) {
	  case '\0':
	    found = TRUE;
	    break;
	  case ' ':
	  case '\t':
	  case '\n':
	    index++;
	    break;
	  default:
	    found = TRUE;
	    break;
	}
    } while (!found);
}

//
// return the next token in Entry.
//
HTML_Entry *
HTML::Next_token(int &index, Pad_Bool &eof)
{
    HTML_Entry *token;
    
    Skip_white_space(index);
    switch (source[index]) {
      case '\0':
	token = NULL;
	break;
      case '<':			// Defining a tag
	token = Parse_tag(index, eof);
        break;
      default:			// Defining text
	token = Parse_text(index, eof);
        break;
    }
    return(token);
}

//
// Parse the following characters as a tag.
// Fill in the list of Keys (type and value).
//
HTML_Entry *
HTML::Parse_tag(int &index, Pad_Bool &eof)
{
    char c;
    Pad_Bool found = FALSE;
    HtmlEntryType type = HTML_ENTRY_UNDEF;
    HTML_EntryTag *tag;
    
    Token = "";
   				// Parse tag along with keywords
    c = source[index++];
    if (c != '<') {
      cerr << "Parse_tag, How did we get here?" << endl;
      return(NULL);
    }

    c = source[index];
    if (c == '/') {
      type = HTML_ENTRY_END;
      index++;
    }
    
    do {
	c = source[index++];
	switch (c) {
	  case '>':
	    found = TRUE;
	    break;
	  case '\0':
	    type = HTML_EOF;
	    eof = TRUE;
	    break;
	  case ' ':
	  case '\t':
	  case '\n':
	    found = TRUE;
	    break;
	  default:
	    Token.Append(c);
	}
    } while ((type != HTML_EOF) && !found);

    if (c != '>') {
	tag = new HTML_EntryTag(this, index);
    } else {
      tag = NULL;
    }
    
                                // the copy constructors that get
                                // called here are all destructive. If
                                // they succeed they will delete tag
                                // and return a new pointer of their
                                // type. If they fail they will return
                                // tag.
    if (type == HTML_ENTRY_END) {
      tag = new HTML_EntryTagEnd();
    } else if (Token == "a") {
      tag = new HTML_EntryTagAnchor(tag);
    } else if (Token == "img") {
      tag = new HTML_EntryTagImage(tag, this);
    } else if (Token == "meta") {
      tag = tag->Parse_meta(this);
    } else {
      tag = new HTML_EntryTagStart(tag);
    }

    tag->Parse(this);
    				// Now find out which tag it was
    return(tag);
}

void 
HTML_EntryTagEnd::Parse(HTML *) 
{
    _endTag = Set_tag_type();
}

void
HTML_EntryTagStart::Parse(HTML *) 
{
    _startTag = Set_tag_type();
}

HTML_EntryTag *
HTML_EntryTag::Parse_meta(HTML *html)
{
    HTML_Key *key;
    HTML_EntryTagTcl *ett;
    HTML_EntryTag *et;
    Pad_Iterator pi;
  
                                // We need to check for the PAD_TCL
                                // key if present we need to create an
                                // HTML_EntryTagTcl set
                                // Pad_tcl_entry to point to it, so
                                // that if we get a tcl create
                                // callback, any Pad_Objects created
                                // in the tcl script, can be stored.
  
    if (_keyList) {
	DOLIST(pi, *_keyList, HTML_Key, key) {
	    switch (key->key){
	      case KEY_PAD_SCALE_BEGIN :
		et = new HTML_EntryTagStart(this);
		return(et);
	      case KEY_PAD_SCALE_END :
		et = new HTML_EntryTagStart(this);
		return(et);
	      case KEY_PAD_TCL :
                                // Pad_tcl_entry must be set before
                                // making the call to Tcl_GlobalEval,
                                // otherwise if an object gets created
                                // by the tcl script it will cause an
                                // error in HTML::Add() when the
                                // create callback happens
		ett = new HTML_EntryTagTcl();
                                // We would actually like to call
                                // HTML_EntryTagTcl (this) to copy the
                                // key_list from this right away, but
                                // we can't, since the eval might
                                // fail, and the 'copy constructor'
                                // for HTML_EntryTag deletes the
                                // parameter
		Pad_tcl_entry = ett;
                                // send key->value.Get() to tcl interp
		if (Tcl_GlobalEval(html->pad->view->win->interp,
				   key->value.Get()) == TCL_ERROR){
		    Tcl_AddErrorInfo(html->pad->view->win->interp, "\n    (MSML meta tag, pad_tcl key)");
		    Pad_Background_error(html->pad->view->win->interp);
		    return(this);
		}
		ett->Set_tag(this); // if the eval succeeds, then we 
				    // need to initialize the key_list of ett
				    // and delete the old tag, since the pointer
				    // we return overwrites the old value of this
		delete this;
		return(ett);
		default :
		break;
	    }
	}
	return(this);
    }
    return(NULL);
}

HtmlEntryTagType 
HTML_EntryTag::Set_tag_type(void) 
{
    int i;

    _tag = Token;
    i = 0;
    while (tagArray[i].name) {
	if (Token == tagArray[i].name) {
	    return(HtmlEntryTagType) tagArray[i].type;
	}
	i++;
    }

    return TAG_UNKNOWN;
}

//
// Parse the following characters as a text element.
// (anything until the next tag).
//
HTML_Entry *
HTML::Parse_text(int &index, Pad_Bool &eof)
{
    char c;
    int type = HTML_ENTRY_UNDEF;
    Pad_Bool found = FALSE;
    Pad_Bool first = TRUE;
    Token.Set("");
    HTML_EntryText *text = new HTML_EntryText();
    
    do {
	c = source[index++];
	switch (c) {
	  case '<':
	    index--;
	    found = TRUE;
	    break;
	  case '\\':
	    c = source[index++];
	    if (c == '\0') {
		type = HTML_EOF;
		found = TRUE;
		eof = TRUE;
		break;
	    }
	    Token.Append(c);
	    first = FALSE;
	    break;
	  case '\0':
	    eof = TRUE;
	    if (first) {
		type = HTML_EOF;
	    } else {
		found = TRUE;
	    }
	    break;
	  default:
	    Token.Append(c);
	    first = FALSE;
	}
    } while ((type != HTML_EOF) && !found);
    text->Set_text(&Token);
    return(text);
}


void
HTML::Delete_display_list_item(HTML_EntryTag *entry) 
{
    HTML_Display *d;
    Pad_Iterator oi;

    DOLIST(oi, displayList, HTML_Display, d) {
	if (d->Get_entry() == entry) {
	    displayList.Remove(d);
	    d->Set_entry(NULL);
	    break;
	}
    }
}

void
HTML::Create_display_list(void) 
{
    Pad_Iterator it;
    HTML_Entry *e;
    HTML_Display *display;
    HTML_Anchor *a;
    Pad_Object *o;
    
    MSML_Reset_left_marg();
    y_curr_pos = 0.0;
    x_curr_pos = 0.0;
    LineSize = DefaultLineSize;
    MaxLineSize = 0.0;
    Anchor_first = NULL;	// Set the current anchor to NULL
    Anchor_current = NULL;
    Anchor_href = "";
    Anchor_name = "";
    MSML_Reset_scale();

    LayingOut = TRUE;

				// Empty the list of members
    while ((o = (Pad_Object *)members.First())) {
	Pad_Group::Remove(o, FALSE);
    }
                                // Delete the old anchors
    while ((a = (HTML_Anchor *)anchorList.Pop())) {
	delete a;
    }
				// Get rid of the old display list
    while ((display = (HTML_Display *)displayList.Pop())) {
	delete display;
    }
    
    Pad_renderer->Set_font(_font); // Necessary to get layout right
    Pad_renderer->Set_font_height(_font.Get_size() * transform.Get_scale());
    DOLIST(it, entryList, HTML_Entry, e) {
	e->Layout(this);
    }   
				// Make room for last item.
    Start_new_line(1);
				// Compute total size, and add room for border
    xmin = 0.0   - HTML_XOFFSET;
    ymin = y_curr_pos  - HTML_YOFFSET + DefaultLineSize;
    xmax = width + HTML_XOFFSET;
    ymax = 0.0   + HTML_YOFFSET + DefaultLineSize;
				// Don't change maxSize if it is unlimited.
    if (maxSize != -1) {
	maxSize = (int)(MAX(ymin, xmax) * 150);
    }
    LineObjs.Make_empty();
    
                                // update the parsing global variables
    
    LayingOut = FALSE;          // objects added after this are from
                                // addgroupmember and not from Layout
    Update();
}

void
HTML::Add_display_item(HTML_Display *di) 
{
    displayList.Push_last(di);
    LineObjs.Push_last(di);
}

//
// Return true to continue rendering other objects, or false if rendering
// was interrupted.
//
Pad_Bool
HTML::Render(void) 
{
    float clip_xmin, clip_ymin, clip_xmax, clip_ymax;
    Pad_Iterator it;
    HTML_Display *i;
    Pad_BBox *htmlBBox, *objBBox;
    Pad_BBox visibleBBox;

				// Compute visible bbox
				// (i.e., for which html components are actually visible)
    Pad_prc->win->activeRestorer->Get_visibility(this, clip_xmin, clip_ymin, clip_xmax, clip_ymax);
    htmlBBox = Get_bbox();
    visibleBBox.Set(LERP(clip_xmin, htmlBBox->Xmin(), htmlBBox->Xmax()),
		    LERP(clip_ymin, htmlBBox->Ymin(), htmlBBox->Ymax()),
		    LERP(clip_xmax, htmlBBox->Xmin(), htmlBBox->Xmax()),
		    LERP(clip_ymax, htmlBBox->Ymin(), htmlBBox->Ymax()));

    if (fillColor.Is_set()) {
	Pad_renderer->Set_color(fillColor);
	Pad_renderer->Draw_filled_box(xmin, ymin, xmax, ymax);
    }

				// Only draw content if not too small
				// (unless refining)
    if ((pixelDim > HTML_MIN_SIZE) || (pad->level > 0)) {
	refineNeeded = FALSE;
	Pad_Group::Render();	// This gets anchors rendered
    
	DOLIST(it, displayList, HTML_Display, i) {
				// Check to see if item is visible
	    objBBox = i->Get_bbox();
	    if (visibleBBox.Overlaps(*objBBox)) {
		i->Render(this);
	    }
	}
    } else {
	Pad_renderer->Set_color(smallColor);
	Pad_renderer->Draw_filled_box(xmin + 10, ymin + 10, xmax - 10, ymax - 10);
	refineNeeded = TRUE;
    }

				// Draw border
    if (border.Is_set()) {
	Pad_Point min, max;

	min.Set(xmin, ymin);
	max.Set(xmax, ymax);
	Pad_renderer->Set_border(border);
	Pad_renderer->Draw_3d_rectangle(min, max, lineWidth, TK_RELIEF_RIDGE);
    }
    
    return(TRUE);
}

//
// True if html object should be picked if non of its members are.
//
Pad_Bool
HTML::Pick_group(Pad_Event *)
{
    return(TRUE);
}

//
// Given a filename relative to this html file, convert it to a complete
// filename with path using the name of this html's URL.
// Return an allocated Pad_String that must be freed by the caller.
//
Pad_String *HTML_EntryTag::Get_pathname(char *newname, HTML *html) 
{
    int index;
    char *slashptr, *dotptr;
    char *cutoffptr;
    char *basename;
    Pad_String *pathname = NULL;
    Pad_Bool slash = FALSE;
    
    // If pathname already complete, leave it alone
    // (i.e., starts with '/' or "http:")
    if ((newname[0] == '/') && (html->url == ""))
      {
	  pathname = new Pad_String(newname);
      } 
    else if (newname[0] == '#')
      {
	  pathname = new Pad_String(html->url);
	  *pathname += newname;
      }
    else if (!strncasecmp(newname, "http://", 7)) 
      {
	  pathname = new Pad_String(newname);
      }
    else 
      {
	  if (!strncasecmp(newname, "http:", 5)) {
	      newname += 5;
	  }
	  if (!strncasecmp(newname, "./", 2)) {
	      newname += 2;
	  }
	  basename = html->url.Get();
	  pathname = new Pad_String(strlen(basename) + strlen(newname) + 2);
	  if (html->url != "") {
	      if (newname[0] == '/') {
		  cutoffptr = strchr(basename + 7, '/');
	      } else {
		  slashptr = strrchr(basename, '/');
		  dotptr = strrchr(basename, '.');
		  if ((slashptr > dotptr) ||
		      ((slashptr - basename) <= 7)) {
		      // Use entire URL as base
		      cutoffptr = basename + strlen(basename) - 1;
		      if (*cutoffptr != '/') {
			  slash = TRUE;
		      }
		  } else {
		      // Cut off last token from URL for base
		      cutoffptr = slashptr;
		  }
	      }
	  } else {
	      cutoffptr = strrchr(basename, '/');
	  }
	  if (cutoffptr) {
	      index = cutoffptr - basename;
	      *pathname = basename;
	      pathname->Insert('\0', index + 1);
	      if (slash) {
		  *pathname += "/";
	      }
	  } else {
	      if (html->url != "") {
		  *pathname = basename;
		  *pathname += "/";
	      } else {
		  *pathname = "";
	      }
	  }
	  if (newname[0] == '/') {	// Don't include extra '/'
	      *pathname += &newname[1];
	  } else {
	      while (!strncmp(newname, "../", 3)) {
				// Need to back up one directory
		  newname += 3;
				// Find last slash
		  slashptr = strrchr(pathname->Get(), '/');
		  if (slashptr) {
		      pathname->Insert('\0', slashptr - pathname->Get());
				// Find second to last slash
		      slashptr = strrchr(pathname->Get(), '/');
		      if (slashptr) {
			  pathname->Insert('\0', slashptr - pathname->Get() + 1);
		      } else {
				// Error - try and recover gracefully
			  *pathname += '/';
		      }
		  }
	      }
	      *pathname += newname;
	  }
      }
    
    return(pathname);
}

void
HTML::Add_anchor(HTML_Anchor *anchor)
{
    anchorList.Push_last(anchor);
    Add(anchor, FALSE);		// Add anchor to HTML (which is really a group).
    anchor->Set_events(TRUE);
    anchor->Add_tag("htmlanchor");
}

void
HTML::Update_line_spacing(void) const 
{
				// Set line size to largest current header size.
    if (H1) {
	LineSize = H1_SIZE * DefaultLineSize;
    } else if (H2) {
	LineSize = H2_SIZE * DefaultLineSize;
    } else if (H3) {
	LineSize = H3_SIZE * DefaultLineSize;
    } else if (H4) {
	LineSize = H4_SIZE * DefaultLineSize;
    } else if (H5) {
	LineSize = H5_SIZE * DefaultLineSize;
    } else if (H6) {
	LineSize = H6_SIZE * DefaultLineSize;
    } else {
	LineSize = DefaultLineSize;
    }
    MaxLineSize = MAX((MSML_Scale * LineSize), MaxLineSize);
}

//
// Generate newlines, depending on the 'request'.
// if 'request' == 0, then generate one newline
// if 'request' > 0, generate enough newlines so that there will
//                   have been a total of request consecutive newlines
//                   taking into account that there may have been recent
//                   previous ones.
//
// Also, move any objects created since the last newline
// down so they are aligned at the bottom with the lowest item.
// (This could happen when an inline image forces all the text to
//  be aligned with the bottom of the image).
//
// Return the distance to re-align objects on this line.
//

float
HTML::Start_new_line(int request) 
{
    int newlines;
    Pad_Iterator oi;
    HTML_Display *di;
    float dy = 0.0;
    
    if ((y_curr_pos == 0.0) &&
	(x_curr_pos == 0.0)) {	// Don't allow extra space at start of page
	return(0.0);
    }
    
    Update_line_spacing();	// If we are parsing a header start tag
				//   we need to set LineSize
    
    if (MaxLineSize != DefaultLineSize) {
				// Relayout vertical position of objects on line
	dy = DefaultLineSize - MaxLineSize;
	DOLIST(oi, LineObjs, HTML_Display, di) {
	    di->Update_line_position(dy);
	}
    }

				// A request of 0 forces a newline
    if (request == 0) {
	request = 1;
	NewLine = 0;
    }
    
    newlines = MAX(0, request - NewLine);
    if (newlines >= 1) {
	y_curr_pos -= MaxLineSize;
    }
    if (newlines >= 2) {
	y_curr_pos -= (newlines - 1) * MSML_Scale * DefaultLineSize;
    }
    x_curr_pos = x_left_marg;
    
    NewLine = request;		// Current number of recent newlines
    
    MaxLineSize = 0.0;		// Line is empty so far
    LineObjs.Make_empty();

    return(dy);
}


//
// Used in html.cpp by various HTML classes
//
Pad_FontRef *
HTML::Get_font_ref()
{
    return &_font;
}


//
// To set the html font, first call the component set_font,
// and then create a new display list so the anchors get
// recreated and layed out with the new font.
//
void
HTML::Set_font(char *fullname)
{
    Pad_Component::Set_font(fullname);
    Create_display_list();
}

void
HTML::Set_font_default(void)
{
    Set_font(HTML_DEFAULT_FONT);
    htmlFlags &= ~HTML_FONT_SET;
}


Pad_Bool
HTML::Add(Pad_Object *obj, Pad_Bool transformp)
{
    Pad_Bool rc;

    rc = Pad_Group::_Add(obj, transformp);
    if (!rc) {
	return(FALSE);
    }

    switch (obj->Type()) {
                                // If this is an HTML_Anchor
                                // then do nothing
      case PAD_HTML_ANCHOR :
      case PAD_HTML_ANCHOR_IMAGE :
      case PAD_HTML_ANCHOR_TEXT :
        return(TRUE);
      default:
        break;
    }
                                // Otherwise check to see if we need
                                // to take special action. We could be
                                // handling a create callback from a tcl
                                // script in the HTML code, or a call
                                // from addgroupmember directly by the
                                // user trying to annotate the page.
    if (Initializing) {
                                // we are adding something to the HTML
                                // inside of Set_url(), so we
                                // add it to Pad_tcl_entry if set
                                // since this must have been created
                                // by a tcl script in the HTML code
	if (Pad_tcl_entry) {
	    Pad_tcl_entry->Add(obj);
	} else {
	    cerr << "Error: trying to add object " << obj;
	    cerr << "in Initialize, but Pad_tcl_entry not set" << endl;
	}
    } else if (!LayingOut) {
                                // We aren't in Create_display_list()
                                // so someone used addgroupmember to
                                // annotate the HTML we will add a new
                                // entry to the end of entryList
	HTML_EntryPadObject *pa = new HTML_EntryPadObject(obj);
	entryList.Push_last(pa);
				// If add an entry after the HTML is
                                // Initialized we need to mark it
                                // damaged so that the new entry will
                                // be read in and the HTML will go
                                // through layout again

				// Must create display list here rather than delay
				// to make HTML's bbox get updated.
	Create_display_list();
    }

    return(rc);
}

//
// Remove obj from group (and return it to the regular pad space).
// If <transformp>, then transform object so it doesn't move.
//
Pad_Bool
HTML::Remove(Pad_Object *obj, Pad_Bool transformp)
{
    Pad_Bool rc;

    rc = Pad_Group::_Remove(obj, transformp);
    if (!rc) {
	return(FALSE);
    }

    if (transformp) {
	Pad_Iterator ei;
	HTML_Entry *entry;

	DOLIST(ei, entryList, HTML_Entry, entry) {
	    if (entry->Remove_obj(obj)) {
		entryList.Remove(entry);
		delete entry;
		break;
	    }
	}
    }

				// Must create display list here rather than delay
				// to make HTML's bbox get updated.
    Create_display_list();
    return(rc);
}

Pad_Bool
HTML::Delete(Pad_Object *) {
                                // do nothing, since we've already
                                // taken care of everthing in
                                // HTML::Create_display_list() and
                                // HTML::Remove()
    return(TRUE);
}

void
HTML::Compute_bounding_box(void) 
{
    float hw;
    
    hw = 0.6 * lineWidth;
    if (!(optionFlags & PAD_WIDTH_SET)) {
        Set_bbox_xmin(xmin - hw);
        Set_bbox_xmax(xmax + hw);
    }
    if (!(optionFlags & PAD_HEIGHT_SET)) {
        Set_bbox_ymin(ymin - hw);
        Set_bbox_ymax(ymax + hw);
    }

    Pad_Object::Compute_bounding_box();
}

Pad_Bool
HTML::Contains_text(char *string) 
{
    Pad_Bool rc;
    Pad_Iterator it;
    HTML_Entry *e;
    
    rc = FALSE;
    DOLIST(it, entryList, HTML_Entry, e) {
	if (e->Contains_text(string)) {
	    rc = TRUE;
	    break;
	}
    }   
    
    return(rc);
}

//
// May contain objects wider than page, so clip.
//
Pad_Bool
HTML::Check_render_clipped(void)
{
    return(TRUE);
}

//
// Flow_string
//
//   Given a string, flow it based on the width of the
//   html object and the current line size.  Generate
//   a list of strings, and return that list.
//
Pad_List *
HTML::Flow_string(Pad_String *ps, Pad_Bool preformatted) 
{
    char *token = NULL;
    char *buf;
    float token_width;
                                // for MSML we need to include the scale
    float textSize =  MSML_Scale * LineSize/ DefaultLineSize;
    float dy;
    Pad_BBox lineBBox;
    Pad_List *stringList = new Pad_List();
    HTML_DisplayString *pds = new HTML_DisplayString();
    float spaceWidth = Pad_renderer->Get_char_width(' ') * textSize;

    if (preformatted) {
	char *cr;
	int len;
				// Preformatted - just parse for carriage returns
	len = ps->Length() + 1;
	buf = new char[len];
	strcpy(buf, ps->Get());
	token = buf;
	while (token) {
	    cr = strchr(token, '\n');
	    if (cr) {
		*cr = '\0';
	    }
	    NewLine = 0;	// No pending newlines
	    stringList->Push_last(pds);
	    dy = Start_new_line(1);
	    pds->Sety(pds->Gety() + dy);   // Re-align text we just created
	    *pds += token;
	    pds = new HTML_DisplayString();      
	    if (cr) {
		token = cr + 1;
	    } else {
		token = NULL;
	    }
	}
	delete [] buf;
	NewLine = 0;	// No pending newlines
    } else {
				// Not preformatted - flow text using line length and text size
	buf = new char[ps->Length() + 1];
	strcpy(buf, ps->Get());

	Pad_renderer->Set_font(_font); 	
	token = strtok(buf, " \t\n\r");
	while (token) {
	    Pad_renderer->Get_string_bbox(token, &lineBBox);
	    token_width = lineBBox.Width() * textSize;
	    if ((x_curr_pos == x_left_marg) || ((x_curr_pos + token_width < width))) {
				// If enough room for this token
		if (pds->Length() > 0) { 
		    *pds += " ";
		    x_curr_pos += spaceWidth;
		}
		*pds += token;
		x_curr_pos += token_width;
		token = strtok(NULL, " \t\n\r");
		NewLine = 0;	// No pending newlines
	    } else {
				// Else, not enough room, start a new line
		stringList->Push_last(pds);
		dy = Start_new_line(1);
		pds->Sety(pds->Gety() + dy);   // Re-align text we just created
		pds = new HTML_DisplayString();      
	    }
	}
	if (pds->Length() > 0) {
	    *pds += " ";		// We need a space after the last token   
	    x_curr_pos += spaceWidth;
	    stringList->Push_last(pds);
	    // Adjust y position of last line
	    dy = MSML_Scale * (DefaultLineSize - LineSize);
	    pds->Sety(pds->Gety() + dy);
	}
	
	delete [] buf;
    }
    
    return(stringList);
}


HTML_DisplayString::HTML_DisplayString() : Pad_String() 
{
    _x = x_curr_pos;
    _y = y_curr_pos;
}

HTML_DisplayString::~HTML_DisplayString() 
{
}

HTML_DisplayString::HTML_DisplayString(Pad_String *ps) : Pad_String(ps) 
{
    _x = x_curr_pos;
    _y = y_curr_pos;
}

HTML_Display::HTML_Display(void) 
{
    _x = 0.0;
    _y = 0.0;
    _scale = 0.0;
    _entry = NULL;
}

HTML_Display::~HTML_Display(void) 
{
    if (_entry) {
	_entry->Remove_display_item();
	_entry = NULL;
    }
}

void
HTML_Display::Update_line_position(float dy)
{
    _y += dy;
    _bbox.Translate(0, dy);
}

Pad_BBox *
HTML_Display::Get_bbox(void)
{
    return(&_bbox);
}

void
HTML_Display::Render(HTML *) {} 

HTML_Entry *
HTML_Display::Get_entry(void)
{
    return(_entry);
}

void
HTML_Display::Set_entry(HTML_Entry *new_entry)
{
    _entry = new_entry;
}

void
HTML_Display::Set_image(Pad_ImageData *)
{
}

float HTML_Display::Get_x(void) const {return _x;}
float HTML_Display::Get_y(void) const {return _y;}
float HTML_Display::Get_size(void) const {return _scale;}
void HTML_Display::Set_x(float xpos) {_x = xpos;}
void HTML_Display::Set_y(float ypos) {_y = ypos;}

HTML_DisplayPadObject::HTML_DisplayPadObject()
{}

HTML_DisplayPadObject::~HTML_DisplayPadObject()
{}

HTML_DisplayPadObject::HTML_DisplayPadObject(HTML *html, Pad_Object *o) 
{
    float width, height;
    Pad_BBox *bb;
    
    _obj = o;
    bb = _obj->Get_global_bbox();
    _x = x_curr_pos;
    _y = y_curr_pos;		// Objects are drawn from their SW corner
    height = bb->Height();
    width = bb->Width();
    _bbox.Set(_x, _y, _x+width, _y+height);
				// Make sure there's enough room on this line
    if ((x_curr_pos != x_left_marg) &&
	((x_curr_pos + width) > html->width)) {
	html->Start_new_line(1);
    }
				// Slide obj to it's new location. 
    html->Add(_obj, FALSE);
    _obj->Slide(x_curr_pos - bb->Xmin(), y_curr_pos - bb->Ymin(), FALSE);
    
    html->Add_display_item(this);
    
				// Update global layout variables
    MaxLineSize = MAX(MaxLineSize, height + (MSML_Scale * HTML_VERTICAL_IMAGE_OFFSET));
    x_curr_pos += width;
    x_curr_pos += Pad_renderer->Get_char_width(' ');  // Add some space
    NewLine = 0;		// No pending newlines
}


void
HTML_DisplayPadObject::Update_line_position(float dy)
{
    _obj->Slide(0.0, dy, FALSE);
    _y += dy;
    _bbox.Translate(0, dy);
}

void
HTML_DisplayPadObject::Render(HTML *) 
{
}

HTML_DisplayRectangle::~HTML_DisplayRectangle(void) 
{
}

HTML_DisplayRectangle::HTML_DisplayRectangle(HTML_Entry *new_entry, HTML *htmlptr, float w, float h)
{
    _entry = new_entry;
    _html = htmlptr;
    _width = w;
    _height = h;
    _scale = 1.0 * MSML_Scale;

    _html->Add_display_item(this);
    MaxLineSize = MAX(MaxLineSize, (MSML_Scale * _height));
				// Make sure there's enough room on this line
    if ((x_curr_pos != x_left_marg) && ((x_curr_pos + (MSML_Scale * _width)) > _html->width)) {
	_html->Start_new_line(1);
    }
    _x = x_curr_pos;
    _y = y_curr_pos;		// Objects are drawn from their SW corner
    _bbox.Set(_x, _y, _x+_width, _y+_height);

    Pad_FontRef *font = htmlptr->Get_font_ref();
    Pad_renderer->Set_font_height(font->Get_size() * MSML_Scale);
    x_curr_pos += (MSML_Scale * _width);
    x_curr_pos += (MSML_Scale * Pad_renderer->Get_char_width(' '));  // Add some space
    NewLine = 0;		// No pending newlines
}

HTML_DisplayRectangle::HTML_DisplayRectangle(HTML_Entry *new_entry, HTML *htmlptr, float w, float h,
					     float xpos, float ypos) 
{
    _entry = new_entry;
    _html = htmlptr;
    _width = w;
    _height = h;
    _scale = 1.0 * MSML_Scale;

    _html->Add_display_item(this);
    MaxLineSize = MAX(MaxLineSize, (MSML_Scale * _height));
				// Make sure there's enough room on this line
    if ((x_curr_pos != x_left_marg) && ((x_curr_pos + (MSML_Scale * _width)) > _html->width)) {
	_html->Start_new_line(1);
    }
    _x = xpos;
    _y = ypos;
    _bbox.Set(_x, _y, _x+_width, _y+_height);

    Pad_FontRef *font = htmlptr->Get_font_ref();
    Pad_renderer->Set_font(*font);
    Pad_renderer->Set_font_height(font->Get_size() * MSML_Scale);
    x_curr_pos += (MSML_Scale * _width);
    x_curr_pos += (MSML_Scale * Pad_renderer->Get_char_width(' '));  // Add some space
    NewLine = 0;		// No pending newlines
}

void
HTML_DisplayRectangle::Render(HTML *) 
{
    if (_penColor.Is_set()) {
	Pad_renderer->Set_color(_penColor);
	Pad_renderer->Draw_filled_box(_x, _y, _x + (_scale * _width), _y + (_scale * _height));
    }
}

void HTML_DisplayRectangle::Set_width(float x) {_width = x;}
void HTML_DisplayRectangle::Set_height(float x) {_height = x;}
void HTML_DisplayRectangle::Set_color(char *name) {_penColor.Set(name); }

HTML_DisplayText::~HTML_DisplayText()
{
    HTML_DisplayString *ds;
    
    if (_stringList) {
	while ((ds = (HTML_DisplayString *) _stringList->Pop())) {
	    delete ds;
	}
	delete _stringList;
	_stringList = NULL;
    }
}

HTML_DisplayText::HTML_DisplayText(HTML *html, Pad_String *text, Pad_Bool preformatted)
{
    float xmax, ymin;
    HTML_DisplayString *str;
    Pad_Iterator oi;

    _scale = (LineSize / DefaultLineSize) * MSML_Scale;
    _x = x_curr_pos;
    _y = y_curr_pos;
    _stringList = html->Flow_string(text, preformatted);
    html->Add_display_item(this);
    NewLine = 0;		// No pending newlines
    if (!_stringList->Is_empty()) {
	str = (HTML_DisplayString *)_stringList->Last();
	ymin = str->Gety();
	xmax = 0;
	DOLIST(oi, *_stringList, HTML_DisplayString, str) {
	    xmax = MAX(xmax, str->Length());
	}
	xmax = _x + xmax*_scale;
    } else {
	xmax = _x;
	ymin = _y;
    }
    _bbox.Set(_x, ymin, xmax, _y + 1);
}

void
HTML_DisplayText::Update_line_position(float dy)
{
    HTML_DisplayString *str;
    
    if (_stringList) {
                                // We only want to update the final line
	str = (HTML_DisplayString *)_stringList->Last();
	if (str) {
	    str->Sety(str->Gety() + dy);
	}
	_bbox.Translate(0, dy);
    }
}

void
HTML_DisplayText::Render(HTML *html) 
{
    Pad_Iterator pi;
    HTML_DisplayString *str;
    int level;
    float mag;

    if (_stringList) {
	Pad_FontRef *font = html->Get_font_ref();
	mag = Pad_renderer->Get_mag() * _scale;
	Pad_renderer->Set_color(html->textColor);
	Pad_renderer->Set_font(*font);
	Pad_renderer->Set_font_height(font->Get_size() * _scale);

	if (mag > 5) {
	    level = 2;
	} else if (mag > 2) {
	    level = 1;
	} else {
	    level = 0;
	}
				// Skip intermediate refinement levels
	if (((Pad_View *)Pad_prc->views.Last())->pad->level > 0) {
	    level = TEXT_LEVELS;
	}

	html->refineNeeded = FALSE;

	DOLIST(pi, *_stringList, HTML_DisplayString, str) {
	    if (Pad_renderer->Draw_string(str->Get(), level, str->Getx(), str->Gety(), 0.0))
	      html->refineNeeded = TRUE;
	}

    }
}

HTML_DisplayImage::~HTML_DisplayImage() 
{
}

HTML_DisplayImage::HTML_DisplayImage(HTML_Entry *new_entry, HTML *html, Pad_ImageData *image, Pad_ImageData *altImage)
{
    _entry = new_entry;
    _image = image;
    _altImage = altImage;
    _scale = IMAGE_SIZE * MSML_Scale;
    _height = image->height;
    _width = image->width;

    html->Add_display_item(this);
    MaxLineSize = MAX(MaxLineSize, ((_scale * _height) + (HTML_VERTICAL_IMAGE_OFFSET * MSML_Scale)));
				// Make sure there's enough room on this line
    if ((x_curr_pos != x_left_marg) && ((x_curr_pos + (_scale * _width)) > html->width)) {
	html->Start_new_line(1);
    }

    Pad_FontRef *font = html->Get_font_ref();
    Pad_renderer->Set_font(*font);
    _x = x_curr_pos;
    _y = y_curr_pos;
    x_curr_pos += (_scale * _image->width);
    x_curr_pos += Pad_renderer->Get_char_width(' ') * MSML_Scale;  // Add some space
    NewLine = 0;		// No pending newlines
    _bbox.Set(_x, _y, _x + (_width * _scale), _y + (_height * _scale));
}

void
HTML_DisplayImage::Render(HTML *html) 
{
    Pad_ImageData *image = NULL;
    float x2, y2;

				// Draw regular image if not too small
				// or alt image if too small.
    if (html->pixelDim > HTML_MIN_SIZE) {
	if (_image) {
	    image = _image;
	} else if (_altImage) {
	    image = _altImage;
	}
    } else {
	if (_altImage) {
	    image = _altImage;
	} else {
	    image = _image;
	}
    }

    if (_image) {
	x2 = _x + (_image->width * _scale);
	y2 = _y + (_image->height * _scale);
    } else if (_altImage) {
	x2 = _x + (_altImage->width * _scale);
	y2 = _y + (_altImage->height * _scale);
    } else {
        x2 = _x;
        y2 = _y;
    }
	
    if (image) {
	html->Draw_image(image, _x, _y, x2, y2);
    }
}

void
HTML_DisplayImage::Set_image(Pad_ImageData *img)
{
    _image = img;
}

HTML_DisplayAnchor::~HTML_DisplayAnchor(void) 
{
    HTML_Anchor *anc;

    Set_image(NULL);		// Make sure associated anchor doesn't render this image
    if (_anchors) {
	while ((anc = (HTML_Anchor *)_anchors->Pop())) {
	    anc->displayAnchor = NULL;
	}
	delete _anchors;
	_anchors = NULL;
    }
}

HTML_DisplayAnchor::HTML_DisplayAnchor(void) 
{
    _x = x_curr_pos;
    _y = y_curr_pos;  
    _anchors = NULL;
}

HTML_DisplayAnchor::HTML_DisplayAnchor(HTML_Entry *new_entry,
				       HTML *html,
				       Pad_ImageData *image,
				       Pad_ImageData *altImage,
				       int w, int h,
				       Pad_Bool ismap,
				       int borderWidth) 
{
    _entry = new_entry;
    HTML_AnchorImage *anchor = new HTML_AnchorImage(html->pad, html, image, altImage, w, h);
    anchor->Set_map(ismap);
    anchor->Set_imageborder_width(borderWidth);
    anchor->displayAnchor = this;
    _anchors = new Pad_List();
    _anchors->Push_last(anchor);
    html->Add_anchor(anchor);
    
    if (!image) {		// Image not yet loaded
	_scale = 1 * MSML_Scale;
	_width = w;
	_height = h;
	anchor->Set_state(HTML_ANCHOR_NOTLOADED);
    } else {
	_scale = IMAGE_SIZE * MSML_Scale;
	_width = image->width;
	_height = image->height;
	anchor->Set_state(HTML_ANCHOR_UNVISITED);
    }
    html->Add_display_item(this);
                                // We need to add an offset to the top of the
                                // image, and we need to take into account
                                // any MSML scaling going on
    MaxLineSize = MAX(MaxLineSize, (_scale * _height) + (HTML_VERTICAL_IMAGE_OFFSET * MSML_Scale));
				// Make sure there's enough room on this line
    if ((x_curr_pos != x_left_marg) && ((x_curr_pos + (_scale * _width)) > html->width)) {
	html->Start_new_line(1);
    }

    Pad_FontRef *font = html->Get_font_ref();
    Pad_renderer->Set_font(*font);
    _x = x_curr_pos;
    _y = y_curr_pos;
    _bbox.Set(_x, _y, _x + (_width * _scale), _y + (_height * _scale));
    x_curr_pos += (_scale * _width);
    x_curr_pos += Pad_renderer->Get_char_width(' ') * MSML_Scale;  // Add some space
    NewLine = 0;
}

HTML_DisplayAnchor::HTML_DisplayAnchor(HTML *html,
				       Pad_String *text) 
{
    _anchors = new Pad_List();
    float orig_x = x_curr_pos;
    float orig_y = y_curr_pos;
    Pad_Iterator pi;
    HTML_DisplayString *str;
    HTML_AnchorText *anchor;
    Pad_List *stringList;
    Pad_Bool dummy_text = FALSE;
    Pad_String fontName;

    if (!text) {
	text = new Pad_String;
	dummy_text = TRUE;
    }

    stringList = html->Flow_string(text, FALSE);
    if (stringList->Is_empty()) {
	str = new HTML_DisplayString();
	stringList->Push(str);
    }

    html->Get_font(fontName);

    DOLIST(pi, *stringList, HTML_DisplayString, str) {
	anchor = new HTML_AnchorText(html->pad, html, str);
	anchor->Set_font(fontName.Get());
	anchor->displayAnchor = this;
	_anchors->Push_last(anchor);
	html->Add_anchor(anchor);
	anchor->Set_state(HTML_ANCHOR_UNVISITED);
    }
    delete stringList;
    if (dummy_text) {
	delete text;
    }
    
    _x = x_curr_pos;
    _y = y_curr_pos;
    _bbox.Set(orig_x, _y, _x, orig_y);
    html->Add_display_item(this);
    NewLine = 0;
}

void
HTML_DisplayAnchor::Remove_anchor(HTML_Anchor *anchor)
{
    _anchors->Remove(anchor);
}

void
HTML_DisplayAnchor::Render(HTML *) 
{
}

void
HTML_DisplayAnchor::Update_line_position(float dy)
{
    Pad_Iterator pi;
    HTML_Anchor *anchor;
    
    DOLIST(pi, *_anchors, HTML_Anchor, anchor) {
	anchor->Update_line_position(dy);
    }
    _bbox.Translate(0, dy);
}

void
HTML_DisplayAnchor::Set_image(Pad_ImageData *img)
{
    Pad_Iterator pi;
    HTML_Anchor *anchor;
    
    DOLIST(pi, *_anchors, HTML_Anchor, anchor) {
	anchor->Set_imagedata(img);
    }
}

void
HTML::Draw_image(Pad_ImageData *image, float x1, float y1, float x2, float y2)
{
    Pad_Bool dither;
    
    if (image->dpy->truecolor) {
	dither = FALSE;
	refineNeeded = FALSE;
    } else {
	if (pad->level > 0) {
	    dither = TRUE;
	} else {
	    dither = FALSE;
	    refineNeeded = TRUE;
	}
    }
  
    Pad_renderer->Draw_image(image, x1, y1, x2, y2, dither);
}


void HTML_DisplayString::Setx(float new_x) {_x = new_x;}
void HTML_DisplayString::Sety(float new_y) {_y = new_y;}
float HTML_DisplayString::Getx(void) const {return _x;}
float HTML_DisplayString::Gety(void) const {return _y;}

HTML_Entry::HTML_Entry(void) {}

HTML_Entry::~HTML_Entry(void) {}

void
HTML_Entry::Remove_display_item(void)
{
}

Pad_Bool
HTML_Entry::Contains_text(char *) 
{
    return FALSE;
}

Pad_Bool
HTML_Entry::Get_text(Pad_String &) 
{
    return(FALSE);
}

Pad_Bool
HTML_Entry::Remove_obj(Pad_Object *)
{
    return FALSE;
}

Pad_Bool
HTML_EntryPadObject::Remove_obj(Pad_Object *o)
{
    if (_obj == o){
	return TRUE;
    }
    return FALSE;
}

HTML_EntryPadObject::~HTML_EntryPadObject(void){}

HTML_EntryPadObject::HTML_EntryPadObject(void)
{
    _obj = NULL;
    _origSize = 0.0;
    _scaledSize = 0.0;
}

HTML_EntryPadObject::HTML_EntryPadObject(Pad_Object *o)
{
    if (o) {
	_obj = o;
	_origSize = _obj->transform.Get_scale();
	_scaledSize = _obj->transform.Get_scale();
    }
}

void
HTML_EntryPadObject::Layout(HTML *html)
{
    HTML_DisplayPadObject *dpo;

				// Check if user changed object's size on us
    if ((_obj->transform.Get_scale() != _origSize) && (_obj->transform.Get_scale() != _scaledSize)) {
	_origSize = _obj->transform.Get_scale();
    }
    _scaledSize = _origSize * MSML_Scale;
    _obj->transform.Set_scale(_scaledSize);
    _obj->Update_global_bounding_box();
    dpo = new HTML_DisplayPadObject(html, _obj);
}

HTML_EntryTag::~HTML_EntryTag(void) 
{
    HTML_Key *k;
    
    if (_keyList) {
	while ((k = (HTML_Key *) _keyList->Pop())) {
	    delete k;
	}
	delete _keyList;
	_keyList = NULL;
    }
}

HTML_EntryTag::HTML_EntryTag(void) 
{
    _keyList = NULL;
}

void
HTML_EntryTag::Set_tag(HTML_EntryTag *et)
{
   HTML_Key *key;
                                // This is a destructive method, it does
                                // not do a copy. So only use this if
                                // you are planning on freeing up et
                                // afterwards....
   if (et->_keyList) {
       _keyList = new Pad_List();
      
       while ((key = (HTML_Key *) et->_keyList->Pop())) {
	   _keyList->Push(key);
       }
   }
}

HTML_EntryTag::HTML_EntryTag(HTML_EntryTag *oldtag) 
{
   if (oldtag){
       Set_tag(oldtag);          // copy the key_list
       
				// This constructor is only called
				// from HTML::Parse_tag, which expects
                                // us to delete oldtag, so we will
       delete oldtag; 
   } else {
       _keyList = NULL;
   }
}

HTML_EntryTag::HTML_EntryTag(HTML *html, int &index)
{
#define KEY_STATE_KEY    0   // Building key
#define KEY_STATE_EQUAL  1   // Looking for '='
#define KEY_STATE_QUOTE  2   // Looking for a quote
#define KEY_STATE_VALUE  3   // Looking for a key value
    int state;
    char c;
    Pad_Bool found = FALSE;
    Pad_Bool quote = FALSE;	// Flag to know if inside quoted string.
    HTML_Key *key = new HTML_Key;
    
    _keyList = new Pad_List();
    
    html->Skip_white_space(index);
    state = KEY_STATE_KEY;
    do {
	c = html->source[index++];
	switch (c) {
	  case '>':
	    if (state == KEY_STATE_KEY) {
		key->Set_type();
	    }
	    state = Finish_key(html, index, key);
	    key = new HTML_Key;
	    found = TRUE;
	    break;
	  case '\0':
	    found = TRUE;
	    break;
	  case '\"':			// hilit19 gets messed up here: "
            if (state == KEY_STATE_QUOTE) {
		quote = TRUE;
		state = KEY_STATE_VALUE;
	    } else if (state == KEY_STATE_VALUE) {
		if (quote == FALSE) {
		    quote = TRUE;
		} else {
		    quote = FALSE;
		    state = Finish_key(html, index, key);
		    key = new HTML_Key;
		}
	    } 
	    break;
	  case ' ':
	  case '\t':
	  case '\n':
	  case '=':
	    if (state == KEY_STATE_KEY) {
		key->Set_type();
		if (c == '=') {
		    state = KEY_STATE_QUOTE;
		} else {
		    state = KEY_STATE_EQUAL;
		}
	    } else if (state == KEY_STATE_EQUAL) {
		if (c == '=') {
		    state = KEY_STATE_QUOTE;
		}
	    } else if (state == KEY_STATE_QUOTE) {
				// Nothing to do here
	    } else {		// Waiting for value
		if (c == '=') {
//		    printf("Parse warning: unexpected '=' in tag value.  Ignoring\n");
		} else {
		    if (quote) {
			key->value.Append(c);
		    } else {
			state = Finish_key(html, index, key);
			key = new HTML_Key;
		    }
		}
	    }
	    break;
	  case '{':
            if (state == KEY_STATE_QUOTE) {
		index--;
		key->Parse_script(html, index);
		break;
            } 
                                // We want to fall through here into the
                                // default case because we weren't expecting
                                // a script
	  default:
	    if (state == KEY_STATE_EQUAL) {
				// More text.  This key must not have a value,
				//   so start a new key.
		state = Finish_key(html, index, key);
		key = new HTML_Key;
	    } else if (state == KEY_STATE_QUOTE) {
		state = KEY_STATE_VALUE;
	    }
	    key->value.Append(c);
	}
    } while (!found);
    delete key;
}

void
HTML_EntryTag::Parse(HTML *)
{}

void
HTML_EntryTag::Layout(HTML *)
{}

HTML_EntryTagStart::~HTML_EntryTagStart(void) {
}

HTML_EntryTagStart::HTML_EntryTagStart(void) :
    HTML_EntryTag()
{
}

HTML_EntryTagStart::HTML_EntryTagStart(HTML_EntryTag *tag) :
    HTML_EntryTag(tag)
{
}

HTML_EntryTagEnd::~HTML_EntryTagEnd(void) {
}

HTML_EntryTagEnd::HTML_EntryTagEnd(void) : HTML_EntryTag() 
{
    _endTag = (HtmlEntryTagType) TAG_UNKNOWN;
}

int HTML_EntryTagEnd::Get_tag(void) 
{
    return _endTag;
}

HTML_EntryText::~HTML_EntryText(void) {
}

HTML_EntryText::HTML_EntryText(void) {
}

Pad_Bool
HTML_EntryText::Contains_text(char *string) {
    Pad_Bool rc;

    if (casestrstr(_text.Get(), string)) {
	rc = TRUE;
    } else {
	rc = FALSE;
    }

    return rc;
}

Pad_Bool
HTML_EntryText::Get_text(Pad_String &string) 
{
    string = "";
    string = _text;
    return(TRUE);
}

void
HTML_EntryText::Set_text(Pad_String *ps) 
{
    _text = ps;
}

HTML_EntryTagImage::~HTML_EntryTagImage(void) 
{
    if (_image) {
	Pad_renderer->Free_image(_image);
	_image = NULL;
    }
    if (_altImage) {
	Pad_renderer->Free_image(_altImage);
	_altImage = NULL;
    }
    if (_displayitem) {
				// Make sure the associated display and anchor items
				// don't try and use this image any more.
	_displayitem->Set_image(NULL);
	_displayitem = NULL;
    }
    if (data) {
	delete [] data;
	data = NULL;
    }
    if (imgUrl) {
	delete imgUrl;
	imgUrl = NULL;
    }
    if (urlItem) {
	html->pad->Url_cancel(urlItem);
	urlItem = NULL;
    }
}

HTML_EntryTagImage::HTML_EntryTagImage(HTML_EntryTag *tag, HTML *htmlptr) : 
    HTML_EntryTag(tag)
{
    html = htmlptr;
    _image = NULL;
    _altImage = NULL;
    _found = FALSE;
    _ismap = FALSE;
    _borderWidth = 2;  
    width = (int)DUMMY_IMAGE_SIZE;
    height = (int)DUMMY_IMAGE_SIZE;
    _displayitem = NULL;
    data = NULL;
    index = 0;
    space = 0;
    imgUrl = NULL;
    urlItem = NULL;
}

void
HTML_EntryTagImage::Remove_display_item(void)
{
    _displayitem = NULL;
}

HTML_EntryTagTcl::HTML_EntryTagTcl(void) :
HTML_EntryTag()
{
    _script = NULL;
    _objects = NULL;
}

HTML_EntryTagTcl::HTML_EntryTagTcl(HTML_EntryTag *oldtag) :
HTML_EntryTag(oldtag)
{
    HTML_Key *key;
    Pad_Iterator pi;
  
    _script = NULL;
    _objects = NULL;
    if (_keyList) {
	DOLIST(pi, *_keyList, HTML_Key, key) {
	    if (key->key == KEY_PAD_TCL) {
		_script = new Pad_String(key->value);
		break;
	    }
	}
    }
}

HTML_EntryTagTcl::~HTML_EntryTagTcl(void) {
    HTML_TclItem *ti;
    
    if (_objects) {
	while ((ti = (HTML_TclItem *)_objects->Pop())) {
	    delete ti->obj;
	}
	delete _objects;
	_objects = NULL;
    }
}

void
HTML_EntryTagTcl::Layout(HTML *html)
{
    HTML_DisplayPadObject *dpo;
    HTML_TclItem *ti;
    Pad_Iterator pi;
    
    if (_objects) {
	DOLIST(pi, *_objects, HTML_TclItem, ti) {
				// Check if user changed object's size on us
	    if ((ti->obj->transform.Get_scale() != ti->origSize) && 
		(ti->obj->transform.Get_scale() != ti->scaledSize)) {
		ti->origSize = ti->obj->transform.Get_scale();
	    }
	    ti->scaledSize = ti->origSize * MSML_Scale;
	    ti->obj->transform.Set_scale(ti->scaledSize);
	    ti->obj->Update_global_bounding_box();
	    dpo = new HTML_DisplayPadObject(html, ti->obj);
	}
    }
}

void
HTML_EntryTagTcl::Add(Pad_Object *o)
{
    HTML_TclItem *ti;

    if (!o) {
	return;
    }

    if (!_objects) {
	_objects = new Pad_List();
    }
    ti = new HTML_TclItem();
    ti->obj = o;
    ti->origSize = o->transform.Get_scale();
    ti->scaledSize = o->transform.Get_scale();
    _objects->Push_last(ti);
} 

Pad_Bool
HTML_EntryTagTcl::Remove_obj(Pad_Object *obj)
{
    HTML_TclItem *ti;
    Pad_Iterator pi;

    if (_objects) {
	DOLIST(pi, *_objects, HTML_TclItem, ti) {
	    if (ti->obj == obj) {
		_objects->Remove(ti);
				// obj was removed from list
		if (_objects->Is_empty()) {
                                // list is now empty so return TRUE
                                // telling HTML::Remove that it should
                                // delete the entry from entryList
		    return(TRUE);
		}
		break;
	    }
	}
	return(FALSE);		// obj either wasn't in the list or
				// the list is not empty, so return FALSE
				// telling HTML::Remove() not to delete
				// this from entryList
    } else {
	return(FALSE);
    }
}

    
void
HTML_EntryTagImage::Set_image(Pad_ImageData *image)
{
    if (image) {
	if (!_image) {
	    _image = image;
	    _found = TRUE;	// Mark it as found for future layouts
	} else {
	    _altImage = image;
	}
    }
}

static Pad_Bool Image_UpdateCB(ClientData clientData, int, char *new_data, int len);
static Pad_Bool Image_DoneCB(ClientData clientData, int);
static Pad_Bool Image_ErrorCB(ClientData clientData, int, char *, int);

//
// Callback to start image URL transfer
//
static void
Pad_InlineimageCB(ClientData clientData)
{
    HTML_EntryTagImage *entry;
    Pad_UrlItem *urlItem;

    entry = (HTML_EntryTagImage *)clientData;
    urlItem = entry->html->pad->Url_fetch(entry->imgUrl->Get(), NULL, NULL, entry, FALSE, 
					   Image_UpdateCB, Image_DoneCB, Image_ErrorCB);
    if (urlItem) {
	entry->urlItem = urlItem;
	if (urlItem->docLength > 0) {
	    entry->space = urlItem->docLength;
	} else {
	    entry->space = entry->width * entry->height;
	}
	entry->data = new unsigned char[entry->space];
	entry->index = 0;
    }
}

//
// Callback for when a URL is updated (i.e., more data arrived)
//
static Pad_Bool
Image_UpdateCB(ClientData clientData, int, char *new_data, int len)
{
    unsigned char *more_data;
    HTML_EntryTagImage *entry = (HTML_EntryTagImage *)clientData;

    if (!strncasecmp(new_data, "<HEAD>", 6)) {
				// We're getting an HTML file when
				// we expected an image.  Probably the
				// image wasn't there.  So, forget about this one.
	Pad_errorString = "Invalid image";
	return(FALSE);
    }

				// Double space if not big enough
    if (len > (entry->space - entry->index)) {
	entry->space = MAX(len, entry->space);
	entry->space *= 2;
	more_data = new unsigned char[entry->space];
	memcpy(more_data, entry->data, entry->index);
	delete [] entry->data;
	entry->data = more_data;
    }
    memcpy(&entry->data[entry->index], new_data, len);
    entry->index += len;

    return(TRUE);
}

//
// Callback for when a URL has completed loading
//
static Pad_Bool
Image_DoneCB(ClientData clientData, int)
{
    int result;
    HTML_EntryTagImage *entry = (HTML_EntryTagImage *)clientData;
    Pad_String cmd;
    Pad_ImageData *imageData;

    entry->urlItem = NULL;

    imageData = Pad_renderer->Alloc_image(entry->data, entry->index);

    entry->Set_image(imageData);
    delete [] entry->data;
    entry->data = NULL;
    entry->html->inlineImages--;

    entry->html->Create_display_list();

    if (entry->altImageUrl != "") {
				// Alt image here, get that one.
	if (entry->imgUrl) {
	    delete entry->imgUrl;
	}
	entry->imgUrl = new Pad_String(entry->Get_pathname(entry->altImageUrl.Get(), entry->html));
	Pad_CreateTimerHandler(IDLE_DELAY, Pad_InlineimageCB, entry);
	entry->html->inlineImages++;
	entry->altImageUrl = "";

	return(TRUE);
    }

				// Call 'update' script.
    if (entry->html->updateScript) {
	cmd = entry->html->updateScript->Get();
	cmd += " ";
	cmd += entry->html->id;
	result = Pad_GlobalEval(entry->html->updateScript->Language(), cmd.Get());
	if (result == TCL_ERROR) {
	    Tcl_AddErrorInfo(entry->html->pad->view->win->interp, "\n    (Pad URL Fetch)");
	    Pad_Background_error(entry->html->pad->view->win->interp);
	}
    }

				// If this was the last in-line image, then html is loaded, so call 
				// "done" script
    if ((entry->html->inlineImages == 0) && (entry->html->doneScript)) {
	cmd = entry->html->doneScript->Get();
	cmd += " ";
	cmd += entry->html->id;
	result = Pad_GlobalEval(entry->html->doneScript->Language(), cmd.Get());
	if (result == TCL_ERROR) {
	    Tcl_AddErrorInfo(entry->html->pad->view->win->interp, "\n    (Pad URL Fetch)");
	    Pad_Background_error(entry->html->pad->view->win->interp);
	}
    }

    return(TRUE);
}

//
// Callback for when there has been an error reading an in-line image.
// Delete the inline image, and relayout the page
//
static Pad_Bool
Image_ErrorCB(ClientData clientData, int, char *, int)
{
    int result;
    Pad_String cmd;
    HTML_EntryTagImage *entry = (HTML_EntryTagImage *)clientData;

    entry->html->Delete_display_list_item(entry);
    entry->html->entryList.Remove(entry);
    entry->html->Damage();
    entry->html->inlineImages--;

				// If this was the last in-line image, then html is loaded, so call 
				// "done" script
    if ((entry->html->inlineImages == 0) && (entry->html->doneScript)) {
	cmd = entry->html->doneScript->Get();
	cmd += " ";
	cmd += entry->html->id;
	result = Pad_GlobalEval(entry->html->doneScript->Language(), cmd.Get());
	if (result == TCL_ERROR) {
	    Tcl_AddErrorInfo(entry->html->pad->view->win->interp, "\n    (Pad URL Fetch)");
	    Pad_Background_error(entry->html->pad->view->win->interp);
	}
    }

    delete entry;

    return(TRUE);
}

//
// Add an image to the current HTML object.
//
void
HTML_EntryTagImage::Parse(HTML *html) 
{
    HTML_Key *key, *src_key;
    Pad_Iterator pi;
    
    src_key = NULL;
    if (_keyList) {
	DOLIST(pi, *_keyList, HTML_Key, key) {
	    if (key->key == KEY_SRC) {
		src_key = key;
	    }
	    if (key->key == KEY_ALT) {
		if ((key->value.Length() >= 4) &&
		    !strcasecmp(key->value.Get(key->value.Length() - 4), ".gif")) {
		    altImageUrl = key->value;
		}
	    }
	    if (key->key == KEY_WIDTH) {
		width = (int)(IMAGE_SIZE * atoi(key->value.Get()));
	    }
	    if (key->key == KEY_HEIGHT) {
		height = (int)(IMAGE_SIZE * atoi(key->value.Get()));
	    }
	    if (key->key == KEY_ISMAP) {
		_ismap = TRUE;
	    }
	    if (key->key == KEY_BORDER) {
		_borderWidth = atoi(key->value.Get());
	    }
	}
    }
    
				// Fetch image from URL
    if (src_key) {
				// Delay image fetching until after HTML text is loaded.
				// This will make text appear faster, and will avoid
				// problems if there is an error loading the image
				// resulting in the html getting deleted.
	imgUrl = new Pad_String(Get_pathname(src_key->value.Get(), html));
	Pad_CreateTimerHandler(IDLE_DELAY, Pad_InlineimageCB, this);
	html->inlineImages++;
    }
}

void
HTML_EntryTagImage::Layout(HTML *html)
{
    HTML_Display *di;

				// If an anchor
    if (Anchor_active) {
	di = new HTML_DisplayAnchor(this, html, _image, _altImage, width, height, 
				    _ismap, _borderWidth);
    } else {
	if (_found) {
				// If image loaded
	    di = new HTML_DisplayImage(this, html, _image, _altImage);
	} else {
				// Else, image not loaded
	    di = new HTML_DisplayRectangle(this, html, width, height);
	    ((HTML_DisplayRectangle *)di)->Set_color("gray50");
	}
    }
    _displayitem = di;
}

HTML_EntryTagAnchor::~HTML_EntryTagAnchor(void) 
{
    if (_href) {
	delete _href;
	_href = NULL;
    }
    if (_name) {
	delete _name;
	_name = NULL;
    }
}

HTML_EntryTagAnchor::HTML_EntryTagAnchor(void) :
    HTML_EntryTag() 
{
    _href = NULL;
    _name = NULL;
}

HTML_EntryTagAnchor::HTML_EntryTagAnchor(HTML_EntryTag *tag) :
    HTML_EntryTag(tag)
{
    _href = NULL;
    _name = NULL;
}

//
// Add a new anchor to the html object
//
void
HTML_EntryTagAnchor::Parse(HTML *html) 
{
    HTML_Key *key;
    
				// Search anchor for href key
				// If an anchor has no href don't do anything
				// this will leave any text or image as just text or image        
    Pad_Iterator pi;
    if (_keyList) {
	DOLIST(pi, *_keyList, HTML_Key, key) {
	    switch (key->key) {
	      case KEY_HREF:
		_href = Get_pathname(key->value.Get(), html);
		break;
	      case KEY_NAME:
		_name = new Pad_String(key->value.Get());
		break;
	    }
	}
    }
}

void
HTML_EntryTagAnchor::Layout(HTML *)
{
    Anchor_first = NULL;
    Anchor_current = NULL;
    Anchor_href = _href;
    Anchor_name = _name;
    Anchor_active = TRUE;
}

void
HTML_EntryTagStart::Layout(HTML *html)
{
    HTML_Key *key;
    Pad_Iterator pi;
    Pad_String colorname;

    switch (_startTag) {
      case TAG_HTML:
	if (_keyList) {
	    DOLIST(pi, *_keyList, HTML_Key, key) {
		colorname = "";
		if (key->key == KEY_WIDTH) {
		    html->width = atoi(key->value.Get());
		}
	    }
	}
	break;
      case TAG_BODY:
	if (_keyList) {
	    DOLIST(pi, *_keyList, HTML_Key, key) {
		colorname = "";
		if (key->key == KEY_BGCOLOR) {
		    if ((key->value.Get())[0] != '#') {
			colorname = "#";
		    }
		    colorname += key->value;
		    html->Set_fill(colorname.Get());
		    html->htmlFlags &= ~HTML_FILL_SET;
		}
		if (key->key == KEY_TEXT) {
		    if ((key->value.Get())[0] != '#') {
			colorname = "#";
		    }
		    colorname += key->value;
		    html->textColor.Set(colorname.Get());
		}
		if (key->key == KEY_LINK) {
		    if ((key->value.Get())[0] != '#') {
			colorname = "#";
		    }
		    colorname += key->value;
		    html->linkColor.Set(colorname.Get());
		}
		if (key->key == KEY_ALINK) {
		    if ((key->value.Get())[0] != '#') {
			colorname = "#";
		    }
		    colorname += key->value;
		    html->alinkColor.Set(colorname.Get());
		}
		if (key->key == KEY_VLINK) {
		    if ((key->value.Get())[0] != '#') {
			colorname = "#";
		    }
		    colorname += key->value;
		    html->vlinkColor.Set(colorname.Get());
		}
	    }
	}
	break;
      case TAG_ADDRESS:
	break;
      case TAG_H1:
	H1 = TRUE;
	html->Start_new_line(2);
	break;
      case TAG_H2:
	H2 = TRUE;
	html->Start_new_line(2);
	break;
      case TAG_H3:
	H3 = TRUE;
	html->Start_new_line(2);
	break;
      case TAG_H4:
	H4 = TRUE;
	html->Start_new_line(2);
	break;
      case TAG_H5:
	H5 = TRUE;
	html->Start_new_line(2);
	break;
      case TAG_H6:
	H6 = TRUE;
	html->Start_new_line(2);
	break;
      case TAG_A:
	break;
      case TAG_IMG:
	break;
      case TAG_P:
	html->Start_new_line(2);
	break;
      case TAG_BR:
	html->Start_new_line(0);
	y_curr_pos -=0.25;
	break;
      case TAG_HR:
	{
	    HTML_DisplayRectangle *dp;
	    html->Start_new_line(1);

	    dp = new HTML_DisplayRectangle(this, html, html->width, 0.25);
	    dp->Set_color("gray60");
	    x_curr_pos = x_left_marg;
	    y_curr_pos -= 0.25;

	    dp = new HTML_DisplayRectangle(this, html, html->width, 0.25);
	    dp->Set_color("gray97");
	    
	    x_curr_pos = html->width;
	    html->Start_new_line(2);
	}
        
	break;
      case TAG_UL:
	x_left_marg = MSML_Push_left_marg(x_left_marg + (MSML_Scale * 3.0));
	html->Start_new_line(2);
	break;
      case TAG_OL:
	break;
      case TAG_LI:
	{
	    HTML_DisplayRectangle *dp;
	    Pad_String colorName;
	    html->Start_new_line(1);
	    dp = new HTML_DisplayRectangle(this, html, 
					  0.5, 0.5,
					  x_left_marg - (MSML_Scale),
					  y_curr_pos + MSML_Scale*((0.5 * (LineSize / DefaultLineSize)) - 0.25));

	    html->textColor.Get(colorName);
	    dp->Set_color(colorName.Get());
	    x_curr_pos = x_left_marg;
	}
	break;
      case TAG_TITLE:
	Title = TRUE;
	// Ignore title for now
	//	Start_new_line(2);
	break;
      case TAG_HEAD:
	break;
      case TAG_EM:
	break;
      case TAG_I:
	break;
      case TAG_B:
	break;
      case TAG_CENTER:
	break;
      case TAG_MENU:
	break;
      case TAG_CITE:
	break;
      case TAG_FONT:
	break;
      case TAG_DL:
	if (HTML_DD) {
	    x_left_marg = MSML_Pop_left_marg();
	    HTML_DD = FALSE;
	}
	html->Start_new_line(2);
	break;
      case TAG_DE:
	break;
      case TAG_DT:
	if (HTML_DD) {
	    x_left_marg = MSML_Pop_left_marg();
	    HTML_DD = FALSE;
	}
	html->Start_new_line(1);
	x_curr_pos = x_left_marg;
	break;
      case TAG_DD:
	if (HTML_DD) {
	    x_left_marg = MSML_Pop_left_marg();
	}
	x_left_marg = MSML_Push_left_marg(x_left_marg + (MSML_Scale * 6.0));
	html->Start_new_line(1);
	x_curr_pos = x_left_marg;
	HTML_DD = TRUE;
	break;
      case TAG_PRE:
        PreFormatted = TRUE;
	break;
      case TAG_KBD:
	break;
      case TAG_STRONG:
	break;
      case TAG_BASE:
	if (_keyList) {
	    DOLIST(pi, *_keyList, HTML_Key, key) {
		if (key->key == KEY_HREF) {
		    html->url = key->value.Get();
		    break;
		}
	    }
	}
	break;
      case TAG_CODE:
	break;
      case TAG_COMMENT1:
	break;
      case TAG_COMMENT2:
	break;
      case TAG_META:
        if (_keyList) {
	    DOLIST(pi, *_keyList, HTML_Key, key) {
              switch (key->key){
                  case KEY_PAD_SCALE_BEGIN :
                                // if we see a begin we want to add the new
                                // value to the stack, and multiply the
                                // current value by the new scale factor
		    MSML_Scale *= MSML_Push_scale(atof(key->value.Get()));
		    break;
                  case KEY_PAD_SCALE_END :
                                // if we see an end, we want to remove the
                                // top of stack value and return the sclae
                                // multiplier to the previous state by
                                // dividing by the old top of stack 
		    MSML_Scale /= MSML_Pop_scale();
		    break;
                  case KEY_PAD_TCL :
                    cerr << "Huh?? How did we get here?";
                    cerr << "This should be an HTML_EntryTagTcl" << endl;
                    break;
                  default :
                    cerr << "Unknown Meta tag" << key->key << endl;
                    break;
              }
	    }
        }
	break;
      case TAG_TEXTAREA:
	break;
      case TAG_FORM:
	break;
      case TAG_INPUT:
	break;
      case TAG_TABLE:
	break;
      case TAG_TR:
	break;
      case TAG_TD:
	break;
      case TAG_TH:
	break;
      case TAG_BLOCKQUOTE:
	break;
      case TAG_BLINK:
	break;
      case TAG_LINK:
	break;
      case TAG_CAPTION:
	break;
      case TAG_HEADER:
	break;
      case TAG_UNKNOWN:
//	printf("Parse warning: Unknown start tag: %s\n", tag.Get());
	break;
      default:
	printf("HTML internal error: Didn't handle start tag: %d\n", _startTag);
    }
}

void
HTML_EntryTagEnd::Layout(HTML *html)
{
    switch (_endTag) {
      case TAG_HTML:
	break;
      case TAG_BODY:
	break;
      case TAG_ADDRESS:
	html->Start_new_line(2);
	break;
      case TAG_H1:
	H1 = FALSE;
	html->Start_new_line(2);
	break;
      case TAG_H2:
	H2 = FALSE;
	html->Start_new_line(2);
	break;
      case TAG_H3:
	H3 = FALSE;
	html->Start_new_line(2);
	break;
      case TAG_H4:
	H4 = FALSE;
	html->Start_new_line(2);
	break;
      case TAG_H5:
	H5 = FALSE;
	html->Start_new_line(2);
	break;
      case TAG_H6:
	H6 = FALSE;
	html->Start_new_line(2);
	break;
      case TAG_P:
	break;
      case TAG_BR:
	break;
      case TAG_HR:
	break;
      case TAG_A:
	  if (!Anchor_current) {
				// No element was made for this anchor.
				// We'll make an empty one in case it
				// is the destination of an index.
	      new HTML_DisplayAnchor(html, NULL);
	  }
	Anchor_first = NULL;
	Anchor_current = NULL;
	Anchor_href = "";
	Anchor_name = "";
	Anchor_active = FALSE;
	break;
      case TAG_IMG:
	break;
      case TAG_UL:
	x_left_marg = MSML_Pop_left_marg();
	html->Start_new_line(2);
	break;
      case TAG_OL:
	break;
      case TAG_LI:
	break;
      case TAG_TITLE:
	Title = FALSE;
	// Ignore title for now
	//	    Start_new_line(2);
	break;
      case TAG_HEAD:
	break;
      case TAG_EM:
	break;
      case TAG_I:
	break;
      case TAG_B:
	break;
      case TAG_CENTER:
	break;
      case TAG_MENU:
	break;
      case TAG_CITE:
	break;
      case TAG_FONT:
	break;
      case TAG_DL:
	if (HTML_DD) {
	    x_left_marg = MSML_Pop_left_marg();
	    HTML_DD = FALSE;
	}
	html->Start_new_line(2);
	break;
      case TAG_DE:
	break;
      case TAG_DT:
	if (HTML_DD) {
	    x_left_marg = MSML_Pop_left_marg();
	    HTML_DD = FALSE;
	}
	break;
      case TAG_DD:
	if (HTML_DD) {
	    x_left_marg = MSML_Pop_left_marg();
	    HTML_DD = FALSE;
	}
	break;
      case TAG_PRE:
        PreFormatted = FALSE;
	break;
      case TAG_KBD:
	break;
      case TAG_STRONG:
	break;
      case TAG_BASE:
	break;
      case TAG_CODE:
	break;
      case TAG_COMMENT1:
	break;
      case TAG_COMMENT2:
	break;
      case TAG_META:
	break;
      case TAG_TEXTAREA:
	break;
      case TAG_FORM:
	break;
      case TAG_INPUT:
	break;
      case TAG_TABLE:
	break;
      case TAG_TR:
	break;
      case TAG_TD:
	break;
      case TAG_TH:
	break;
      case TAG_BLOCKQUOTE:
	break;
      case TAG_BLINK:
	break;
      case TAG_LINK:
	break;
      case TAG_CAPTION:
	break;
      case TAG_HEADER:
	break;
      case TAG_UNKNOWN:
//	printf("Parse warning: Unknown end tag: %s\n", tag.Get());
	break;
      default:
	printf("HTML internal error: Didn't handle end tag: %d\n", _endTag);
    }
}

void
HTML_EntryText::Layout(HTML *html)
{
    if (Title) {		// Titles not implemented yet
	return;
    }  
    html->Update_line_spacing();  
    if (Anchor_active) {
	new HTML_DisplayAnchor(html, &_text);
    } else {
	new HTML_DisplayText(html, &_text, PreFormatted);
    }
}

//
// Finish up this key, and add it to the list of keys.
//
int
HTML_EntryTag::Finish_key(HTML *html, int &index, HTML_Key *key)
{
    HTML_Key *copy = new HTML_Key(key);
    
    if (!_keyList) {
	_keyList = new Pad_List();
    }
    _keyList->Push(copy);

    html->Skip_white_space(index);

    return(KEY_STATE_KEY);
}

Pad_String *
HTML_Entry::ParseStartTag(int tag) 
{
    Pad_String *string = new Pad_String();
    string->Set(StartTagArray[tag]);
    return(string);
}

Pad_String *
HTML_Entry::ParseEndTag(int tag) 
{
    Pad_String *string = new Pad_String();
    string->Set(StartTagArray[tag]);
    return(string);
}

Pad_List *
HTML_EntryTag::Get_keys(void) 
{
    return(_keyList);
}

HTML_Key::~HTML_Key(void) {
}

HTML_Key::HTML_Key(void) {
    key = KEY_UNKNOWN;
}

//
// Set the key's type based on its current value.
// Reset the value to an empty string. 
//
void
HTML_Key::Set_type()
{
    int i;
    Pad_Bool keyfound;

    i = 0;
    keyfound = FALSE;
    key = KEY_UNKNOWN;
    while ((keyArray[i].name) && !keyfound) {
	if (!strcasecmp(value.Get(),
			keyArray[i].name)) {
	    key = keyArray[i].type;
	    keyfound = TRUE;
	}
	i++;
    }
    value.Set("");
}

HTML_Key::HTML_Key(HTML_Key *original) 
{
    key = original->key;
    value = original->value;
}

void
HTML_Key::Parse_script(HTML *html, int &index)
{
    int open;
    char c;
    Pad_Bool Quote = FALSE;	// Flag to know if inside quoted string.

    c = html->source[index++];
    if (c != '{'){
      cerr << "Error: Parse_script got called without a script" << endl;
      return;
    }
    open = 1;
    do {
      c = html->source[index++];
      switch (c) {
	  case (char)EOF:
	    open = 0;
	    break;
	  case '\\':
	    c = html->source[index++];
	    value.Append(c);
	    break;
	  case '{':
            if (!Quote){
              open++;
            }
	    value.Append(c);
	    break;
          case '}':
            if (!Quote){
              open--;
            }
            if (open){
              value.Append(c); // don't add the last '}'
            }
	    break;
          case '\"':			// hilit19 gets messed up here: "
            Quote = ((Quote) ? FALSE : TRUE);
	    value.Append(c);
	    break;
	  default:
	    value.Append(c);
	}
    } while (open);
}

////////////////////////////////////////////////////////////////////
//
// HTML Anchors
//
////////////////////////////////////////////////////////////////////

#define ANCHOR_DEFAULT_IMAGEDATA         NULL

HTML_Anchor::~HTML_Anchor(void) 
{
    if (displayAnchor) {
	displayAnchor->Remove_anchor(this);
    }
}

HTML_Anchor::HTML_Anchor(Pad *pad, HTML *htmlptr) :
Pad_Object(pad)
{
    _type = PAD_HTML_ANCHOR;
    anchorFlags = PAD_NO_MASK;
    html = htmlptr;
    next = NULL;
    state = HTML_ANCHOR_NOTLOADED;
    displayAnchor = NULL;
    Set_font_default();
}

//
// Set the object's <option> to the value specified by <sval>.
// Note that each option has a specific setter and getter method
// if the correct argument type is available.
//
Pad_Bool
HTML_Anchor::Set_option(Pad_Option *option, char *sval)
{
    Pad_Bool rc = TRUE;

    switch (option->id) {
      case PAD_HTMLOPTION:
	Pad_errorString = "Error: -html is a read-only option";
	rc = FALSE;
	break;
      case PAD_IMAGEDATA:
	{
	    Pad_ImageData *imageData;

	    if (sval == NULL) {
		Set_imagedata(NULL);
	    } else {
		imageData = (Pad_ImageData *)Pad_imageTokenTable.Get(Tk_GetUid(sval));
		if (imageData) {
		    Set_imagedata(imageData);
		} else {
		    Pad_errorString = "Imagetoken not allocated by allocimage: ";
		    rc = FALSE;
		}
	    }
	}
	break;
      case PAD_ISMAP:
	Pad_errorString = "Error: -ismap is a read-only option";
	rc = FALSE;
	break;
      case PAD_URL:
	rc = Set_url(sval);
	break;
      case PAD_STATE:
	if (!strcmp(sval, "unvisited")) {
	    Set_state(HTML_ANCHOR_UNVISITED);
	} else if (!strcmp(sval, "active")) {
	    Set_state(HTML_ANCHOR_ACTIVE);
	} else if (!strcmp(sval, "visited")) {
	    Set_state(HTML_ANCHOR_VISITED);
	} else if (!strcmp(sval, "notloaded")) {
	    Set_state(HTML_ANCHOR_NOTLOADED);
	} else {
	    Pad_errorString = "Error: Invalid state: ";
	    Pad_errorString += sval;
	    rc = FALSE;
	}
	break;
      default:
	rc = Pad_Object::Set_option(option, sval);
	break;
    }

    return(rc);
}

//
// Set the object's <option> to the default value.
//
Pad_Bool
HTML_Anchor::Set_option_default(Pad_Option *option)
{
    Pad_Bool rc = TRUE;

    switch (option->id) {
      case PAD_HTMLOPTION:
	Pad_errorString = "Error: -html is a read-only option";
	rc = FALSE;
	break;
      case PAD_IMAGEDATA:
	Set_imagedata_default();
	break;
      case PAD_ISMAP:
	Pad_errorString = "Error: -ismap is a read-only option";
	rc = FALSE;
	break;
      case PAD_URL:
	rc = Set_url_default();
	break;
      case PAD_STATE:
	Set_state_default();
	break;
      default:
	rc = Pad_Object::Set_option_default(option);
	break;
    }

    return(rc);
}

//
// Return the name.
//
Pad_Bool
HTML::Get_name(Pad_String &result) 
{
    result = "";
    result = url;
    return(TRUE);
}


//
// Return the value of the object's <option> as a string
// in <result>.
//
Pad_Bool
HTML_Anchor::Get_option(Pad_Option *option, Pad_String &result, Pad_Bool &modified)
{
    Pad_Bool rc = TRUE;

    modified = FALSE;
    switch (option->id) {
      case PAD_HTMLOPTION:
	{
	    HTML *html_item;

	    html_item = Get_html();
	    result = html_item->id;
	}
	break;
      case PAD_IMAGEDATA:
	{
	    Pad_ImageData *imageData;
	    
	    imageData = Get_imagedata();
	    if (imageData) {
		result = imageData->Get_token();
	    } else {
		result = "";
	    }
	}
	break;
      case PAD_ISMAP:
	result = Get_map();
	break;
      case PAD_URL:
	result = Get_url();
	break;
      case PAD_STATE:
	switch (Get_state()) {
	  case HTML_ANCHOR_UNVISITED:
	    result = "unvisited";
	    break;
	  case HTML_ANCHOR_ACTIVE:
	    result = "active";
	    break;
	  case HTML_ANCHOR_VISITED:
	    result = "visited";
	    break;
	  case HTML_ANCHOR_NOTLOADED:
	    result = "notloaded";
	    break;
	default:
	    result = "";
	    break;
	}
	break;
      default:
	rc = Pad_Object::Get_option(option, result, modified);
	break;
    }

    return(rc);
}

//
// Return true to continue rendering other objects, or false if rendering
// was interrupted.
//
Pad_Bool
HTML_Anchor::Render(void)
{
    return(TRUE);
}

void
HTML_Anchor::Update_line_position(float dy)
{
    transform.Set_offset_y(transform.Get_offset_y() + dy);
    Update_global_bounding_box();
}

//
// Setters and getters for html item
//
HTML *
HTML_Anchor::Get_html(void)
{
    return(html);
}

//
// Setters and getters for ismap
//
Pad_Bool
HTML_Anchor::Get_map(void)
{
    return(FALSE);
}

Pad_Bool
HTML_AnchorImage::Get_map(void)
{
    return(_ismap);
}

void
HTML_AnchorImage::Set_map(Pad_Bool map)
{
    _ismap = map;
}

//
// Setters and getters for url
//
Pad_Bool
HTML_Anchor::Set_url(char *) 
{
    Pad_errorString = "Error: -url is readonly for htmlanchor's";
    return(FALSE);
}

Pad_Bool
HTML_Anchor::Set_url_default(void)
{
    Pad_errorString = "Error: -url is readonly for htmlanchor's";
    return(FALSE);
}

char *
HTML_Anchor::Get_url(void)
{
    return(href.Get());
}

//
// Getter for anchor name
//
char *
HTML_Anchor::Get_anchor_name(void)
{
    return(name.Get());
}

//
// Setters and getters for anchor image item.
//
void
HTML_Anchor::Set_imagedata(Pad_ImageData *) 
{
}

void
HTML_Anchor::Set_imagedata_default(void)
{
}

Pad_ImageData *
HTML_Anchor::Get_imagedata(void)
{
    return(NULL);
}

//
// Setters and getters for anchor state
//
Pad_Bool
HTML_Anchor::Set_state(int flag)
{
    Pad_Bool rc = TRUE;
    HTML_Anchor *anchor = this;
    HTML_Anchor *first = this;
    
    do {
	switch (flag) {
	  case HTML_ANCHOR_UNVISITED:
	    anchor->penColor.Set(html->linkColor);
	    break;
	  case HTML_ANCHOR_ACTIVE:
	    anchor->penColor.Set(html->alinkColor);
	    break;
	  case HTML_ANCHOR_VISITED:
	    anchor->penColor.Set(html->vlinkColor);
	    break;
	  case HTML_ANCHOR_NOTLOADED:
	    anchor->penColor.Set(html->ulinkColor);
	    break;
	  default:
	    rc = FALSE;
	    break;
	}
	if (!rc) {
	    break;
	}
	anchor->state = (unsigned char)flag;
	if (anchor->pad) {
	    if ((anchor->Type() == PAD_HTML_ANCHOR_TEXT) ||
		((anchor->Type() == PAD_HTML_ANCHOR_TEXT) && 
		 (((HTML_AnchorImage *)anchor)->Get_imageborder_width() > 0))) {
		anchor->Damage();
	    }
	}
	anchor = anchor->next;
    } while(anchor != first);

    return(rc);
}

void
HTML_Anchor::Set_state_default(void)
{
    Set_state(HTML_DEFAULT_STATE);
}

int
HTML_Anchor::Get_state(void)
{
    return(state);
}

HTML_AnchorImage::~HTML_AnchorImage(void) 
{
    Generate_delete();

    _image = NULL;		// Image not freed because it will be
				// deleted later because it is still on
				// the entry lists.
    _altImage = NULL;
}

HTML_AnchorImage::HTML_AnchorImage(Pad *pad, HTML *html, Pad_ImageData *image, Pad_ImageData *altImage, int w, int h) :
HTML_Anchor(pad, html)
{
    _type = PAD_HTML_ANCHOR_IMAGE;
    _image = image;
    _altImage = altImage;
    if (_image) {
	transform.Set_scale(IMAGE_SIZE * MSML_Scale);
	_width = _image->width;
	_height = _image->height;
    } else {
	transform.Set_scale(MSML_Scale);
	_width = w;
	_height = h;
    }
    
    transform.Set_offset(x_curr_pos, y_curr_pos);

    Update_and_compute_anchor();

    if (!Anchor_first) {
	Anchor_first = this;
	Anchor_current = this;
    }
    href = Anchor_href;
    name = Anchor_name;
    Anchor_current->next = this;
    Anchor_current = this;
    next = Anchor_first;
}

//
// Return true to continue rendering other objects, or false if rendering
// was interrupted.
//
Pad_Bool
HTML_AnchorImage::Render(void)
{
    float bb[4];
    Pad_Bool dither;
    Pad_ImageData *image;

    Pad_renderer->Set_color(penColor);
    
    if (_image) {
	if (_image->dpy->truecolor) {
	    dither = FALSE;
	} else {
	    if (pad->level > 0) {
		dither = TRUE;
	    } else {
		dither = FALSE;
	    }
	}
				// Draw regular image if not too small
				// or alt image if too small.
	if (html->pixelDim > HTML_MIN_SIZE) {
	    if (_image) {
		image = _image;
	    } else if (_altImage) {
		image = _altImage;
	    } else {
	        image = NULL;
	    }
	} else {
	    if (_altImage) {
		image = _altImage;
	    } else {
		image = _image;
	    }
	}

        Get_bbox(bb);
	Pad_renderer->Draw_image(image, bb[XMIN], bb[YMIN], bb[XMAX], bb[YMAX], dither);
	if (_borderWidth > 0) {
	    Pad_renderer->Set_line_width(_borderWidth);
	    Pad_renderer->Draw_box(0.0, 0.0, _width, _height);
	}
    } else {
	Pad_renderer->Draw_filled_box(0.0, 0.0, _width, _height);
    }
    
    return(TRUE);
}

//
// Returns true if object needs more refinement.
//
Pad_Bool 
HTML_AnchorImage::Continue_refinement(void)
{
    return(((Pad_View *)Pad_prc->views.Last())->pad->level < 1);
}

void
HTML_AnchorImage::Compute_bounding_box(void) 
{
    if (!(optionFlags & PAD_WIDTH_SET)) {
        Set_bbox_xmin(0.0);
        Set_bbox_xmax(_width);
    }
    if (!(optionFlags & PAD_HEIGHT_SET)) {
        Set_bbox_ymin(0.0);
        Set_bbox_ymax(_height);
    }

    Pad_Object::Compute_bounding_box();
}

void
HTML_AnchorImage::Set_imageborder_width(int b)
{
    _borderWidth = b;
}

int
HTML_AnchorImage::Get_imageborder_width(void)
{
    return(_borderWidth);
}

float
HTML_AnchorImage::Get_width(void)
{
    return(_width);
}

int
HTML_AnchorImage::Get_height(void)
{
    return(_height);
}

//
// Setters and getters for anchor image item.
//
void
HTML_AnchorImage::Set_imagedata(Pad_ImageData *newImage) 
{
    Damage();
    anchorFlags |= ANCHOR_IMAGEDATA_SET;
    if (_image) {
	_image->Remove_object(this);
	_image = NULL;
    }
    if (_image) {
	_altImage = newImage;
    } else {
	_image = newImage;
    }
    if (_image) {
	_image->Add_object(this);
    }
}

void
HTML_AnchorImage::Set_imagedata_default(void)
{
    Set_imagedata(ANCHOR_DEFAULT_IMAGEDATA);
    anchorFlags &= ~ANCHOR_IMAGEDATA_SET;
}

Pad_ImageData *
HTML_AnchorImage::Get_imagedata(void)
{
    return(_image);
}

HTML_AnchorText::~HTML_AnchorText(void) {
    Generate_delete();
}

HTML_AnchorText::HTML_AnchorText(Pad *pad, HTML *html, HTML_DisplayString *str) :
HTML_Anchor(pad, html)
{
    _type = PAD_HTML_ANCHOR_TEXT;
    _text = str;
    transform.Set_offset(str->Getx(), str->Gety());
    Set_font_default();
    transform.Set_scale((LineSize / DefaultLineSize) * MSML_Scale);
    Update_global_bounding_box();
    if (!Anchor_first) {
	Anchor_first = this;
	Anchor_current = this;
    }
    href = Anchor_href;
    name = Anchor_name;
    Anchor_current->next = this;
    Anchor_current = this;
    next = Anchor_first;
}

Pad_Bool
HTML_AnchorText::Contains_text(char *string) {
    Pad_Bool rc;


    if (casestrstr(_text.Get(), string)) {
	rc = TRUE;
    } else {
	rc = FALSE;
    }

    return rc;
}

Pad_Bool
HTML_AnchorText::Get_text(Pad_String &string) 
{
    string = "";
    string = _text;
    return (TRUE);

}

void
HTML_Anchor::Set_font(char *fullname)
{
    anchorFlags |= ANCHOR_FONT_SET;
    _font.Set(fullname);
    Compute_bounding_box();
}

void
HTML_Anchor::Set_font_default()
{
    _font.Set(HTML_DEFAULT_FONT);
    anchorFlags &= ~ANCHOR_FONT_SET;
}


void
HTML_Anchor::Get_font(Pad_String &name)
{
    _font.Get(name);
}

//
// Return true to continue rendering other objects, or false if rendering
// was interrupted.
//
Pad_Bool
HTML_AnchorText::Render(void)
{
    float mag = Pad_renderer->Get_mag() * transform.Get_scale();
    
    Pad_renderer->Set_color(penColor);
    Pad_renderer->Set_font(_font);

    if (mag > 5) {
	level = 2;
    } else if (mag > 2) {
	level = 1;
    } else {
	level = 0;
    }
				// Skip intermediate refinement levels
    if (pad->level > 0) {
	level = TEXT_LEVELS;
    }

    _refineNeeded = Pad_renderer->Draw_string(_text.Get(), level, 0.0, 0.0, 0.0);
    
    return(TRUE);
}

//
// Returns true if object needs more refinement.
//
Pad_Bool 
HTML_AnchorText::Continue_refinement(void)
{
    return(_refineNeeded);
}

void
HTML_AnchorText::Compute_bounding_box(void) 
{
    float max_width, line_height;
    Pad_BBox lineBBox;
    float line_spacing = 1.0;
    float extra_width = 0.0;
    
    Pad_renderer->Set_font(_font);
    extra_width = Pad_renderer->Get_char_width(' ');
    Pad_renderer->Get_string_bbox(_text.Get(), &lineBBox);
    
    max_width = lineBBox.Width(); 
    
    line_height = line_spacing * Pad_renderer->Get_font_height();
    
    if (!(optionFlags & PAD_WIDTH_SET)) {
        Set_bbox_xmin(-0.2);
        Set_bbox_xmax(max_width);
    }
    if (!(optionFlags & PAD_HEIGHT_SET)) {
        Set_bbox_ymin(-0.3);
        Set_bbox_ymax(line_height);
    }

    Pad_Object::Compute_bounding_box();
}

//
// Don't write out anchors
//
Pad_Bool
HTML_Anchor::Write_obj_pre(char *, Pad_String &)
{
    return(FALSE);
}

