/*
"(c) Copyright 1993-1997 Pad++ Consortium {University of New Mexico (UNM),
and New York University (NYU)}, All Rights Reserved."  
Licensee can not remove or obscure any of the
copyright or trademark notices in this software.

IN NO EVENT SHALL THE AUTHORS OR DISTRIBUTORS BE LIABLE TO ANY PARTY
FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
ARISING OUT OF THE USE OF THIS SOFTWARE, ITS DOCUMENTATION, OR ANY
DERIVATIVES THEREOF, EVEN IF THE AUTHORS HAVE BEEN ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.

THE AUTHORS AND DISTRIBUTORS SPECIFICALLY DISCLAIM ANY WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE, AND NON-INFRINGEMENT.  THIS SOFTWARE
IS PROVIDED ON AN "AS IS" BASIS, AND THE AUTHORS AND DISTRIBUTORS HAVE
NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR
MODIFICATIONS.

See the file "License" for general information on usage and
redistribution, and the file "LicenseTerms" for the specific license
agreement on usage and redistribution of this file, and the Pad++
software in general.
*/

#ifdef HAS_JPEG

#include "jpeg.h"
#include "defs.h"

#include <stdio.h>
#include <setjmp.h>
#include <string.h>

#undef EXTERN        // prevents warning from jpeglib.h
#undef HAVE_STDLIB_H // prevents warning from jpeglib.h

#ifdef HAVE_JINCLUDE
extern "C" {
#include <jinclude.h>
};
#else
#ifndef SIZEOF
#define SIZEOF(x) (sizeof(x))
#endif
#endif

extern "C" {
#include <jpeglib.h>
#include <jerror.h>
};

struct my_error_mgr {
  struct jpeg_error_mgr pub;	/* "public" fields */

  jmp_buf setjmp_buffer;	/* for return to caller */
};

typedef struct my_error_mgr * my_error_ptr;

/*
 * Here's the routine that will replace the standard error_exit method:
 */

static void
my_error_exit (j_common_ptr cinfo)
{
  /* cinfo->err really points to a my_error_mgr struct, so coerce pointer */
  my_error_ptr myerr = (my_error_ptr) cinfo->err;

  /* Always display the message. */
  /* We could postpone this until after returning, if we chose. */
  (*cinfo->err->output_message) (cinfo);

  /* Return control to the setjmp point */
  longjmp(myerr->setjmp_buffer, 1);
}


/* Code to support a memory object as a data source */

typedef struct {
  struct jpeg_source_mgr pub;	/* public fields */

  unsigned char *indata;		/* source stream */
  int len; /* length of buffer */

  JOCTET * buffer;		/* start of buffer */
  boolean start_of_file;	/* have we gotten any data yet? */
} my_source_mgr;

typedef my_source_mgr * my_src_ptr;

#define INPUT_BUF_SIZE  4096	/* choose an efficiently fread'able size */

/*
 * Initialize source --- called by jpeg_read_header
 * before any data is actually read.
 */

static void
init_source (j_decompress_ptr cinfo)
{
  my_src_ptr src = (my_src_ptr) cinfo->src;

  /* We reset the empty-input-file flag for each image,
   * but we don't clear the input buffer.
   * This is correct behavior for reading a series of images from one source.
   */
  src->start_of_file = TRUE;
}


/*
 * Fill the input buffer --- called whenever buffer is emptied.
 *
 */

static boolean
fill_input_buffer (j_decompress_ptr cinfo)
{
  my_src_ptr src = (my_src_ptr) cinfo->src;
  size_t nbytes;

  if (src->len < INPUT_BUF_SIZE) {
      nbytes = src->len;
  } else {  
      nbytes = INPUT_BUF_SIZE;
  }

  if (nbytes > 0)
    memcpy((char*)src->buffer, (char*)src->indata, nbytes);

  src->indata += nbytes;
  src->len -= nbytes;

  if (nbytes <= 0) {
    if (src->start_of_file)	/* Treat empty input file as fatal error */
      ERREXIT(cinfo, JERR_INPUT_EMPTY);
    WARNMS(cinfo, JWRN_JPEG_EOF);
    /* Insert a fake EOI marker */
    src->buffer[0] = (JOCTET) 0xFF;
    src->buffer[1] = (JOCTET) JPEG_EOI;
    nbytes = 2;
  }
  src->pub.next_input_byte = src->buffer;
  src->pub.bytes_in_buffer = nbytes;
  src->start_of_file = FALSE;

  return TRUE;
}


/*
 * Skip data --- used to skip over a potentially large amount of
 * uninteresting data (such as an APPn marker).
 *
 */

static void
skip_input_data (j_decompress_ptr cinfo, long num_bytes)
{
  my_src_ptr src = (my_src_ptr) cinfo->src;

  if (num_bytes > 0) {
    while (num_bytes > (long) src->pub.bytes_in_buffer) {
      num_bytes -= (long) src->pub.bytes_in_buffer;
      (void) fill_input_buffer(cinfo);
      /* note we assume that fill_input_buffer will never return FALSE,
       * so suspension need not be handled.
       */
    }
    src->pub.next_input_byte += (size_t) num_bytes;
    src->pub.bytes_in_buffer -= (size_t) num_bytes;
  }


}



static void
term_source (j_decompress_ptr )
{
  /* no work necessary here */
}

/*
 * Prepare for input from a stdio stream.
 * The caller must have already opened the stream, and is responsible
 * for closing it after finishing decompression.
 */

static void
memory_src (j_decompress_ptr cinfo, unsigned char * indata, int len)
{
  my_src_ptr src;

  if (cinfo->src == NULL) {	/* first time for this JPEG object? */
    cinfo->src = (struct jpeg_source_mgr *)
      (*cinfo->mem->alloc_small) ((j_common_ptr) cinfo, JPOOL_PERMANENT,
				  SIZEOF(my_source_mgr));
    src = (my_src_ptr) cinfo->src;
    src->buffer = (JOCTET *)
      (*cinfo->mem->alloc_small) ((j_common_ptr) cinfo, JPOOL_PERMANENT,
				  INPUT_BUF_SIZE * SIZEOF(JOCTET));
  }

  src = (my_src_ptr) cinfo->src;
  src->pub.init_source = init_source;
  src->pub.fill_input_buffer = fill_input_buffer;
  src->pub.skip_input_data = skip_input_data;
  src->pub.resync_to_restart = jpeg_resync_to_restart; /* use default method */
  src->pub.term_source = term_source;
  src->indata = indata;
  src->len = len;
  src->pub.bytes_in_buffer = 0; /* forces fill_input_buffer on first read */
  src->pub.next_input_byte = NULL; /* until buffer loaded */
}



/*
 * Sample routine for JPEG decompression.  We assume that the source file name
 * is passed in.  We want to return 1 on success, 0 on error.
 */

#define MODE_FILE 1
#define MODE_DATA 2

static
unsigned long *LoadJPEG(int mode, /* MODE_FILE or MODE_DATA */
			char *filename, 			
			unsigned char *indata,
			int len,
			int *ret_width, int *ret_height)
{
  struct jpeg_decompress_struct cinfo;
  struct my_error_mgr jerr;
  FILE * infile;		/* source file */
  JSAMPARRAY buffer;		/* Output row buffer */
  int row_stride;		/* physical row width in output buffer */
  unsigned int x, y;
  unsigned char r, g, b;
  unsigned long *data, *dataptr;

  if (mode == MODE_FILE) {
      if ((infile = fopen(filename, "rb")) == NULL) {
	  fprintf(stderr, "can't open %s\n", filename);
	  return NULL;
      }
  }

  /* set up the normal JPEG error routines, then override error_exit. */
  cinfo.err = jpeg_std_error(&jerr.pub);
  jerr.pub.error_exit = my_error_exit;

  if (setjmp(jerr.setjmp_buffer)) {
    /* If we get here, the JPEG code has signaled an error.
     * We need to clean up the JPEG object, close the input file, and return.
     */
    jpeg_destroy_decompress(&cinfo);
    if (mode == MODE_FILE) 
      fclose(infile);
    fprintf(stderr, "error reading JPEG file");
    return NULL;
  }

  /* initialize the JPEG decompression object. */
  jpeg_create_decompress(&cinfo);

  /* standard FILE source */
  if (mode == MODE_FILE) {
      jpeg_stdio_src(&cinfo, infile);
  } else { /* MODE_DATA */
      memory_src(&cinfo, indata, len);
  }

  (void) jpeg_read_header(&cinfo, TRUE);
  (void) jpeg_start_decompress(&cinfo);


  /* JSAMPLEs per row in output buffer */
  row_stride = cinfo.output_width * cinfo.output_components;

  /* Make a one-row-high sample array that will go away when done with image */
  buffer = (*cinfo.mem->alloc_sarray)
		((j_common_ptr) &cinfo, JPOOL_IMAGE, row_stride, 1);

  data = new unsigned long[cinfo.output_width * cinfo.output_height];

  /* Use the library's state variable cinfo.output_scanline as the
   * loop counter, so that we don't have to keep track ourselves.
   */
  dataptr = data;


  for (y = 0; y < cinfo.output_height; y++) {
      JSAMPLE *src;

      (void) jpeg_read_scanlines(&cinfo, buffer, 1);

      switch (cinfo.output_components) {
	case 1:
	  for (src = buffer[0], x = 0; x < cinfo.output_width; x++) {
	      r = g = b = *src++;
	      *dataptr++ = CC_RGBA(r, g, b, 0);
	  }
	  break;
	case 3:
	  for (src = buffer[0], x = 0; x < cinfo.output_width; x++) {
	      r = *src++;
	      g = *src++;
	      b = *src++;
	      *dataptr++ = CC_RGBA(r, g, b, 0);
	  }
	  break;
      }
  }

  *ret_width = cinfo.output_width;
  *ret_height = cinfo.output_height;

  (void) jpeg_finish_decompress(&cinfo);
  jpeg_destroy_decompress(&cinfo);

  if (mode == MODE_FILE)
    fclose(infile);

  return data;
}


unsigned long *Pad_LoadJPEG(char *filename, int *ret_width, int *ret_height)
{
  return LoadJPEG(MODE_FILE, filename, NULL, 0, ret_width, ret_height);
}

unsigned long *Pad_LoadJPEG(unsigned char *data, int len, 
			    int *ret_width, int *ret_height)
{
  return LoadJPEG(MODE_DATA, NULL, data, len, ret_width, ret_height);
}

#endif /* HAS_JPEG */
