/*
"(c) Copyright 1993-1997 Pad++ Consortium {University of New Mexico (UNM),
and New York University (NYU)}, All Rights Reserved."  
Licensee can not remove or obscure any of the
copyright or trademark notices in this software.

IN NO EVENT SHALL THE AUTHORS OR DISTRIBUTORS BE LIABLE TO ANY PARTY
FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
ARISING OUT OF THE USE OF THIS SOFTWARE, ITS DOCUMENTATION, OR ANY
DERIVATIVES THEREOF, EVEN IF THE AUTHORS HAVE BEEN ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.

THE AUTHORS AND DISTRIBUTORS SPECIFICALLY DISCLAIM ANY WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE, AND NON-INFRINGEMENT.  THIS SOFTWARE
IS PROVIDED ON AN "AS IS" BASIS, AND THE AUTHORS AND DISTRIBUTORS HAVE
NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR
MODIFICATIONS.

See the file "License" for general information on usage and
redistribution, and the file "LicenseTerms" for the specific license
agreement on usage and redistribution of this file, and the Pad++
software in general.
*/

#include "defs.h"
#include "misc.h"
#include "plist.h"
#include "point.h"
#include "list.h"
#include "win.h"
#include "global.h"
#include "restorer.h"
#include "tclInt.h"

#include <stdlib.h>
#include <math.h>

#if defined(sun)
extern "C" {
long random();
}
#endif

#ifndef HAVE_RANDOM
#  define random rand
#endif

#ifdef PAD_WIN
#   define random rand
#endif

static Pad_CreateTimerProc *createProc = NULL;
static Pad_DeleteTimerProc *deleteProc = NULL;

//
// Pad TimerHandlers either call Tcl handlers, or some
// other handlers that are defined elsewhere.
//
Pad_TimerToken Pad_CreateTimerHandler(int delay_in_milliseconds,
				      Pad_TimerCallback *timerCallback,
				      void *clientData)
{
    Pad_TimerToken token;

    if (createProc) 
      token = createProc(delay_in_milliseconds, timerCallback, clientData);
    else
      token = Tcl_CreateTimerHandler(delay_in_milliseconds, timerCallback, clientData);

    return(token);
}


void Pad_DeleteTimerHandler(Pad_TimerToken token)
{
    if (deleteProc) 
      deleteProc(token);
    else 
      Tcl_DeleteTimerHandler((Tcl_TimerToken)token);
}

void Pad_SetTimerProcs(Pad_CreateTimerProc *cProc, Pad_DeleteTimerProc *dProc)
{
    createProc = cProc;
    deleteProc = dProc;
}



//
// System independent way of getting the current
// time in seconds and microseconds.
//
void
Pad_Time::Update(void)
{

#if TCL_MAJOR_VERSION == 7 &&  TCL_MINOR_VERSION == 5 && !defined(TCL_PATCH_P1)
    TclGetTime(&_time);
#else
    TclpGetTime(&_time);
#endif

}

long
Pad_Time::Get_sec(void)
{
    return(_time.sec);
}

long
Pad_Time::Get_usec(void)
{
    return(_time.usec);
}

//
// Convert a string to a Pad_Bool
//
Pad_Bool
Pad_Atob(char *string)
{
    int ival;
    Pad_Bool rc;

    ival = atoi(string);
    if (ival == 0) {
	rc = FALSE;
    } else {
	rc = TRUE;
    }

    return(rc);
}

//
// Return a random integer between <min> and <max>, inclusively
//
int
Pad_Random(int min, int max)
{
    int r;

    r = min + (random() % (max - min + 1));

    return(r);
}

//
// Map the input linear lerp (a linear interpolation from (0-1) to
// a slow-in, slow-out lerp.
//
float 
Pad_Siso_lerp(float t)
{
    float siso, t1, t2, l;

    t1 = t * t;
    t2 = 1 - (1 - t) * (1 - t);
    l = LERP(t, t1, t2);
    siso = LERP(l, t1, t2);
    
    return(siso);
}

//
// Output the specified message to the current debug output place.
// This may be stdout, or a Tcl variable, depending on the value
// of win->debugOut.  A carriage return is automatically added
// to the end of each message.
//
void
Pad_Debug_output(Tcl_Interp *interp, Pad_Win *win, char *msg) {
    Tcl_Channel outChannel;

    if (win->debugOut) {
	Tcl_SetVar(interp, win->debugOut, msg,
		   TCL_GLOBAL_ONLY | TCL_APPEND_VALUE);
	Tcl_SetVar(interp, win->debugOut, "\n",
		   TCL_GLOBAL_ONLY | TCL_APPEND_VALUE);
    } else {
	outChannel = Tcl_GetStdChannel(TCL_STDOUT);
	Tcl_Write(outChannel, msg, -1);
	Tcl_Write(outChannel, "\n", -1);
	Tcl_Flush(outChannel);
    }
}

//
// Pad_Rotate receives a Pad_PList "coord_plist",
// an angle in degrees "theta", and a center of 
// rotation "center", and transforms coord_plist
// by rotating each of its points about "center"
// by angle "theta".  Positive rotate is counter-clockwise.
//  
// This is a utility function used by the "Rotate"
// functions for each object for which rotation is 
// defined.
//
void
Pad_Rotate(Pad_PList &coordPlist,
	   float theta,
	   Pad_Point &center) 
{

  Pad_Point *tmpPt;
  int pt, lenPlist;
  double thetaRad;
  float oldTmpx,oldTmpy;

  thetaRad = -DEG2RAD(theta);
  lenPlist = coordPlist.Length();

  for (pt=0;pt<lenPlist;pt++)
    {

      tmpPt = coordPlist.Nth(pt);
      oldTmpx = tmpPt->x;
      oldTmpy = tmpPt->y;

      tmpPt->x =
	center.x +
	((oldTmpx)-center.x)*cos(thetaRad) +
	((oldTmpy)-center.y)*sin(thetaRad);

      tmpPt->y =
	center.y +
	((oldTmpy)-center.y)*cos(thetaRad) -
	((oldTmpx)-center.x)*sin(thetaRad);
    }
}

//
// Compute the miter length between the three specified points
// with the specified line width
// 
float
Pad_Miter_length(float lineWidth, float x1, float y1, float x2, float y2, float x3, float y3)
{
    int cdeg;
    double c;
    double a2, b2, c2;
#define S 0.00001

				// Use cosine rule to convert lengths a,b,c to angle ABC
    a2 = (double)(POW2(x3 - x2) + POW2(y3 - y2));
    b2 = (double)(POW2(x2 - x1) + POW2(y2 - y1));
    c2 = (double)(POW2(x3 - x1) + POW2(y3 - y1));
    c = acos((a2 + b2 - c2) / (2.0 * sqrt(a2 + S) * sqrt(b2 + S)));
	
				// Coerce angle C to degrees to test for angles < 11
    cdeg = (int)((c * 360.0) / PITIMES2);
	
				// Miter length is line width * (1 / (sin(C / 2))
    c = sin(c / 2.0);
	
				// Cut off a bit early to be careful
    if ((cdeg % 360 <= 9) || (cdeg % 360 > 85) || (c == 0.0)) {
	return(0.0);
    } else {
	return(lineWidth * (0.5 / c));
    }
}

//
// Expand <name> into an absolute pathname, taking care
// of ~-expansion, and collapsing '.'s and '..'s.
// This fills in <fullName>, and returns TRUE if successfull,
// or FALSE otherwise.
//
Pad_Bool
Pad_Expand_pathname(Tcl_Interp *interp, char *name, Pad_String &fullName)
{
    int i, j;
    int index;
    Pad_String buf;
    Tcl_DString buffer;

				// Take care of ~-expansion
    fullName = Tcl_TranslateFileName(interp, name, &buffer);
	
#ifdef PAD_WIN
				// don't need to convert to backward slash, so set back.
	fullName.Set(name);
#endif

    if (fullName != "") {
	Tcl_DStringFree(&buffer);
    } else {
	Pad_errorString = interp->result;
	return(FALSE);
    }

				// Expand relative pathname
    if (fullName.Get_char(0) != '/' &&
		( (fullName.Get_char(1) != ':') || (fullName.Get_char(2) != '/')) ) {
	buf = TclpGetCwd(interp, &buffer);
	Tcl_DStringFree(&buffer);
	buf += "/";
	buf += fullName;
	fullName = buf;
    }

				// Take care of '.'s and '..'s
    index = 0;
    do {
				// Find delimiting '/'s
	i = fullName.Strchr('/', index);
	j = fullName.Strchr('/', i + 1);
	if (i == -1) {
	    break;
	}
	if (j == -1) {
	    j = i + 1;
	}
	if ((j == i+2) && (fullName.Get_char(i+1) == '.')) {
				// '.' found
	    buf = fullName.Get(j);
	    fullName.Insert(&buf, i);
	    index = i;
	} else if ((j == i+3) && 
		   (fullName.Get_char(i+1) == '.') &&
		   (fullName.Get_char(i+2) == '.')) {
				// '..' found
	    
	    buf = fullName.Get(j);
	    fullName.Insert('\0', i);
	    i = fullName.Strrchr('/');
	    if (i == -1) {
		Pad_errorString = "Invalid filename, there are too many '..'s";
		return(FALSE);
	    }
	    fullName.Insert('\0', i);
	    fullName.Append(buf);
	    index = i;
	} else {
	    index = j;
	}
    } while(1);

    return(TRUE);
}

//
// Return the relative pathname of <newName> relative to <dir>.
// Put result in <relativeName>.  Returns TRUE if succcessfull,
// or FALSE otherwise.
//
Pad_Bool
Pad_Get_relative_pathname(Tcl_Interp *interp, char *dir, char *newName, Pad_String &relativeName)
{
    int i, j, k;
    int common;
    Pad_Bool rc;
    Pad_String name;
    Pad_String currentDir;

				// Compare filename with directory, one
				// dir at a time, finding which shared
				// dirs there are.
    currentDir = dir;
    rc = Pad_Expand_pathname(interp, newName, name);  // Turn into a standard absolute pathname
    if (!rc) {
	return(FALSE);
    }
    common = 0;
    k = 0;
    do {
	i = currentDir.Strchr('/', k);
	j = name.Strchr('/', k);
	if ((i == j) && (i != -1) && (!currentDir.Strncmp(name, i))) {
	    common = i;
	    k = i + 1;
	} else {
	    break;
	}
    } while(1);

				// Now, build result, backing up as many
				// directories as are not shared
    k = common + 1;
    relativeName = "";
    do {
	i = currentDir.Strchr('/', k);
	if (i == -1) {
	    break;
	} else {
	    relativeName += "../";
	    k = i + 1;
	}
    } while (1);
    relativeName += name.Get(common + 1);

    return(TRUE);
}

//
// Data used by Pad_Background_error
//
static Pad_String errorStr;

//
// Callback for Pad_Background_error
//
static void
errorCB(ClientData clientData)
{
    int result;
    Tcl_Interp *interp = (Tcl_Interp *)clientData;
    Pad_String error_cmd;

    Tcl_SetVar(interp, "Pad_Error", "1", TCL_GLOBAL_ONLY);
    error_cmd = "bgerror {";
    error_cmd += errorStr;
    error_cmd += "}";
    result = Tcl_GlobalEval(interp, error_cmd.Get());
    if (result != TCL_OK) {
	if (strcmp(interp->result, "\"bgerror\" is an invalid command name or ambiguous abbreviation") == 0) {
	    fprintf(stderr, "%s\n", error_cmd.Get());
	} else {
	    fprintf(stderr, "bgerror failed to handle background error.\n");
	    fprintf(stderr, "    Original error: %s\n", errorStr.Get());
	    fprintf(stderr, "    Error in bgerror: %s\n", interp->result);
	}
    }
    Tcl_SetVar(interp, "Pad_Error", "0", TCL_GLOBAL_ONLY);
    Tcl_ResetResult(interp);
}

//
// Pad_Background_error
//
//   Similar to Tk_BackgroundError, but works within a render by
//   forcing the actual background error to occur after a timer
//   (of zero seconds) expires.  Thus, the error occurs at the top
//   level.
//
void Pad_Background_error(Tcl_Interp *interp)
{
    char *errorInfo;

    errorInfo = Tcl_GetVar(interp, "errorInfo", TCL_GLOBAL_ONLY);
    errorStr = "";
    if (errorInfo) {
	errorStr.Append_with_escapes(errorInfo);
    }
    Pad_CreateTimerHandler(IDLE_DELAY, errorCB, (ClientData)interp);
}

//
// A version of Tk_GetUid that takes a const char * as an argument.
//
#define NAMESIZE 512

char *
Pad_GetUid(const char *name)
{
    // must copy name into nameCopy, to preserve -const- status of name.
    char nameCopy[NAMESIZE];
    nameCopy[0] = 0;
    strncat(nameCopy, name, NAMESIZE - 1);

    return (char*)Tk_GetUid(nameCopy);
}

//
// Swap the two specified numbers
//
void
Pad_Swap(float &a, float &b)
{
    float temp;

    temp = a;
    a = b;
    b = temp;
}
