/*
"(c) Copyright 1993-1997 Pad++ Consortium {University of New Mexico (UNM),
and New York University (NYU)}, All Rights Reserved."  
Licensee can not remove or obscure any of the
copyright or trademark notices in this software.

IN NO EVENT SHALL THE AUTHORS OR DISTRIBUTORS BE LIABLE TO ANY PARTY
FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
ARISING OUT OF THE USE OF THIS SOFTWARE, ITS DOCUMENTATION, OR ANY
DERIVATIVES THEREOF, EVEN IF THE AUTHORS HAVE BEEN ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.

THE AUTHORS AND DISTRIBUTORS SPECIFICALLY DISCLAIM ANY WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE, AND NON-INFRINGEMENT.  THIS SOFTWARE
IS PROVIDED ON AN "AS IS" BASIS, AND THE AUTHORS AND DISTRIBUTORS HAVE
NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR
MODIFICATIONS.

See the file "License" for general information on usage and
redistribution, and the file "LicenseTerms" for the specific license
agreement on usage and redistribution of this file, and the Pad++
software in general.
*/

#include "defs.h"
#include "list.h"
#include "pad.h"
#include "point.h"
#include "image.h"
#include "view.h"
#include "line.h"
#include "text.h"
#include "pad-tcl.h"
#include "renderer.h"
#include "noisedata.h"
#include "callback.h"
#include "win.h"
#include "global.h"
#include "trait.h"
#include "group.h"

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <errno.h>

#ifdef PAD_UNIX
#  include <unistd.h>
#endif

//////////////////////////////////////////////
//              db definitions
//////////////////////////////////////////////  

//
// Decide if Object should be written out to file.
// Generates <Write> even to get extra code to be written into file.
// Whatever string <Write> returns will be written into the file directly
// after the line that creates the object.
//
// However, if <Write> returns the "continue" return code, then the
// system-generated code will not be written out.  The user-generated
// code gets written out regardless.
//
// Returns true if object should be written to file, or false if not.
//
Pad_Bool
Pad_Object::Write_obj_callback(Pad_String &cb_str)
{
    int rc;
    char *writeInfo;
    Pad_Bool writeRc;
    Pad_String result;

    Tcl_SetVar(pad->view->win->interp, "Pad_Write", "1", TCL_GLOBAL_ONLY);
    rc = Generate_event(Pad_WriteNotify, result);
    writeInfo = Tcl_GetVar(pad->view->win->interp, "Pad_Write", TCL_GLOBAL_ONLY);

    if (!strcmp(writeInfo, "0")) {
	writeRc = FALSE;
    } else {
	writeRc = TRUE;
    }
    if (rc != TCL_ERROR) {
	cb_str += result;
    }

    return(writeRc);
}

//
// Write the Tcl command necessary to recreate the object into a string.
// This is a base class which writes the code common to all classes.
// Derived classes must write the object specific itemconfigure options, and
// then call this function.
//

void
Pad_Object::Write_obj(char *padname, Pad_String &str)
{
    int t;
    Pad_Iterator oi;
    Pad_Type *type;
    Pad_Option *option;
    Tcl_Interp *interp;
    Tk_Uid typeUid;
    Pad_Bool rc = TRUE;
    Pad_String *userType;

    interp = pad->view->win->interp;
    t = Type();

				// Don't write out pad
    if (t != PAD_PAD) {
        userType = Get_usertype();
	if (!userType) {
	    rc = Write_obj_pre(padname, str);
	}
	if (!rc) {
	    return;
	}
	    
	if (userType) {
	    typeUid = Tk_GetUid(userType->Get());
	} else {
	    typeUid = Tk_GetUid(Type_name());
	}

	str << "set Pad_ID [";
	str += padname;
	str += " create ";
	str += typeUid;
	str += " ";
	
	if (!userType) {
	    Write_obj_args(padname, str);
	}
	// First, extract generic options
	type = (Pad_Type *)pad->typeTable->Get(Pad_allUid);
	DOLIST(oi, type->options, Pad_Option, option) {
	    Pad_AppendOption(interp, this, option, str, TRUE, FALSE);
	}
	// Then, extract options for this object's type
	type = (Pad_Type *)pad->typeTable->Get(typeUid);
	DOLIST(oi, type->options, Pad_Option, option) {
	    Pad_AppendOption(interp, this, option, str, TRUE, FALSE);
	}
	str += "]\n";
	str += "set Pad_NewID(";
	str += id;
	str += ") $Pad_ID\n";

	if (!userType) {
	    Write_obj_post(padname, str);
	}

	Set_written(TRUE, FALSE);

	if (!group) {
	    str += "lappend Pad_ObjectList $Pad_NewID(";
	    str += id;
	    str += ")\n";
	}
    }
}

//
// This gets called just after ".pad create type", but before options
// are specified.  Items can override this if there are required args.
//
void
Pad_Object::Write_obj_args(char *, Pad_String &)
{
}

//
// This gets called just before the object is written so the item type
// can override this definition, and write some item-specific information.
// If this returns FALSE, the object won't get written out at all.
//
Pad_Bool
Pad_Object::Write_obj_pre(char *, Pad_String &)
{
    return(TRUE);
}

//
// This gets called just after the object is written so the item type
// can override this definition, and write some item-specific information.
//
void
Pad_Object::Write_obj_post(char *, Pad_String &)
{
}


//
// Writes the binary data necessary to recreate the object.
// This base class method writes the data common to all classes.
// Derived classes must call this method first and then write type 
// specific data.
//
Pad_Bool 
Pad_Object::Write_obj(char *padname, Pad_StreamData &sd) 
{
    Pad_Bool rc=TRUE;
    int t = Type();
    Pad_String str;

    if (t == PAD_PAD) {
        return TRUE;
    }

    // for tcl, kpl and user defined objs just write their script
    if ( t == PAD_KPL || t == PAD_TCL || t == PAD_USER_TYPE ) {
	Write_obj(padname, str);
	sd << PAD_OBJ_SCRIPT;
	sd << str.Get();

    } else {
        // type
        sd << t;

	// flags and options
	sd << flags << optionFlags;
	
	// transformation & anchor point
	float farr[5];
	farr[0] = transform.Get_offset_x(); 
	farr[1] = transform.Get_offset_y();
	farr[2] = transform.Get_scale();
	farr[3] = anchorpt.x; 
	farr[4] = anchorpt.y;
	sd.Marshal(farr, 5);

	// bbox
	farr[0] = bbox.Xmin();
	farr[1] = bbox.Ymin();
	farr[2] = bbox.Xmax();
	farr[3] = bbox.Ymax();
	sd.Marshal(farr, 4);
	    
	// options
	rc = _Options_stream(sd, TRUE);
	
	// obj id so the tcl Pad_NewID var can be set upon reading 
	sd << id;
    }

    Set_written(TRUE, FALSE);

    return rc;
}

//
// Reads the binary data for the object from the stream.  This base class
// method reads data common to all classes.  Derived classes must call
// this method first and then read object specific data.
// Note: The tcl variables Pad_ID and Pad_NewID are set here.
//
Pad_Bool 
Pad_Object::Read_obj(Pad_StreamData &sd) 
{
    Pad_Bool rc=TRUE;
    unsigned int n=0;

    if (_type != PAD_USER_TYPE) {

      // flags and options
      rc &= !sd.Reconstruct(flags);
      rc &= !sd.Reconstruct(optionFlags);
      
      // transformation & anchor point
      float *farr=NULL;
      if (!sd.Reconstruct(farr, &n) && (n == 5)) {
	transform.Set_offset(farr[0], farr[1]);
	transform.Set_scale(farr[2]);
	anchorpt.x = farr[3];
	anchorpt.y = farr[4];
      } else {
	 rc = FALSE;
      }
      delete[] farr;
      
      // bbox
      farr = NULL;
      if (!sd.Reconstruct(farr, &n) && (n == 4)) {
	bbox.Set(farr[0], farr[1], farr[2], farr[3]);
      } else {
	rc = FALSE;
      }
      delete[] farr;

      Update_global_bounding_box();

      if (!rc) {
	Pad_errorString << "ERR: problems reading core of object: " << id
			 << " of type: " << Type_name() << "\n";
      }

      // options
      rc &= _Options_stream(sd, FALSE);

      // set the Pad_ID tcl var
      Pad_String var, val;
      val = id;
      Tcl_SetVar(pad->view->win->interp, "Pad_ID", val.Get(),TCL_GLOBAL_ONLY);

      // need append Pad_ID to Pad_ObjectList

      // set the tcl Pad_NewID var
      int oldid;
      if (!sd.Reconstruct(oldid)) {
	var += oldid;
	val = id;
	Tcl_SetVar2(pad->view->win->interp, "Pad_NewID", 
		    var.Get(), val.Get(), TCL_GLOBAL_ONLY);
      } else {
	Pad_errorString << "ERR: Could not read oldid of object: " << id
			 << " of type: " << Type_name() << "\n";
	rc = FALSE;
      }

    }

    return rc;
}
    
//
// Read or Write the object options (traits, attributes) from/to the stream.
// This method operates on options common to all classes.
// Note: sticky option is not dealt with here due to the
// special case of the "label" sticky option.  It is assumed the io
// driver (e.g. Pad::Write and Read) has a way of dealing with this.
//
Pad_Bool 
Pad_Object::_Options_stream(Pad_StreamData &sd, Pad_Bool write) 
{
    Pad_Bool rc=TRUE;
    
    if (write) {
      Pad_Type *type = (Pad_Type *)pad->typeTable->Get(Pad_allUid);
      Pad_Iterator oi;
      Pad_Option *option;
      DOLIST(oi, type->options, Pad_Option, option) {      
	rc &= _Option_stream(sd, option->id, write, option);
      }
      sd << PAD_ENDOBJ_OPTIONS;

    } else {
      int optionid=0;
      while (!sd.Reconstruct(optionid) &&
	     (optionid != PAD_ENDOBJ_OPTIONS)) {
	rc &= _Option_stream(sd, optionid, write);
      }

      if (optionid != PAD_ENDOBJ_OPTIONS) {
	Pad_errorString << "ERR: Could not get to end of options of object: "
			 << id << " of type: " << Type_name() << "\n";
	rc = FALSE;
      }
    }
      
    return rc;
}

//
// Read or write an option from/to the stream.
// When writing:
//    int option_id;    // option's unique integer id
//    int total_bytes;  // total number of bytes to follow
//    <data items>;     // zero or more data items for the option
//
// When writing, in order to get the total number of bytes we write
// the option's items in a automatic stream, then get its total length
// and move it to the main stream.
//
 
Pad_Bool 
Pad_Object::_Option_stream(Pad_StreamData &sd, int optionid, Pad_Bool write,
			   Pad_Option *option) 
{
    Pad_Bool rc;
    Pad_Iterator it;
    char *str;
    Pad_Timer *timer = Get_timerscript();
    Pad_StreamData opsd(sd.Get_rep(), 0, sd.Get_istream(), sd.Get_ostream());
    int optLength=0;
    unsigned int n=0;
    unsigned int i;

    if (write) {
      rc = TRUE;
    } else {
      rc = FALSE;
      // get option's total length
      sd.Reconstruct(optLength);
    }

    switch (optionid) {
    case PAD_ALIASES:
      // Pad_List *aliases
      rc = TRUE;
      break;

    case PAD_ANCHOR:
      // Tk_Anchor anchor
      if (write) {
	if (optionFlags & PAD_ANCHOR_SET) {
	  sd << PAD_ANCHOR;
	  opsd << (int)anchor;  // is this ok?  yes...
	}
      } else {
	int value;
	if (!sd.Reconstruct(value)) {
	  anchor = (Tk_Anchor)value;
	  rc = TRUE;
	}
      }
      break;

    case PAD_ANGLE:
      if (write) {
	if (optionFlags & PAD_ANGLE_SET) {
	  sd << PAD_ANGLE;
	  opsd << Get_angle();
	}
      } else {
	float value;
	if (!sd.Reconstruct(value)) {
	  // No point in rotating now,
	  // there is no data yet.  It is up
	  // to the object to rotate if necessary later.
	  _Set_angle_data(value);
	  rc = TRUE;
	}
      }
      break;
	
    case PAD_CLIPPING:
      // clipping value is in the flags 
      break;

    case PAD_FADERANGE:
      if (write) {
	if (optionFlags & PAD_FADERANGE_SET) {
	  sd << PAD_FADERANGE;
	  opsd << fadeRange;
	}
      } else {
	if (!sd.Reconstruct(fadeRange)) {
	  rc = TRUE;
	}
      }
      break;
      
    case PAD_INFO:
      if (write) {
	if (optionFlags & PAD_INFO_SET) {
	  sd << PAD_INFO;
	  char *info = Get_info();
	  opsd << info;
	}
      } else {
	if (!sd.Reconstruct(str)) {
	  Set_info(str);
	  delete[] str;
	  rc = TRUE;
	}
      }
      break;

    case PAD_LAYER:
      if (write) {
	if (optionFlags & PAD_LAYER_SET) {
	  sd << PAD_LAYER;
	  opsd << Get_layer();
	}
      } else {
	if (!sd.Reconstruct(str)) {
	  Set_layer(str);
	  delete[] str;
	  rc = TRUE;
	}
      }
      break;
      
    case PAD_MINSIZE:
      if (write) {
	if (optionFlags & PAD_MINSIZE_SET) {
	  sd << PAD_MINSIZE;
	  opsd << minSize;
	}
      } else {
	if (!sd.Reconstruct(minSize)) {
	  rc = TRUE;
	}
      }
      break;

    case PAD_MAXSIZE:
      if (write) {
	if (optionFlags & PAD_MAXSIZE_SET) {
	  sd << PAD_MAXSIZE;
	  opsd << maxSize;
	}
      } else {
	if (!sd.Reconstruct(maxSize)) {
	  rc = TRUE;
	}
      }
      break;

    case PAD_NOISEDATA:
      Pad_NoiseData *nd;
      if (write) {
	if (optionFlags & PAD_NOISEDATA_SET) {
	  sd << PAD_NOISEDATA;
	  nd = Get_noisedata();
	  float farr[4];
	  farr[0]=nd->pos; 
	  farr[1]=nd->freq; 
	  farr[2]=nd->amp; 
	  farr[3]=nd->steps;
	  opsd.Marshal(farr, 4);
	}
      } else {
	float *farr=NULL;
	if (!sd.Reconstruct(farr) && (n == 4)) {
	  Set_noisedata(farr[0], farr[1], farr[2], farr[3]);
	  delete[] farr;
	  rc = TRUE;
	}
      }
      break;
      
    case PAD_RENDERSCRIPT:
      Pad_Callback *renderScript;
      if (write) {
	if (optionFlags & PAD_RENDERSCRIPT_SET) {
	  sd << PAD_RENDERSCRIPT;
	  renderScript = Get_renderscript();
	  opsd << renderScript->Language()->name.Get();  // needed?
	  opsd << renderScript->Get(); 
	}
      } else {
	if (!sd.Reconstruct(str)) {  // skip lang name for now
	  delete[] str;
	  str = NULL;
	  sd.Reconstruct(str);
	  Set_renderscript(new Pad_Callback(str, Pad_callbackLanguage));
	  delete[] str;
	  rc = TRUE;
	}
      }
      break;
      
    case PAD_TAGS:
      // Pad_List tags
      if (write) {
	if (optionFlags & PAD_TAGS_SET) {
	  sd << PAD_TAGS;
	  opsd << tags.Length();
	  DOLIST(it, tags, char, str) {
	    opsd << str;
	  }
	}
      } else {
	if (!sd.Reconstruct(n)) {
	  for (i = 0; i < n; i++) {
	    sd.Reconstruct(str);
	    Add_tag(str);
	    delete[] str;
	  }
	  rc = TRUE;
	}
      }
      break;

    case PAD_TIMERSCRIPT:
      if (write) {
	if (optionFlags & PAD_TIMERSCRIPT_SET) {
	  sd << PAD_TIMERSCRIPT;
	  opsd << timer->timerScript->Language()->name.Get();  // needed?
	  opsd << timer->timerScript->Get();
	}
      } else {
	if (!sd.Reconstruct(str)) {
	  delete[] str;
	  sd.Reconstruct(str);
	  Set_timerscript(str);
	  delete[] str;
	  rc = TRUE;
	}
      }
      break;

    case PAD_TIMERRATE:
      if (write) {
	if (optionFlags & PAD_TIMERRATE_SET) {
	  sd << PAD_TIMERRATE;
	  i = Get_timerrate();
	  opsd << i;
	}
      } else {
	if (!sd.Reconstruct(i)) {
	  Set_timerrate(i);
	  rc = TRUE;
	}
      }
      break;
      
    case PAD_TRANSPARENCY:
      if (write) {
	if (optionFlags & PAD_TRANSPARENCY_SET) {
	  sd << PAD_TRANSPARENCY;
	  opsd << Get_transparency();
	}
      } else {
	float transparency;
	if (!sd.Reconstruct(transparency)) {
	  Set_transparency(transparency);
	  rc = TRUE;
	}
      }
      break;
      
    case PAD_VIEWSCRIPT:
      Pad_Callback *viewScript;
      if (write) {
	if (optionFlags & PAD_VIEWSCRIPT_SET) {
	  sd << PAD_VIEWSCRIPT;
	  viewScript = Get_viewscript();
	  opsd << viewScript->Language()->name.Get();
	  opsd << viewScript->Get();
	}
      } else {
	if (!sd.Reconstruct(str)) {
	  delete[] str;
	  sd.Reconstruct(str);
	  Set_viewscript(str);
	  delete[] str;
	  rc = TRUE;
	}
      }
      break;
      
    case PAD_ZOOMACTION:
      Pad_ZoomAction *zact;
      if (write) {
	if (optionFlags & PAD_ZOOMACTION_SET) {
	  sd << PAD_ZOOMACTION;
	  Pad_List *zoomActions = Get_zoomaction();
	  opsd << zoomActions->Length();
	  DOLIST(it, *zoomActions, Pad_ZoomAction, zact) {
	    opsd << zact->userActionSize << zact->actionSize;
	    opsd << zact->growScript->Language()->name.Get();
	    opsd << zact->growScript->Get();
	    opsd << zact->shrinkScript->Language()->name.Get();
	    opsd << zact->shrinkScript->Get();
	  }
	}
      } else {
	if (!sd.Reconstruct(n)) {
	  for (i = 0; i < n; i++) {
	    zact = new Pad_ZoomAction;
	    sd >> zact->userActionSize >> zact->actionSize;
	    sd >> str;
	    Pad_Language *lang = NULL;
	    Pad_GetLanguage(str, &lang);
	    delete[] str;
	    sd >> str;
	    zact->growScript = new Pad_Callback(str, lang);
	    delete[] str;
	    
	    sd >> str;
	    lang = NULL;
	    Pad_GetLanguage(str, &lang);
	    delete[] str;
	    sd >> str;
	    zact->shrinkScript = new Pad_Callback(str, lang);
	    delete[] str;

	    Modify_zoomaction(zact);
	  }
	  rc = TRUE;
	}
      }
      break;

    case PAD_USER_OPTION:
      rc = _UserOption_stream(sd, opsd, optionid, write, option, 
			      Pad_allUid, optLength);
      break;

    default:
      Pad_errorString << "ERR: Unknown option id: " << optionid 
		       << " of object: " << id << " of type: " << Type_name();
      if (!write) {
	Pad_errorString << " -- SKIPPING " << optLength << " BYTES";
	sd.SkipInput(optLength);
      }
      Pad_errorString << '\n';
      rc = FALSE;
    }
    
    if (write && rc) {
      if (opsd.Items()) {
	sd << opsd.ItemsLength();
	sd.MoveItems(opsd);
      }
    }
    
    return rc;
}

//
// Convinient method for binary read/write of user-defined options.
// This is mostly used by type specific _Option_stream() methods.
// For writing, the option name and its values are written as strings.
//
Pad_Bool
Pad_Object::_UserOption_stream(Pad_StreamData &sd, Pad_StreamData &opsd,
				int optionid, Pad_Bool write, 
				Pad_Option *option, Tk_Uid typeUid, int optLength)
{
  Pad_Bool rc = !write;
  char *str;
 
  if (optionid == PAD_USER_OPTION) {
    if (write) {
      // if the option is writable then write its name and value as strings
      Pad_String pstr;
      Pad_Bool mod;
      if (option->write && (rc = Get_option(option, pstr, mod))) {
	if (pstr != option->defaultValue) {
	  sd << PAD_USER_OPTION;
	  opsd << (char*) option->option;
	  opsd << pstr.Get();
	}
      }

    } else {
      // read the option name and value strings, get its option structure
      // set its value
      char *opname=NULL;
      Pad_OptionKey optionKey;
      Pad_Option *opt=NULL;
      if (!sd.Reconstruct(opname)) {  // get option name
	optionKey.optionUid = Tk_GetUid(opname);
	optionKey.typeUid = typeUid;
	opt = (Pad_Option*)pad->optionTable->Get(&optionKey);
	if (opt) {
	  str = NULL;
	  if (!sd.Reconstruct(str)) {  // get option value
	    rc = Set_option(opt, str);
	  }
	  delete[] str;
	} else {
	  Pad_errorString << "ERR: Unknown option: " << opname
			  << " of object: " << id << " of type: " << Type_name();
	  Pad_errorString << " -- SKIPPING " << optLength << " BYTES\n";
	  sd.SkipInput(optLength);
	}
      }
      delete[] opname;
    }  
  } else {
    rc = FALSE;
  }

  return rc;
}

//
// Writes a non-group object along with its application write-callback
// script string to a string or stream.
// This is a convinience routine that container objects (such as a pad)
// can use to write their objects.
// For the script format the object script is added to result string.
// For binary formats the object is marshaled in the stream
// and the stream is then flushed to the file.
//
Pad_Bool
Pad_Object::Write_obj_wrapper(int format, char *padname,
			      Pad_String &result, Pad_StreamData &sd,
			      Pad_Bool ignoreGroup, Pad_Bool ignoreWritten,
			      Pad_Bool flushStream)
{
    Pad_Bool rc=TRUE;
    Pad_String cb_str = "";
    Pad_String str = "";
    unsigned int n = sd.Items();

    if ((ignoreWritten || !Get_written()) &&
	(ignoreGroup || (group == NULL)) &&
	Write_obj_callback(cb_str)) {

        if (format == PAD_SCRIPT_FORMAT) {
	    Write_obj(padname, str);
	    result += str;
	} else {
	    rc = Write_obj(padname, sd);
	}
    }
    
    result += cb_str;
    if (format != PAD_SCRIPT_FORMAT) {
        char *str = result.Get();
	if (str && ((str[0] != '\0') && (str[0] != '\n' || strlen(str) > 2))) {
	    if (n == sd.Items()) { // nothing was written for object
	        sd << PAD_OBJ_SCRIPT;
	    }
	    sd << str;
	}

	if (flushStream) {
	    if (n != sd.Items()) {
	        sd.Flush();
	    } else {
	        sd.Clear();
	    }
	}
	result.Set((char*)0);
    }

    return rc;
}

Pad_Bool
Pad_Object::Get_written(void)
{
   return ((flags & PAD_MASK_WRITTEN) ? TRUE : FALSE);
}

Pad_Bool
Pad_Object::Set_written(Pad_Bool val, Pad_Bool /*groupMembers*/)
{
   if (val) {
       flags |= PAD_MASK_WRITTEN;
   } else {
       flags &= ~PAD_MASK_WRITTEN;
   }

   return TRUE;
}

//
// Methods for dealing with traits

//
// Returns a trait.  If the trait is not set
// then NULL is returned.
//
Pad_Trait*
Pad_Object::_Get_trait(int id)
{
    Pad_Trait *trait=NULL;
    Pad_Iterator iter;
    if (_traitFlags & id) {
        DOLIST(iter, _traitList, Pad_Trait, trait) {
	   if (trait->Id() & id) {
	       break;
	   }
	}
    } 
    return trait;
}

//
// Removes a trait from the traitList and
// deallocates it.  Return TRUE if all goes well.
//
Pad_Bool
Pad_Object::_Remove_trait(int id)
{
    int rc=TRUE;
    Pad_Trait *trait = _Get_trait(id);
    if (trait) {
        _traitList.Remove(trait);
	delete trait;
	_traitFlags &= ~id;
    }
    return rc;
}


//
// Convinient method for getting the value of a string trait.
// If the trait does not exist then the value is set to NULL.
// Returns the trait if it's set and NULL otherwise.
//
Pad_Trait*
Pad_Object::_Get_trait(int id, char *&val)
{
    Pad_Trait *trait = _Get_trait(id);
    val = NULL;
    if (trait) {
        trait->Get_value(val);
    }

    return trait;
}

//
// Convinient method for setting the value of a string trait.
// It creates a trait for the option if it does not already exist.
// Returns TRUE if all goes well (if the trait could be set to
// the given value).
//
Pad_Bool
Pad_Object::_Set_trait(int id, char *val)
{
    Pad_Trait *trait = _Get_trait(id);
    Pad_Bool rc=TRUE;

    if (trait) {
        rc = trait->Set_value(val);
    } else {
        trait = new Pad_StringTrait(id, val);
	if (trait) {
	    _traitList.Push(trait);
            _traitFlags |= id;
	} else {
	    rc = FALSE;
	}
    }
    
    return rc;
}

Pad_Trait*
Pad_Object::_Get_trait(int id, Pad_String *&val)
{
    Pad_Trait *trait = _Get_trait(id);
    val = NULL;
    if (trait) {
        trait->Get_value(val);
    }

    return trait;
}

Pad_Bool
Pad_Object::_Set_trait(int id, Pad_String *val)
{
    Pad_Trait *trait = _Get_trait(id);
    Pad_Bool rc=TRUE;

    if (trait) {
        rc = trait->Set_value(val);
    } else {
        trait = new Pad_StringTrait(id, val);
	if (trait) {
	    _traitList.Push(trait);
            _traitFlags |= id;
	} else {
	    rc = FALSE;
	}
    }
    
    return rc;
}

Pad_Trait*
Pad_Object::_Get_trait(int id, Pad_Callback *&val)
{
    Pad_Trait *trait = _Get_trait(id);
    val = NULL;
    if (trait) {
        trait->Get_value(val);
    }

    return trait;
}

Pad_Bool
Pad_Object::_Set_trait(int id, Pad_Callback *val)
{
    Pad_Trait *trait = _Get_trait(id);
    Pad_Bool rc=TRUE;

    if (trait) {
        rc = trait->Set_value(val);
    } else {
        trait = new Pad_CallbackTrait(id, *val);
	if (trait) {
	    _traitList.Push(trait);
            _traitFlags |= id;
	} else {
	    rc = FALSE;
	}
    }
    
    return rc;
}

Pad_Trait*
Pad_Object::_Get_trait(int id, Pad_List *&val)
{
    Pad_Trait *trait = _Get_trait(id);
    val = NULL;
    if (trait) {
        trait->Get_value(val);
    }

    return trait;
}

Pad_Bool
Pad_Object::_Set_trait(int id, Pad_List *val)
{
    Pad_Trait *trait = _Get_trait(id);
    Pad_Bool rc=TRUE;

    if (trait) {
        rc = trait->Set_value(val);
    } else {
        trait = new Pad_ZoomActionTrait(id, *val);
	if (trait) {
	    _traitList.Push(trait);
            _traitFlags |= id;
	} else {
	    rc = FALSE;
	}
    }
    
    return rc;
}
