#ifdef HAS_SCHEME

/*
"(c) Copyright 1993-1995 Pad++ Consortium {University of New Mexico (UNM),
and New York University (NYU)}, All Rights Reserved."  
Licensee can not remove or obscure any of the
copyright or trademark notices in this software.

IN NO EVENT SHALL THE AUTHORS OR DISTRIBUTORS BE LIABLE TO ANY PARTY
FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
ARISING OUT OF THE USE OF THIS SOFTWARE, ITS DOCUMENTATION, OR ANY
DERIVATIVES THEREOF, EVEN IF THE AUTHORS HAVE BEEN ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.

THE AUTHORS AND DISTRIBUTORS SPECIFICALLY DISCLAIM ANY WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE, AND NON-INFRINGEMENT.  THIS SOFTWARE
IS PROVIDED ON AN "AS IS" BASIS, AND THE AUTHORS AND DISTRIBUTORS HAVE
NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR
MODIFICATIONS.

See the file "License" for general information on usage and
redistribution, and the file "LicenseTerms" for the specific license
agreement on usage and redistribution of this file, and the Pad++
software in general.
*/

#include <scheme.h>       // Must be first
#undef Print                    // scheme.h defines this and pad.h
                                // defines a funtion Print() that conflict
#include "pad-scheme.h"
#include "pad-tcl.h"
#include "global.h"
#include "defs.h"
#include "hashtab.h"
#include "tkwin.h"
#include "callback.h"
#include "api.h"

#include <stdlib.h>
#include <string.h>

#ifdef PAD_UNIX
#  include <unistd.h>
#endif

extern "C" char *Elk_Eval(char *expr);
extern "C" void elk_init_eval(void);

static int    SchemeEval(ClientData clientData, Tcl_Interp *interp, int argc, char **);
static Object Scheme_ExecCommand(int argc, Object *argv_obj);
static Object Scheme_TopLevelCommand(int argc, Object *argv_obj);
static int    SchemeLoad(ClientData clientData, Tcl_Interp *interp, int argc, char **);
static void   Scheme_TclString(char *str);


struct SchemeData {
    Tcl_Interp *interp;
    Pad_TkWin *tkwin;
};

static Pad_HashTable *Scheme_Table;

//
// Called upon system startup to initialize the scheme interpreter.
//
void
Scheme_InitProc(Pad_Language *, int, char **argv)
{
    char *eav[1];
    int eac = 1;
    
    char *initfile = NULL;
    int initflag = 0; 
    
    Set_App_Name(argv[0]);
    eav[0] = argv[0];
    
    Elk_Init(eac, eav, initflag, initfile);
    
    elk_init_eval();
    
    Scheme_Table = new Pad_HashTable(PAD_STRING_TABLE);

				// Define settoplevel command
    Define_Primitive((Object (*)(...))Scheme_TopLevelCommand, "settoplevel", 0, MANY, VARARGS);
}

//
// Called every time a new pad widget is created.  It creates a new
// scheme command associated with the pad widget, and registers the
// pad widget name in a hash table for future access.
//
void
Scheme_CreateProc(Pad_Language *language, char *pathname, Pad_TkWin *tkwin)
{
    SchemeData *data = new SchemeData();

    data->tkwin = tkwin;
    data->interp = language->interp;
    Scheme_Table->Set(pathname, data);
    Define_Primitive((Object (*)(...))Scheme_ExecCommand, "pad", 0, MANY, VARARGS);
}

//
// Gets called to determine if <cmd> is complete and is ready to be evaluated.
// For scheme, this means a balanced number parens.
//
int
Scheme_CompleteProc(char *cmd)
{
    int parens = 0;
    char *p;
    Pad_Bool rc;
    
    p = cmd;
    while (*p) {
	if (*p == '(') {
	    parens++;
	} else if (*p == ')') {
	    parens--;
	}
	p++;
    }

    if (parens < 0) {
	fputs("ignoring extra parens\n", stderr);
    }

    rc = (parens <= 0) ? TRUE : FALSE;

    return(rc);
}

//
// Gets called to evaluate a scheme command
//
int
Scheme_EvalProc(Pad_Language *language, char *cmd)
{
    char *result;

    if (!strcmp(cmd, "") || !strcmp(cmd, "\n")) {
	return(TCL_OK);
    }

    result = Elk_Eval(cmd);
    if (result) {
	Tcl_ResetResult(language->interp);
	Tcl_AppendResult(language->interp, result, NULL);
    }

    return(TCL_OK);
}

//
// Implement pad "scm" command
//
int
Scheme_CommandProc(ClientData clientData, Tcl_Interp *interp, int argc, char **argv)
{
    int result;
    
    if (argc != 3) {
	TCL_RESULT(interp, "wrong # args: should be \"pathName scm command arg\"");
	return(TCL_ERROR);
    }

    if (!strncmp(argv[1], "eval", 1)){
	result = SchemeEval(clientData, interp, argc, argv);
    } else if (!strncmp(argv[1], "load", 1)) {
	result = SchemeLoad(clientData, interp, argc, argv);
    } else  {
	TCL_RESULT(interp, "invalid command: should be \"eval or load\"");
	return(TCL_ERROR);
    }

    return result;
}

//
// Generate a Scheme prompt
//
void
Scheme_PromptProc(Pad_Language *, int partial, Pad_String &prompt)
{
    if (!partial) {
	prompt = "> ";
    }
}

//
// Implement pad "scm load" command
//
static int
SchemeLoad(ClientData, Tcl_Interp *interp, int argc, char **argv)
{
    if (argc != 3) {
	TCL_RESULT(interp, "wrong # args: should be \"pathName scm load filename\"");
	return(TCL_ERROR);
    }

    Load_File(argv[2]);
  
    return(TCL_OK);
}

//
// Implement pad "scm eval" command
//
static int
SchemeEval(ClientData, Tcl_Interp *interp, int argc, char **argv)
{
    char *result;
    
    if (argc != 3) {
	TCL_RESULT(interp, "wrong # args: should be \"pathName scm eval scm_command\"");
	return(TCL_ERROR);
    }

    result = Elk_Eval(argv[2]);

    if (result) {
	Tcl_ResetResult(interp);
	Tcl_AppendResult(interp, result, NULL);
    }
  
    return(TCL_OK);
}

//
// We need to allocate the data here instead of calling
// Get_Strsym_Stack because it cyclically uses its data,
// and it can not be freed.  Since we free a whole bunch
// of stuff, we want to be able to free this too.
//
static char *
Scheme_GetStrsym(Object str) 
{
    char *new_data;
    char *alloc_data;

    if (TYPE(str) == T_Symbol)
	str = SYMBOL(str)->name;
    else if (TYPE(str) != T_String)
	Wrong_Type_Combination(str, "string or symbol");

    new_data = Get_String(str);
    alloc_data = new char[strlen(new_data) + 1];
    strcpy(alloc_data, new_data);

    return(alloc_data);
}

//
// Implements the pad widget command in scheme.
//
static Object
Scheme_ExecCommand(int argc, Object *argv_obj)
{
    SchemeData *data;
    char **argv, ctmp;
    int i, itmp;
    Pad_TkWin *tkwin;
    Tcl_Interp *interp;
    double dtmp;
    typedef char *charPtr;
    Object return_str;

    argv = new charPtr[argc + 1];
    for (i = 0; i < argc; i++) {
	switch (TYPE(argv_obj[i])) {
	case T_Symbol :
	case T_String :
	    argv[i] = Scheme_GetStrsym(argv_obj[i]);
	    break;
	case T_Fixnum  :
	    itmp = FIXNUM(argv_obj[i]);
	    argv[i] = new char[(itmp / 10) + 2];
	    sprintf(argv[i],"%d", itmp);
	    break;
	case T_Flonum :
	    dtmp = FLONUM(argv_obj[i])->val;
	    argv[i] = new char[32];
	    sprintf(argv[i],"%g", dtmp);
	    break;        
	case T_Character :
	    ctmp = CHAR(argv_obj[i]);
	    argv[i] = new char[2];
	    strcpy(argv[i], &ctmp);
	    break;
	case T_Bignum :
	case T_Null :
	case T_Boolean :
	case T_Unbound :
	case T_Special :
	case T_Pair :
	case T_Environment :
	case T_Vector :
	case T_Primitive :
	case T_Compound :
	case T_Control_Point :
	case T_Promise :
	case T_Port :
	case T_End_Of_File :
	case T_Autoload :
	case T_Macro :
	case T_Broken_Heart :
	default :
	    fprintf(stderr, "Error: Scheme_ExecCommand: Wrong arg type %d\n", TYPE(argv_obj[i]));
	    return_str = Make_String("", 0);
	    argv[i] = new char[2];
	    strcpy(argv[i], "");
	    break;
	}
	Scheme_TclString(argv[i]);
    }
    argv[i] = 0;
  
    data = (SchemeData *)Scheme_Table->Get(argv[0]);
    if (data) {
	tkwin = data->tkwin;
	interp = data->interp;
	
	Pad_Command((ClientData)tkwin, interp, argc, argv);

	return_str = Make_String(interp->result, strlen(interp->result));
    } else {
	fprintf(stderr, "Error: non-existent pad widget: %s\n", argv[0]);
	return_str = Make_String("", 0);
    }

    for (i=0; i < argc; i++) {
	delete [] argv[i];
    }
    delete [] argv;

    return(return_str);
}

//
// Implements the scheme settoplevel command
//
static Object
Scheme_TopLevelCommand(int argc, Object *argv_obj)
{
    Object return_str;
    char *language_str;
    Pad_Language *language;
    Pad_Bool rc;

    if (argc != 1) {
	fprintf(stderr, "wrong # args: should be \"settoplevel language\"\n");
    } else {
	switch (TYPE(argv_obj[0])) {
	case T_Symbol :
	case T_String :
	    language_str = Scheme_GetStrsym(argv_obj[0]);
	    rc = Pad_GetLanguage(language_str, &language);
	    if (!rc) {
		fprintf(stderr, Pad_errorString.Get());
		fprintf(stderr, "\n");
	    } else {
		Pad_topLevelLanguage = language;
	    }
	    delete [] language_str;
	    break;
	default :
	    fprintf(stderr, "wrong arg type: language should be a symbol\n");
	    break;
	}
    }

    return_str = Make_String("", 0);

    return(return_str);
}

//
// Deal with the quoting problems
//   (currently unimplemented)
//
static void
Scheme_TclString(char *)
{
}

#endif /* HAS_SCHEME */
