/*
"(c) Copyright 1993-1997 Pad++ Consortium {University of New Mexico (UNM),
and New York University (NYU)}, All Rights Reserved."  
Licensee can not remove or obscure any of the
copyright or trademark notices in this software.

IN NO EVENT SHALL THE AUTHORS OR DISTRIBUTORS BE LIABLE TO ANY PARTY
FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
ARISING OUT OF THE USE OF THIS SOFTWARE, ITS DOCUMENTATION, OR ANY
DERIVATIVES THEREOF, EVEN IF THE AUTHORS HAVE BEEN ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.

THE AUTHORS AND DISTRIBUTORS SPECIFICALLY DISCLAIM ANY WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE, AND NON-INFRINGEMENT.  THIS SOFTWARE
IS PROVIDED ON AN "AS IS" BASIS, AND THE AUTHORS AND DISTRIBUTORS HAVE
NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR
MODIFICATIONS.

See the file "License" for general information on usage and
redistribution, and the file "LicenseTerms" for the specific license
agreement on usage and redistribution of this file, and the Pad++
software in general.
*/

#include "defs.h"
#include "pad-string.h"
#include "pad-tcl.h"
#include "tkwin.h"
#include "view.h"
#include "list.h"
#include "pad.h"
#include "global.h"
#include "tclInt.h"

#include <stdlib.h>
#include <ctype.h>

static Pad_Bool
_Is_quote(char c)
{
    Pad_Bool rc;

    switch (c) {
    case '"':
    case '{':
    case '}':
	rc = TRUE;
	break;
    default:
	rc = FALSE;
	break;
    }

    return(rc);
}

//
// Determine if the specified <str> is a single item,
// and can be specified in Tcl as just the string,
// or if it has embedded stuff, and thus needs to be quoted.
//
static Pad_Bool
_Single_item(Tcl_Interp *interp, char *str)
{
    int len, llen;
    Pad_Bool single = TRUE;
    Pad_String cmd;

    len = strlen(str);
    if (isspace(str[0]) || isspace(str[len-1]) ||
	_Is_quote(str[0]) || _Is_quote(str[len-1])) {
	single = FALSE;
    } else {
	cmd = "llength";
	cmd += str;
	if (Tcl_Eval(interp, cmd.Get()) == TCL_OK) {
	    llen = atoi(interp->result);
	    if (llen == 1) {
		single = TRUE;
	    } else {
		single = FALSE;
	    }
	} else {
	    single = FALSE;
	}
    }

    return(single);
}

//
// Pad_AppendTclElement
//
//   This appends <element> to <result> as a single Tcl list element.
//   Thus, if <element> consists of either zero, or more than one token,
//   it is wrapped with braces.
//
void
Pad_AppendTclElement(Tcl_Interp *interp, char *element, Pad_String &result)
{
    Pad_Bool single;

    if (result.Length() > 0) {
	result += " ";
    }
    single = _Single_item(interp, element);

    if (!single) {
	result += "\"";
    }
    result.Append_with_escapes(element);
    if (!single) {
	result += "\"";
    }
}

//
// Pad_AppendOption
//
//   This is a helper function for Padtcl_itemconfigure.
//   It gets the specified option, and appends the value
//   onto the specified result string, properly creating a Tcl list
//   (i.e., it puts braces around multi-token values).
//
void
Pad_AppendOption(Tcl_Interp *interp, Pad_Object *obj, Pad_Option *option, Pad_String &result, 
		 Pad_Bool nondefault_only, Pad_Bool braces)
{
    Pad_String value;
    Pad_Bool modified;

    if (option->write) 
	{
	    obj->Get_option(option, value, modified);
	    if ((nondefault_only && modified) || !nondefault_only) {
		if (result.Length() > 0) {
		    result += " ";
		}
		if (braces) {
		    result += "{";
		}
		result += option->option;
		if ((option->id == PAD_FILE) && (obj->pad->writeRelative)) {
		    Pad_Get_relative_pathname(interp, obj->pad->writeFiledir.Get(), value.Get(), value);
		}
		Pad_AppendTclElement(interp, value.Get(), result);
		if (braces) {
		    result += "}";
		}
	    }
	}
}

//
// Padtcl_itemconfigure:
// 
//   Configure an item on a pad widget.  Any options that could be specified when the
//   object was created can be specified here.  
// 
//   If the last option doesn't have an associated value, then that value is 
//   returned.
//
int 
Padtcl_itemconfigure(ClientData clientData, Tcl_Interp *interp, int argc, char **argv) 
{
    int rc = TCL_OK;
    int return_value;
    Pad_TkWin *tkwin = (Pad_TkWin *)clientData;
    Pad *pad;
    Pad_Object *obj;
    Pad_List objs;
    Pad_Iterator oi;
    Pad_Type *type;
    Pad_Option *option;
    Pad_String value;
    Pad_String result;
    Pad_Bool non_defaults;
    Tk_Uid typeUid;
    Pad_String *userType;
    
    non_defaults = FALSE;
    if ((argc >= 2) && !strcmp(argv[1], "-nondefaults")) {
	non_defaults = TRUE;
	argc--;
	argv++;
    }
    
    if (argc <= 1)
      {
	  TCL_ERR(interp, "wrong # args: should be \"pathName itemconfigure ?-nondefaults? tagOrId ?option value ...?\"");
      }
    
    pad = tkwin->view->pad;

    tkwin->view->pad->Find_with_tagorid(argv[1], objs, TRUE);
    
    if (argc == 2) {
	if (!objs.Is_empty()) {
				// No options specified, return list of pairs
				// of options and current values.
	    obj = (Pad_Object *)objs.First();
				// First, extract options from all types
	    type = (Pad_Type *)pad->typeTable->Get(Pad_allUid);
	    DOLIST(oi, type->options, Pad_Option, option) {
		Pad_AppendOption(interp, obj, option, result, non_defaults, TRUE);
	    }
				// Then, extract options for this object's type
            userType = obj->Get_usertype();
	    if (userType) {
		typeUid = Tk_GetUid(userType->Get());
	    } else {
		typeUid = Tk_GetUid(obj->Type_name());
	    }
	    type = (Pad_Type *)pad->typeTable->Get(typeUid);
	    DOLIST(oi, type->options, Pad_Option, option) {
		Pad_AppendOption(interp, obj, option, result, non_defaults, TRUE);
	    }

	    Tcl_ResetResult(interp);
	    Tcl_AppendResult(interp, result.Get(), NULL);
	}
    } else {
				// For each object in tagOrId, configure it
	if (argc > 4) {
	    return_value = 0;
	} else {
	    return_value = 1;
	}
	argc -= 2;
	argv += 2;
	DOLIST(oi, objs, Pad_Object, obj)
	  {
	      rc = Pad_ItemConfigure(tkwin, interp, obj, argc, argv);
	      if ((rc == TCL_OK) && (argc >= 2)) {
		  obj->Generate_event(Pad_ModifyNotify, NULL, "itemconfigure");
	      }
	  }
	if (!return_value && (rc == TCL_OK)) {
	    Tcl_ResetResult(interp);
	}
    }
    
    return(rc);
}

//
// Process pad object configuration.  This assumes that the object already
// exists. argc and argv specify pairs of options and values.
// 
// If the last option doesn't have an associated value, then that value is 
// returned.
//
int 
Pad_ItemConfigure(Pad_TkWin *tkwin, Tcl_Interp *interp, Pad_Object *obj, 
		  int argc, char **argv)
{
    int type;
    int len;
    Pad_Bool rc;
    Pad_Bool found;
    Pad_Bool modified;
    Pad_String item_result;
    Pad_Option *option;
    Pad_OptionKey optionKey;
    Pad *pad = tkwin->view->pad;
    Pad_String *userType;
    
    type = obj->Type();

    while (argc > 0) 
      {
	  rc = TRUE;
	  found = TRUE;
	  len = strlen(argv[0]);

				// Look up option from table.  
				// First try type-specific options.
				// If none found, and is a user-defined type, try base type options.
				// If none found, try generic options
	  optionKey.optionUid = Tk_GetUid(argv[0]);
          userType = obj->Get_usertype();
	  if (userType) {
	      optionKey.typeUid = Tk_GetUid(userType->Get());
	  } else {
	      optionKey.typeUid = Tk_GetUid(obj->Type_name());
	  }
	  option = (Pad_Option *)pad->optionTable->Get(&optionKey);
	  if (!option && userType) {
	      optionKey.typeUid = Tk_GetUid(obj->Type_name());
	      option = (Pad_Option *)pad->optionTable->Get(&optionKey);
	  }	    
	  if (!option) {
	      optionKey.typeUid = Pad_allUid;
	      option = (Pad_Option *)pad->optionTable->Get(&optionKey);
	  }
	  
	  if (option) {
	      Pad_errorString = "";
				// Set option
	      if (obj->Get_lock() && (option->id != PAD_LOCK)) {
		  rc = TRUE; // Don't modify locked items (except lock status)
	      } else {
		  if (argc >= 2) {
		      if (argv[1][0] == '\0') {
			  rc = obj->Set_option_default(option);
		      } else {
			  rc = obj->Set_option(option, argv[1]);
		      }
		  }
	      }
				// Get new value
				// (If user-defined option and just set option, 
				// then setting the option also returned the value, 
				// so just use that instead of calling the script again).
	      if (rc) {
		  if ((argc < 2) || (option->id != PAD_USER_OPTION)) {
		      rc = obj->Get_option(option, item_result, modified);
		      if (rc) {
		          TCL_RESULT(interp, item_result.Get());
		      }
		  }
	      }
	      if (!rc) {
		  TCL_ERR(interp, Pad_errorString.Get());
	      }
	  } else {
				// Don't generate errors for unknown options
				// while reading from a file.  Instead, write
				// a warning to stderr
	    if (!(pad->padFlags & PADROOT_READING)) {
  	        Pad_errorString = "Unknown option: ";
		Pad_errorString += argv[0];
		TCL_ERR(interp, Pad_errorString.Get());
	    } else {
		cerr << "Warning: Unknown option: " << argv[0] << endl;
	    }
	  }
	  argc -= 2;
	  argv += 2;
      }
    
    return(TCL_OK);
}
