/*
"(c) Copyright 1993-1997 Pad++ Consortium {University of New Mexico (UNM),
and New York University (NYU)}, All Rights Reserved."  
Licensee can not remove or obscure any of the
copyright or trademark notices in this software.

IN NO EVENT SHALL THE AUTHORS OR DISTRIBUTORS BE LIABLE TO ANY PARTY
FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
ARISING OUT OF THE USE OF THIS SOFTWARE, ITS DOCUMENTATION, OR ANY
DERIVATIVES THEREOF, EVEN IF THE AUTHORS HAVE BEEN ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.

THE AUTHORS AND DISTRIBUTORS SPECIFICALLY DISCLAIM ANY WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE, AND NON-INFRINGEMENT.  THIS SOFTWARE
IS PROVIDED ON AN "AS IS" BASIS, AND THE AUTHORS AND DISTRIBUTORS HAVE
NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR
MODIFICATIONS.

See the file "License" for general information on usage and
redistribution, and the file "LicenseTerms" for the specific license
agreement on usage and redistribution of this file, and the Pad++
software in general.
*/

#include "defs.h"
#include "pad-string.h"
#include "tkwin.h"
#include "view.h"
#include "portal.h"
#include "pad-tcl.h"
#include "text.h"
#include "point.h"
#include "list.h"
#include "pad.h"
#include "html.h"
#include "image.h"
#include "line.h"
#include "grid.h"
#include "group.h"
#include "alias.h"
#include "callback.h"
#include "global.h"

#include <stdlib.h>
#include <ctype.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/stat.h>

#ifdef PAD_UNIX
#  include <unistd.h>
#endif

/*
  Padtcl_create:

    Add any of the following objects to the pad from tcl.
    Returns an id for the object.

	   * alias      - An alias to an existing object
	   * grid       - A container with a layout engine that manages members
	   * group      - Group of objects
	   * html       - HTML document from disk
	   * image      - Image from disk
	   * kpl        - A kpl script that gets called to render itself
	   * line       - Straight lines between pairs of points
	   * oval       - Oval specified by ll and ur corners
	   * polygon    - Polygon between sequence of points
	   * portal     - Viewport to another place on the pad
	   * rectangle  - Rectangle specified by ll and ur corners
	   * spline     - Curve, curently beziers only
	   * tcl        - A tcl script that gets called to render itself
	   * text       - Multi-line text
	   * textfile   - Multi-line text file loaded from disk

	      And these widgets

           * button     - A basic button
	   * canvas     - A rectangle with a background cover meant to have its paint method overridden
	   * frame      - A container that looks and acts like a top-level window
	   * label      - A basic label
	   * panel      - A container
	   * scrollbar  - Horizontal or vertical scrollbar
	   * textfield  - A one line entry form
	   * textarea   - A multi-line entry form with scrollbar
	   * window     - Like a frame without window dressing

	   * User-defined types.  
	     For this, there must be a new type must have been previously
	     created with the addtype command.

    See the reference documentation for details.

    Note: This does not use the Pad++ API, but goes directly to the class methods
          because it does some specialized things.  Part of the issue is that
          user-defined object types are Tcl only, and eventually should work
          in any language.
*/

int 
Padtcl_create(ClientData clientData, Tcl_Interp *interp, int argc, char **argv)
{
    int id;
    int rc = TCL_OK;
    int processed_args;
    Pad_Object *obj = NULL;
    Pad_Type *type;
    Pad_String cmd;
    Tk_Uid typeUid;
    Pad_TkWin *tkwin = (Pad_TkWin *)clientData;
    Pad *pad = tkwin->view->pad;
    Pad_HashTableIterator hi;
    Pad_String *userType;

    if (argc <= 1)
      {
	  DOTABLE(hi, *pad->typeTable, Pad_Type, type) {
	      if ((type->name != Pad_allUid) && (type->name != Pad_padUid)) {
		  Tcl_AppendElement(interp, type->name);
	      }
	  }
	  return(TCL_OK);
      }

    typeUid = Tk_GetUid(argv[1]);
    type = (Pad_Type *)pad->typeTable->Get(typeUid);

    argc -= 2;
    argv += 2;

    if (type) {
	pad->Set_create_event_control(TRUE);    // We'll generate the <Create> event ourselves
        obj = pad->Create_object(type->id);
	if ( !obj ) {

				// If no built in widget, try user-defined type
	      if (type->createScript) {
		  cmd = type->createScript->Get();
		  cmd += " ";
		  cmd += tkwin->Get_name();
				// Add on any extra arguments to the create command
				// before options.
		  while ((argc > 0) && (argv[0][0] != '-')) {
		      cmd += " ";
		      cmd += argv[0];
		      argc--;
		      argv++;
		  }
				// Evaluate create script
		  rc = Pad_GlobalEval(type->createScript->Language(), cmd.Get());
		  if (rc == TCL_ERROR) {
		      pad->Set_create_event_control(FALSE);    // Return to default
		      return(TCL_ERROR);
		  }
		  id = atoi(interp->result);
		  obj = pad->Get_object_from_id(id);
		  if (obj) {
		      userType = new Pad_String(typeUid);
		      obj->Set_usertype(userType);
                      delete userType;
		  } else {
		      TCL_RESULT(tkwin->interp, "User-defined type didn't return object id");
		      pad->Set_create_event_control(FALSE);    // Return to default
		      return(TCL_ERROR);
		  }
	      }
	}
	pad->Set_create_event_control(FALSE);    // Return to default
    } else {
	Tcl_ResetResult(interp);
	Tcl_AppendResult(interp, "unknown type. Try:\n", NULL);
	DOTABLE(hi, *pad->typeTable, Pad_Type, type) {
	    if (type->name != Pad_allUid) {
		Tcl_AppendResult(interp, type->name, " ", NULL);
	    }
	}
	return(TCL_ERROR);
    }
				// Process extra command-line args before options
    processed_args = obj->Create_obj_args(argc, argv);
    if (processed_args == -1) {
	obj->Create_obj_error(); // Give objects an opportunity to clean up before object is deleted
	delete obj;
	TCL_RESULT(tkwin->interp, Pad_errorString.Get());
	return(TCL_ERROR);
    }
    argc -= processed_args;
    argv += processed_args;
    
				// Process itemconfigure requests
    rc = Pad_ItemConfigure(tkwin, interp, obj, argc, argv);

    if (rc == TCL_ERROR) {
	obj->Create_obj_error(); // Give objects an opportunity to clean up before object is deleted
	delete obj;
	return(TCL_ERROR);
    }
				// If within a coord frame, and -place hasn't been configured, 
				// and not a group, then transform object by it.
    if (!Pad_coordFrames.Is_empty() && 
	!(obj->optionFlags & PAD_PLACE_SET) &&
	!(obj->Is_group())) {
	obj->Set_rel_position_z(1.0);
    }

    obj->Generate_event(Pad_CreateNotify, NULL);

    sprintf(Pad_result, "%d", obj->id);
    TCL_RESULT(interp, Pad_result);
    
    return(TCL_OK);
}

//
// Reads a script stream and possibly evaluates it.
// The script can be either from a file or from a stream.
//
Pad_Bool
Read_script(istream *fp, Pad_StreamData &sd, Tcl_Interp *interp) 
{
   char *str = NULL;
   Pad_Bool rc=TRUE;

   if (fp) {      // read from file
     (*fp) >> sd;
   }

   if (sd.NextType() == _no_item) {   // nothing to get
       if (interp) sd.Clear();
       return FALSE;
   }

   if (!sd.Reconstruct(str)) {
     if (interp) {        // evalute script
       if (str && ((str[0] != '\0') && (str[0] != '\n' || strlen(str) > 2))) {
	   Pad_String orig_errorString = Pad_errorString;
	   Pad_errorString = "";
	   Tcl_GlobalEval(interp, str);
	   Pad_errorString = orig_errorString;
       }
     }
     delete[] str;
   } else {
     rc = FALSE;
   }

   sd.Clear();
   return TRUE;
}

