/*
"(c) Copyright 1993-1997 Pad++ Consortium {University of New Mexico (UNM),
and New York University (NYU)}, All Rights Reserved."  
Licensee can not remove or obscure any of the
copyright or trademark notices in this software.

IN NO EVENT SHALL THE AUTHORS OR DISTRIBUTORS BE LIABLE TO ANY PARTY
FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
ARISING OUT OF THE USE OF THIS SOFTWARE, ITS DOCUMENTATION, OR ANY
DERIVATIVES THEREOF, EVEN IF THE AUTHORS HAVE BEEN ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.

THE AUTHORS AND DISTRIBUTORS SPECIFICALLY DISCLAIM ANY WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE, AND NON-INFRINGEMENT.  THIS SOFTWARE
IS PROVIDED ON AN "AS IS" BASIS, AND THE AUTHORS AND DISTRIBUTORS HAVE
NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR
MODIFICATIONS.

See the file "License" for general information on usage and
redistribution, and the file "LicenseTerms" for the specific license
agreement on usage and redistribution of this file, and the Pad++
software in general.
*/

#ifndef HAS_SOUND
#include "api.h"
#include "defs.h"
#include "pad-string.h"
#include "pad-tcl-sound.h"
#include "pad-tcl.h"
#include "list.h"
#include "object.h"
#include "global.h"
#include "flist.h"
#include "ilist.h"
#include "pad-sound.h"

#include <iostream.h>

//
// Sound is not enabled so these functions return an error
//
void
Padtcl_sound_init(void)
{
    Pad_errorString = "Sound not enabled";
    cerr << "Sound not enabled" << endl;
}

//
// Padtcl_sound only supports the config -enabled option which returns FALSE
//    when sound support is not enabled.
//
int 
Padtcl_sound(ClientData, Tcl_Interp *interp, int argc, char **argv) 
{
    if (argc < 3) {
        TCL_RESULT(interp, "wrong # args: should be \"pathName sound config -enabled\"");
        return(TCL_ERROR);
    }

    if (!strncmp(argv[1], "config", 2)) {
        if (!strncmp(argv[2], "-enabled", 2)) {
            Pad_resultString = "0";
            TCL_RESULT(interp, Pad_resultString.Get());
            return(TCL_OK);
        }
    }
    Pad_errorString = "Sound not enabled";
    TCL_RESULT(interp, Pad_errorString.Get());
    return(TCL_ERROR);
}


#else /* HAS_SOUND */

#include "api.h"
#include "defs.h"
#include "pad-string.h"
#include "pad-tcl-sound.h"
#include "pad-tcl.h"
#include "list.h"
#include "object.h"
#include "global.h"
#include "flist.h"
#include "ilist.h"
#include "pad-sound.h"
#include <stdlib.h>
#include <string.h>

static int Padtcl_loadToken = 0;
static Pad_HashTable *Padtcl_soundTable = NULL;
static Pad_String Tcl_errorString;
static Pad_String optionUsage = "Should be one of -direction, -gain,  -materia, -maxWork, -minWork, -ratio, -size, -speed, -hardness, -impact, -length, -loop, -roughness, -scrape, or -volume";

static int Padtcl_sound_play  (ClientData, Tcl_Interp *, int, char **);
static int Padtcl_sound_config(ClientData, Tcl_Interp *, int, char **);
static int Padtcl_sound_load  (ClientData, Tcl_Interp *, int, char **);
static int Padtcl_sound_stop  (ClientData, Tcl_Interp *, int, char **);
static int Padtcl_sound_volume(ClientData, Tcl_Interp *, int, char **);

static char *Padtcl_load_token(void);

static int      Padtcl_config_sound(ClientData, Tcl_Interp *, int, char **);
static Pad_Bool Get_sound_option(char*, Pad_SoundOption&);
static Pad_Bool Get_sound_options(int, char**, Pad_List&);
static void     Print_sound_options(Pad_List&);

//
// Padtcl_sound_init() initializes sound support for tcl
//
void
Padtcl_sound_init(void)
{
    Padtcl_soundTable = new Pad_HashTable(PAD_STRING_TABLE);
}

//
// Padtcl_sound_create
//
int 
Padtcl_sound_create(ClientData, Tcl_Interp *interp, int argc, char **argv) 
{
#ifndef  HAS_NCD_SOUND
    int junk1 = argc;  argc = junk1;   // Just to stop compiler warnings
    char **junk2 = argv; argv = junk2;
    TCL_RESULT(interp, "creation of sounds only enabled for NCDSoundDevice, and not for native sound support");
    return(TCL_ERROR);
#else
    if (argc < 3) {
        TCL_RESULT(interp, "wrong # args: should be \"pathName sound create soundtype ?arg?\"");
        return(TCL_ERROR);
    }
    int numOptions;
    Pad_SoundHandle *sound = new Pad_SoundHandle;
    if (!strncmp(argv[2], "object", 2)){
        sound->Create(SoundObject);
    } else if (!strncmp(argv[2], "machine", 2)){
        sound->Create(SoundMachine);
    } else if (!strncmp(argv[2], "lift", 2)){
        sound->Create(SoundLift);
    } else {
        Pad_errorString = "sound create: error: unknown sound type: ";
        Pad_errorString += argv[2];
        Pad_errorString += "\n";
        Pad_errorString += "should be one of: lift, object, or machine";

            // we're not going to use <sound> so free it
        delete sound;
        
        TCL_RESULT(interp, Pad_errorString.Get());
        return(TCL_ERROR);
    }

        // We are creating a sound, so we need an even number of
        // option value pairs.
    if (numOptions % 2 != 0) {
        Pad_errorString = "sound create: error: should be even number of option value pairs";
        TCL_RESULT(interp, Pad_errorString.Get());
        return(TCL_ERROR);
    }

    Pad_List options;
    for (int i=3;i<argc;i+=2) {
        if (!strncmp(argv[i], "-material", 2)){
            Pad_optionSoundMaterial.data = atoi(argv[i+1]);
            options.Push(&Pad_optionSoundMaterial);
        } else if (!strncmp(argv[i], "-size", 2)){
            Pad_optionSoundSize.data = atoi(argv[i+1]);
            options.Push(&Pad_optionSoundSize);
        } else if (!strncmp(argv[i], "-speed", 2)){
            Pad_optionSoundSpeed.data = atoi(argv[i+1]);
            options.Push(&Pad_optionSoundSpeed);
        } else if (!strncmp(argv[i], "-direction", 2)){
            Pad_optionSoundDirection.data = atoi(argv[i+1]);
            options.Push(&Pad_optionSoundDirection);
        } else if (!strncmp(argv[i], "-ratio", 2)){
            Pad_optionSoundRatio.data = atoi(argv[i+1]);
            options.Push(&Pad_optionSoundRatio);
        } else if (!strncmp(argv[i], "-minWork", 2)){
            Pad_optionSoundMinWork.data = atoi(argv[i+1]);
            options.Push(&Pad_optionSoundMinWork);
        } else if (!strncmp(argv[i], "-maxWork", 2)){
            Pad_optionSoundMaxWork.data = atoi(argv[i+1]);
            options.Push(&Pad_optionSoundMaxWork);
        } else if (!strncmp(argv[i], "-gain", 2)){
            Pad_optionSoundGain.data = atoi(argv[i+1]);
            options.Push(&Pad_optionSoundGain);
        } else {
            Pad_errorString = "sound create: error: unknown configure option: ";
            Pad_errorString += argv[i];
            Pad_errorString += "\n";
            Pad_errorString += optionUsage;
            TCL_RESULT(interp, Pad_errorString.Get());
            return(TCL_ERROR);
        }
    }

    Pad_String token;
    int rc;
    if (!(rc = sound->Configure(options))) {
        Pad_errorString = "Couldn't configure sound ";
        TCL_RESULT(interp, Pad_errorString.Get());
        return(TCL_ERROR);
    }
    
    token = Padtcl_load_token();
    Padtcl_soundTable->Set(token.Get(),sound);
    TCL_RESULT(interp, token.Get());
    return TCL_OK;
#endif
}

//
// Padtcl_sound main
//
int 
Padtcl_sound(ClientData clientData, Tcl_Interp *interp, int argc, char **argv) 
{
    if (argc < 2) {
        TCL_RESULT(interp, "wrong # args: should be \"pathName sound subcommand ?arg?\"");
        return(TCL_ERROR);
    }

    if (!strncmp(argv[1], "config", 2)){
        return Padtcl_sound_config (clientData, interp, argc, argv);
    } else if (!strncmp(argv[1], "create", 2)){
        return Padtcl_sound_create (clientData, interp, argc, argv);
    } else if (!strncmp(argv[1], "load", 2)){
        return Padtcl_sound_load (clientData, interp, argc, argv);
    } else if (!strncmp(argv[1], "play", 2)){
        return Padtcl_sound_play (clientData, interp, argc, argv);
    } else if (!strncmp(argv[1], "stop", 2)){
        return Padtcl_sound_stop (clientData, interp, argc, argv);
    } else {
        Pad_resultString = "unknown sound subcommand: ";
        Pad_resultString += argv[1];
        Pad_resultString += "\n";
        Pad_resultString += "Should be one of: config, load, play, or stop";
        TCL_RESULT(interp, Pad_resultString.Get());
        return(TCL_ERROR);
    }  
}


//
// Padtcl_sound_load
//    loads a sound file into memory and stores it 
//    for later playing
//
int
Padtcl_sound_load(ClientData, 
		  Tcl_Interp *interp, 
		  int argc, char **argv) 
{
    Pad_String token;
    Pad_SoundHandle *sound = new Pad_SoundHandle;
    int rc;
    
    if (argc != 3) {
        TCL_RESULT(interp, "wrong # args: should be \"pathName sound load filename\"");
        return(TCL_ERROR);
    }
        // Load the sound into Pad++
    if (!(rc = sound->Load(argv[2]))) {
        TCL_RESULT(interp, Pad_errorString.Get());
        return TCL_ERROR;
    }

    token = Padtcl_load_token();
    Padtcl_soundTable->Set(token.Get(),sound);
    TCL_RESULT(interp, token.Get());
    return TCL_OK;
}

//
// Padtcl_sound_volume
//    Queries or sets the audio device volume settting
//    Currently supported options are:
//
//      -volume            - gets the volume for the right and left channels
//      -volume master     - sets the volume for the right and left channels
//      -volume right left - sets the the right and left channels seperately
//
int 
Padtcl_sound_volume(ClientData,
                    Tcl_Interp *interp,
                    int argc,
                    char **argv) 
{
    int rc;
    
        // Padtcl_sound_config has already checked to see if argc <= 3
    if (argc > 5) {
        TCL_RESULT(interp, "wrong # args: should be \"pathName sound config -volume right_vol left_vol\nor sound config -volume master_vol\"");
        return(TCL_ERROR);
    }
    
    Pad_UniverseMgr *mgr = Pad_UniverseMgr::Get_instance();

        // this is the getter function, so we return the volumes
        // of the left and right channels in a list.
    switch (argc) {
        case 3:
            int right;
            int left;
            if ((rc = mgr->Get_volume(left, right))){
                Pad_resultString = left;
                Pad_resultString += " ";
                Pad_resultString += right;
                TCL_RESULT(interp, Pad_resultString.Get());
            }
            break;
        case 4:
                // This sets both right and left channels to the same value
            if ((rc = mgr->Set_volume(atoi(argv[3])))) {
                TCL_RESULT(interp, argv[3]);
            }
            break;
        case 5:
                // Set the channels seperately
            if ((rc = mgr->Set_volume(atoi(argv[3]), atoi(argv[4])))) {
                Pad_resultString = argv[3];
                Pad_resultString += " ";
                Pad_resultString += argv[4];
                TCL_RESULT(interp, Pad_resultString.Get());
            }
            break;
    }

    if (!rc) {
        TCL_RESULT(interp, Pad_errorString.Get());
        return TCL_ERROR;
    } else {
        return TCL_OK;
    }
}

int
Padtcl_sound_print(void)
{
    Pad_SoundHandle *sound;
    Pad_HashTableIterator si;

        // go through the list of entries in the hash table and
        // print them out
    DOTABLE (si, *Padtcl_soundTable, Pad_SoundHandle, sound){
            //
        cout << "<" << (char*)si.key << ">" << *sound << endl;
    }
    return TCL_OK;
}

//
// Padtcl_sound_config
//    Configures the audio device
//    Currently supported options are:
//
//      -volume   - sets or gets the volume for the right and left channels
//      -sounds   - prints out the list of available sounds to stdout
//      -enabled  - return TRUE if sound support is enabled (which it is)
//
int 
Padtcl_sound_config(ClientData clientData,
                    Tcl_Interp *interp,
                    int argc,
                    char **argv) 
{
    if (argc < 3) {
        TCL_RESULT(interp, "wrong # args: should be \"pathName sound config subcommand ?args?\"");
        return(TCL_ERROR);
    }

    if (!strncmp(argv[2], "-volume", 2)){
        return Padtcl_sound_volume (clientData,interp,argc,argv);
    } else if (!strncmp(argv[2], "-sounds", 2)){
        return Padtcl_sound_print ();
    } else if (!strncmp(argv[2], "-enabled", 2)){
        Pad_resultString = "1";
        TCL_RESULT(interp, Pad_resultString.Get());
        return(TCL_OK);
    } else if (!strncmp(argv[2], "sound", 2)){
        return Padtcl_config_sound(clientData,interp,argc,argv);
    } else {
        Pad_resultString = "Unknown command: ";
        Pad_resultString << argv[2] << "\n"
                         << "should be one of: -sounds, -volume, or -enabled";
        TCL_RESULT(interp, Pad_resultString.Get());
        return(TCL_ERROR);
    }
}

//
// Get_sound_options
//    parses a list of option value pairs, building up a Pad_List
//    of the corresponding Pad_SoundOption's
//
Pad_Bool
Get_sound_options(int numOptions, char **options, Pad_List &conf)
{
    Pad_SoundOption *config;
    
        // set up the options array
    for (int i=0;i<numOptions;i+=2) {
        if (!strncmp(options[i], "-speed", 4)){
            config = &Pad_optionSoundSpeed;
        } else if (!strncmp(options[i], "-size", 4)){
            config = &Pad_optionSoundSize;
        } else if (!strncmp(options[i], "-ratio", 4)){
            config = &Pad_optionSoundRatio;
        } else if (!strncmp(options[i], "-maxWork", 4)){
            config = &Pad_optionSoundMaxWork;
        } else if (!strncmp(options[i], "-minWork", 4)){
            config = &Pad_optionSoundMinWork;
        } else if (!strncmp(options[i], "-roughness", 4)){
            config = &Pad_optionSoundRoughness;
        } else if (!strncmp(options[i], "-hardness", 4)){
            config = &Pad_optionSoundHardness;
        } else if (!strncmp(options[i], "-direction", 4)){
            config = &Pad_optionSoundDirection;
        } else if (!strncmp(options[i], "-material", 4)){
            config = &Pad_optionSoundMaterial;
        } else if (!strncmp(options[i], "-gain", 4)){
            config = &Pad_optionSoundGain;
        } else {
            Pad_errorString = "unknown configuration option: ";
            Pad_errorString += options[i];
            Pad_errorString += "\n";
            Pad_errorString += optionUsage;
            return FALSE;
        }
        config->data = atoi(options[i+1]);
        conf.Push(config);
    }
    return TRUE;
}

//
// Get_sound_option
//   return the corresponding SoundConfigType for the string <option>
//
Pad_Bool
Get_sound_option(char *option, Pad_SoundOption &conf)
{
    if (!strncmp(option, "-speed", 4)){
        conf.type = SoundConfigSpeed;
        return TRUE;
    } else if (!strncmp(option, "-size", 4)){
        conf.type = SoundConfigSize;
        return TRUE;
    } else if (!strncmp(option, "-ratio", 4)){
        conf.type = SoundConfigRatio;
        return TRUE;
    } else if (!strncmp(option, "-maxWork", 4)){
        conf.type = SoundConfigMaxWork;
        return TRUE;
    } else if (!strncmp(option, "-minWork", 4)){
        conf.type = SoundConfigMinWork;
        return TRUE;
    } else if (!strncmp(option, "-roughness", 4)){
        conf.type = SoundConfigRoughness;
        return TRUE;
    } else if (!strncmp(option, "-hardness", 4)){
        conf.type = SoundConfigHardness;
        return TRUE;
    } else if (!strncmp(option, "-direction", 4)){
        conf.type = SoundConfigDirection;
        return TRUE;
    } else if (!strncmp(option, "-material", 4)){
        conf.type = SoundConfigMaterial;
        return TRUE;
    } else if (!strncmp(option, "-gain", 4)){
        conf.type = SoundConfigGain;
        return TRUE;
    } else {
        return FALSE;
    }
}

// Print_sound_options
//   given a list of Pad_SoundOption's, print them
//   out in the Tcl {-option value} fashion
//
//   Pad_resultString is set to the output value
void
Print_sound_options(Pad_List &config)
{
    Pad_SoundOption *conf;

    Pad_resultString = "";
    while ((conf = (Pad_SoundOption*) config.Pop())) {
        switch (conf->type) {
            case SoundConfigDirection :
                Pad_resultString << "{-direction " << conf->data << "} ";
                break;
            case SoundConfigGain :
                Pad_resultString << "{-gain " << conf->data << "} ";
                break;
            case SoundConfigMaterial :
                Pad_resultString << "{-material " << conf->data << "} ";
                break;
            case SoundConfigMaxWork :
                Pad_resultString << "{-maxWork " << conf->data << "} ";
                break;
            case SoundConfigMinWork :
                Pad_resultString << "{-minWork " << conf->data << "} ";
                break;
            case SoundConfigRatio :
                Pad_resultString << "{-ratio " << conf->data << "} ";
                break;
            case SoundConfigSize :
                Pad_resultString << "{-size " << conf->data << "} ";
                break;
            case SoundConfigSpeed :
                Pad_resultString << "{-speed " << conf->data << "} ";
                break;
            case SoundConfigHardness :
                Pad_resultString << "{-hardness " << conf->data << "} ";
                break;
            case SoundConfigImpact :
                Pad_resultString << "{-impact " << conf->data << "} ";
                break;
            case SoundConfigLength :
                Pad_resultString << "{-length " << conf->data << "} ";
                break;
            case SoundConfigLoop :
                Pad_resultString << "{-loop " << conf->data << "} ";
                break;
            case SoundConfigRoughness :
                Pad_resultString << "{-roughness " << conf->data << "} ";
                break;
            case SoundConfigScrape :
                Pad_resultString << "{-scrape " << conf->data << "} ";
                break;
            case SoundConfigVolume :
                Pad_resultString << "{-volume " << conf->data << "} ";
                break;
            default :
                break;
        }
    }
}

int 
Padtcl_config_sound(ClientData /* clientData */,
                     Tcl_Interp *interp,
                     int argc,
                     char **argv) 
{
    Pad_SoundHandle *sound;
    Pad_List configList;
    Pad_SoundOption soundOption;
    int numOptions;

    numOptions = argc - 3;
    sound = (Pad_SoundHandle *) Padtcl_soundTable->Get(argv[2]);
    if (argc == 3) {
            // print out the current configuration of the sound
        if (sound->Get_configuration(configList)) {
            Print_sound_options(configList);
            TCL_RESULT(interp, Pad_resultString.Get());
            return(TCL_OK);
        } else {
            return(TCL_ERROR);
        }
    } else if (argc == 4) {
            // print out a single configuration option
        if (Get_sound_option (argv[3],soundOption)) {
            if (sound->Get_configuration(soundOption)) {
                Pad_resultString = soundOption.data;
                TCL_RESULT(interp, Pad_resultString.Get());
                return(TCL_OK);
            }
        } else {
            Pad_errorString = "Bad option: ";
            Pad_errorString << argv[3] << "\n" << optionUsage;
            TCL_RESULT(interp, Pad_errorString.Get());
            return(TCL_ERROR);
        }
    } else if ((numOptions) % 2 != 0) {
        TCL_RESULT(interp, "wrong # args: should be even number of option value pairs");
        return(TCL_ERROR);
    } else {
            // set a list of options
        Get_sound_options(numOptions, argv+3, configList);
        if (!sound->Configure (configList)) {
            TCL_RESULT(interp, Pad_errorString.Get());
            return(TCL_ERROR);
        }
        if (numOptions == 2) {
                // we were only setting a single option so we need to
                // echo the result. We don't need to check error codes
                // on these calls, because we know that we just
                // successfully set the option
            Get_sound_option (argv[3],soundOption);
            sound->Get_configuration(soundOption);
            Pad_resultString = soundOption.data;
            TCL_RESULT(interp, Pad_resultString.Get());
            return(TCL_OK);
        }
    }
    return(TCL_OK);
}

//
// Padtcl_sound_stop
//    stops the currently playing sound
//
int 
Padtcl_sound_stop(ClientData /* clientData */,
                  Tcl_Interp *interp,
                  int argc,
                  char **argv) 
{
    if (argc != 3) {
        TCL_RESULT(interp, "wrong # args: should be \"pathName sound stop token\"");
        return(TCL_ERROR);
    }
    
    Pad_SoundHandle *sound;
    sound = (Pad_SoundHandle *) Padtcl_soundTable->Get(argv[2]);
    if (!sound) {
        Pad_resultString = "Bad token: ";
        Pad_resultString += argv[2];
        TCL_RESULT(interp, Pad_resultString.Get());
        return(TCL_ERROR);
    }
    if (!sound->Stop()) {
        Pad_resultString = "Couldn't stop sound: ";
        Pad_resultString += argv[2];
        TCL_RESULT(interp, Pad_resultString.Get());
        return(TCL_ERROR);
    } else {
        return TCL_OK;
    }    
}

//
// Padtcl_sound_play
//    plays a sound from a previously loaded token
//
int 
Padtcl_sound_play(ClientData,
                  Tcl_Interp *interp,
                  int argc,
                  char **argv) 
{
    Pad_String token;
    Pad_SoundHandle *sound;
    static char *usage = "wrong # args: should be \"pathName sound play FileOrToken ?-loop? ?-impact? ?-scrape? ?-direction dir? ?-gain gain? ?-material mat? ?-maxWork max? ?-minWork min? ?-ration rat? ?-size size? ?-speed speed? ?-hardness hard? ?-length len? ?-roughness rough? ?-volume vol?\"";
    
    if (!(argc >= 3)) {
        TCL_RESULT(interp, usage);
        return(TCL_ERROR);
    }

    sound = (Pad_SoundHandle*) Padtcl_soundTable->Get(argv[2]);
    if (!sound) {
        TCL_RESULT(interp, "not a valid sound token");
        return(TCL_ERROR);
    }

        // Get all of the configuration options of the command
    Pad_SoundConfig *config = Pad_SoundConfig::Get_instance();
        // unset all the values 
    config->Reset();
    for (int i=3;i<argc;i++) {
        if (!strncmp(argv[i], "-loop", 4)){
            config->loop->data = 1;
        } else if (!strncmp(argv[i], "-scrape", 4)){
            config->scrape->data = 1;
        } else if (!strncmp(argv[i], "-impact", 4)){
            config->impact->data = 1;
        } else if (!strncmp(argv[i], "-direction", 4)){
            if (argc > i+1) {
                i++;
                config->direction->data = atoi(argv[i]);
            } else {
                TCL_RESULT(interp, usage);
                return(TCL_ERROR);
            }
        } else if (!strncmp(argv[i], "-gain", 4)){
            if (argc > i+1) {
                i++;
                config->gain->data = atoi(argv[i]);
            } else {
                TCL_RESULT(interp, usage);
                return(TCL_ERROR);
            }
        } else if (!strncmp(argv[i], "-material", 4)){
            if (argc > i+1) {
                i++;
                config->material->data = atoi(argv[i]);
            } else {
                TCL_RESULT(interp, usage);
                return(TCL_ERROR);
            }
        } else if (!strncmp(argv[i], "-minWork", 4)){
            if (argc > i+1) {
                i++;
                config->minWork->data = atoi(argv[i]);
            } else {
                TCL_RESULT(interp, usage);
                return(TCL_ERROR);
            }
        } else if (!strncmp(argv[i], "-maxWork", 4)){
            if (argc > i+1) {
                i++;
                config->maxWork->data = atoi(argv[i]);
            } else {
                TCL_RESULT(interp, usage);
                return(TCL_ERROR);
            }
        } else if (!strncmp(argv[i], "-ratio", 4)){
            if (argc > i+1) {
                i++;
                config->ratio->data = atoi(argv[i]);
            } else {
                TCL_RESULT(interp, usage);
                return(TCL_ERROR);
            }
        } else if (!strncmp(argv[i], "-size", 4)){
            if (argc > i+1) {
                i++;
                config->size->data = atoi(argv[i]);
            } else {
                TCL_RESULT(interp, usage);
                return(TCL_ERROR);
            }
        } else if (!strncmp(argv[i], "-speed", 4)){
            if (argc > i+1) {
                i++;
                config->speed->data = atoi(argv[i]);
            } else {
                TCL_RESULT(interp, usage);
                return(TCL_ERROR);
            }
        } else if (!strncmp(argv[i], "-hardness", 4)){
            if (argc > i+1) {
                i++;
                config->hardness->data = atoi(argv[i]);
            } else {
                TCL_RESULT(interp, usage);
                return(TCL_ERROR);
            }
        } else if (!strncmp(argv[i], "-length", 4)){
            if (argc > i+1) {
                i++;
                config->length->data = atoi(argv[i]);
            } else {
                TCL_RESULT(interp, usage);
                return(TCL_ERROR);
            }
        } else if (!strncmp(argv[i], "-roughness", 4)){
            if (argc > i+1) {
                i++;
                config->roughness->data = atoi(argv[i]);
            } else {
                TCL_RESULT(interp, usage);
                return(TCL_ERROR);
            }
        } else if (!strncmp(argv[i], "-volume", 4)){
            if (argc > i+1) {
                i++;
                config->volume->data = atoi(argv[i]);
            } else {
                TCL_RESULT(interp, usage);
                return(TCL_ERROR);
            }
        } else {
            Pad_errorString = "unknown play option: ";
            Pad_errorString += argv[i];
            Pad_errorString += "\n";
            Pad_errorString += optionUsage;
            TCL_RESULT(interp, Pad_errorString.Get());
            return(TCL_ERROR);
        }
    }

    if (! sound->Play(config)) {
        TCL_RESULT(interp, Pad_errorString.Get());
        return(TCL_ERROR);
    } else {
        return TCL_OK;
    }
}

char *
Padtcl_load_token(void)
{
    static Pad_String token;

    Padtcl_loadToken++;
    token = "sound";
    token += Padtcl_loadToken;
    return token.Get();
}

#endif /* HAS_SOUND */
