/*
"(c) Copyright 1993-1997 Pad++ Consortium {University of New Mexico (UNM),
and New York University (NYU)}, All Rights Reserved."  
Licensee can not remove or obscure any of the
copyright or trademark notices in this software.

IN NO EVENT SHALL THE AUTHORS OR DISTRIBUTORS BE LIABLE TO ANY PARTY
FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
ARISING OUT OF THE USE OF THIS SOFTWARE, ITS DOCUMENTATION, OR ANY
DERIVATIVES THEREOF, EVEN IF THE AUTHORS HAVE BEEN ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.

THE AUTHORS AND DISTRIBUTORS SPECIFICALLY DISCLAIM ANY WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE, AND NON-INFRINGEMENT.  THIS SOFTWARE
IS PROVIDED ON AN "AS IS" BASIS, AND THE AUTHORS AND DISTRIBUTORS HAVE
NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR
MODIFICATIONS.

See the file "License" for general information on usage and
redistribution, and the file "LicenseTerms" for the specific license
agreement on usage and redistribution of this file, and the Pad++
software in general.
*/

// ************************************************************************
// Pad_StreamData class implementation
//
// Implements operations to manipulate, store and retrieve marshaled data.
// ************************************************************************

#include <string.h>
#include <iostream.h>

#ifdef PAD_UNIX
#include <strstream.h>
#endif

#ifdef PAD_WIN
#include <strstrea.h>
#endif

#include "defs.h"
#include "streamdata.h"
#include "xdrintf.h"


// ************************************************************************
// constructors, etc.
// ************************************************************************

Pad_StreamData::Pad_StreamData(int rep, unsigned int initialSize, 
			       istream *inputfp, ostream *outputfp)
  : data(NULL)
{
  xrep = rep;
  switch (xrep) {
  case PAD_XDR_FORMAT:
    drep = new Pad_XDRDataRep;
    break;
  case PAD_NATIVE_FORMAT:
    drep = new Pad_NativeDataRep;
    break;
  case PAD_CHAR_FORMAT:
    drep = new Pad_CharDataRep;
    break;
  default:
    cerr << "ERR: Ignoring unknown stream data rep: " << xrep << endl;
    drep = new Pad_XDRDataRep;
    break;
  }

  items = 0;
  currentItem = 0;
  size = initialSize;
  if (size) data = new Pad_StreamItem*[size];
  ifp = inputfp;
  ofp = outputfp;
  totalBytes = bytesRead = 0;
}

Pad_StreamData::Pad_StreamData(const Pad_StreamData& m)
  : data(NULL)
{
  xrep = m.xrep;
  items = m.items;
  currentItem = m.currentItem;
  size = m.size;
  if (size) data = new Pad_StreamItem*[size];
  for (unsigned int i=0; i<items; i++) {
    data[i] = new Pad_StreamItem;
    *data[i] = *m.data[i];
  }
}

Pad_StreamData& Pad_StreamData::operator= (const Pad_StreamData& m)
{
  if (this==&m) return *this;
  Clear();
  if (size != m.size){
    if (data) delete [] data;
    size = m.size;
    if (size) data = new Pad_StreamItem*[size];
    else data = NULL;
  };

  items = m.items;
  currentItem = m.currentItem;
  for (unsigned int i=0; i<items; i++) {
    data[i] = new Pad_StreamItem;
    *data[i] = *m.data[i];
  };
  return *this;
}

//
// Move encoded items from another stream to here.
//
int Pad_StreamData::MoveItems(Pad_StreamData &src)
{
    unsigned int i, j;

    if (size < (items + src.items)) {
	Pad_StreamItem **newdata = new Pad_StreamItem*[items + src.items];
	for (i=0; i < items; i++) {
	    newdata[i] = data[i];
	}
	delete[] data;
	data = newdata;
	size = items + src.items;
    }

    for (i=0, j = items; i < src.items; i++, j++) {
	data[j] = src.data[i];
	src.data[i] = NULL;
    }

    items += src.items;
    src.items = 0;  // so they don't get deleted
    src.currentItem=0;
    return 0;
}

Pad_StreamData::~Pad_StreamData()
{
  Clear();
  delete[] data;
  delete drep;
}

int Pad_StreamData::UseRep(int rep)
{
  int rc = noErr;
  if (items==0) {
    switch (rep) {
    case PAD_XDR_FORMAT:
      delete drep;
      xrep=rep;
      drep = new Pad_XDRDataRep;
      break;
    case PAD_NATIVE_FORMAT:
      delete drep;
      xrep=rep;
      drep = new Pad_NativeDataRep;
      break;
    case PAD_CHAR_FORMAT:
      delete drep;
      xrep=rep;
      drep = new Pad_CharDataRep;
      break;
    default:
      rc = objectErr;
      break;
    }

  } else if (xrep != rep) {
    rc = objectErr;
  }

  return rc;
}

istream* Pad_StreamData::Get_istream()
{
  return ifp;
}

ostream* Pad_StreamData::Get_ostream()
{
  return ofp;
}

// ************************************************************************
// flattening and expansion methods
// ************************************************************************

int Pad_StreamData::Flatten(unsigned char*& output, unsigned int& length)
{
  ostrstream out;
  out << *this;
  length = out.pcount();
  output = (unsigned char*)out.str();
  if (out.good())
    return noErr;
  else
    return objectErr;
}

int Pad_StreamData::Expand(unsigned char* input, unsigned int length,
			     unsigned int& bytes_read)
{
  Clear();
  bytes_read = 0;
  istrstream in((char*)input, length);
  in >> *this;
  bytes_read = in.gcount();
  in.rdbuf()->freeze(1);  // avoid deallocation by stream destructor
  if (in.good())
    return noErr;
  else
    return objectErr;
}


// ************************************************************************
// fundamental marshaling and reconstruction member functions
// ************************************************************************

void Pad_StreamData::marshal(Pad_StreamItem* mi)
{
  // just put the item in the table.  This is where other marshaling
  // processing could take place (such as encryption, etc).

  if (items == size) {
    Pad_StreamItem** temp;
    temp = data;
    size = items+1;
    data = new Pad_StreamItem*[size];
    for (unsigned int i=0; i<items; i++) {
      data[i] = temp[i];
    }
    if (temp) delete [] temp;
  }

  data[items] = mi;
  items++;
}

int Pad_StreamData::reconstruct(Pad_StreamItem *&mi)
{
  // just return the next item in the stream

  if (currentItem < items) {
    mi = data[currentItem];
    currentItem++;
    return noErr;
  }
  else
    return itemErr;
}


// ************************************************************************
// marshaling functions
// ************************************************************************

int Pad_StreamData::Marshal(const char* input)
{
  int err=noErr;
  Pad_StreamItem* item = new Pad_StreamItem;

  item->type = _char | _vector;

  if (input) {
    item->dataLength=strlen(input)+1;
    if (item->dataLength > 1) {
      err = drep->EncodeString(input, item->info, item->infoLength);

    } else {
      // hack, xdr encoding for length of 1 seems broken
      char buf[2];
      buf[0] = buf[1] = 0;
      item->dataLength = 2;
      err = drep->EncodeString(buf, item->info, item->infoLength);
    }
  }

  if (!err) marshal(item);
  return err;
}

int Pad_StreamData::Marshal(const short* input, unsigned int length)
{
  Pad_StreamItem* item=new Pad_StreamItem;
  item->type=_short;
  if (length > 1) item->type |= _vector;
  if (length) {
    item->dataLength=length;
    if (drep) {
      int err=drep->EncodeArray(input, length, item->info, item->infoLength);
      if (err) return err;
    }
    else {
      return objectErr;
    };
  };
  marshal(item);
  return noErr;
}

int Pad_StreamData::Marshal(const int* input, unsigned int length)
{
  Pad_StreamItem* item=new Pad_StreamItem;
  item->type=_int;
  if (length > 1) item->type |= _vector;
  if (length) {
    item->dataLength=length;
    if (drep) {
      int err=drep->EncodeArray(input, length, item->info, item->infoLength);
      if (err) return err;
    }
    else {
      return objectErr;
    }
  }
  marshal(item);
  return noErr;
}

int Pad_StreamData::Marshal(const long* input, unsigned int length)
{
  Pad_StreamItem* item=new Pad_StreamItem;
  item->type=_long;
  if (length > 1) item->type |= _vector;
  if (length) {
    item->dataLength=length;
    if (drep) {
      int err=drep->EncodeArray(input, length, item->info, item->infoLength);
      if (err) return err;
    }
    else {
      return objectErr;
    }
  }
  marshal(item);
  return noErr;
}

int Pad_StreamData::Marshal(const unsigned char* input, unsigned int length)
{
  Pad_StreamItem* item=new Pad_StreamItem;
  item->type=_unsigned_char;
  if (length > 1) item->type |= _vector;
  if (length) {
    item->dataLength=length;
    if (drep) {
      int err = drep->EncodeArray(input, length, item->info, item->infoLength);
      if (err) return err;
    } else {
      return objectErr;
    }
  }	
  marshal(item);
  return noErr;
}

int Pad_StreamData::Marshal(const unsigned short* input, unsigned int length)
{
  Pad_StreamItem* item=new Pad_StreamItem;
  item->type=_unsigned_short;
  if (length > 1) item->type |= _vector;
  if (length)
  {
    item->dataLength=length;
    if (drep) {
      int err=drep->EncodeArray(input, length, item->info, item->infoLength);
      if (err) return err;
    }
    else {
      return objectErr;
    }
  }
  marshal(item);
  return noErr;
}

int Pad_StreamData::Marshal(const unsigned int* input, unsigned int length)
{
  Pad_StreamItem* item=new Pad_StreamItem;
  item->type=_unsigned_int;
  if (length > 1) item->type |= _vector;
  if (length)  {
    item->dataLength=length;
    if (drep) {
      int err=drep->EncodeArray(input, length, item->info, item->infoLength);
      if (err) return err;
    }
    else {
      return objectErr;
    }
  }
  marshal(item);
  return noErr;
}

int Pad_StreamData::Marshal(const unsigned long* input, unsigned int length)
{
  Pad_StreamItem* item=new Pad_StreamItem;
  item->type=_unsigned_long;
  if (length > 1) item->type |= _vector;
  if (length) {
    item->dataLength=length;
    if (drep) {
      int err=drep->EncodeArray(input, length, item->info, item->infoLength);
      if (err) return err;
    }
    else {
      return objectErr;
    }
  }
  marshal(item);
  return noErr;
}
  
int Pad_StreamData::Marshal(const float* input, unsigned int length)
{
  Pad_StreamItem* item=new Pad_StreamItem;
  item->type=_float;
  if (length > 1) item->type |= _vector;
  if (length) {
    item->dataLength=length;
    if (drep) {
      int err=drep->EncodeArray(input, length, item->info, item->infoLength);
      if (err) return err;
    }
    else {
      return objectErr;
    }
  }
  marshal(item);
  return noErr;
}

int Pad_StreamData::Marshal(const double* input, unsigned int length)
{
  Pad_StreamItem* item=new Pad_StreamItem;
  item->type=_double;
  if (length > 1) item->type |= _vector;
  if (length) {
    item->dataLength=length;
    if (drep) {
      int err=drep->EncodeArray(input, length, item->info, item->infoLength);
      if (err) return err;
    }
    else {
      return objectErr;
    }
  }
  marshal(item);
  return noErr;
}


// ************************************************************************
// reconstruction functions
// ************************************************************************

int Pad_StreamData::Reconstruct(char* output, unsigned int length)
{
  Pad_StreamItem *item;
  int err=reconstruct(item);
  if (err) return err;
  item->type = (length > 1) ? (_char | _vector) : (_char);
  item->dataLength = length;
  if (item->Read(*ifp, this)) {
    err=drep->DecodeString(item->info, item->infoLength, output, length);
  } else {
    err = lengthErr;
  }
  return err;
}

int Pad_StreamData::Reconstruct(short* output, unsigned int length)
{
  Pad_StreamItem *item;
  int err=reconstruct(item);
  if (err) return err;
  item->type = (length > 1) ? (_short | _vector) : (_short);
  item->dataLength = length;
  if (item->Read(*ifp, this)) {
    err = drep->DecodeArray(item->info, item->infoLength, output, length);
  } else {
    err = lengthErr;
  }
  return err;
}

int Pad_StreamData::Reconstruct(int* output, unsigned int length)
{
  Pad_StreamItem *item;
  int err=reconstruct(item);
  if (err) return err;
  item->type = (length > 1) ? (_int | _vector) : (_int);
  item->dataLength = length;
  if (item->Read(*ifp, this)) {
    err = drep->DecodeArray(item->info, item->infoLength, output, length);
  } else {
    err = lengthErr;
  }
  return err;
}

int Pad_StreamData::Reconstruct(long* output, unsigned int length)
{
  Pad_StreamItem *item;
  int err=reconstruct(item);
  if (err) return err;
  item->type = (length > 1) ? (_long | _vector) : (_long);
  item->dataLength = length;
  if (item->Read(*ifp, this)) {
    err = drep->DecodeArray(item->info, item->infoLength, output, length);
  } else {
    err = lengthErr;
  }
  return err;
}

int Pad_StreamData::Reconstruct(unsigned char* output, unsigned int length)
{
  Pad_StreamItem *item;
  int err=reconstruct(item);
  if (err) return err;
  item->type = (length > 1) ? (_unsigned_char | _vector) : (_unsigned_char);
  item->dataLength = length;
  if (item->Read(*ifp, this)) {
    err = drep->DecodeArray(item->info, item->infoLength, output, length);
  } else {
    err = lengthErr;
  }
  return err;
}

int Pad_StreamData::Reconstruct(unsigned short* output, unsigned int length)
{
  Pad_StreamItem *item;
  int err=reconstruct(item);
  if (err) return err;
  item->type = (length > 1) ? (_unsigned_short | _vector) : (_unsigned_short);
  item->dataLength = length;
  if (item->Read(*ifp, this)) {
    err = drep->DecodeArray(item->info, item->infoLength, output, length);
  } else {
    err = lengthErr;
  }
  return err;
}

int Pad_StreamData::Reconstruct(unsigned int* output, unsigned int length)
{
  Pad_StreamItem *item;
  int err=reconstruct(item);
  if (err) return err;
  item->type = (length > 1) ? (_unsigned_int | _vector) : (_unsigned_int);
  item->dataLength = length;
  if (item->Read(*ifp, this)) {
    err = drep->DecodeArray(item->info, item->infoLength, output, length);
  } else {
    err = lengthErr;
  }
  return err;
}

int Pad_StreamData::Reconstruct(unsigned long* output, unsigned int length)
{
  Pad_StreamItem *item;
  int err=reconstruct(item);
  if (err) return err;
  item->type = (length > 1) ? (_unsigned_long | _vector) : (_unsigned_long);
  item->dataLength = length;
  if (item->Read(*ifp, this)) {
    err = drep->DecodeArray(item->info, item->infoLength, output, length);
  } else {
    err = lengthErr;
  }
  return err;
}

int Pad_StreamData::Reconstruct(float* output, unsigned int length)
{
  Pad_StreamItem *item;
  int err=reconstruct(item);
  if (err) return err;
  item->type = (length > 1) ? (_float | _vector) : (_float);
  item->dataLength = length;
  if (item->Read(*ifp, this)) {
    err = drep->DecodeArray(item->info, item->infoLength, output, length);
  } else {
    err = lengthErr;
  }
  return err;
}

int Pad_StreamData::Reconstruct(double* output, unsigned int length)
{
  Pad_StreamItem *item;
  int err=reconstruct(item);
  if (err) return err;
  item->type = (length > 1) ? (_double | _vector) : (_double);
  item->dataLength = length;
  if (item->Read(*ifp, this)) {
    err = drep->DecodeArray(item->info, item->infoLength, output, length);
  } else {
    err = lengthErr;
  }
  return err;
}


// ************************************************************************
// convenience methods that allocate output buffer
// ************************************************************************

int Pad_StreamData::Reconstruct(char*& output, unsigned int *len)
{
  int length=NextLength();
  if (len) *len = (unsigned int)length;
  if (length < 0) {
    return itemErr;
  }

  if (!length) {
    output=NULL;
    ++currentItem;
    return noErr;
  }

  output=new char[length];
  int rc=Reconstruct(output, length);
  if (rc) {
    delete[] output; 
    output=NULL;
  }
  return rc;
}

int Pad_StreamData::Reconstruct(short*& output, unsigned int *len)
{
  int length=NextLength();
  if (len) *len = (unsigned int)length;
  if (length < 0) {
    return itemErr;
  }

  if (!length) {
    output=NULL;
    ++currentItem;
    return noErr;
  }

  output=new short[length];
  int rc=Reconstruct(output, length);
  if (rc) {
    delete[] output; 
    output=NULL;
  }
  return rc;
}

int Pad_StreamData::Reconstruct(int*& output, unsigned int *len)
{
  int length=NextLength();
  if (len) *len = (unsigned int)length;
  if (length < 0) {
    return itemErr;
  }

  if (!length) {
    output=NULL;
    ++currentItem;
    return noErr;
  }

  output=new int[length];
  int rc=Reconstruct(output, length);
  if (rc) {
    delete[] output; 
    output=NULL;
  }
  return rc;
}

int Pad_StreamData::Reconstruct(long*& output, unsigned int *len)
{
  int length=NextLength();
  if (len) *len = (unsigned int)length;
  if (length < 0) {
    return itemErr;
  }

  if (!length) {
    output=NULL;
    ++currentItem;
    return noErr;
  }

  output=new long[length];
  int rc=Reconstruct(output, length);
  if (rc) {
    delete[] output; 
    output=NULL;
  }
  return rc;
}

int Pad_StreamData::Reconstruct(unsigned char*& output, unsigned int *len)
{
  int length=NextLength();
  if (len) *len = (unsigned int)length;
  if (length < 0) {
    return itemErr;
  }

  if (!length) {
    output=NULL;
    ++currentItem;
    return noErr;
  }

  output=new unsigned char[length];
  int rc=Reconstruct(output, length);
  if (rc) {
    delete[] output; 
    output=NULL;
  }
  return rc;
}

int Pad_StreamData::Reconstruct(unsigned short*& output, unsigned int *len)
{
  int length=NextLength();
  if (len) *len = (unsigned int)length;
  if (length < 0) {
    return itemErr;
  }

  if (!length) {
    output=NULL;
    ++currentItem;
    return noErr;
  }

  output=new unsigned short[length];
  int rc=Reconstruct(output, length);
  if (rc) {
    delete[] output; 
    output=NULL;
  }
  return rc;
}

int Pad_StreamData::Reconstruct(unsigned int*& output, unsigned int *len)
{
  int length=NextLength();
  if (len) *len = (unsigned int)length;
  if (length < 0) {
    return itemErr;
  }

  if (!length) {
    output=NULL;
    ++currentItem;
    return noErr;
  }

  output=new unsigned int[length];
  int rc=Reconstruct(output, length);
  if (rc) {
    delete[] output; 
    output=NULL;
  }
  return rc;
}

int Pad_StreamData::Reconstruct(unsigned long*& output, unsigned int *len)
{
  int length=NextLength();
  if (len) *len = (unsigned int)length;
  if (length < 0) {
    return itemErr;
  }

  if (!length) {
    output=NULL;
    ++currentItem;
    return noErr;
  }

  output=new unsigned long[length];
  int rc=Reconstruct(output, length);
  if (rc) {
    delete[] output; 
    output=NULL;
  }
  return rc;
}

int Pad_StreamData::Reconstruct(float*& output, unsigned int *len)
{
  int length=NextLength();
  if (len) *len = (unsigned int)length;
  if (length < 0) {
    return itemErr;
  }

  if (!length) {
    output=NULL;
    ++currentItem;
    return noErr;
  }

  output=new float[length];
  int rc=Reconstruct(output, length);
  if (rc) {
    delete[] output; 
    output=NULL;
  }
  return rc;
}

int Pad_StreamData::Reconstruct(double*& output, unsigned int *len)
{
  int length=NextLength();
  if (len) *len = (unsigned int)length;
  if (length < 0) {
    return itemErr;
  }

  if (!length) {
    output=NULL;
    ++currentItem;
    return noErr;
  }

  output=new double[length];
  int rc=Reconstruct(output, length);
  if (rc) {
    delete[] output; 
    output=NULL;
  }

  return rc;
}


// ************************************************************************
// miscellaneous public member functions
// ************************************************************************

//
// Free the item table and skip over any unread bytes in the stream.
//
void Pad_StreamData::Clear()
{
  for (unsigned int i=0; i<items; i++) {
    delete data[i];
  }
  items=0;
  currentItem=0;
  if (totalBytes > 0) {  // read the rest of input
    if (bytesRead < totalBytes) {
      if (ifp) ifp->ignore(totalBytes - bytesRead);
    }
  }
  totalBytes = bytesRead = 0;
}


// ************************************************************************
// _full methods
// ************************************************************************

int Pad_StreamData::marshal_full(Pad_StreamData& m) const
{
  int err=m.Marshal(xrep);
  if (err) return err;
  err=m.Marshal(items);
  if (err) return err;
  for (unsigned int i=0; i<items; i++) {
    err=data[i]->marshal(m);
    if (err) return err;
  }
  return noErr;
}

int Pad_StreamData::reconstruct_full(Pad_StreamData& m)
{
  Clear();
  currentItem=0;
  int err=m.Reconstruct(xrep);
  if (err) return err;
  err=m.Reconstruct(items);
  if (err) return err;
 
 if (items>size) {
    if (data) delete [] data;
    size=items;
    if (size) data=new Pad_StreamItem*[size];
    else data=NULL;
  }

  for (unsigned int i=0; i<items; i++) {
    data[i]=new Pad_StreamItem;
    err=data[i]->reconstruct(m);
    if (err) {
      items=i+1;
      return err;
    }
  }
  return noErr;
}

int Pad_StreamData::NextType() const
{
  //if (currentItem<items) return data[currentItem]->type; else return _no_item;

  // this isn't really returning the next type but an indication
  // for its existence.

  if (bytesRead < totalBytes) 
    return _no_type;
  else
    return _no_item;
}

// ************************************************************************
// iostream operators
// ************************************************************************

//
// Read an integer from the input and assume it is the
// length of the vector to follow.
// 
int Pad_StreamData::NextLength()
{
  int length=0;
  unsigned int inbuf_len;
  unsigned char *inbuf=NULL;
  int rc;

  if ((rc = drep->ReadData(*ifp, _int, 1, inbuf_len, inbuf)) > 0) {
    bytesRead += rc;
    drep->DecodeArray(inbuf, inbuf_len, &length, 1);
    delete[] inbuf;
    rc = length;
  } else {
    rc = itemErr;
  }

  return rc;
}

//
// Write the stream and its items.  The format is as follows:
//   int num_items;        // items to expect in the stream
//   int total_bytes;      // number of bytes to follow
//   <zero or more items>    // See Pad_StreamItem::Write.
//

ostream& operator<< (ostream& os, Pad_StreamData& m)
{
    int inbuf[4];
    int inbuf_sz;
    unsigned int i;
    unsigned char *outbuf=NULL;
    unsigned int outbuf_len=0;
    Pad_StreamItem *item;

    // number of items
    inbuf[0] = m.items;
    inbuf_sz = 1;
    m.drep->EncodeArray(inbuf, inbuf_sz, outbuf, outbuf_len);
    os.write(outbuf, outbuf_len);
    delete[] outbuf;
    outbuf_len = 0;
    if (m.xrep == PAD_CHAR_FORMAT) os << ' ';

    inbuf[0] = m.ItemsLength();
    if (m.xrep == PAD_CHAR_FORMAT) inbuf[0]++;
    m.drep->EncodeArray(inbuf, inbuf_sz, outbuf, outbuf_len);
    os.write(outbuf, outbuf_len);
    delete[] outbuf;
    outbuf_len = 0;
    if (m.xrep == PAD_CHAR_FORMAT) os << endl;

    // Now do the items
    for (i=0; i < m.items; i++) {
	item = m.data[i];
	item->Write(os, &m);
    }

    return os;
}

istream& operator>> (istream& is, Pad_StreamData& m)
{
  int numItems, bytes;
  unsigned int inbuf_len;
  unsigned char *inbuf=NULL;

  if ( !m.drep->ReadData(is, _int, 1, inbuf_len, inbuf) ) {
    return is;
  }
  m.drep->DecodeArray(inbuf, inbuf_len, &numItems, 1);
  delete[] inbuf;
  m.items = numItems;

  if ( !m.drep->ReadData(is, _int, 1, inbuf_len, inbuf) ) {
    return is;
  }
  m.drep->DecodeArray(inbuf, inbuf_len, &bytes, 1);
  delete[] inbuf;
  m.totalBytes = bytes;

  m.bytesRead = 0;
  if (m.xrep == PAD_CHAR_FORMAT) m.bytesRead=1;

  // Allocate space for items
  if (m.items > m.size) {
    if (m.data) delete [] m.data;
    m.size=m.items;
    if (m.size) m.data=new Pad_StreamItem*[m.size];
    else m.data=NULL;
  }

  for (unsigned int i=0; i<m.items; i++) {
    m.data[i]=new Pad_StreamItem;
  }

  return is;
}

//
// Methods for skipping over input bytes.
//
int Pad_StreamData::SkipInput(int bytes)
{
  int rc=1;
  if (ifp) {
    ifp->ignore(bytes);
    bytesRead += bytes;
    if (drep->Type() == PAD_CHAR_FORMAT) {
      ifp->ignore(1);
      bytesRead++;
    }
    rc = ifp->good();
  }

  return rc;
}

// Compute the total length of byutes required for writting the items.
//
int Pad_StreamData::ItemsLength() const
{
    unsigned int sum=0, i;

    for (i=0; i < items; i++) {
	sum += data[i]->OutputLength(drep);
    }

    return sum;
}

//
// If have an ostream then write the items to it and flush it.
// Clears the stream too.
//
int Pad_StreamData::Flush() {
  int rc= noErr;
  if (ofp) {
    if (items > 0) {
      (*ofp) << *this;
      rc = ofp->good();
    }
  }

  Clear();
  return rc;
}

//
// If have an istream then read the stream header info from it.
//
int Pad_StreamData::Fill() {
  int rc = noErr;
  if (ifp) {
    (*ifp) >> *this;
    rc = ifp->good();
  }

  return rc;
}

//
// DataRep method for reading in raw data before it is decoded.
//

int Pad_DataRep::ReadData(istream &is, unsigned int type, unsigned int dataLength,
			  unsigned int &infoLength, unsigned char *&info)
{
  // compute size of expected data
  int bytesRead=0;
  infoLength = dataLength * DatumLength(type & ~_vector);
  info = new unsigned char[infoLength];
  is.read(info, infoLength);
  bytesRead = infoLength;
  if (is.good())
    return bytesRead;
  else 
    return itemErr;
}

//
// Methods for getting bytes legnth for decoded basic data types
//
int 
Pad_DataRep::DatumLength(int dtype)
{
  int rc = 0;

  switch (dtype & ~_vector) {
  case _char:
    rc = extend(sizeof(char));
    break;
  case _short:
    rc = extend(sizeof(short));
    break;
  case _int:
    rc = extend(sizeof(int));
    break;
  case _long:
    rc = extend(sizeof(long));
    break;
  case _unsigned_char:
    rc = extend(sizeof(unsigned char));
    break;
  case _unsigned_short:
    rc = extend(sizeof(unsigned short));
    break;
  case _unsigned_int:
    rc = extend(sizeof(unsigned int));
    break;
  case _unsigned_long:
    rc = extend(sizeof(unsigned long));
    break;
  case _float:
    rc = extend(sizeof(float));
    break;
  case _double:
    rc = extend(sizeof(double));
    break;
  default:
    break;
  }

  return rc;
}


//
// Silly method for reading ascii data before it is decoded.
//
int Pad_CharDataRep::ReadData(istream &is, unsigned int dtype, unsigned int dataLength,
			  unsigned int &infoLength, unsigned char *&info)
{
    unsigned int i;
    ostrstream ostr;
    int ival;
    unsigned int uival;
    short sval;
    unsigned short usval;
    long lval;
    unsigned long ulval;
    float fval;
    double dval;
    char cval;
    unsigned char ucval;

    unsigned int datumType = (dtype & ~_vector);

    if (datumType == _char || datumType == _unsigned_char) {
	is.ignore(1);
    }

    for (i=0; i < dataLength; i++) {
	switch (datumType) {
	case _char:
	    is.read(&cval, sizeof(char));
	    ostr << cval;
	    break;
	case _short:
	    is >> sval;
	    ostr << sval << " ";
	    break;
	case _int:
	    is >> ival;
	    ostr << ival << " ";
	    break;
	case _long:
	    is >> lval;
	    ostr << lval << " ";
	    break;
	case _unsigned_char:
	    is.read(&ucval, sizeof(unsigned char));
	    ostr << ucval << " ";
	    break;
	case _unsigned_short:
	    is >> usval;
	    ostr << usval << " ";
	    break;
	case _unsigned_int:
	    is >> uival;
	    ostr << uival << " ";
	    break;
	case _unsigned_long:
	    is >> ulval;
	    ostr << ulval << " ";
	    break;
	case _float:
	    is >> fval;
	    ostr << fval << " ";
	    break;
	case _double:
	    is >> dval;
	    ostr << dval << " ";
	    break;
	default:
	    break;
	}
    }

    // get the string buffer....
    infoLength = ostr.pcount();
    info = (unsigned char*) ostr.str();
    int bytesRead = ostr.pcount();  // hack, assume a 1-1 mapping from input
    if (datumType == _char && (dataLength > 1)) bytesRead++; // for the newline
    if (is.good())
	return bytesRead;
    else
	return itemErr;
}

int Pad_NativeDataRep::Type()
{
  return PAD_NATIVE_FORMAT;
}

int Pad_NativeDataRep::EncodeString(const char* input, unsigned char*& output,
				    unsigned int& length) {
  length = strlen(input)+1;
  output = new unsigned char[length];
  memcpy(output, input, length);
  return 0;
}

int Pad_NativeDataRep::EncodeArray(const short* input, unsigned int inputLength,
				   unsigned char*& output, unsigned int& outputLength) {
  outputLength = inputLength*sizeof(short);
  output = new unsigned char[outputLength];
  memcpy(output, input, outputLength);
  return 0;
}


int Pad_NativeDataRep::EncodeArray(const int* input, unsigned int inputLength,
				   unsigned char*& output, unsigned int& outputLength) {
  outputLength = inputLength*sizeof(int);
  output = new unsigned char[outputLength];
  memcpy(output, input, outputLength);
  return 0;
}

int Pad_NativeDataRep::EncodeArray(const long* input, unsigned int inputLength,
				   unsigned char*& output, unsigned int& outputLength) {
  outputLength = inputLength*sizeof(long);
  output = new unsigned char[outputLength];
  memcpy(output, input, outputLength);
  return 0;
}

int Pad_NativeDataRep::EncodeArray(const unsigned char* input, unsigned int inputLength,
				   unsigned char*& output, unsigned int& outputLength){
  outputLength = inputLength*sizeof(unsigned char);
  output = new unsigned char[outputLength];
  memcpy(output, input, outputLength);
  return 0;
}

int Pad_NativeDataRep::EncodeArray(const unsigned short* input, unsigned int inputLength,
				   unsigned char*& output, unsigned int& outputLength) {
  outputLength = inputLength*sizeof(unsigned short);
  output = new unsigned char[outputLength];
  memcpy(output, input, outputLength);
  return 0;
}

int Pad_NativeDataRep::EncodeArray(const unsigned int* input, unsigned int inputLength,
				   unsigned char*& output, unsigned int& outputLength) {
  outputLength = inputLength*sizeof(unsigned int);
  output = new unsigned char[outputLength];
  memcpy(output, input, outputLength);
  return 0;
}

int Pad_NativeDataRep::EncodeArray(const unsigned long* input, unsigned int inputLength,
				   unsigned char*& output, unsigned int& outputLength) {
  outputLength = inputLength*sizeof(unsigned long);
  output = new unsigned char[outputLength];
  memcpy(output, input, outputLength);
  return 0;
}

int Pad_NativeDataRep::EncodeArray(const float* input, unsigned int inputLength,
				   unsigned char*& output, unsigned int& outputLength) {
  outputLength = inputLength*sizeof(float);
  output = new unsigned char[outputLength];
  memcpy(output, input, outputLength);
  return 0;
}

int Pad_NativeDataRep::EncodeArray(const double* input, unsigned int inputLength,
			       unsigned char*& output, unsigned int& outputLength) {
  outputLength = inputLength*sizeof(double);
  output = new unsigned char[outputLength];
  memcpy(output, input, outputLength);
  return 0;
}
  
// Decoding routines
int Pad_NativeDataRep::DecodeString(const unsigned char* input, unsigned int ,
				char* output, unsigned int outputLength) {
  memcpy(output, input, outputLength*sizeof(char));
  return 0;
}

int Pad_NativeDataRep::DecodeArray(const unsigned char* input, unsigned int ,
			       short* output, unsigned int outputLength) {
  memcpy(output, input, outputLength*sizeof(short));
  return 0;
}

int Pad_NativeDataRep::DecodeArray(const unsigned char* input, unsigned int ,
			  int* output, unsigned int outputLength) {
  memcpy(output, input, outputLength*sizeof(int));
  return 0;
}

int Pad_NativeDataRep::DecodeArray(const unsigned char* input, unsigned int ,
			  long* output, unsigned int outputLength) {
  memcpy(output, input, outputLength*sizeof(long));
  return 0;
}

int Pad_NativeDataRep::DecodeArray(const unsigned char* input, unsigned int ,
			       unsigned char* output, unsigned int outputLength) {
  memcpy(output, input, outputLength*sizeof(unsigned char));
  return 0;
}

int Pad_NativeDataRep::DecodeArray(const unsigned char* input, unsigned int ,
			       unsigned short* output, unsigned int outputLength) {
  memcpy(output, input, outputLength*sizeof(unsigned short));
  return 0;
}

int Pad_NativeDataRep::DecodeArray(const unsigned char* input, unsigned int ,
			       unsigned int* output, unsigned int outputLength) {
  memcpy(output, input, outputLength*sizeof(unsigned int));
  return 0;
}

int Pad_NativeDataRep::DecodeArray(const unsigned char* input, unsigned int ,
			       unsigned long* output, unsigned int outputLength) {
  memcpy(output, input, outputLength*sizeof(unsigned long));
  return 0;
}

int Pad_NativeDataRep::DecodeArray(const unsigned char* input, unsigned int ,
			       float* output, unsigned int outputLength) {
  memcpy(output, input, outputLength*sizeof(float));
  return 0;
}

int Pad_NativeDataRep::DecodeArray(const unsigned char* input, unsigned int ,
			       double* output, unsigned int outputLength) {
  memcpy(output, input, outputLength*sizeof(double));
  return 0;
}

// Encoding routines for flattening
    
int Pad_NativeDataRep::flatEncodeString(const char* input, unsigned char* output,
				    unsigned int length) {
  return EncodeString(input, output, length);
}

int Pad_NativeDataRep::flatEncodeArray(const short* input, unsigned int inputLength,
				   unsigned char* output, unsigned int outputLength) {
  return EncodeArray(input, inputLength, output, outputLength);
}

int Pad_NativeDataRep::flatEncodeArray(const int* input, unsigned int inputLength,
				   unsigned char* output, unsigned int outputLength) {
  return EncodeArray(input, inputLength, output, outputLength);
}

int Pad_NativeDataRep::flatEncodeArray(const long* input, unsigned int inputLength,
				   unsigned char* output, unsigned int outputLength) {
  return EncodeArray(input, inputLength, output, outputLength);
}

int Pad_NativeDataRep::flatEncodeArray(const unsigned char* input, unsigned int inputLength,
				   unsigned char* output, unsigned int outputLength) {
  return EncodeArray(input, inputLength, output, outputLength);
}

int Pad_NativeDataRep::flatEncodeArray(const unsigned short* input, unsigned int inputLength,
			      unsigned char* output, unsigned int outputLength) {
  return EncodeArray(input, inputLength, output, outputLength);
}

int Pad_NativeDataRep::flatEncodeArray(const unsigned int* input, unsigned int inputLength,
				   unsigned char* output, unsigned int outputLength) {
  return EncodeArray(input, inputLength, output, outputLength);
}

int Pad_NativeDataRep::flatEncodeArray(const unsigned long* input, unsigned int inputLength,
				   unsigned char* output, unsigned int outputLength) {
  return EncodeArray(input, inputLength, output, outputLength);
}

int Pad_NativeDataRep::flatEncodeArray(const float* input, unsigned int inputLength,
			      unsigned char* output, unsigned int outputLength) {
  return EncodeArray(input, inputLength, output, outputLength);
}

int Pad_NativeDataRep::flatEncodeArray(const double* input, unsigned int inputLength,
			      unsigned char* output, unsigned int outputLength) {
  return EncodeArray(input, inputLength, output, outputLength);
}

int Pad_NativeDataRep::extend(int length) {
  return length;
}

Pad_NativeDataRep::Pad_NativeDataRep() {
  BlockSize=0;
}


//
// char-based rep
//

int Pad_CharDataRep::Type()
{
  return PAD_CHAR_FORMAT;
}

int Pad_CharDataRep::EncodeString(const char* input, unsigned char*& output,
				unsigned int& length) {
  length = strlen(input)+1;
  output = new unsigned char[length];
  memcpy(output, input, length);
  return 0;
}

int Pad_CharDataRep::EncodeArray(const short* input, unsigned int inputLength,
			       unsigned char*& output, unsigned int& outputLength) {
  ostrstream str;
  if (inputLength) str << input[0];
  for(unsigned int i=1; i < inputLength; i++) {
    str << ' ' << input[i];
  }
  outputLength = str.pcount();
  output = (unsigned char*)str.str();
  return 0;
}


int Pad_CharDataRep::EncodeArray(const int* input, unsigned int inputLength,
			       unsigned char*& output, unsigned int& outputLength) {
  ostrstream str;
  if (inputLength) str << input[0];
  for(unsigned int i=1; i < inputLength; i++) {
    str << ' ' << input[i];
  }
  outputLength = str.pcount();
  output = (unsigned char*)str.str();
  return 0;
}

int Pad_CharDataRep::EncodeArray(const long* input, unsigned int inputLength,
			       unsigned char*& output, unsigned int& outputLength) {
  ostrstream str;
  if (inputLength) str << input[0];
  for(unsigned int i=1; i < inputLength; i++) {
    str << ' ' << input[i];
  }
  outputLength = str.pcount();
  output = (unsigned char*)str.str();
  return 0;
}

int Pad_CharDataRep::EncodeArray(const unsigned char* input, unsigned int inputLength,
			       unsigned char*& output, unsigned int& outputLength){
  outputLength = inputLength*sizeof(unsigned char);
  output = new unsigned char[outputLength];
  memcpy(output, input, outputLength);
  return 0;
}

int Pad_CharDataRep::EncodeArray(const unsigned short* input, unsigned int inputLength,
			       unsigned char*& output, unsigned int& outputLength) {
  ostrstream str;
  if (inputLength) str << input[0];
  for(unsigned int i=1; i < inputLength; i++) {
    str << ' ' << input[i];
  }
  outputLength = str.pcount();
  output = (unsigned char*)str.str();
  return 0;
}

int Pad_CharDataRep::EncodeArray(const unsigned int* input, unsigned int inputLength,
			       unsigned char*& output, unsigned int& outputLength) {
  ostrstream str;
  if (inputLength) str << input[0];
  for(unsigned int i=1; i < inputLength; i++) {
    str << ' ' << input[i];
  }
  outputLength = str.pcount();
  output = (unsigned char*)str.str();
  return 0;
}

int Pad_CharDataRep::EncodeArray(const unsigned long* input, unsigned int inputLength,
			       unsigned char*& output, unsigned int& outputLength) {
  ostrstream str;
  if (inputLength) str << input[0];
  for(unsigned int i=1; i < inputLength; i++) {
    str << ' ' << input[i];
  }
  outputLength = str.pcount();
  output = (unsigned char*)str.str();
  return 0;
}

int Pad_CharDataRep::EncodeArray(const float* input, unsigned int inputLength,
			       unsigned char*& output, unsigned int& outputLength) {
  ostrstream str;
  if (inputLength) str << input[0];
  for(unsigned int i=1; i < inputLength; i++) {
    str << ' ' << input[i];
  }
  outputLength = str.pcount();
  output = (unsigned char*)str.str();
  return 0;
}

int Pad_CharDataRep::EncodeArray(const double* input, unsigned int inputLength,
			       unsigned char*& output, unsigned int& outputLength) {
  ostrstream str;
  if (inputLength) str << input[0];
  for(unsigned int i=1; i < inputLength; i++) {
    str << ' ' << input[i];
  }
  outputLength = str.pcount();
  output = (unsigned char*)str.str();
  return 0;
}
  
// Decoding routines
int Pad_CharDataRep::DecodeString(const unsigned char* input, unsigned int ,
				char* output, unsigned int outputLength) {
  memcpy(output, input, outputLength*sizeof(char));
  return 0;
}

int Pad_CharDataRep::DecodeArray(const unsigned char* input, unsigned int inputLength,
			       short* output, unsigned int outputLength) {
  istrstream str((char*)input, inputLength);
  for(unsigned int i=0; i < outputLength; i++) {
    str >> output[i];
  }
  str.rdbuf()->freeze(1);  // to avoid deallocation of input buffer
  return 0;
}

int Pad_CharDataRep::DecodeArray(const unsigned char* input, unsigned int inputLength,
				 int* output, unsigned int outputLength) {
  istrstream str((char*)input, inputLength);
  for(unsigned int i=0; i < outputLength; i++) {
    str >> output[i];
  }
  return 0;
}

int Pad_CharDataRep::DecodeArray(const unsigned char* input, unsigned int inputLength,
				 long* output, unsigned int outputLength) {
  istrstream str((char*)input, inputLength);
  for(unsigned int i=0; i < outputLength; i++) {
    str >> output[i];
  }
  str.rdbuf()->freeze(1);  // to avoid deallocation of input buffer
  return 0;
}

int Pad_CharDataRep::DecodeArray(const unsigned char* input, unsigned int ,
				 unsigned char* output, unsigned int outputLength) {
  memcpy(output, input, outputLength*sizeof(unsigned char));
  return 0;
}

int Pad_CharDataRep::DecodeArray(const unsigned char* input, unsigned int inputLength,
				 unsigned short* output, unsigned int outputLength) {
  istrstream str((char*)input, inputLength);
  for(unsigned int i=0; i < outputLength; i++) {
    str >> output[i];
  }
  str.rdbuf()->freeze(1);  // to avoid deallocation of input buffer
  return 0;
}

int Pad_CharDataRep::DecodeArray(const unsigned char* input, unsigned int inputLength,
				 unsigned int* output, unsigned int outputLength) {
  istrstream str((char*)input, inputLength);
  for(unsigned int i=0; i < outputLength; i++) {
    str >> output[i];
  }
  str.rdbuf()->freeze(1);  // to avoid deallocation of input buffer
  return 0;
}

int Pad_CharDataRep::DecodeArray(const unsigned char* input, unsigned int inputLength,
				 unsigned long* output, unsigned int outputLength) {
  istrstream str((char*)input, inputLength);
  for(unsigned int i=0; i < outputLength; i++) {
    str >> output[i];
  }
  str.rdbuf()->freeze(1);  // to avoid deallocation of input buffer
  return 0;
}

int Pad_CharDataRep::DecodeArray(const unsigned char* input, unsigned int inputLength,
			       float* output, unsigned int outputLength) {
  istrstream str((char*)input, inputLength);
  for(unsigned int i=0; i < outputLength; i++) {
    str >> output[i];
  }
  str.rdbuf()->freeze(1);  // to avoid deallocation of input buffer
  return 0;
}

int Pad_CharDataRep::DecodeArray(const unsigned char* input, unsigned int inputLength,
			       double* output, unsigned int outputLength) {
  istrstream str((char*)input, inputLength);
  for(unsigned int i=0; i < outputLength; i++) {
    str >> output[i];
  }
  str.rdbuf()->freeze(1);  // to avoid deallocation of input buffer
  return 0;
}

// Encoding routines for flattening
    
int Pad_CharDataRep::flatEncodeString(const char* input, unsigned char* output,
				    unsigned int length) {
  return EncodeString(input, output, length);
}

int Pad_CharDataRep::flatEncodeArray(const short* input, unsigned int inputLength,
				   unsigned char* output, unsigned int outputLength) {
  return EncodeArray(input, inputLength, output, outputLength);
}

int Pad_CharDataRep::flatEncodeArray(const int* input, unsigned int inputLength,
				   unsigned char* output, unsigned int outputLength) {
  return EncodeArray(input, inputLength, output, outputLength);
}

int Pad_CharDataRep::flatEncodeArray(const long* input, unsigned int inputLength,
				   unsigned char* output, unsigned int outputLength) {
  return EncodeArray(input, inputLength, output, outputLength);
}

int Pad_CharDataRep::flatEncodeArray(const unsigned char* input, unsigned int inputLength,
				   unsigned char* output, unsigned int outputLength) {
  return EncodeArray(input, inputLength, output, outputLength);
}

int Pad_CharDataRep::flatEncodeArray(const unsigned short* input, unsigned int inputLength,
			      unsigned char* output, unsigned int outputLength) {
  return EncodeArray(input, inputLength, output, outputLength);
}

int Pad_CharDataRep::flatEncodeArray(const unsigned int* input, unsigned int inputLength,
				   unsigned char* output, unsigned int outputLength) {
  return EncodeArray(input, inputLength, output, outputLength);
}

int Pad_CharDataRep::flatEncodeArray(const unsigned long* input, unsigned int inputLength,
				   unsigned char* output, unsigned int outputLength) {
  return EncodeArray(input, inputLength, output, outputLength);
}

int Pad_CharDataRep::flatEncodeArray(const float* input, unsigned int inputLength,
			      unsigned char* output, unsigned int outputLength) {
  return EncodeArray(input, inputLength, output, outputLength);
}

int Pad_CharDataRep::flatEncodeArray(const double* input, unsigned int inputLength,
			      unsigned char* output, unsigned int outputLength) {
  return EncodeArray(input, inputLength, output, outputLength);
}

int Pad_CharDataRep::extend(int length) {
  return length;
}

Pad_CharDataRep::Pad_CharDataRep() {
  BlockSize=0;
}
