/*
"(c) Copyright 1993-1997 Pad++ Consortium {University of New Mexico (UNM),
and New York University (NYU)}, All Rights Reserved."  
Licensee can not remove or obscure any of the
copyright or trademark notices in this software.

IN NO EVENT SHALL THE AUTHORS OR DISTRIBUTORS BE LIABLE TO ANY PARTY
FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
ARISING OUT OF THE USE OF THIS SOFTWARE, ITS DOCUMENTATION, OR ANY
DERIVATIVES THEREOF, EVEN IF THE AUTHORS HAVE BEEN ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.

THE AUTHORS AND DISTRIBUTORS SPECIFICALLY DISCLAIM ANY WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE, AND NON-INFRINGEMENT.  THIS SOFTWARE
IS PROVIDED ON AN "AS IS" BASIS, AND THE AUTHORS AND DISTRIBUTORS HAVE
NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR
MODIFICATIONS.

See the file "License" for general information on usage and
redistribution, and the file "LicenseTerms" for the specific license
agreement on usage and redistribution of this file, and the Pad++
software in general.
*/

//
// Classes for basic marshaling and reconstruction (encoding & decoding) 
// of data items.
//

#ifndef streamdata_h
#define streamdata_h

#include <iostream.h>
#include "defs.h"

// symbols and values for format of pad data files
#define PAD_SCRIPT_FORMAT  0
#define PAD_XDR_FORMAT     1
#define PAD_NATIVE_FORMAT  2
#define PAD_CHAR_FORMAT    3

class Pad_DataRep;
class Pad_StreamItem;

// Common data types

// Error codes
const int noErr=0;
const int xdrErr=-1024;
const int typeErr=-1025;
const int itemErr=-1026;
const int linkErr=-1027;
const int lengthErr=-1028;
const int corruptErr=-1030;
const int objectErr=-1031;

// Data item type codes
typedef unsigned int BaseType;
const          int _no_item=-1;
const unsigned int _no_type=0;
const unsigned int _vector=(1<<0);
const unsigned int _char=(1<<1);
const unsigned int _short=(1<<2);
const unsigned int _int=(1<<3);
const unsigned int _long=(1<<4);
const unsigned int _unsigned_char=(1<<5);
const unsigned int _unsigned_short=(1<<6);
const unsigned int _unsigned_int=(1<<7);
const unsigned int _unsigned_long=(1<<8);
const unsigned int _float=(1<<9);
const unsigned int _double=(1<<10);

//
// The Pad_StreamData class has methods for marshaling and reconstructing data 
// items using particular data representations.
//

class Pad_StreamData {
public:
        // Constructor for specifying which data rep to use,
	// anticipated number of items that may be put on the stream,
        // and io streams that may be used to fill or flush the stream.

  Pad_StreamData (int rep=PAD_XDR_FORMAT, unsigned int initialSize=0, 
		istream *inputfp=NULL, ostream *outputfp=NULL);

        // Constructor from another data stream.  This will copy the
	// other items.

  Pad_StreamData(const Pad_StreamData&);

        // Deallocate all memory used for data items.

  virtual ~Pad_StreamData();

        // Assignment operator for copying anther stream.

  Pad_StreamData& operator= (const Pad_StreamData&);


        // Data representation to use.  This is only allowed if no data 
        // items have been placed on the stream.

  int UseRep(int rep=PAD_XDR_FORMAT);

        // Deallocate all data items in the stream

  void Clear();

        // Set the item counter to the first item in the stream

  void Reset();

        // Move the item counter.

  void Skip(unsigned int stride=1);

        // Get the number of data items in the stream.

  unsigned int Items() const;

        // Get the next data item type.

  int NextType() const;

        // Get the next data item's length.

  int NextLength();

        // Read data stream from the input file.

  int Fill();

        // Write stream to the input file.

  int Flush();

        // Flattening and expansion methods for data items in the stream.

  int Flatten(unsigned char*& output, unsigned int& length);
  int Expand(unsigned char* input, unsigned int length,
	     unsigned int& bytes_read);

        // Misc methods for dealing with data rep, io streams and output length
  int Get_rep();
  istream *Get_istream();
  ostream *Get_ostream();
  int ItemsLength() const;
  int OutputLength();
  int DatumLength(int);
  int SkipInput(int);
  int MoveItems(Pad_StreamData&);

        // Data encoding and decoding methods

  int Marshal(const char* input);
  int Marshal(const short* input, unsigned int length);
  int Marshal(const int* input, unsigned int length);
  int Marshal(const long* input, unsigned int length);
  int Marshal(const unsigned char* input, unsigned int length);
  int Marshal(const unsigned short* input, unsigned int length);
  int Marshal(const unsigned int* input, unsigned int length);
  int Marshal(const unsigned long* input, unsigned int length);
  int Marshal(const float* input, unsigned int length);
  int Marshal(const double* input, unsigned int length);

  int Reconstruct(char* output, unsigned int length);
  int Reconstruct(short* output, unsigned int length);
  int Reconstruct(int* output, unsigned int length);
  int Reconstruct(long* output, unsigned int length);
  int Reconstruct(unsigned char* output, unsigned int length);
  int Reconstruct(unsigned short* output, unsigned int length);
  int Reconstruct(unsigned int* output, unsigned int length);
  int Reconstruct(unsigned long* output, unsigned int length);
  int Reconstruct(float* output, unsigned int length);
  int Reconstruct(double* output, unsigned int length);

         // Convenience encoding and decoding methods and operators

  int Marshal(const short& input);
  int Marshal(const int& input);
  int Marshal(const long& input);
  int Marshal(const unsigned char& input);
  int Marshal(const unsigned short& input);
  int Marshal(const unsigned int& input);
  int Marshal(const unsigned long& input);
  int Marshal(const float& input);
  int Marshal(const double& input);

  Pad_StreamData& operator<< (const char*);
  Pad_StreamData& operator<< (const short&);
  Pad_StreamData& operator<< (const int&);
  Pad_StreamData& operator<< (const long&);
  Pad_StreamData& operator<< (const unsigned char&);
  Pad_StreamData& operator<< (const unsigned short&);
  Pad_StreamData& operator<< (const unsigned int&);
  Pad_StreamData& operator<< (const unsigned long&);
  Pad_StreamData& operator<< (const float&);
  Pad_StreamData& operator<< (const double&);

  int Reconstruct(short& output);
  int Reconstruct(int& output);
  int Reconstruct(long& output);
  int Reconstruct(unsigned char& output);
  int Reconstruct(unsigned short& output);
  int Reconstruct(unsigned int& output);
  int Reconstruct(unsigned long& output);
  int Reconstruct(float& output);
  int Reconstruct(double& output);

  Pad_StreamData& operator>> (short&);
  Pad_StreamData& operator>> (int&);
  Pad_StreamData& operator>> (long&);
  Pad_StreamData& operator>> (unsigned char&);
  Pad_StreamData& operator>> (unsigned short&);
  Pad_StreamData& operator>> (unsigned int&);
  Pad_StreamData& operator>> (unsigned long&);
  Pad_StreamData& operator>> (float&);
  Pad_StreamData& operator>> (double&);

         // Convenience decoding methods that allocate output buffer

  int Reconstruct(char*& output, unsigned int *len=0);
  int Reconstruct(short*& output, unsigned int *len=0);
  int Reconstruct(int*& output, unsigned int *len=0);
  int Reconstruct(long*& output, unsigned int *len=0);
  int Reconstruct(unsigned char*& output, unsigned int *len=0);
  int Reconstruct(unsigned short*& output, unsigned int *len=0);
  int Reconstruct(unsigned int*& output, unsigned int *len=0);
  int Reconstruct(unsigned long*& output, unsigned int *len=0);
  int Reconstruct(float*& output, unsigned int *len=0);
  int Reconstruct(double*& output, unsigned int *len=0);

  Pad_StreamData& operator>> (char*&);
  Pad_StreamData& operator>> (short*&);
  Pad_StreamData& operator>> (int*&);
  Pad_StreamData& operator>> (long*&);
  Pad_StreamData& operator>> (unsigned char*&);
  Pad_StreamData& operator>> (unsigned short*&);
  Pad_StreamData& operator>> (unsigned int*&);
  Pad_StreamData& operator>> (unsigned long*&);
  Pad_StreamData& operator>> (float*&);
  Pad_StreamData& operator>> (double*&);


         // Input and output stream operators for reading and writing
         // data items from files.

  friend ostream& operator<< (ostream& os, Pad_StreamData& m);
  friend istream& operator>> (istream& is, Pad_StreamData& m);
  friend class Pad_StreamItem;

private:

  // Fundamental marshaling methods
  void marshal(Pad_StreamItem* mi);
  int reconstruct(Pad_StreamItem *&mi);

  // _full methods
  int marshal_full(Pad_StreamData&) const;
  int reconstruct_full(Pad_StreamData&);

  int     xrep;
  Pad_DataRep *drep;
  unsigned int items;
  unsigned int currentItem;
  unsigned int size;
  unsigned int totalBytes;
  unsigned int bytesRead;
  Pad_StreamItem** data;
  istream *ifp;
  ostream *ofp;
};

//
// Class for data items representation.
//

class Pad_StreamItem {
public:
  Pad_StreamItem();
  Pad_StreamItem(const Pad_StreamItem&);
  Pad_StreamItem& operator= (const Pad_StreamItem&);
  virtual ~Pad_StreamItem();

  Pad_Bool Null();
  int marshal(Pad_StreamData&) const;
  int reconstruct(Pad_StreamData&);

  ostream& Write(ostream &os, Pad_StreamData *sd);
  istream& Read(istream &is, Pad_StreamData *sd);

  int OutputLength(Pad_DataRep*);

protected:

  // _full methods
  virtual int marshal_full(Pad_StreamData&) const;
  virtual int reconstruct_full(Pad_StreamData&);

public:
  BaseType type;
  unsigned int dataLength;
  unsigned int infoLength;
  unsigned char* info;
};


//
// Abstract base class for wrappers around underlying system 
// calls for encoding and decoding of data in some format (e.g. xdr,
// native, etc).  These will be used by Pad_StreamData objects
// for encoding and decoding of data items.
//

class Pad_DataRep {
public:
  int BlockSize;

  // Encoding routines
  virtual int EncodeString(const char* input, unsigned char*& output,
			   unsigned int& length)=0;
  virtual int EncodeArray(const short* input, unsigned int inputLength,
			  unsigned char*& output, unsigned int& outputLength)=0;
  virtual int EncodeArray(const int* input, unsigned int inputLength,
			  unsigned char*& output, unsigned int& outputLength)=0;
  virtual int EncodeArray(const long* input, unsigned int inputLength,
			  unsigned char*& output, unsigned int& outputLength)=0;
  virtual int EncodeArray(const unsigned char* input, unsigned int inputLength,
			  unsigned char*& output, unsigned int& outputLength)=0;
  virtual int EncodeArray(const unsigned short* input, unsigned int inputLength,
			  unsigned char*& output, unsigned int& outputLength)=0;
  virtual int EncodeArray(const unsigned int* input, unsigned int inputLength,
			  unsigned char*& output, unsigned int& outputLength)=0;
  virtual int EncodeArray(const unsigned long* input, unsigned int inputLength,
			  unsigned char*& output, unsigned int& outputLength)=0;
  virtual int EncodeArray(const float* input, unsigned int inputLength,
			  unsigned char*& output, unsigned int& outputLength)=0;
  virtual int EncodeArray(const double* input, unsigned int inputLength,
			  unsigned char*& output, unsigned int& outputLength)=0;
  
  
  // Decoding routines
  virtual int DecodeString(const unsigned char* input, unsigned int inputLength,
			   char* output, unsigned int outputLength)=0;
  virtual int DecodeArray(const unsigned char* input, unsigned int inputLength,
			  short* output, unsigned int outputLength)=0;
  virtual int DecodeArray(const unsigned char* input, unsigned int inputLength,
			  int* output, unsigned int outputLength)=0;
  virtual int DecodeArray(const unsigned char* input, unsigned int inputLength,
			  long* output, unsigned int outputLength)=0;
  virtual int DecodeArray(const unsigned char* input, unsigned int inputLength,
			  unsigned char* output, unsigned int outputLength)=0;
  virtual int DecodeArray(const unsigned char* input, unsigned int inputLength,
			  unsigned short* output, unsigned int outputLength)=0;
  virtual int DecodeArray(const unsigned char* input, unsigned int inputLength,
			  unsigned int* output, unsigned int outputLength)=0;
  virtual int DecodeArray(const unsigned char* input, unsigned int inputLength,
			  unsigned long* output, unsigned int outputLength)=0;
  virtual int DecodeArray(const unsigned char* input, unsigned int inputLength,
			  float* output, unsigned int outputLength)=0;
  virtual int DecodeArray(const unsigned char* input, unsigned int inputLength,
			  double* output, unsigned int outputLength)=0;
  
  // Encoding routines for flattening
    
  virtual int flatEncodeString(const char* input, unsigned char* output,
			       unsigned int length)=0;
  virtual int flatEncodeArray(const short* input, unsigned int inputLength,
			      unsigned char* output, unsigned int outputLength)=0;
  virtual int flatEncodeArray(const int* input, unsigned int inputLength,
			      unsigned char* output, unsigned int outputLength)=0;
  virtual int flatEncodeArray(const long* input, unsigned int inputLength,
			      unsigned char* output, unsigned int outputLength)=0;
  virtual int flatEncodeArray(const unsigned char* input, unsigned int inputLength,
			      unsigned char* output, unsigned int outputLength)=0;
  virtual int flatEncodeArray(const unsigned short* input, unsigned int inputLength,
			      unsigned char* output, unsigned int outputLength)=0;
  virtual int flatEncodeArray(const unsigned int* input, unsigned int inputLength,
			      unsigned char* output, unsigned int outputLength)=0;
  virtual int flatEncodeArray(const unsigned long* input, unsigned int inputLength,
			      unsigned char* output, unsigned int outputLength)=0;
  virtual int flatEncodeArray(const float* input, unsigned int inputLength,
			      unsigned char* output, unsigned int outputLength)=0;
  virtual int flatEncodeArray(const double* input, unsigned int inputLength,
			      unsigned char* output, unsigned int outputLength)=0;

 virtual int extend(int length)=0;

 virtual int ReadData(istream &is, unsigned int type, unsigned int dataLength,
                      unsigned int &infoLength, unsigned char *&info);
 virtual int DatumLength(int dtype);

 virtual int Type()=0;

protected:
  Pad_DataRep();
};


//
// A trivial class for streaming data in native machine binary format.
// The Pad_StreamData object could use a data rep. object of this type.
//

class Pad_NativeDataRep : public Pad_DataRep {
public:
  Pad_NativeDataRep();

  // Encoding routines
  virtual int EncodeString(const char* input, unsigned char*& output,
			   unsigned int& length);
  virtual int EncodeArray(const short* input, unsigned int inputLength,
			  unsigned char*& output, unsigned int& outputLength);
  virtual int EncodeArray(const int* input, unsigned int inputLength,
			  unsigned char*& output, unsigned int& outputLength);
  virtual int EncodeArray(const long* input, unsigned int inputLength,
			  unsigned char*& output, unsigned int& outputLength);
  virtual int EncodeArray(const unsigned char* input, unsigned int inputLength,
			  unsigned char*& output, unsigned int& outputLength);
  virtual int EncodeArray(const unsigned short* input, unsigned int inputLength,
			  unsigned char*& output, unsigned int& outputLength);
  virtual int EncodeArray(const unsigned int* input, unsigned int inputLength,
			  unsigned char*& output, unsigned int& outputLength);
  virtual int EncodeArray(const unsigned long* input, unsigned int inputLength,
			  unsigned char*& output, unsigned int& outputLength);
  virtual int EncodeArray(const float* input, unsigned int inputLength,
			  unsigned char*& output, unsigned int& outputLength);
  virtual int EncodeArray(const double* input, unsigned int inputLength,
			  unsigned char*& output, unsigned int& outputLength);
  
  // Decoding routines
  virtual int DecodeString(const unsigned char* input, unsigned int inputLength,
			   char* output, unsigned int outputLength);
  virtual int DecodeArray(const unsigned char* input, unsigned int inputLength,
			  short* output, unsigned int outputLength);
  virtual int DecodeArray(const unsigned char* input, unsigned int inputLength,
			  int* output, unsigned int outputLength);
  virtual int DecodeArray(const unsigned char* input, unsigned int inputLength,
			  long* output, unsigned int outputLength);
  virtual int DecodeArray(const unsigned char* input, unsigned int inputLength,
			  unsigned char* output, unsigned int outputLength);
  virtual int DecodeArray(const unsigned char* input, unsigned int inputLength,
			  unsigned short* output, unsigned int outputLength);
  virtual int DecodeArray(const unsigned char* input, unsigned int inputLength,
			  unsigned int* output, unsigned int outputLength);
  virtual int DecodeArray(const unsigned char* input, unsigned int inputLength,
			  unsigned long* output, unsigned int outputLength);
  virtual int DecodeArray(const unsigned char* input, unsigned int inputLength,
			  float* output, unsigned int outputLength);
  virtual int DecodeArray(const unsigned char* input, unsigned int inputLength,
			  double* output, unsigned int outputLength);
  
  // Encoding routines for flattening
    
  virtual int flatEncodeString(const char* input, unsigned char* output,
			       unsigned int length);
  virtual int flatEncodeArray(const short* input, unsigned int inputLength,
			      unsigned char* output, unsigned int outputLength);
  virtual int flatEncodeArray(const int* input, unsigned int inputLength,
			      unsigned char* output, unsigned int outputLength);
  virtual int flatEncodeArray(const long* input, unsigned int inputLength,
			      unsigned char* output, unsigned int outputLength);
  virtual int flatEncodeArray(const unsigned char* input, unsigned int inputLength,
			      unsigned char* output, unsigned int outputLength);
  virtual int flatEncodeArray(const unsigned short* input, unsigned int inputLength,
			      unsigned char* output, unsigned int outputLength);
  virtual int flatEncodeArray(const unsigned int* input, unsigned int inputLength,
			      unsigned char* output, unsigned int outputLength);
  virtual int flatEncodeArray(const unsigned long* input, unsigned int inputLength,
			      unsigned char* output, unsigned int outputLength);
  virtual int flatEncodeArray(const float* input, unsigned int inputLength,
			      unsigned char* output, unsigned int outputLength);
  virtual int flatEncodeArray(const double* input, unsigned int inputLength,
			      unsigned char* output, unsigned int outputLength);

 virtual int extend(int length);
 virtual int Type();
};

//
// A trivial class for streaming data in character format.
// The Pad_StreamData object could use a data rep. object of this type.
// This can be good for debugging purposes.
//

class Pad_CharDataRep : public Pad_DataRep {
public:
  Pad_CharDataRep();

  // Encoding routines
  virtual int EncodeString(const char* input, unsigned char*& output,
			   unsigned int& length);
  virtual int EncodeArray(const short* input, unsigned int inputLength,
			  unsigned char*& output, unsigned int& outputLength);
  virtual int EncodeArray(const int* input, unsigned int inputLength,
			  unsigned char*& output, unsigned int& outputLength);
  virtual int EncodeArray(const long* input, unsigned int inputLength,
			  unsigned char*& output, unsigned int& outputLength);
  virtual int EncodeArray(const unsigned char* input, unsigned int inputLength,
			  unsigned char*& output, unsigned int& outputLength);
  virtual int EncodeArray(const unsigned short* input, unsigned int inputLength,
			  unsigned char*& output, unsigned int& outputLength);
  virtual int EncodeArray(const unsigned int* input, unsigned int inputLength,
			  unsigned char*& output, unsigned int& outputLength);
  virtual int EncodeArray(const unsigned long* input, unsigned int inputLength,
			  unsigned char*& output, unsigned int& outputLength);
  virtual int EncodeArray(const float* input, unsigned int inputLength,
			  unsigned char*& output, unsigned int& outputLength);
  virtual int EncodeArray(const double* input, unsigned int inputLength,
			  unsigned char*& output, unsigned int& outputLength);
  
  // Decoding routines
  virtual int DecodeString(const unsigned char* input, unsigned int inputLength,
			   char* output, unsigned int outputLength);
  virtual int DecodeArray(const unsigned char* input, unsigned int inputLength,
			  short* output, unsigned int outputLength);
  virtual int DecodeArray(const unsigned char* input, unsigned int inputLength,
			  int* output, unsigned int outputLength);
  virtual int DecodeArray(const unsigned char* input, unsigned int inputLength,
			  long* output, unsigned int outputLength);
  virtual int DecodeArray(const unsigned char* input, unsigned int inputLength,
			  unsigned char* output, unsigned int outputLength);
  virtual int DecodeArray(const unsigned char* input, unsigned int inputLength,
			  unsigned short* output, unsigned int outputLength);
  virtual int DecodeArray(const unsigned char* input, unsigned int inputLength,
			  unsigned int* output, unsigned int outputLength);
  virtual int DecodeArray(const unsigned char* input, unsigned int inputLength,
			  unsigned long* output, unsigned int outputLength);
  virtual int DecodeArray(const unsigned char* input, unsigned int inputLength,
			  float* output, unsigned int outputLength);
  virtual int DecodeArray(const unsigned char* input, unsigned int inputLength,
			  double* output, unsigned int outputLength);
  
  // Encoding routines for flattening
    
  virtual int flatEncodeString(const char* input, unsigned char* output,
			       unsigned int length);
  virtual int flatEncodeArray(const short* input, unsigned int inputLength,
			      unsigned char* output, unsigned int outputLength);
  virtual int flatEncodeArray(const int* input, unsigned int inputLength,
			      unsigned char* output, unsigned int outputLength);
  virtual int flatEncodeArray(const long* input, unsigned int inputLength,
			      unsigned char* output, unsigned int outputLength);
  virtual int flatEncodeArray(const unsigned char* input, unsigned int inputLength,
			      unsigned char* output, unsigned int outputLength);
  virtual int flatEncodeArray(const unsigned short* input, unsigned int inputLength,
			      unsigned char* output, unsigned int outputLength);
  virtual int flatEncodeArray(const unsigned int* input, unsigned int inputLength,
			      unsigned char* output, unsigned int outputLength);
  virtual int flatEncodeArray(const unsigned long* input, unsigned int inputLength,
			      unsigned char* output, unsigned int outputLength);
  virtual int flatEncodeArray(const float* input, unsigned int inputLength,
			      unsigned char* output, unsigned int outputLength);
  virtual int flatEncodeArray(const double* input, unsigned int inputLength,
			      unsigned char* output, unsigned int outputLength);

 virtual int extend(int length);
 virtual int ReadData(istream &is, unsigned int type, unsigned int dataLength,
                      unsigned int &infoLength, unsigned char *&info);
 virtual int Type();

};

// ************************************************************************
// inline methods
// ************************************************************************

inline Pad_DataRep::Pad_DataRep() {}

inline int Pad_StreamData::Get_rep() {return xrep;}

inline Pad_Bool Pad_StreamItem::Null() {return !info;}

inline void Pad_StreamData::Reset() {currentItem=0;}

inline void Pad_StreamData::Skip(unsigned int stride) {currentItem+=stride;}

inline unsigned int Pad_StreamData::Items() const {return items;}

// ************************************************************************
// inline convenience methods
// ************************************************************************

inline int Pad_StreamData::Marshal(const short& input)
{
  return Marshal(&input, 1);
}

inline int Pad_StreamData::Marshal(const int& input)
{
  return Marshal(&input, 1);
}

inline int Pad_StreamData::Marshal(const long& input)
{
  return Marshal(&input, 1);
}

inline int Pad_StreamData::Marshal(const unsigned char& input)
{
  return Marshal(&input, 1);
}

inline int Pad_StreamData::Marshal(const unsigned short& input)
{
  return Marshal(&input, 1);
}

inline int Pad_StreamData::Marshal(const unsigned int& input)
{
  return Marshal(&input, 1);
}

inline int Pad_StreamData::Marshal(const unsigned long& input)
{
  return Marshal(&input, 1);
}

inline int Pad_StreamData::Marshal(const float& input)
{
  return Marshal(&input, 1);
}

inline int Pad_StreamData::Marshal(const double& input)
{
  return Marshal(&input, 1);
}

inline Pad_StreamData& Pad_StreamData::operator<< (const char* input)
{
  Marshal(input);
  return *this;
}

inline Pad_StreamData& Pad_StreamData::operator<< (const short& input)
{
  Marshal(&input, 1);
  return *this;
}

inline Pad_StreamData& Pad_StreamData::operator<< (const int& input)
{
  Marshal(&input, 1);
  return *this;
}

inline Pad_StreamData& Pad_StreamData::operator<< (const long& input)
{
  Marshal(&input, 1);
  return *this;
}

inline Pad_StreamData& Pad_StreamData::operator<< (const unsigned char& input)
{
  Marshal(&input, 1);
  return *this;
}

inline Pad_StreamData& Pad_StreamData::operator<< (const unsigned short& input)
{
  Marshal(&input, 1);
  return *this;
}

inline Pad_StreamData& Pad_StreamData::operator<< (const unsigned int& input)
{
  Marshal(&input, 1);
  return *this;
}

inline Pad_StreamData& Pad_StreamData::operator<< (const unsigned long& input)
{
  Marshal(&input, 1);
  return *this;
}

inline Pad_StreamData& Pad_StreamData::operator<< (const float& input)
{
  Marshal(&input, 1);
  return *this;
}

inline Pad_StreamData& Pad_StreamData::operator<< (const double& input)
{
  Marshal(&input, 1);
  return *this;
}

inline int Pad_StreamData::Reconstruct(short& output)
{
  return Reconstruct(&output, 1);
}

inline int Pad_StreamData::Reconstruct(int& output)
{
  return Reconstruct(&output, 1);
}

inline int Pad_StreamData::Reconstruct(long& output)
{
  return Reconstruct(&output, 1);
}

inline int Pad_StreamData::Reconstruct(unsigned char& output)
{
  return Reconstruct(&output, 1);
}

inline int Pad_StreamData::Reconstruct(unsigned short& output)
{
  return Reconstruct(&output, 1);
}

inline int Pad_StreamData::Reconstruct(unsigned int& output)
{
  return Reconstruct(&output, 1);
}

inline int Pad_StreamData::Reconstruct(unsigned long& output)
{
  return Reconstruct(&output, 1);
}

inline int Pad_StreamData::Reconstruct(float& output)
{
  return Reconstruct(&output, 1);
}

inline int Pad_StreamData::Reconstruct(double& output)
{
  return Reconstruct(&output, 1);
}

inline Pad_StreamData& Pad_StreamData::operator>> (short& output)
{
  Reconstruct(&output, 1);
  return *this;
}

inline Pad_StreamData& Pad_StreamData::operator>> (int& output)
{
  Reconstruct(&output, 1);
  return *this;
}

inline Pad_StreamData& Pad_StreamData::operator>> (long& output)
{
  Reconstruct(&output, 1);
  return *this;
}

inline Pad_StreamData& Pad_StreamData::operator>> (unsigned char& output)
{
  Reconstruct(&output, 1);
  return *this;
}

inline Pad_StreamData& Pad_StreamData::operator>> (unsigned short& output)
{
  Reconstruct(&output, 1);
  return *this;
}

inline Pad_StreamData& Pad_StreamData::operator>> (unsigned int& output)
{
  Reconstruct(&output, 1);
  return *this;
}

inline Pad_StreamData& Pad_StreamData::operator>> (unsigned long& output)
{
  Reconstruct(&output, 1);
  return *this;
}

inline Pad_StreamData& Pad_StreamData::operator>> (float& output)
{
  Reconstruct(&output, 1);
  return *this;
}

inline Pad_StreamData& Pad_StreamData::operator>> (double& output)
{
  Reconstruct(&output, 1);
  return *this;
}


// ************************************************************************
// inline convenience methods that allocate output buffer
// ************************************************************************

inline Pad_StreamData& Pad_StreamData::operator>> (char*& output)
{
  Reconstruct(output);
  return *this;
}

inline Pad_StreamData& Pad_StreamData::operator>> (short*& output)
{
  Reconstruct(output);
  return *this;
}

inline Pad_StreamData& Pad_StreamData::operator>> (int*& output)
{
  Reconstruct(output);
  return *this;
}

inline Pad_StreamData& Pad_StreamData::operator>> (long*& output)
{
  Reconstruct(output);
  return *this;
}

inline Pad_StreamData& Pad_StreamData::operator>> (unsigned char*& output)
{
  Reconstruct(output);
  return *this;
}

inline Pad_StreamData& Pad_StreamData::operator>> (unsigned short*& output)
{
  Reconstruct(output);
  return *this;
}

inline Pad_StreamData& Pad_StreamData::operator>> (unsigned int*& output)
{
  Reconstruct(output);
  return *this;
}

inline Pad_StreamData& Pad_StreamData::operator>> (unsigned long*& output)
{
  Reconstruct(output);
  return *this;
}

inline Pad_StreamData& Pad_StreamData::operator>> (float*& output)
{
  Reconstruct(output);
  return *this;
}

inline Pad_StreamData& Pad_StreamData::operator>> (double*& output)
{
  Reconstruct(output);
  return *this;
}


#endif

