/*
"(c) Copyright 1993-1997 Pad++ Consortium {University of New Mexico (UNM),
and New York University (NYU)}, All Rights Reserved."  
Licensee can not remove or obscure any of the
copyright or trademark notices in this software.

IN NO EVENT SHALL THE AUTHORS OR DISTRIBUTORS BE LIABLE TO ANY PARTY
FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
ARISING OUT OF THE USE OF THIS SOFTWARE, ITS DOCUMENTATION, OR ANY
DERIVATIVES THEREOF, EVEN IF THE AUTHORS HAVE BEEN ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.

THE AUTHORS AND DISTRIBUTORS SPECIFICALLY DISCLAIM ANY WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE, AND NON-INFRINGEMENT.  THIS SOFTWARE
IS PROVIDED ON AN "AS IS" BASIS, AND THE AUTHORS AND DISTRIBUTORS HAVE
NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR
MODIFICATIONS.

See the file "License" for general information on usage and
redistribution, and the file "LicenseTerms" for the specific license
agreement on usage and redistribution of this file, and the Pad++
software in general.
*/

#ifdef HAS_TIFF

#include "tiff.h"

//
// LIBTIFF Wrappers for loading tiff images.
//

#include <string.h>
#include <tiffio.h>

#ifdef PAD_UNIX
#  include <unistd.h>
#endif



// global vars used when reading tiff images from memory buffers
static unsigned char *tiff_databuf=NULL;
static tsize_t        tiff_databuf_size = 0;
static toff_t         tiff_databuf_curpos = 0;

// io hooks for libtiff to use when reading images from buffer
static tsize_t tiffReadProc (thandle_t clientdata, tdata_t data, tsize_t size);
static toff_t tiffSeekProc(thandle_t, toff_t offset, int whence);
static int tiffCloseProc(thandle_t);
static toff_t tiffSizeProc(thandle_t);
static int tiffMapFileProc(thandle_t, tdata_t* pbase, toff_t* psize);
static void tiffUnmapFileProc(thandle_t, tdata_t, toff_t);
static void tiffReadRGBAImage(TIFF *tif, unsigned long *ret_w, unsigned long *ret_h,
			 unsigned long **raster_buf, int stopOnError);
static void tiffErrorHandler(const char* module, const char* fmt, va_list ap);

static TIFFErrorHandler operror = TIFFSetErrorHandler(tiffErrorHandler);

//
// Loads a tiff file and constructs a buffer for the first image
// in the data.  Returns NULL if the file could not be opened or is
// not in tiff format.
// Note that tiff files can contain more than one image and we only
// deal with the first image on its directory.
//
unsigned long *Pad_LoadTIFF(char *filename, int *ret_width, int *ret_height) {
  TIFF* tif = TIFFOpen(filename, "r");
  unsigned long *raster=NULL, w, h;

  *ret_width = *ret_height = 0;
  tiffReadRGBAImage(tif, &w, &h, &raster, 1);
  *ret_width = w;
  *ret_height = h;
  return raster;
}

//
// Constructs an image from the given tiff data buffer.
//
unsigned long *Pad_LoadTIFF(unsigned char *datain, int lenin, int *ret_width, int *ret_height) {
  tiff_databuf = datain;
  tiff_databuf_size = lenin;
  tiff_databuf_curpos = 0;
  TIFF* tif = TIFFClientOpen("", "r", tiff_databuf,
			     tiffReadProc, NULL, tiffSeekProc,
			     tiffCloseProc, tiffSizeProc, 
			     tiffMapFileProc, tiffUnmapFileProc);
  unsigned long *raster=NULL, w, h;
  *ret_width = *ret_height = 0;
  
  tiffReadRGBAImage(tif, &w, &h, &raster, 1);
  *ret_width = w;
  *ret_height = h;
  return raster;
}

//
// Reads the first tiff image and converts it for pad.
//
static void tiffReadRGBAImage(TIFF *tif, unsigned long *ret_w, unsigned long *ret_h,
			 unsigned long **raster_buf, int stopOnError) {
  if (tif) {
    size_t npixels;
    unsigned long w, h;
    TIFFGetField(tif, TIFFTAG_IMAGEWIDTH, &w);
    TIFFGetField(tif, TIFFTAG_IMAGELENGTH, &h);
    *ret_w = w;
    *ret_h = h;
    npixels = w * h;
    //raster = (unsigned long*) _TIFFmalloc(npixels * sizeof(unsigned long));
    unsigned long *raster = new unsigned long[npixels];
    *raster_buf = raster;
    if (raster != NULL) {
      if (TIFFReadRGBAImage(tif, w, h, raster, stopOnError)) {
	//...process raster data...
	unsigned long *dataptr = raster;
	uint32 pixel;
	for (int i = 0; i < npixels; i++) {
	  pixel = (uint32)*dataptr;
	  *dataptr++ = CC_RGBA(TIFFGetR(pixel), TIFFGetG(pixel), TIFFGetB(pixel), 0);
	}

	// flip rows
	unsigned long *tmp = new unsigned long[w];
	unsigned long *r1, *r2;
	int rowsize = w*sizeof(unsigned long);
	for (int y = 0; y < h/2; y++) {
	  r1 = raster+y*w;
	  r2 = raster+(h-y-1)*w;
	  memcpy(tmp, r2, rowsize);
	  memcpy(r2, r1, rowsize);
	  memcpy(r1, tmp, rowsize);
	}
	delete tmp;
	//_TIFFfree(raster);
      }
    }
    TIFFClose(tif);
  }
}

//
// The following routines will be used by libtiff when reading images
// from an in memory data buffer.
//

tsize_t tiffReadProc (thandle_t, tdata_t data, tsize_t size) {
  tsize_t opsize = size;
  if (tiff_databuf_size - tiff_databuf_curpos < size)
    opsize = tiff_databuf_size - tiff_databuf_curpos;

  memcpy((char*)data, tiff_databuf + tiff_databuf_curpos, opsize);
  return opsize;
}

toff_t tiffSeekProc(thandle_t, toff_t offset, int whence) {
  if (whence == SEEK_SET) {
    tiff_databuf_curpos = offset;
  } else if (whence == SEEK_CUR) {
    tiff_databuf_curpos += offset;
  } else { //SEEK_END
    return -1;
  }

  return tiff_databuf_curpos;
}

int tiffCloseProc(thandle_t) {
  return 0;
}

toff_t tiffSizeProc(thandle_t) {
  return tiff_databuf_size;
}

int tiffMapFileProc(thandle_t, tdata_t* pbase, toff_t* psize) {
  *psize = tiff_databuf_size;
  *pbase = tiff_databuf;
  return 1;
}

void tiffUnmapFileProc(thandle_t, tdata_t, toff_t) {
}

void tiffErrorHandler(const char* module, const char* fmt, va_list ap) {
  if ( !strstr(fmt,"bad magic number") ) {
    operror(module, fmt, ap);
  }
}

#endif  /* HAS_TIFF */
