/*
"(c) Copyright 1993-1997 Pad++ Consortium {University of New Mexico (UNM),
and New York University (NYU)}, All Rights Reserved."  
Licensee can not remove or obscure any of the
copyright or trademark notices in this software.

IN NO EVENT SHALL THE AUTHORS OR DISTRIBUTORS BE LIABLE TO ANY PARTY
FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
ARISING OUT OF THE USE OF THIS SOFTWARE, ITS DOCUMENTATION, OR ANY
DERIVATIVES THEREOF, EVEN IF THE AUTHORS HAVE BEEN ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.

THE AUTHORS AND DISTRIBUTORS SPECIFICALLY DISCLAIM ANY WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE, AND NON-INFRINGEMENT.  THIS SOFTWARE
IS PROVIDED ON AN "AS IS" BASIS, AND THE AUTHORS AND DISTRIBUTORS HAVE
NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR
MODIFICATIONS.

See the file "License" for general information on usage and
redistribution, and the file "LicenseTerms" for the specific license
agreement on usage and redistribution of this file, and the Pad++
software in general.
*/

#include "defs.h"
#include "misc.h"
#include "tkwin.h"
#include "callback.h"
#include "init.h"
#include "global.h"
#include "tkkpl.h"
#include "display.h"
#include "renderer.h"
#include "effect.h"
#include "api.h"
#include "imagedata.h"
#include <string.h>
#include <stdlib.h>
#include <X11/X.h>

//
// Forward reference static procedures
//
static void Create_tcl_proc(Pad_Language *language, char *pathname, Pad_TkWin *tkwin);
static void Destroy_window(char *clientData);
static int  Eval_tcl_proc(Pad_Language *language, char *cmd);
static int  New_tcl_pad(ClientData clientData, Tcl_Interp *interp, int argc, char **argv);
static void Prompt_tcl_proc(Pad_Language *language, int partial, Pad_String &prompt);
static void Event_proc(ClientData clientData, XEvent *eventPtr);
static void Initialize_unit_factors(Pad_TkWin *tkwin);

////////////////////////////////////////////////////////////////////////////
//
// Top-level procedures to interact directly with Tk
//
////////////////////////////////////////////////////////////////////////////

//
// Called upon system startup to create "pad" command in Tcl.
//
extern "C" int 
Pad_Init(Tcl_Interp *interp)
{

    Tcl_CreateCommand(interp, "pad", New_tcl_pad, (ClientData)Tk_MainWindow(interp), NULL);

    Tcl_SetVar(interp, "Pad_Error", "0", TCL_GLOBAL_ONLY);
    Tcl_SetVar(interp, "Pad_Version", Pad_version, TCL_GLOBAL_ONLY);

    if (Kpl_Init(interp) == TCL_ERROR) {
	return TCL_ERROR;
    }

				// Initialize Tcl language
    new Pad_Language(interp, "tcl", "tcl", (Pad_InitProc *)NULL, Create_tcl_proc,
		     (Pad_CommandProc *)NULL, Prompt_tcl_proc, Tcl_CommandComplete, 
		     Eval_tcl_proc);

				// Start with default 'tcl' language
    Pad_GetLanguage("tcl", &Pad_topLevelLanguage);

    return(TCL_OK);
}

static void
Create_tcl_proc(Pad_Language *language, char *pathname, Pad_TkWin *tkwin)
{
    Tcl_CreateCommand(language->interp, pathname, Pad_Command, (ClientData)tkwin, 
		      (void (*)(ClientData)) NULL);
}

static int
Eval_tcl_proc(Pad_Language *language, char *cmd)
{
    int rc;

				// Should really call Tcl_RecordAndEval here,
				// but can't because it seems to reset Tcl's
				// recursion level - resulting in an error
				// if break is executed from an event binding.
				// The only loss is that history isn't recorded.
    rc = Tcl_GlobalEval(language->interp, cmd);

    return(rc);
}

//
// Issue a prompt on standard output, or invoke a script
// to issue the prompt.
//
static void
Prompt_tcl_proc(Pad_Language *language, int partial, Pad_String &prompt)
{
    char *promptCmd;
    int code;

    promptCmd = Tcl_GetVar(language->interp,
	(char *)(partial ? "tcl_prompt2" : "tcl_prompt1"), TCL_GLOBAL_ONLY);
    if (promptCmd == NULL) {
	defaultPrompt:
	if (!partial) {
	    prompt = "% ";
	}
    } else {
	code = Tcl_Eval(language->interp, promptCmd);
	if (code != TCL_OK) {
	    Tcl_AddErrorInfo(language->interp,
		    "\n    (script that generates prompt)");
	    fprintf(stderr, "%s\n", language->interp->result);
	    goto defaultPrompt;
	}
    }
}

//
// Create a new pad widget within a top-level Tk window.
// <padName> must specify a valid Tk path to a new window.
// The pad widget must still be managed by a Tk Geometry Manager
// (such as pack or grid before it will appear).
//
Pad_Bool
Pad_Create_win(Tcl_Interp *interp, char *padName)
{
    int rc;
    char *args[2];

    args[0] = "pad";
    args[1] = padName;
    rc = New_tcl_pad((ClientData)Tk_MainWindow(interp), interp, 2, args);

    return((rc == TCL_OK) ? TRUE : FALSE);
}

//
// Called when Pad is executed - creation of an Pad.  This does several things.
//   * If no pathName is given, a unique name is created.
//   * Creates a new window
//   * Creates an 'Pad' data structure
//   * Creates a Pad_View and attaches it to this Pad
//   * Creates an event handler for this window
//   * Creates a command that handles this object
//   * Configures this Pad for the given arguments
//
static int
New_tcl_pad(ClientData clientData, Tcl_Interp *interp, int argc, char **argv)
{
    char *name;
    Pad_String newName;
    Tk_Window main = (Tk_Window)clientData;
    Pad_TkWin *tkwin;
    Tk_Window tkwindow;
    Pad_HashTableIterator hi;
    Pad_Language *language;

    static int padId = 0;	// Counter for generating unique pad names

    //
    // Create the window.
    //
    if (argc > 1) {
	name = argv[1];
    }
    else {
	newName.Printf(".pad%d", padId++);
	name = newName.Get();
    }
    tkwindow = Tk_CreateWindowFromPath(interp, main, name, (char *) NULL);
    if (tkwindow == NULL) {
	return(TCL_ERROR);
    }
    Tk_SetClass(tkwindow, "Pad");

				// Make a Pad window structure
    tkwin = new Pad_TkWin(interp, tkwindow);
    
    Tk_CreateEventHandler(tkwindow,
			  ExposureMask | StructureNotifyMask,
			  Event_proc,
			  (ClientData)tkwin);

    Tk_CreateEventHandler(tkwindow,
			  KeyPressMask | KeyReleaseMask |
			  ButtonPressMask | ButtonReleaseMask | EnterWindowMask |
			  LeaveWindowMask | PointerMotionMask,
			  Pad_Bind_proc,
			  (ClientData)tkwin);

				// Call the create function for each registered language
    DOTABLE(hi, Pad_languages, Pad_Language, language) {
	if (language->createProc) {
	    language->createProc(language, name, tkwin);
	}
    }

    if (Pad_Configure(tkwin, interp, argc-2, argv+2) == TCL_ERROR)
      {
	  Tk_DestroyWindow(tkwin->Get_tkwindow());
	  return(TCL_ERROR);
      }

    tkwin->Make_window_restorer();

    TCL_RESULT(interp, name);
    return TCL_OK;
}

////////////////////////////////////////////////////////////////////////////
//
// Definition of Pad_TkWin class
//
////////////////////////////////////////////////////////////////////////////

Pad_TkWin::Pad_TkWin(Tcl_Interp *interp, Tk_Window tkwindow) :
Pad_Win(interp, Tk_Display(tkwindow), Tk_Screen(tkwindow), Tk_Visual(tkwindow), 
	Tk_Colormap(tkwindow), Tk_Depth(tkwindow), Tk_PathName(tkwindow))
{
    Set_tkwindow(tkwindow);
				// Tk needs to be informed about window size
    Tk_GeometryRequest(tkwindow, width, height);
}

Pad_TkWin::~Pad_TkWin()
{
    Tcl_DeleteCommand(interp, Get_name());
    Set_tkwindow(NULL);

    Tk_FreeOptions(Pad_configSpecs, (char *)this, dpy->display, 0);
}

//
// Set Tk window pointer
//
void
Pad_TkWin::Set_tkwindow(Tk_Window tkwin)
{
    _tkwin = tkwin;
}

//
// Return name of window
//
Tk_Window
Pad_TkWin::Get_tkwindow(void)
{
    return(_tkwin);
}


Window
Pad_TkWin::Get_toplevel_window_id()
{
    Tk_Window win = _tkwin;
    while (!Tk_IsTopLevel(win)) win = Tk_Parent(win);
    return Tk_WindowId(win);
}


//
// Window's configuration has changed
//
void
Pad_TkWin::Configure(int newWidth, int newHeight)
{
    Pad_Win::Configure(newWidth, newHeight);

    Initialize_unit_factors(this);
}

//
// Window has appeared on screen
//
void
Pad_TkWin::Map(void)
{
    // BBB - Delete these
    //    printf("Pad_Map: dim=(%d, %d), Tk_dim(%d, %d)\n",
    //	   width, height, Tk_Width(Get_tkwindow()), Tk_Height(Get_tkwindow()));

    width = Tk_Width(Get_tkwindow());
    height = Tk_Height(Get_tkwindow());

				// Tk does not actually create a window until it
				// is mapped to the screen.
    id = Tk_WindowId(Get_tkwindow());

    Pad_Win::Map();
}

////////////////////////////////////////////////////////////////////////////
//
// Tk configuration of pad widget
//
////////////////////////////////////////////////////////////////////////////

				// True when window needs to be reconfigured
static Pad_Bool configurationNeeded = FALSE;
				// True when window needs to be redrawn
static Pad_Bool damageNeeded = FALSE;

int
Pad_Configure_widget(ClientData clientData, Tcl_Interp *interp, int argc, char *argv[])
{
    Pad_TkWin *tkwin = (Pad_TkWin *)clientData;

    if (Tk_ConfigureWidget(interp, tkwin->Get_tkwindow(), Pad_configSpecs,
			   argc, argv, (char *)tkwin, TK_CONFIG_ARGV_ONLY) == TCL_ERROR) {
	return(TCL_ERROR);
    }
				// Some options need the window to be reconfigured
    if (configurationNeeded) {
	configurationNeeded = FALSE;
	tkwin->Configure(tkwin->width, tkwin->height);
    }

				// Some options need the window to be redrawn
    if (damageNeeded) {
	damageNeeded = FALSE;
	tkwin->view->Damage();
    }
    
    return(TCL_OK);
}

//		
// This gets called to handle Pad window configuration events
//
static void 
Event_proc(ClientData clientData, XEvent *eventPtr)
{
    Pad_TkWin *tkwin = (Pad_TkWin *)clientData;

    switch (eventPtr->type)
      {
	case Expose:
	  if (eventPtr->xexpose.count == 0) {
	      tkwin->Expose_win();
	  }
	  break;
        case ConfigureNotify:
	  tkwin->Configure(eventPtr->xconfigure.width, eventPtr->xconfigure.height);
	  break;
	case DestroyNotify:
	  Tcl_EventuallyFree((ClientData)tkwin, Destroy_window);
	  break;
	case MapNotify:
	  tkwin->Map();
	  break;
	case UnmapNotify:
	  tkwin->Unmap();
	  break;
      }
}

//
// Call when Pad_TkWin is destroyed
//
static void 
Destroy_window(char *clientData)
{
    Pad_TkWin *tkwin = (Pad_TkWin *)clientData;
    
    delete tkwin;
}

//
// Custom handling of window width configuation
//
static int 
width_parseProc(ClientData, Tcl_Interp *, Tk_Window, char *value, char *widgRec, int)
{
    Pad_TkWin *tkwin = (Pad_TkWin *)widgRec;

    tkwin->width = atoi(value);
    Tk_GeometryRequest(tkwin->Get_tkwindow(), tkwin->width, tkwin->height);
    configurationNeeded = TRUE;

    return(TCL_OK);
}

static char *
width_printProc(ClientData, Tk_Window, char *widgRec,
		 int, Tcl_FreeProc **)
{
    static char buf[30];
    Pad_TkWin *tkwin = (Pad_TkWin *)widgRec;

    sprintf(buf, "%d", tkwin->width);
    return(buf);
}

//
// Custom handling of window height configuation
//
static int 
height_parseProc(ClientData, Tcl_Interp *, Tk_Window, char *value, char *widgRec, int)
{
    Pad_TkWin *tkwin = (Pad_TkWin *)widgRec;

    tkwin->height = atoi(value);
    Tk_GeometryRequest(tkwin->Get_tkwindow(), tkwin->width, tkwin->height);
    configurationNeeded = TRUE;

    return(TCL_OK);
}

static char *
height_printProc(ClientData, Tk_Window, char *widgRec,
		 int, Tcl_FreeProc **)
{
    static char buf[30];
    Pad_TkWin *tkwin = (Pad_TkWin *)widgRec;

    sprintf(buf, "%d", tkwin->height);
    return(buf);
}

//
// Custom handling of refinement delay configuation
//
static int 
refinementdelay_parseProc(ClientData, Tcl_Interp *, Tk_Window, char *value, char *widgRec, int)
{
    Pad_TkWin *tkwin = (Pad_TkWin *)widgRec;

    tkwin->refinementDelay = atoi(value);

    return(TCL_OK);
}

static char *
refinementdelay_printProc(ClientData, Tk_Window, char *widgRec, int, Tcl_FreeProc **)
{
    static char buf[30];
    Pad_TkWin *tkwin = (Pad_TkWin *)widgRec;

    sprintf(buf, "%d", tkwin->refinementDelay);
    return(buf);
}

//
// Custom handling of window double buffer configuation
//
static int 
doublebuffer_parseProc(ClientData, Tcl_Interp *, Tk_Window, char *value, char *widgRec, int)
{
    Pad_TkWin *tkwin = (Pad_TkWin *)widgRec;

    tkwin->doubleBuffer = Pad_Atob(value);
    configurationNeeded = TRUE;

    return(TCL_OK);
}

static char *
doublebuffer_printProc(ClientData, Tk_Window, char *widgRec,
		       int, Tcl_FreeProc **)
{
    static char buf[30];
    Pad_TkWin *tkwin = (Pad_TkWin *)widgRec;

    sprintf(buf, "%d", tkwin->doubleBuffer);
    return(buf);
}

//
// Custom handling of window sync configuation
//
static int 
sync_parseProc(ClientData, Tcl_Interp *, Tk_Window, char *value, char *widgRec, int)
{
    Pad_TkWin *tkwin = (Pad_TkWin *)widgRec;

    tkwin->Set_sync(Pad_Atob(value));

    return(TCL_OK);
}

static char *
sync_printProc(ClientData, Tk_Window, char *widgRec,
		       int, Tcl_FreeProc **)
{
    static char buf[30];
    Pad_TkWin *tkwin = (Pad_TkWin *)widgRec;

    sprintf(buf, "%d", tkwin->Is_synced());
    return(buf);
}

//
// Custom handling of window renderlevel configuation
//
static int 
renderlevel_parseProc(ClientData, Tcl_Interp *interp, Tk_Window, char *value, char *widgRec, int)
{
    int level;
    Pad_TkWin *tkwin = (Pad_TkWin *)widgRec;

    level = atoi(value);
    if (level >= 0) {
	tkwin->defaultRenderLevel = level;
	damageNeeded = TRUE;
	return(TCL_OK);
    } else {
	TCL_ERR(interp, "-defaultrenderlevel must be greater or equal to 0");
    }
}

static char *
renderlevel_printProc(ClientData, Tk_Window, char *widgRec,
		 int, Tcl_FreeProc **)
{
    static char buf[30];
    Pad_TkWin *tkwin = (Pad_TkWin *)widgRec;

    sprintf(buf, "%d", tkwin->defaultRenderLevel);
    return(buf);
}

//
// Custom handling of default event handlers for the main view
//
static int 
eventhandlers_parseProc(ClientData, Tcl_Interp *interp, Tk_Window, char *value, char *widgRec, int)
{
    int level;
    Pad_TkWin *tkwin = (Pad_TkWin *)widgRec;

    level = atoi(value);
    if (level == 0) {
	tkwin->view->pad->Init_events(FALSE);
    } else if (level == 1) {
	tkwin->view->pad->Init_events(TRUE);
    } else {
	TCL_ERR(interp, "-defaultEventHandlers must be 1 or 0");
    }

    return(TCL_OK);
}

static char *
eventhandlers_printProc(ClientData, Tk_Window, char *widgRec,
		 int, Tcl_FreeProc **)
{
    static char buf[30];
    Pad_TkWin *tkwin = (Pad_TkWin *)widgRec;

    sprintf(buf, "%d", tkwin->view->pad->defaultEvents);
    return(buf);
}

//
// Custom handling of window framerate configuation
//
static int 
framerate_parseProc(ClientData, Tcl_Interp *interp, Tk_Window, char *value, char *widgRec, int)
{
    int frameRate;
    Pad_TkWin *tkwin = (Pad_TkWin *)widgRec;

    frameRate = atoi(value);
    if (frameRate >= 1) {
	tkwin->desiredFrameRate = frameRate;
	damageNeeded = TRUE;
	return(TCL_OK);
    } else {
	TCL_ERR(interp, "-defaultFrameRate must be greater or equal to 1");
    }
}

static char *
framerate_printProc(ClientData, Tk_Window, char *widgRec,
		 int, Tcl_FreeProc **)
{
    static char buf[30];
    Pad_TkWin *tkwin = (Pad_TkWin *)widgRec;

    sprintf(buf, "%d", tkwin->desiredFrameRate);
    return(buf);
}

//
// Custom handling of window debugbb configuation
//
static int 
debugbb_parseProc(ClientData, Tcl_Interp *, Tk_Window, char *value, char *widgRec, int)
{
    Pad_TkWin *tkwin = (Pad_TkWin *)widgRec;

    tkwin->debugBB = Pad_Atob(value);
    damageNeeded = TRUE;

    return(TCL_OK);
}

static char *
debugbb_printProc(ClientData, Tk_Window, char *widgRec,
		 int, Tcl_FreeProc **)
{
    static char buf[30];
    Pad_TkWin *tkwin = (Pad_TkWin *)widgRec;

    sprintf(buf, "%d", tkwin->debugBB);
    return(buf);
}

//
// Custom handling of window debugregion configuation
//
static int 
debugregion_parseProc(ClientData, Tcl_Interp *, Tk_Window, char *value, char *widgRec, int)
{
    Pad_TkWin *tkwin = (Pad_TkWin *)widgRec;

    tkwin->debugRegion = Pad_Atob(value);
    damageNeeded = TRUE;

    return(TCL_OK);
}

static char *
debugregion_printProc(ClientData, Tk_Window, char *widgRec,
		 int, Tcl_FreeProc **)
{
    static char buf[30];
    Pad_TkWin *tkwin = (Pad_TkWin *)widgRec;

    sprintf(buf, "%d", tkwin->debugRegion);
    return(buf);
}

//
// Custom handling of window font cache size configuation
//
static int 
fontcache_parseProc(ClientData, Tcl_Interp *, Tk_Window, char *value, char *widgRec, int)
{
    Pad_TkWin *tkwin = (Pad_TkWin *)widgRec;

    tkwin->dpy->Set_fontcache_size((atoi(value) + 99) / 100);
    tkwin->fontCacheSize = tkwin->dpy->Get_fontcache_size() * 100;

    return(TCL_OK);
}

static char *
fontcache_printProc(ClientData, Tk_Window, char *widgRec,
		    int, Tcl_FreeProc **)
{
    static char buf[30];
    Pad_TkWin *tkwin = (Pad_TkWin *)widgRec;

    sprintf(buf, "%d", tkwin->debugRegion);
    return(buf);
}

//
// Set up conversion factors between input and world units
//
static void 
Initialize_unit_factors(Pad_TkWin *tkwin)
{
    int widthmm, heightmm;
    float ppmx, ppmy;		// pixels/mm

				// If user doesn't set screen size, then use X Server values.
    if (tkwin->dpy->widthmmofscreen == 0) {
	widthmm = WidthMMOfScreen(tkwin->dpy->screen);
    } else {
	widthmm = tkwin->dpy->widthmmofscreen;
    }
    if (tkwin->dpy->heightmmofscreen == 0) {
	heightmm = HeightMMOfScreen(tkwin->dpy->screen);
    } else {
	heightmm = tkwin->dpy->heightmmofscreen;
    }

    ppmx = (float)WidthOfScreen(tkwin->dpy->screen) / widthmm;
    ppmy = (float)HeightOfScreen(tkwin->dpy->screen) / heightmm;
    switch (tkwin->units) {
      case MM:
	tkwin->xfac = ppmx;
	tkwin->yfac = ppmy;
	break;
      case POINTS:
	tkwin->xfac = ppmx * 25.4 / 72.0;
	tkwin->yfac = ppmy * 25.4 / 72.0;
	break;
      case INCHES:
	tkwin->xfac = ppmx * 25.4;
	tkwin->yfac = ppmy * 25.4;
	break;
      case PIXELS:
	tkwin->xfac = 1.0;
	tkwin->yfac = 1.0;
	break;
    }
}

static int 
unit_parseProc(ClientData, Tcl_Interp *interp, Tk_Window, char *value, char *widgRec, int offset)
{
    if (!strcasecmp(value, "points")) {
	*((int *)(widgRec + offset)) = POINTS;
    } else if (!strcasecmp(value, "mm")) {
	*((int *)(widgRec + offset)) = MM;
    } else if (!strcasecmp(value, "inches")) {
	*((int *)(widgRec + offset)) = INCHES;
    } else if (!strcasecmp(value, "pixels")) {
	*((int *)(widgRec + offset)) = PIXELS;
    } else {
	TCL_ERR(interp, "Invalid unit, should be: points, mm, inches, or pixels");
    }
    Initialize_unit_factors((Pad_TkWin *)widgRec);
    return(TCL_OK);
}

static char *
unit_printProc(ClientData, Tk_Window, char *widgRec, int offset, Tcl_FreeProc **)
{
    char *rc;
    switch (*((int *)(widgRec + offset))) {
      case POINTS:
	rc = "points";
	break;
      case MM:
	rc = "mm";
	break;
      case INCHES:
	rc = "inches";
	break;
      case PIXELS:
	rc = "pixels";
	break;
      default:
	rc = "<bad unit>";
	break;
    }
    return(rc);
}

//
// Custom handling of widthmmofscreen configuration
// Used when the X server's notion of the physical screen size is incorrect.
//

static int 
widthmmofscreen_parseProc(ClientData, Tcl_Interp *, Tk_Window, char *value, char *widgRec, int)
{
    Pad_Display *dpy = ((Pad_TkWin *)widgRec)->dpy;

    if (strcmp(value, "")) {
	dpy->widthmmofscreen = atoi(value);
	Initialize_unit_factors((Pad_TkWin *)widgRec);
    }
    return(TCL_OK);
}

static char *
widthmmofscreen_printProc(ClientData, Tk_Window, char *widgRec, int, Tcl_FreeProc **)
{
    Pad_Display *dpy = ((Pad_TkWin *)widgRec)->dpy;

    static char buf[30];
    sprintf(buf, "%d", dpy->widthmmofscreen);
    return(buf);
}

//
// Custom handling of heightmmofscreen configuration
// Used when the X server's notion of the physical screen size is incorrect.
//

static int 
heightmmofscreen_parseProc(ClientData, Tcl_Interp *, Tk_Window, char *value, char *widgRec, int)
{
    Pad_Display *dpy = ((Pad_TkWin *)widgRec)->dpy;

    if (strcmp(value, "")) {
	dpy->heightmmofscreen = atoi(value);
	Initialize_unit_factors((Pad_TkWin *)widgRec);
    }
    return(TCL_OK);
}

static char *
heightmmofscreen_printProc(ClientData, Tk_Window, char *widgRec,
				int, Tcl_FreeProc **)
{
    Pad_Display *dpy = ((Pad_TkWin *)widgRec)->dpy;

    static char buf[30];
    sprintf(buf, "%d", dpy->heightmmofscreen);
    return(buf);
}

//
// Custom handling of gamma configuration
//

static int 
gamma_parseProc(ClientData, Tcl_Interp *, Tk_Window, char *value, char *widgRec, int)
{
    Pad_TkWin *tkwin = (Pad_TkWin *)widgRec;
    Pad_Display *dpy = tkwin->dpy;

    dpy->Set_gamma(atof(value));

    return(TCL_OK);
}

static char *
gamma_printProc(ClientData, Tk_Window, char *widgRec,
		int, Tcl_FreeProc **)
{
    static char buf[30];
    sprintf(buf, "%f", ((Pad_TkWin *)widgRec)->dpy->Get_gamma());
    return(buf);
}

//
// Custom handling of background color
//
static int 
backgroundParseProc(ClientData, Tcl_Interp *interp, Tk_Window, char *value, char *widgRec, int)
{
    Pad_TkWin *tkwin = (Pad_TkWin *)widgRec;
    if (strchr(value, ' ')) {

        // JM's experimental tiling facility:
        // if the background consists of "filename1 filename2 size1 size2"
	// where filename1, filename2 and are the image files containing the image
        // to tile (the images should be the same size). Only filename1 and filename2 actually
        // contain image data.

	Pad_String name;
	Pad_ImageData *image1 = NULL, *image2 = NULL;
	char *c;

	// get the first filename
	if (!(c = strchr(value, ' '))) return(TCL_OK);
	*c = '\0';
	if (Pad_Expand_pathname(interp, value, name)) {
	    image1 = tkwin->renderer->Alloc_image(name.Get());
	}
	*c = ' '; value = c + 1;

	// get the second filename
	if (!(c = strchr(value, ' '))) return(TCL_OK);
	*c = '\0';
	if (Pad_Expand_pathname(interp, value, name)) {
	    image2 = tkwin->renderer->Alloc_image(name.Get());
	}
	*c = ' '; value = c + 1;

	// get the two sizes
	sscanf(value, "%f %f", &tkwin->tilemin, &tkwin->tilemax);

	if (image1 && image2) {
	    tkwin->tile1 = image1;
	    tkwin->tile2 = image2;
	    tkwin->tiletmp = new long[image1->width * image1->height];
	}
    } else {
	tkwin->background.Set(value);
    }
    tkwin->view->Damage();

    return(TCL_OK);
}

static char *
backgroundPrintProc(ClientData, Tk_Window, char *widgRec, int, Tcl_FreeProc **)
{
    static char buf[30];
    Pad_String backgroundName;
    Pad_TkWin *tkwin = (Pad_TkWin *)widgRec;

    if (tkwin->background.Is_set()) {
	tkwin->background.Get(backgroundName);
    } else {
	backgroundName = "none";
    }
    sprintf(buf, "%s", backgroundName.Get());

    return(buf);
}

static Tk_CustomOption sync_struct = {sync_parseProc, sync_printProc, NULL};
static Tk_CustomOption renderlevel_struct = {renderlevel_parseProc, renderlevel_printProc, NULL};
static Tk_CustomOption eventhandlers_struct = {eventhandlers_parseProc, eventhandlers_printProc, NULL};
static Tk_CustomOption framerate_struct = {framerate_parseProc, framerate_printProc, NULL};
static Tk_CustomOption debugbb_struct = {debugbb_parseProc, debugbb_printProc, NULL};
static Tk_CustomOption debugregion_struct = {debugregion_parseProc, debugregion_printProc, NULL};
static Tk_CustomOption fontcache_struct = {fontcache_parseProc, fontcache_printProc, NULL};
static Tk_CustomOption width_struct = {width_parseProc, width_printProc, NULL};
static Tk_CustomOption height_struct = {height_parseProc, height_printProc, NULL};
static Tk_CustomOption doublebuffer_struct = {doublebuffer_parseProc, doublebuffer_printProc, NULL};
static Tk_CustomOption unit_struct = {unit_parseProc, unit_printProc, NULL};
static Tk_CustomOption gamma_struct = {gamma_parseProc, gamma_printProc, NULL};
static Tk_CustomOption backgroundStruct = {backgroundParseProc, backgroundPrintProc, NULL};
static Tk_CustomOption refinementdelay_struct = {refinementdelay_parseProc, refinementdelay_printProc, NULL};
static Tk_CustomOption widthmmofscreen_struct = {widthmmofscreen_parseProc, \
						 widthmmofscreen_printProc, NULL};
static Tk_CustomOption heightmmofscreen_struct = {heightmmofscreen_parseProc, \
						  heightmmofscreen_printProc, NULL};

//
// Tk Pad widget configuration stuff here
//
#define DEF_PAD_CURSOR           ((char *)NULL)
#define DEF_PAD_DEBUG_EVENT      ""
#define DEF_PAD_DEBUG_GEN        "0"
#define DEF_PAD_DEBUG_OUT        ""
#define DEF_PAD_DISSOLVE_SPEED   "2"
#define DEF_PAD_GAMMA            "1.0"      // Default gamma
#define DEF_PAD_HEIGHTMM         ""         // Default height of screen in mm
#define DEF_PAD_REFINEMENTDELAY  "1000"
#define DEF_PAD_WIDTHMM          ""         // Default width of screen in mm

Tk_ConfigSpec Pad_configSpecs[] = {
    {TK_CONFIG_CUSTOM, "-background", "background", "Background",
     NULL, 0, 0, &backgroundStruct},

    {TK_CONFIG_SYNONYM, "-bg", "background", NULL,
     NULL, 0, 0, NULL},

    {TK_CONFIG_DOUBLE, "-closeEnough", "closeEnough", "CloseEnough",
     NULL, Tk_Offset(Pad_TkWin, closeEnough), 0, NULL},

    {TK_CONFIG_ACTIVE_CURSOR, "-cursor", "cursor", "Cursor",
     DEF_PAD_CURSOR, Tk_Offset(Pad_TkWin, cursor), TK_CONFIG_NULL_OK, NULL},

    {TK_CONFIG_CUSTOM, "-debugBB", "debugBB", "DebugBB",
     NULL, 0, 0, &debugbb_struct},

    {TK_CONFIG_STRING, "-debugEvent", "debugEvent", "DebugEvent",
     DEF_PAD_DEBUG_EVENT, Tk_Offset(Pad_TkWin, debugEvent), TK_CONFIG_NULL_OK, NULL},

    {TK_CONFIG_STRING, "-debugOut", "debugOut", "DebugOut",
     DEF_PAD_DEBUG_OUT, Tk_Offset(Pad_TkWin, debugOut), TK_CONFIG_NULL_OK, NULL},

    {TK_CONFIG_CUSTOM, "-debugRegion", "debugRegion", "DebugRegion",
     NULL, 0, 0, &debugregion_struct},

    {TK_CONFIG_CUSTOM, "-defaultEventHandlers", "defaultEventHandlers", "DefaultEventHandlers",
     NULL, 0, 0, &eventhandlers_struct},

    {TK_CONFIG_CUSTOM, "-defaultRenderLevel", "defaultRenderLevel", "DefaultRenderLevel",
     NULL, 0, 0, &renderlevel_struct},

    {TK_CONFIG_CUSTOM, "-desiredFrameRate", "desiredFrameRate", "DesiredFrameRate",
     NULL, 0, 0, &framerate_struct},

    {TK_CONFIG_INT, "-dissolveSpeed", "dissolveSpeed", "DissolveSpeed",
     DEF_PAD_DISSOLVE_SPEED, Tk_Offset(Pad_TkWin, refineDissolveSpeed), 0, NULL},

    {TK_CONFIG_INT, "-smallObjSize", "smallObjSize", "SmallObjSize",
     NULL, Tk_Offset(Pad_TkWin, smallObjSize), 0, NULL},

    {TK_CONFIG_INT, "-mediumObjSize", "mediumObjSize", "MediumObjSize",
     NULL, Tk_Offset(Pad_TkWin, mediumObjSize), 0, NULL},

    {TK_CONFIG_CUSTOM, "-doubleBuffer", "doubleBuffer", "DoubleBuffer",
     NULL, 0, TK_CONFIG_ARGV_ONLY, &doublebuffer_struct},

    {TK_CONFIG_INT, "-fastPan", "fastPan", "FastPan",
     NULL, Tk_Offset(Pad_TkWin, fastPan), 0, NULL},

    {TK_CONFIG_CUSTOM, "-fontCacheSize", "fontCacheSize", "FontCacheSize",
     NULL, 0, 0, &fontcache_struct},

    {TK_CONFIG_CUSTOM, "-gamma", "gamma", "Gamma",
     NULL, 0, TK_CONFIG_ARGV_ONLY, &gamma_struct},

    {TK_CONFIG_CUSTOM, "-height", "height", "Height",
     NULL, 0, 0, &height_struct},

    {TK_CONFIG_CUSTOM, "-heightmmofscreen", "heightmmofscreen", "HeightMMOfScreen",
     DEF_PAD_HEIGHTMM, 0, 0, &heightmmofscreen_struct},

    {TK_CONFIG_INT, "-interruptible", "interruptible", "Interruptible",
     NULL, Tk_Offset(Pad_TkWin, interruptible), 0, NULL},

    {TK_CONFIG_DOUBLE, "-maxzoom", "maxZoom", "MaxZoom",
     NULL, Tk_Offset(Pad_TkWin, maxZoom), 0, NULL},

    {TK_CONFIG_CUSTOM, "-refinementDelay", "refinementDelay", "RefinementDelay",
     NULL, 0, 0, &refinementdelay_struct},

    {TK_CONFIG_CUSTOM, "-sync", "sync", "Sync",
     NULL, 0, 0, &sync_struct},

    {TK_CONFIG_CUSTOM, "-units", "units", "Units",
     NULL, Tk_Offset(Pad_TkWin, units), 0, &unit_struct},

    {TK_CONFIG_CUSTOM, "-width", "width", "Width",
     NULL, 0, 0, &width_struct},

    {TK_CONFIG_CUSTOM, "-widthmmofscreen", "widthmmofscreen", "WidthMMOfScreen",
     DEF_PAD_WIDTHMM, 0, 0, &widthmmofscreen_struct},

    {TK_CONFIG_END, (char *) NULL, (char *) NULL, (char *) NULL,
     (char *) NULL, 0, 0, NULL}
};

