#ifdef HAS_LINUX_SOUND 
/*
"(c) Copyright 1993-1997 Pad++ Consortium {University of New Mexico (UNM),
and New York University (NYU)}, All Rights Reserved."  
Licensee can not remove or obscure any of the
copyright or trademark notices in this software.

IN NO EVENT SHALL THE AUTHORS OR DISTRIBUTORS BE LIABLE TO ANY PARTY
FOR DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES
ARISING OUT OF THE USE OF THIS SOFTWARE, ITS DOCUMENTATION, OR ANY
DERIVATIVES THEREOF, EVEN IF THE AUTHORS HAVE BEEN ADVISED OF THE
POSSIBILITY OF SUCH DAMAGE.

THE AUTHORS AND DISTRIBUTORS SPECIFICALLY DISCLAIM ANY WARRANTIES,
INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE, AND NON-INFRINGEMENT.  THIS SOFTWARE
IS PROVIDED ON AN "AS IS" BASIS, AND THE AUTHORS AND DISTRIBUTORS HAVE
NO OBLIGATION TO PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR
MODIFICATIONS.

See the file "License" for general information on usage and
redistribution, and the file "LicenseTerms" for the specific license
agreement on usage and redistribution of this file, and the Pad++
software in general.
*/

#include "linux-sound.h"
#include "../generic/pad-sound.h"
#include "sound-tools.h"
#include "sound-lib.h"
#include <tcl.h>
#include <stdlib.h>
#include <iostream.h>
#include <unistd.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include "../generic/global.h"
#include "soundcard.h"
#include <stdarg.h>

#undef DEBUG_SOUND
//#define DEBUG_SOUND 1

static int FRAGMENT_SIZE;

static int old_left, old_right;
static int oldTime = 0;         // used in the play timer
static Tcl_TimerToken playToken = 0;

static char *MIXER_DEV = "/dev/mixer";
FILE *dspfile;

static Pad_Sound const *currentLoopSound = NULL;

//
// The SoundDevice class is a singleton class - only one instance
//    can exist at any time. To access that instance a Get_instance()
//    fucntion is used. Each Subclass of SoundDevice implements its
//    own version of SoundDevice::Get_instance(), and only one version
//    is compiled in at load time.
//
SoundDevice *SoundDevice::_instance = NULL;

SoundDevice *
SoundDevice::Get_instance(void) {
    if (! _instance) {
        _instance = new LinuxSoundDevice ("/dev/dsp");
    }
    return _instance;
}

//
// Constructors for LinuxSoundDevice should never be called except
//    from Get_instance()
//
LinuxSoundDevice::LinuxSoundDevice(){
    _dspfd = -1;
}

LinuxSoundDevice::LinuxSoundDevice(char *dspname){

    _deviceName = new Pad_String(dspname);
    _bufferSize = 0;

        // Open the audio device - should this be done
        // at play time instead???
    _dspfd = open (dspname, O_WRONLY|O_NONBLOCK, 0);
    if (_dspfd == -1) {
        perror (dspname);
    }

        // Set the fragment size for the internal audio buffer
    int frags = 0x7fff << 16;
    int size = 11 | frags;

//    cout << hex << size << endl;
    
    if (ioctl(_dspfd, SNDCTL_DSP_SETFRAGMENT, &size) == -1) {
        perror(dspname);
    }

        // configure the Pad++ audio buffer
    if (ioctl (_dspfd, SNDCTL_DSP_GETBLKSIZE, &FRAGMENT_SIZE) == -1) {
        perror (dspname);
    }
    _audioBuffer = new char[FRAGMENT_SIZE*2];
    _Set_bufsize();
}

//
// The destructor. As of yet no data needs to be special deleted
//
LinuxSoundDevice::~LinuxSoundDevice(){
}

//
// Load
//
Pad_Bool
LinuxSoundDevice::Load(Pad_Sound &sound){
        // this function allocates memory
    LinuxAudioFormat *format = LinuxAudioFormat::Get_sound_file_format(sound.Get_name());
    if(format == NULL){
        return FALSE;
    }
        // shouldn't need this case...
    if (*format->Get_Filename() == "") {
        Pad_errorString = "Couldn't load sound file: ";
        Pad_errorString += sound.Get_name();
        Pad_errorString += "\n";
            // free up the allocated memory
        delete format;
        return FALSE;
    }
    format->Read();
    sound.Set_format(format);
    return TRUE;
}

//
// _Flush() flushes the Pad++ audio buffer.
//
int
LinuxSoundDevice::_Flush(void){
    if (write (_dspfd, _audioBuffer, _bufferCount) != _bufferCount) {
        perror (_deviceName->Get());

    }
    _bufferCount = 0;
    return TRUE;
}

//
// _Set_channels configures the audio device for either stereo or mono
//    sound playback. If <channels> == 1, stereo sound is configured, if
//    <channels> == 0, mono sound is configured. Yeah I know this is dumb,
//    but that's what linux wants.
//
Pad_Bool
LinuxSoundDevice::_Set_channels(int channels)
{
    int tmp = channels;
    
    if (ioctl (_dspfd, SNDCTL_DSP_STEREO, &tmp) == -1) {
        fprintf(stderr, "Couldn't set to %s\n", channels?  "stereo":"mono");
        return FALSE;
    }
    return TRUE;
}

//
// _Set_samplesize configures the audio device to play either 8 or 16 bit
//    sound. In a stroke of genius, Linux expects either 8 or 16 as input
//    to the function call...
//
Pad_Bool
LinuxSoundDevice::_Set_samplesize(int samplesize)
{
    int tmp = samplesize;
    
    if (ioctl (_dspfd, SNDCTL_DSP_SAMPLESIZE, &tmp) == -1) {  
        fprintf(stderr, "Unable to set the sample size to %d\n", samplesize);
        return FALSE;
    }
    return TRUE;
}

//
// _Set_rate sets the playback rate on the audio device.
//
Pad_Bool
LinuxSoundDevice::_Set_rate(int rate)
{
    int tmp = rate;
    
    ioctl (_dspfd, SNDCTL_DSP_SPEED, &tmp);
    if (rate != tmp) {
        if (rate - tmp > tmp/10 || tmp - rate > tmp/10)
            fprintf (stderr, "Unable to set audio speed to %d (set to %d)\n",
                     rate, tmp);
    }
    return TRUE;
}

//
// _Set_bufsize() configures the Pad++ audio buffer before a Play(). If the
//    internal audio buffer cannot accept a full fragment (or two) of
//    samples this call returns FALSE and expects the calling function
//    to set a timer and call Play() later when there is more space in
//    the internal audio buffer.
//
Pad_Bool
LinuxSoundDevice::_Set_bufsize(void)
{
    audio_buf_info info;

    ioctl(_dspfd, SNDCTL_DSP_GETOSPACE, &info);
        // this sets the buffer size to be the number of bytes that
        // can be written without blocking.
    if (info.fragsize != FRAGMENT_SIZE) {
        cout << "_Set_bufsize: fragsize differs" << endl;
    }
    
    if (info.bytes < FRAGMENT_SIZE) {
        return FALSE;
    }
    if (info.fragments >= 2) {
        _bufferSize = FRAGMENT_SIZE * 2;
        return TRUE;
    }     
    return FALSE;
}

//
// _Init_play() configures all of the necessary parameters of the audio
//    device before a sound can be played. It is important that the
//    attributes are configured in the proper order:
//
//     1) Sample size.
//     2) Channels
//     3) Rate
//
int
LinuxSoundDevice::_Init_play(Pad_AudioFormat *format){
    int samplesize = 8;
    soundstream * ft = (soundstream *) format->Get_private();

    if (ft->info.rate == 0.0) ft->info.rate = 8000;
    if (ft->info.size == -1) ft->info.size = BYTE;
    if (ft->info.size == BYTE) samplesize = 8;
    else if (ft->info.size == WORD) samplesize = 16;

    ft->info.style = UNSIGNED;
    if (ft->info.channels == -1) ft->info.channels = 1;
    else if (ft->info.channels > 2) ft->info.channels = 2;

        // set the sample size
    _Set_samplesize(samplesize);

        // set the number of channels
    if (ft->info.channels == 2) {
            // linux wants 1 to set sound to stereo. Go figure...
        if (!_Set_channels(1)) {
            ft->info.channels = 1;
        }
    } else {
        _Set_channels(0);
    }
    
        // set the sample rate
    _Set_rate(ft->info.rate);

        // Set the buffer size
    _Set_bufsize();
  
        // Reset the audio device - for now a soft reset is used,
        // this only cancels any play timers for other sounds and
        // doesn't actually reset the audio device.
    _Reset(FALSE);

    return TRUE;
}

//
// _Reset() cancels any play timers for sounds being played. If
//    <hard> is TRUE a hard reset is performed immediately cancelling
//    any samples that may be in the internal audio buffer - this has
//    the drawback of causing a lag time between when the first sound
//    is canceled and the new sound starts playing.
//
void
LinuxSoundDevice::_Reset(Pad_Bool hard)
{
        // get rid of any timers dealing with asynchronous play
    if (playToken) {
        Tcl_DeleteTimerHandler (playToken);
    }

    if (hard) {
            // Do a hard reset on the audio device
        ioctl(_dspfd, SNDCTL_DSP_RESET, 0);
    }
}

//
// _Stop_play() is called after a sound has finished playing to alert
//    the audio device that a period of inactivity will result. This
//    call flushes both the Pad++ audio buffer, as well as the internal
//    audio buffer. It also resets the residual time counter used by
//    play timers.
//
void
LinuxSoundDevice::_Stop_play(void){
    _Flush();

//    cout << "_Stop_play: called" << endl << endl;
    
        // Reset the residual time counter
    oldTime = 0;
    
        // Tell the audio device to expect a pause
        // This also flushes the output (very important)
    ioctl(_dspfd, SNDCTL_DSP_POST, 0);

        // This is where we should set a timer to close the
        // audio port when it is done
}

//
// _Set_play_timer() configures a Tcl timer to continue playing a sound that
//    was too long to fit the complete sound into the Pad++ audio buffer.
//
void
LinuxSoundDevice::_Set_play_timer(long *buf,
                                  long nsamp,
                                  Pad_AudioFormat *format,
                                  Pad_PlayState *state)
{
    int time,total;
    Play_struct *data = new Play_struct;

    data->nsamp = nsamp - _bufferCount;
    data->state = state;
    data->format = format;
    data->buf = buf;
    
    total = int((float(_bufferCount) / format->Get_rate()) * 1000);
    time = int(total * 0.75) + oldTime;

#ifdef DEBUG_SOUND    
    cout << "_Set_play_timer: count = " << _bufferCount << endl;
    cout << "_Set_play_timer: nsamp = " << nsamp << endl;
    cout << "_Set_play_timer: time = " << time << endl;
    cout << "_Set_play_timer: oldTime = " << oldTime << endl;
#endif
    
    oldTime = int(total*0.2);
    playToken = Tcl_CreateTimerHandler(time, _Handle_play, (void*)data);

        // write's to the audio device and resets _bufferCount
    _Flush();
}

void
LinuxSoundDevice::_Set_play_timer(Pad_PlayState *state)
{
    int time,total;
    Play_struct *data = new Play_struct;

    data->nsamp = currentLoopSound->Get_length();
    data->state = state;
    data->format = currentLoopSound->Get_format();
    data->buf = (long*) currentLoopSound->Get_data();
    
    total = int((float(_bufferCount) / data->format->Get_rate()) * 1000);
    time = int(total * 0.75);

#ifdef DEBUG_SOUND    
    cout << "_Set_play_timer: count = " << _bufferCount << endl;
    cout << "_Set_play_timer: nsamp = " << data->nsamp << endl;
    cout << "_Set_play_timer: time = " << time << endl;
    cout << "_Set_play_timer: oldTime = " << oldTime << endl << endl;
#endif
    
    oldTime = int(total*0.2);
    playToken = Tcl_CreateTimerHandler(time, _Handle_play, (void*)data);

        // write's to the audio device and resets _bufferCount
    _Flush();
}

//
// This version is used by _Handle_play() when there is not enough space
//    in the internal audio buffer.
//
void
LinuxSoundDevice::_Set_play_timer(Play_struct *data)

{
    int time;
    
    time = 100;

#ifdef DEBUG_SOUND    
    cout << "_Set_play_timer: count = " << _bufferCount << endl;
    cout << "_Set_play_timer: nsamp = " << data->nsamp << endl;
    cout << "_Set_play_timer: time = " << time << endl;
    cout << "_Set_play_timer: oldTime = " << oldTime << endl;
#endif
    
    oldTime = 0;

    playToken = Tcl_CreateTimerHandler(time, _Handle_play, (void*)data);

        // write's to the audio device and resets _bufferCount
    _Flush();
}

//
//  _Handle_play() is the callback called by the Tcl timer to continue
//     playing a sound which was too long to completely fit into the
//     Pad++ audio buffer.
//
void
LinuxSoundDevice::_Handle_play (ClientData clientdata)
{

    Play_struct *data = (Play_struct*)clientdata;
    LinuxSoundDevice *dev = (LinuxSoundDevice *)SoundDevice::Get_instance();

#ifdef DEBUG_SOUND    
    cout << "_Handle_play: nsamp = " << data->nsamp << endl;
#endif
    
    if (!dev->_Set_bufsize()) {
        dev->_Set_play_timer(data);
    }
    
    if(!dev->_Play(data->format,data->buf,data->nsamp,data->state)) {
          // the sound hasn't finished playing so don't call _Stop_play()
      return;
    }
        // Clean up
    dev->_Stop_play();
    return;
}

//
// These are helper functions used by _Play() to store data in the
//   Pad++ audio buffer.
//
void
LinuxSoundDevice::_Write_short(const soundstream * ft, unsigned int ui){
    int c;
    unsigned short us;
  
    us = ui;
    if (ft->swap)
        us = swapw(us);
    c = (us >> 8) & 0xff;
    _Put(c);
    c = us & 0xff;
    _Put(c);
}

void
LinuxSoundDevice::_Put(int c){
    *(_audioBuffer + _bufferCount) = c;
    _bufferCount++;
}

//
// _Play() is the guts of the LinuxSoundDevice. It converts samples one at
//    a time and puts them into the Pad++ audio buffer. If the buffer gets
//    full and there are still sample left, a timer is created to handle
//    the rest of the sound later.
//
//    Note: a lot of this code should be removed, and only one internal
//      format should be supported. Something for a rainy day...
//
Pad_Bool
LinuxSoundDevice::_Play(Pad_AudioFormat *format,
                        long *buffer,
                        long nsamp,
                        Pad_PlayState *state){

#ifdef DEBUG_SOUND    
    cout << "_Play: nsamp = " << nsamp << endl;
#endif
    
    register int datum;
    int done = 0;
    const soundstream * ft = (soundstream*) format->Get_private();
    long *buf = (long *) buffer;
    
    float dampen = 0.01;

    _bufferCount = 0;
  
    switch(ft->info.size) {
        case BYTE: switch(ft->info.style) {
            case SIGN2:
                while(done < nsamp) {
                    if (_bufferCount + 1 > _bufferSize) {
                        _Set_play_timer(buf,nsamp,format,state);
                        return FALSE;
                    }
                                /* scale signed up to long's range */
                    datum = RIGHT(*buf++, 24);
                    _Put(datum);
                    done++;
                }
                break;
            case UNSIGNED:
                while(done < nsamp) {
                    if (_bufferCount + 1 > _bufferSize) {
                        _Set_play_timer(buf,nsamp,format,state);
                        return FALSE;
                    }
                                /* scale signed up to long's range */
                    datum = RIGHT(*buf++, 24);
                                /* Convert to unsigned */
                    datum ^= 128;
                    _Put(datum);
                    done++;
                }
                break;
            case ULAW:
                if (_bufferCount + 1 > _bufferSize) {
                    _Set_play_timer(buf,nsamp,format,state);
                    return FALSE;
                }
				/* grab table from Posk stuff */
                while(done < nsamp) {
                                /* scale signed up to long's range */
                    datum = RIGHT(*buf++, 16);
                    datum = st_linear_to_ulaw(datum);
                    _Put(datum);
                    done++;
                }
                break;
            case ALAW:
                fail("No A-Law support");
                return TRUE;
        }
        case WORD: switch(ft->info.style) {
            case SIGN2:
                while(done < nsamp) {
                    if (_bufferCount + 2 > _bufferSize) {
                        _Set_play_timer(buf,nsamp,format,state);
                        return FALSE;
                    }
                                /* scale signed up to long's range */
                    datum = RIGHT(*buf++, 16);
                    _Write_short(ft, datum);
                    done++;
                }
                break;
            case UNSIGNED:
                while(done < nsamp) {
                    if (_bufferCount + 2 > _bufferSize) {
                        _Set_play_timer(buf,nsamp,format,state);
                        return FALSE;
                    }
                                /* scale signed up to long's range */
              
                    datum = RIGHT(*buf++, 16);
                        // for some reason I need to dampen 16 bit sound
                        // or it sounds awful when you try to play it. I've
                        // made sure that I'm setting audio device sample format
                        // properly but it doesn't seem to matter...
                    datum = int (datum * dampen); // keep the compiler quiet
                                /* Convert to unsigned */
                    datum ^= 0x8000;
                    _Write_short(ft, datum);
                    done++;
                }
                break;
            case ULAW:
                fail("No U-Law support for shorts");
                return TRUE;
            case ALAW:
                fail("No A-Law support");
                return TRUE;
        }
        if (state->Loop()) {
                // this sound should loop until stopped
            _Set_play_timer(state);
        }
    }
    return TRUE;
}

int
LinuxSoundDevice::Play(Pad_SoundConfig *config, Pad_PlayState *state)
{
    Pad_Sound *sound = state->_source;
    int rc;
    
        // Set up the audio port
    if(! _Init_play(sound->Get_format())){
        Pad_errorString = "";
        return FALSE;
    }
    
    if (config->volume->data != Pad_Sound::CONFIG_UNSET) {
          // record the old volume
      Get_volume(old_left,old_right);
      Set_volume(config->volume->data);
    }
    
        // Now really play the sound
    rc = _Play(sound->Get_format(),
               (long*) sound->Get_data(),
               sound->Get_length(),
               state);

    if (config->volume->data != Pad_Sound::CONFIG_UNSET) {
          // reset the volume
      Set_volume(old_left,old_right);
    }
    
    if (!rc) {
            // the sound hasn't finished playing so don't call _Stop_play()
        return TRUE;
    }
    
  
        // Clean up
    _Stop_play();
    return TRUE;
}

  
//
// _Stop() is the interface function that stops any sounds being played.
//   It currently performs a hard reset on the audio device.
//
int
LinuxSoundDevice::Stop(){
    _Reset(TRUE);

        // in case we were looping, null out the pointer
    currentLoopSound = NULL;

        // Reset the residual time counter
    oldTime = 0;
    
    return TRUE;
}
 
//
// Set_volume() sets either the both channels to the same volume, or sets
//    them independantly. If a sound is being played a Tcl timer is created
//    to set the volume when the device is idle.
//
// This procedure sets Pad_resultString on success and Pad_errorString
//   on failure.
//
int
LinuxSoundDevice::Set_volume (int &vol){
    int mixerFD;                 // the file descriptor for the mixer
    int left, right;
    
    if (! Is_Ready()) {
            // something is being played so create a timer
            // to set the volume later
        Volume_struct *volStruct = new Volume_struct;
        volStruct->right = vol;
        volStruct->left = vol;
        Tcl_CreateTimerHandler(100,_Handle_set_volume,(void*)volStruct);
    }
    
    if (vol < 0 || vol > 100){
        Pad_errorString = "Bad Value: accepted range is 0-100.";
        return FALSE;
    }

        // Don't reset the volume when there is a sound playing,
        // so we block until it's done - this allows any residual samples
        // to finish playing (I don't completely trust Is_Ready() ...)
    ioctl(_dspfd,SNDCTL_DSP_SYNC,0);

        // manipulate vol so it is in the form the linux expects
    _Set_stereo_volumes (vol);
  
        // open the mixer
    mixerFD = open (MIXER_DEV, O_WRONLY, 0);
    if (mixerFD == -1){
        Pad_errorString = "Couldn't open mixer device: ";
        Pad_errorString += MIXER_DEV;
        return FALSE;
    }

        // configure the volume
    if (ioctl(mixerFD, SOUND_MIXER_WRITE_PCM, &vol)==-1){
        Pad_errorString = "No volume mixer supported by this sound card.";
        close (mixerFD);
        return FALSE;
    }

        // we store the actual volume in the output variable <vol>
    _Get_stereo_volumes(left,right,vol);
    vol = left;
        // close the mixer
    close (mixerFD);
    return TRUE;
}

int
LinuxSoundDevice::Set_volume (int &right, int &left){
    int mixerFD;                 // the file descriptor for the mixer
    int vol;

    if (! Is_Ready()) {
            // something is being played so create a timer
            // to set the volume later
        Volume_struct *vol = new Volume_struct;
        vol->right = right;
        vol->left = left;
        Tcl_CreateTimerHandler(100,_Handle_set_volume,(void*)vol);
    }
    
    if (right < 0 || right > 100 || left < 0 || left > 100){
        Pad_errorString = "Bad Value: accepted range is 0-100.";
        return FALSE;
    }

        // Don't reset the volume when there is a sound playing,
        // so we block until it's done - this allows any residual samples
        // to finish playing (I don't completely trust Is_Ready() ...)
    ioctl(_dspfd,SNDCTL_DSP_SYNC,0);

        // manipulate vol so it is in the form the linux expects
    _Set_stereo_volumes (left, right, vol);
  
        // open the mixer
    mixerFD = open (MIXER_DEV, O_WRONLY, 0);
    if (mixerFD == -1){
        Pad_errorString = "Couldn't open mixer device: ";
        Pad_errorString += MIXER_DEV;
        return FALSE;
    }

        // configure the volume
    if (ioctl(mixerFD, SOUND_MIXER_WRITE_PCM, &vol)==-1){
        Pad_errorString = "No volume mixer supported by this sound card.";
        close (mixerFD);
        return FALSE;
    }

        // we store the actual volume in the output variables
        // <left> and <right>
    _Get_stereo_volumes(left,right,vol);

        // close the mixer
    close (mixerFD);
    return TRUE;
}

//
// _Handle_set_volume is called by the Tcl timer to set the audio volume
//   when the audio device is busy.
//
void
LinuxSoundDevice::_Handle_set_volume (ClientData clientdata)
{
    Volume_struct *vol = (Volume_struct*) clientdata;
    LinuxSoundDevice *dev = (LinuxSoundDevice*)SoundDevice::Get_instance();
    
    if(dev->Is_Ready()) {
            // ready to be set
        dev->Set_volume(vol->right,vol->left);
            // this memory was allocated when the file handler was created
            // so it must be deleted here.
        delete vol;
    } else {
            // not ready to be set so make a new timer
        Tcl_CreateTimerHandler(100, _Handle_set_volume, (void*)vol);
    }
}

//
// Is_Ready() checks if the audio device is ready for playing or if
//   it is in use. It returns 0 if in use. This procedure should be
//   checking if info.fragments == info.fragstotal, but that didn't work.
//   It always returned false - it seems that there is always one fragment.
//   unavailable...
//
//   Maybe if the audio device was open()'ed at every play
//   and close()'ed after the play had completed this would be easier...
//
int
LinuxSoundDevice::Is_Ready(void) const
{
    audio_buf_info info;

    ioctl(_dspfd, SNDCTL_DSP_GETOSPACE, &info);
        // this doesn't seem to get set right, we should be comparing
        // for strict equality, but it doesn't work...
    if ((info.fragments == info.fragstotal) ||
        (info.fragments == info.fragstotal - 1)) {
        return 1;
    }
    return 0;
}

//
// Get_volume() returns the hardware volume. returns -1 on error.
//
int
LinuxSoundDevice::Get_volume (int &left, int &right) const{
    int mixerFD;                 // the file descriptor for the mixer
    int vol;
  
        // open the mixer
    mixerFD = open (MIXER_DEV, O_RDONLY, 0);
    if (mixerFD == -1){
        Pad_errorString = "Couldn't open mixer device: ";
        Pad_errorString += MIXER_DEV;
        return -1;
    }
  
        // configure the volume
    if (ioctl(mixerFD, SOUND_MIXER_READ_PCM, &vol)==-1){
        Pad_errorString = "No volume mixer supported by this sound card.";
        close (mixerFD);
        return -1;
    }
        // close the mixer
    close (mixerFD);
  
    _Get_stereo_volumes (left, right, vol);
    return TRUE;
}

//
// _Get_stereo_volumes() demangles the packed int <vol> and stores
//    the unpacked values in <left> and <right>.
//
void
LinuxSoundDevice::_Get_stereo_volumes (int &left,
                                       int &right,
                                       int vol) const{
    const int channelMask = 0x7f;

      // compute the volumes for both channels
      // the left volume is in the least significant 8 bits
      // the right volume is in the next 8 bits
      // the most significant 16 bits are unreliable
    left = (unsigned char) channelMask & vol;
    right = (unsigned char) channelMask & (vol >> 8);
}

//
// _Set_stereo_volumes() packs <vol> into the format the Linux audio
// device requires to set the volume. Given a single volume we
// duplicate value in the least significant (upper) 8 bits - which is
// the left channel into the next 8 bits as well for the right
// channel.
//
void
LinuxSoundDevice::_Set_stereo_volumes (int &vol){
    int left;

    left = vol;
    vol |= left << 8;
}

//
// _Set_stereo_volumes
//    given two seperate volumes we store the left value in the least
//    significant (upper) 8 bits, and the right volume in the next
//    8 bits. The value is stored in the int &vol which is used by the
//    calling function.
//
void
LinuxSoundDevice::_Set_stereo_volumes (int left, int right, int &vol){
    vol = left;
    vol |= right << 8;
}

////////////////////////////////////////////////////////////////
//
// LinuxAudioFormat Class
//   This class implements an abstraction around the different
//   input sound file types. It is currently just a wrapper
//   around the code in libst.a, but in the future it would
//   be nice to have it all in Pad++.
//
////////////////////////////////////////////////////////////////

extern autostartread(soundstream *);
extern auread(soundstream * ft, long *buf, long samp);
extern wavread(soundstream * ft, long *buf, long samp);
extern aiffread(soundstream * ft, long *buf, long samp);


//
// Get_sound_file_format() This function does the brunt of the work. It
// determines the type of the audio file by calling autostartread from
// the SoX library. Autostartread() in turn calls the startread()
// function appropriate to the file format. After initializing the
// soundstream * structure, we can now create a new LinuxAudioFormat
// of the appropriate type, initialize it with the soundstream *, and
// return it to the calling function.
//
LinuxAudioFormat *
LinuxAudioFormat::Get_sound_file_format(char *filename){
    FILE *fp;
    soundstream * format;
    Filetype type;

        // open the file for binary reading
    if ((fp = fopen(filename, "rb"))) {
            // if we successfully opened it 
        format = (soundstream *) malloc (sizeof (soundstream));

            /// It's OK if filename is of local scope and goes away once this
            // function returns. We only need it for opening the file, and
            // assigning it to LinuxAudioFormat::_filename.
        format->filename = filename;
        format->filetype = FILETYPE_UNDEF;
        format->swap = 0;
        format->seekable = 1;
        format->info.style = -1;    // setting the info fields became 
        format->info.channels = -1; // important for using .wav
        format->info.rate = -0;
        format->info.size = -1;    
        format->fp = fp;
    } else {
        Pad_errorString = "Couldn't open sound file: ";
        Pad_errorString = filename;
        return NULL;
    }
        // This will initialize the soundstream * struct with the data
        // from the file header. It will then call the sound file format specific
        // startread function from sox.
    autostartread(format);

        // Now that we know what kind of sound file this is,
        // we can return a pointer to the appropriate type.
    type = format->filetype;
    if (type == FILETYPE_AU){
        return  new LinuxAUAudioFormat(format);
    } else if (type == FILETYPE_AIFF){
        return  new LinuxAIFFAudioFormat(format);
    } else if (type == FILETYPE_WAV){
        return  new LinuxWAVAudioFormat(format);
    } else {
        Pad_errorString = "Unsupported sound file format";
        return NULL;
    }
}

//
// Constructors
//
LinuxAudioFormat::LinuxAudioFormat(soundstream * format) : Pad_AudioFormat (format->filename) {
        // the _data and _length fields will get set during _Read().
    _soundstream = format;
    _fp          = _soundstream->fp;
}

LinuxAudioFormat::LinuxAudioFormat(){
    _fp          = stdin;
    _data        = NULL;
    _filename    = NULL;
    _soundstream = NULL;
}

LinuxAudioFormat::LinuxAudioFormat(char *filename){
    _fp        = stdin;
    if ((_fp = fopen(filename, READBINARY))) {
        _filename  = new Pad_String(filename);
        _soundstream = new soundstream;
        _soundstream->filename = _filename->Get();
        _soundstream->filetype = FILETYPE_UNDEF;
        _soundstream->swap = 0;
        _soundstream->seekable = 1;
        _soundstream->fp = _fp;
    } else {
        Pad_errorString = "Couldn't open sound file: ";
        Pad_errorString += filename;
    }
}

//
// Destructor
//
LinuxAudioFormat::~LinuxAudioFormat(){
    if (_soundstream){
        delete _soundstream;
        _soundstream = NULL;
    }
    if (_fp){
        delete _fp;
        _fp = NULL;
    }
}

//
// Get function for the private data
//
void *
LinuxAudioFormat::Get_private() const {
    return _soundstream;
}

//
// Get function for the sound data sample stlye
//
Pad_SoundFormat
LinuxAudioFormat::Get_style() const {
    return (Pad_SoundFormat)_soundstream->info.style;
}

//
// Get function for the sound data sample frequency
//
long 
LinuxAudioFormat::Get_rate() const {
    return _soundstream->info.rate;
}

//
// Get function for the number of samples
//
long 
LinuxAudioFormat::Get_length() const {
    return  _length;
}

long 
LinuxAudioFormat::Get_bits_per_sample() const {
    return (long) ((_soundstream->info.size == 1) ? 8 : 16);
}

//
// Get function for the number of channels in the sound
//
long 
LinuxAudioFormat::Get_channels() const {
    return (long) _soundstream->info.channels;
}

//
// Get the internal representation of the sound data
//
const soundstream *
LinuxAudioFormat::Get_format()const {
    return _soundstream;
}

//
// Called after a read is complete
//
void
LinuxAudioFormat::_Stop_read(void){
    fclose (_fp);
}

//
// Read() loads in sound data from a file into Pad++
//
long
LinuxAudioFormat::Read(){
    long bufsize;
    struct stat file_info;

        // figure out how big the file is so that we know how much
        // info to read in
    fstat(fileno(_fp), &file_info);
        // a long is 4 bytes, so this tells us how many longs we
        // need to hold the data...
    bufsize = file_info.st_size * 4;
  
        // read in the data and store the actuall data size
    _data  = new char[bufsize];
    _length = _Read(_soundstream, _data, bufsize); 
        // cleanup
    _Stop_read();
        // return the data length
    return _length;

}

////////////////////////////////////////////////////////////////
//
// LinuxAUAudioFormat Class
//    Specific _Read() function for .au files
//
////////////////////////////////////////////////////////////////



LinuxAUAudioFormat::LinuxAUAudioFormat(){}
LinuxAUAudioFormat::~LinuxAUAudioFormat(){}
LinuxAUAudioFormat::LinuxAUAudioFormat(char *filename) : LinuxAudioFormat(filename){
}
LinuxAUAudioFormat::LinuxAUAudioFormat(soundstream * format) : LinuxAudioFormat(format){
}

Pad_Bool
LinuxAUAudioFormat::Open_format(char*)
{
    return(FALSE);
}

int
LinuxAUAudioFormat::_Read(soundstream * ft, char *data, long size){
    return ft->h->read (ft, (long *) data, size); 
}

////////////////////////////////////////////////////////////////
//
// LinuxAIFFAudioFormat Class
//    Specific _Read() function for .aiff files
//
////////////////////////////////////////////////////////////////

int
LinuxAIFFAudioFormat::_Read(soundstream * ft, char *data, long size){
    return ft->h->read (ft, (long*) data, size); 
}

LinuxAIFFAudioFormat::LinuxAIFFAudioFormat(){}
LinuxAIFFAudioFormat::~LinuxAIFFAudioFormat(){}
LinuxAIFFAudioFormat::LinuxAIFFAudioFormat(char *filename) : LinuxAudioFormat(filename){
}
LinuxAIFFAudioFormat::LinuxAIFFAudioFormat(soundstream * format) : LinuxAudioFormat(format){
}

Pad_Bool
LinuxAIFFAudioFormat::Open_format(char*)
{
    return(FALSE);
}

////////////////////////////////////////////////////////////////
//
// LinuxWAVAudioFormat Class
//    Specific _Read() function for .wav files
//
////////////////////////////////////////////////////////////////

int
LinuxWAVAudioFormat::_Read(soundstream * ft, char *data, long size){
    return ft->h->read (ft, (long*) data, size); 
}

LinuxWAVAudioFormat::LinuxWAVAudioFormat(){}
LinuxWAVAudioFormat::~LinuxWAVAudioFormat(){}
LinuxWAVAudioFormat::LinuxWAVAudioFormat(char *filename) : LinuxAudioFormat(filename){
}
LinuxWAVAudioFormat::LinuxWAVAudioFormat(soundstream * format) : LinuxAudioFormat(format){
}

Pad_Bool
LinuxWAVAudioFormat::Open_format(char*)
{
    return(FALSE);
}

#endif /* HAS_LINUX_SOUND */
